"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createCryptoStoreCacheCallbacks = createCryptoStoreCacheCallbacks;
exports.requestKeysDuringVerification = requestKeysDuringVerification;
exports.DeviceTrustLevel = exports.UserTrustLevel = exports.CrossSigningLevel = exports.CrossSigningInfo = void 0;

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _assertThisInitialized2 = _interopRequireDefault(require("@babel/runtime/helpers/assertThisInitialized"));

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _olmlib = require("./olmlib");

var _events = require("events");

var _logger = require("../logger");

var _indexeddbCryptoStore = require("../crypto/store/indexeddb-crypto-store");

var _aes = require("./aes");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

var KEY_REQUEST_TIMEOUT_MS = 1000 * 60;

function publicKeyFromKeyInfo(keyInfo) {
  // `keys` is an object with { [`ed25519:${pubKey}`]: pubKey }
  // We assume only a single key, and we want the bare form without type
  // prefix, so we select the values.
  return Object.values(keyInfo.keys)[0];
}

var CrossSigningInfo = /*#__PURE__*/function (_EventEmitter) {
  (0, _inherits2["default"])(CrossSigningInfo, _EventEmitter);

  var _super = _createSuper(CrossSigningInfo);

  /**
   * Information about a user's cross-signing keys
   *
   * @class
   *
   * @param {string} userId the user that the information is about
   * @param {object} callbacks Callbacks used to interact with the app
   *     Requires getCrossSigningKey and saveCrossSigningKeys
   * @param {object} cacheCallbacks Callbacks used to interact with the cache
   */
  function CrossSigningInfo(userId, callbacks, cacheCallbacks) {
    var _this;

    (0, _classCallCheck2["default"])(this, CrossSigningInfo);
    _this = _super.call(this); // you can't change the userId

    Object.defineProperty((0, _assertThisInitialized2["default"])(_this), 'userId', {
      enumerable: true,
      value: userId
    });
    _this._callbacks = callbacks || {};
    _this._cacheCallbacks = cacheCallbacks || {};
    _this.keys = {};
    _this.firstUse = true; // This tracks whether we've ever verified this user with any identity.
    // When you verify a user, any devices online at the time that receive
    // the verifying signature via the homeserver will latch this to true
    // and can use it in the future to detect cases where the user has
    // become unverifed later for any reason.

    _this.crossSigningVerifiedBefore = false;
    return _this;
  }

  (0, _createClass2["default"])(CrossSigningInfo, [{
    key: "toStorage",
    value: function toStorage() {
      return {
        keys: this.keys,
        firstUse: this.firstUse,
        crossSigningVerifiedBefore: this.crossSigningVerifiedBefore
      };
    }
    /**
     * Calls the app callback to ask for a private key
     *
     * @param {string} type The key type ("master", "self_signing", or "user_signing")
     * @param {string} expectedPubkey The matching public key or undefined to use
     *     the stored public key for the given key type.
     * @returns {Array} An array with [ public key, Olm.PkSigning ]
     */

  }, {
    key: "getCrossSigningKey",
    value: function () {
      var _getCrossSigningKey = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(type, expectedPubkey) {
        var shouldCache, validateKey, privkey, cacheresult, result;
        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                validateKey = function _validateKey(key) {
                  if (!key) return;
                  var signing = new global.Olm.PkSigning();
                  var gotPubkey = signing.init_with_seed(key);

                  if (gotPubkey === expectedPubkey) {
                    return [gotPubkey, signing];
                  }

                  signing.free();
                };

                shouldCache = ["master", "self_signing", "user_signing"].indexOf(type) >= 0;

                if (this._callbacks.getCrossSigningKey) {
                  _context.next = 4;
                  break;
                }

                throw new Error("No getCrossSigningKey callback supplied");

              case 4:
                if (expectedPubkey === undefined) {
                  expectedPubkey = this.getId(type);
                }

                if (!(this._cacheCallbacks.getCrossSigningKeyCache && shouldCache)) {
                  _context.next = 9;
                  break;
                }

                _context.next = 8;
                return this._cacheCallbacks.getCrossSigningKeyCache(type, expectedPubkey);

              case 8:
                privkey = _context.sent;

              case 9:
                cacheresult = validateKey(privkey);

                if (!cacheresult) {
                  _context.next = 12;
                  break;
                }

                return _context.abrupt("return", cacheresult);

              case 12:
                _context.next = 14;
                return this._callbacks.getCrossSigningKey(type, expectedPubkey);

              case 14:
                privkey = _context.sent;
                result = validateKey(privkey);

                if (!result) {
                  _context.next = 21;
                  break;
                }

                if (!(this._cacheCallbacks.storeCrossSigningKeyCache && shouldCache)) {
                  _context.next = 20;
                  break;
                }

                _context.next = 20;
                return this._cacheCallbacks.storeCrossSigningKeyCache(type, privkey);

              case 20:
                return _context.abrupt("return", result);

              case 21:
                if (privkey) {
                  _context.next = 23;
                  break;
                }

                throw new Error("getCrossSigningKey callback for " + type + " returned falsey");

              case 23:
                throw new Error("Key type " + type + " from getCrossSigningKey callback did not match");

              case 24:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function getCrossSigningKey(_x, _x2) {
        return _getCrossSigningKey.apply(this, arguments);
      }

      return getCrossSigningKey;
    }()
    /**
     * Check whether the private keys exist in secret storage.
     * XXX: This could be static, be we often seem to have an instance when we
     * want to know this anyway...
     *
     * @param {SecretStorage} secretStorage The secret store using account data
     * @returns {object} map of key name to key info the secret is encrypted
     *     with, or null if it is not present or not encrypted with a trusted
     *     key
     */

  }, {
    key: "isStoredInSecretStorage",
    value: function () {
      var _isStoredInSecretStorage = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(secretStorage) {
        var stored, intersect, _i2, _arr, type;

        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                intersect = function _intersect(s) {
                  for (var _i = 0, _Object$keys = Object.keys(stored); _i < _Object$keys.length; _i++) {
                    var k = _Object$keys[_i];

                    if (!s[k]) {
                      delete stored[k];
                    }
                  }
                };

                _context2.next = 3;
                return secretStorage.isStored("m.cross_signing.master", false);

              case 3:
                _context2.t0 = _context2.sent;

                if (_context2.t0) {
                  _context2.next = 6;
                  break;
                }

                _context2.t0 = {};

              case 6:
                stored = _context2.t0;
                _i2 = 0, _arr = ["self_signing", "user_signing"];

              case 8:
                if (!(_i2 < _arr.length)) {
                  _context2.next = 21;
                  break;
                }

                type = _arr[_i2];
                _context2.t1 = intersect;
                _context2.next = 13;
                return secretStorage.isStored("m.cross_signing.".concat(type), false);

              case 13:
                _context2.t2 = _context2.sent;

                if (_context2.t2) {
                  _context2.next = 16;
                  break;
                }

                _context2.t2 = {};

              case 16:
                _context2.t3 = _context2.t2;
                (0, _context2.t1)(_context2.t3);

              case 18:
                _i2++;
                _context2.next = 8;
                break;

              case 21:
                return _context2.abrupt("return", Object.keys(stored).length ? stored : null);

              case 22:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2);
      }));

      function isStoredInSecretStorage(_x3) {
        return _isStoredInSecretStorage.apply(this, arguments);
      }

      return isStoredInSecretStorage;
    }()
    /**
     * Store private keys in secret storage for use by other devices. This is
     * typically called in conjunction with the creation of new cross-signing
     * keys.
     *
     * @param {Map} keys The keys to store
     * @param {SecretStorage} secretStorage The secret store using account data
     */

  }, {
    key: "isStoredInKeyCache",

    /**
     * Check whether the private keys exist in the local key cache.
     *
     * @param {string} [type] The type of key to get. One of "master",
     * "self_signing", or "user_signing". Optional, will check all by default.
     * @returns {boolean} True if all keys are stored in the local cache.
     */
    value: function () {
      var _isStoredInKeyCache = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(type) {
        var cacheCallbacks, types, _i3, _types, t;

        return _regenerator["default"].wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                cacheCallbacks = this._cacheCallbacks;

                if (cacheCallbacks) {
                  _context3.next = 3;
                  break;
                }

                return _context3.abrupt("return", false);

              case 3:
                types = type ? [type] : ["master", "self_signing", "user_signing"];
                _i3 = 0, _types = types;

              case 5:
                if (!(_i3 < _types.length)) {
                  _context3.next = 14;
                  break;
                }

                t = _types[_i3];
                _context3.next = 9;
                return cacheCallbacks.getCrossSigningKeyCache(t);

              case 9:
                if (_context3.sent) {
                  _context3.next = 11;
                  break;
                }

                return _context3.abrupt("return", false);

              case 11:
                _i3++;
                _context3.next = 5;
                break;

              case 14:
                return _context3.abrupt("return", true);

              case 15:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function isStoredInKeyCache(_x4) {
        return _isStoredInKeyCache.apply(this, arguments);
      }

      return isStoredInKeyCache;
    }()
    /**
     * Get cross-signing private keys from the local cache.
     *
     * @returns {Map} A map from key type (string) to private key (Uint8Array)
     */

  }, {
    key: "getCrossSigningKeysFromCache",
    value: function () {
      var _getCrossSigningKeysFromCache = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4() {
        var keys, cacheCallbacks, _i4, _arr2, type, privKey;

        return _regenerator["default"].wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                keys = new Map();
                cacheCallbacks = this._cacheCallbacks;

                if (cacheCallbacks) {
                  _context4.next = 4;
                  break;
                }

                return _context4.abrupt("return", keys);

              case 4:
                _i4 = 0, _arr2 = ["master", "self_signing", "user_signing"];

              case 5:
                if (!(_i4 < _arr2.length)) {
                  _context4.next = 16;
                  break;
                }

                type = _arr2[_i4];
                _context4.next = 9;
                return cacheCallbacks.getCrossSigningKeyCache(type);

              case 9:
                privKey = _context4.sent;

                if (privKey) {
                  _context4.next = 12;
                  break;
                }

                return _context4.abrupt("continue", 13);

              case 12:
                keys.set(type, privKey);

              case 13:
                _i4++;
                _context4.next = 5;
                break;

              case 16:
                return _context4.abrupt("return", keys);

              case 17:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function getCrossSigningKeysFromCache() {
        return _getCrossSigningKeysFromCache.apply(this, arguments);
      }

      return getCrossSigningKeysFromCache;
    }()
    /**
     * Get the ID used to identify the user. This can also be used to test for
     * the existence of a given key type.
     *
     * @param {string} type The type of key to get the ID of.  One of "master",
     * "self_signing", or "user_signing".  Defaults to "master".
     *
     * @return {string} the ID
     */

  }, {
    key: "getId",
    value: function getId(type) {
      type = type || "master";
      if (!this.keys[type]) return null;
      var keyInfo = this.keys[type];
      return publicKeyFromKeyInfo(keyInfo);
    }
    /**
     * Create new cross-signing keys for the given key types. The public keys
     * will be held in this class, while the private keys are passed off to the
     * `saveCrossSigningKeys` application callback.
     *
     * @param {CrossSigningLevel} level The key types to reset
     */

  }, {
    key: "resetKeys",
    value: function () {
      var _resetKeys = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5(level) {
        var privateKeys, keys, masterSigning, masterPub, _yield$this$getCrossS, _yield$this$getCrossS2, sskSigning, sskPub, uskSigning, uskPub;

        return _regenerator["default"].wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                if (this._callbacks.saveCrossSigningKeys) {
                  _context5.next = 2;
                  break;
                }

                throw new Error("No saveCrossSigningKeys callback supplied");

              case 2:
                if (!(level === undefined || level & CrossSigningLevel.MASTER || !this.keys.master)) {
                  _context5.next = 6;
                  break;
                }

                level = CrossSigningLevel.MASTER | CrossSigningLevel.USER_SIGNING | CrossSigningLevel.SELF_SIGNING;
                _context5.next = 8;
                break;

              case 6:
                if (!(level === 0)) {
                  _context5.next = 8;
                  break;
                }

                return _context5.abrupt("return");

              case 8:
                privateKeys = {};
                keys = {};
                _context5.prev = 10;

                if (!(level & CrossSigningLevel.MASTER)) {
                  _context5.next = 18;
                  break;
                }

                masterSigning = new global.Olm.PkSigning();
                privateKeys.master = masterSigning.generate_seed();
                masterPub = masterSigning.init_with_seed(privateKeys.master);
                keys.master = {
                  user_id: this.userId,
                  usage: ['master'],
                  keys: (0, _defineProperty2["default"])({}, 'ed25519:' + masterPub, masterPub)
                };
                _context5.next = 24;
                break;

              case 18:
                _context5.next = 20;
                return this.getCrossSigningKey("master");

              case 20:
                _yield$this$getCrossS = _context5.sent;
                _yield$this$getCrossS2 = (0, _slicedToArray2["default"])(_yield$this$getCrossS, 2);
                masterPub = _yield$this$getCrossS2[0];
                masterSigning = _yield$this$getCrossS2[1];

              case 24:
                if (level & CrossSigningLevel.SELF_SIGNING) {
                  sskSigning = new global.Olm.PkSigning();

                  try {
                    privateKeys.self_signing = sskSigning.generate_seed();
                    sskPub = sskSigning.init_with_seed(privateKeys.self_signing);
                    keys.self_signing = {
                      user_id: this.userId,
                      usage: ['self_signing'],
                      keys: (0, _defineProperty2["default"])({}, 'ed25519:' + sskPub, sskPub)
                    };
                    (0, _olmlib.pkSign)(keys.self_signing, masterSigning, this.userId, masterPub);
                  } finally {
                    sskSigning.free();
                  }
                }

                if (level & CrossSigningLevel.USER_SIGNING) {
                  uskSigning = new global.Olm.PkSigning();

                  try {
                    privateKeys.user_signing = uskSigning.generate_seed();
                    uskPub = uskSigning.init_with_seed(privateKeys.user_signing);
                    keys.user_signing = {
                      user_id: this.userId,
                      usage: ['user_signing'],
                      keys: (0, _defineProperty2["default"])({}, 'ed25519:' + uskPub, uskPub)
                    };
                    (0, _olmlib.pkSign)(keys.user_signing, masterSigning, this.userId, masterPub);
                  } finally {
                    uskSigning.free();
                  }
                }

                Object.assign(this.keys, keys);

                this._callbacks.saveCrossSigningKeys(privateKeys);

              case 28:
                _context5.prev = 28;

                if (masterSigning) {
                  masterSigning.free();
                }

                return _context5.finish(28);

              case 31:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this, [[10,, 28, 31]]);
      }));

      function resetKeys(_x5) {
        return _resetKeys.apply(this, arguments);
      }

      return resetKeys;
    }()
    /**
     * unsets the keys, used when another session has reset the keys, to disable cross-signing
     */

  }, {
    key: "clearKeys",
    value: function clearKeys() {
      this.keys = {};
    }
  }, {
    key: "setKeys",
    value: function setKeys(keys) {
      var signingKeys = {};

      if (keys.master) {
        if (keys.master.user_id !== this.userId) {
          var error = "Mismatched user ID " + keys.master.user_id + " in master key from " + this.userId;

          _logger.logger.error(error);

          throw new Error(error);
        }

        if (!this.keys.master) {
          // this is the first key we've seen, so first-use is true
          this.firstUse = true;
        } else if (publicKeyFromKeyInfo(keys.master) !== this.getId()) {
          // this is a different key, so first-use is false
          this.firstUse = false;
        } // otherwise, same key, so no change


        signingKeys.master = keys.master;
      } else if (this.keys.master) {
        signingKeys.master = this.keys.master;
      } else {
        throw new Error("Tried to set cross-signing keys without a master key");
      }

      var masterKey = publicKeyFromKeyInfo(signingKeys.master); // verify signatures

      if (keys.user_signing) {
        if (keys.user_signing.user_id !== this.userId) {
          var _error = "Mismatched user ID " + keys.master.user_id + " in user_signing key from " + this.userId;

          _logger.logger.error(_error);

          throw new Error(_error);
        }

        try {
          (0, _olmlib.pkVerify)(keys.user_signing, masterKey, this.userId);
        } catch (e) {
          _logger.logger.error("invalid signature on user-signing key"); // FIXME: what do we want to do here?


          throw e;
        }
      }

      if (keys.self_signing) {
        if (keys.self_signing.user_id !== this.userId) {
          var _error2 = "Mismatched user ID " + keys.master.user_id + " in self_signing key from " + this.userId;

          _logger.logger.error(_error2);

          throw new Error(_error2);
        }

        try {
          (0, _olmlib.pkVerify)(keys.self_signing, masterKey, this.userId);
        } catch (e) {
          _logger.logger.error("invalid signature on self-signing key"); // FIXME: what do we want to do here?


          throw e;
        }
      } // if everything checks out, then save the keys


      if (keys.master) {
        this.keys.master = keys.master; // if the master key is set, then the old self-signing and
        // user-signing keys are obsolete

        this.keys.self_signing = null;
        this.keys.user_signing = null;
      }

      if (keys.self_signing) {
        this.keys.self_signing = keys.self_signing;
      }

      if (keys.user_signing) {
        this.keys.user_signing = keys.user_signing;
      }
    }
  }, {
    key: "updateCrossSigningVerifiedBefore",
    value: function updateCrossSigningVerifiedBefore(isCrossSigningVerified) {
      // It is critical that this value latches forward from false to true but
      // never back to false to avoid a downgrade attack.
      if (!this.crossSigningVerifiedBefore && isCrossSigningVerified) {
        this.crossSigningVerifiedBefore = true;
      }
    }
  }, {
    key: "signObject",
    value: function () {
      var _signObject = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6(data, type) {
        var _yield$this$getCrossS3, _yield$this$getCrossS4, pubkey, signing;

        return _regenerator["default"].wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                if (this.keys[type]) {
                  _context6.next = 2;
                  break;
                }

                throw new Error("Attempted to sign with " + type + " key but no such key present");

              case 2:
                _context6.next = 4;
                return this.getCrossSigningKey(type);

              case 4:
                _yield$this$getCrossS3 = _context6.sent;
                _yield$this$getCrossS4 = (0, _slicedToArray2["default"])(_yield$this$getCrossS3, 2);
                pubkey = _yield$this$getCrossS4[0];
                signing = _yield$this$getCrossS4[1];
                _context6.prev = 8;
                (0, _olmlib.pkSign)(data, signing, this.userId, pubkey);
                return _context6.abrupt("return", data);

              case 11:
                _context6.prev = 11;
                signing.free();
                return _context6.finish(11);

              case 14:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this, [[8,, 11, 14]]);
      }));

      function signObject(_x6, _x7) {
        return _signObject.apply(this, arguments);
      }

      return signObject;
    }()
  }, {
    key: "signUser",
    value: function () {
      var _signUser = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee7(key) {
        return _regenerator["default"].wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                if (this.keys.user_signing) {
                  _context7.next = 3;
                  break;
                }

                _logger.logger.info("No user signing key: not signing user");

                return _context7.abrupt("return");

              case 3:
                return _context7.abrupt("return", this.signObject(key.keys.master, "user_signing"));

              case 4:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function signUser(_x8) {
        return _signUser.apply(this, arguments);
      }

      return signUser;
    }()
  }, {
    key: "signDevice",
    value: function () {
      var _signDevice = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee8(userId, device) {
        return _regenerator["default"].wrap(function _callee8$(_context8) {
          while (1) {
            switch (_context8.prev = _context8.next) {
              case 0:
                if (!(userId !== this.userId)) {
                  _context8.next = 2;
                  break;
                }

                throw new Error("Trying to sign ".concat(userId, "'s device; can only sign our own device"));

              case 2:
                if (this.keys.self_signing) {
                  _context8.next = 5;
                  break;
                }

                _logger.logger.info("No self signing key: not signing device");

                return _context8.abrupt("return");

              case 5:
                return _context8.abrupt("return", this.signObject({
                  algorithms: device.algorithms,
                  keys: device.keys,
                  device_id: device.deviceId,
                  user_id: userId
                }, "self_signing"));

              case 6:
              case "end":
                return _context8.stop();
            }
          }
        }, _callee8, this);
      }));

      function signDevice(_x9, _x10) {
        return _signDevice.apply(this, arguments);
      }

      return signDevice;
    }()
    /**
     * Check whether a given user is trusted.
     *
     * @param {CrossSigningInfo} userCrossSigning Cross signing info for user
     *
     * @returns {UserTrustLevel}
     */

  }, {
    key: "checkUserTrust",
    value: function checkUserTrust(userCrossSigning) {
      // if we're checking our own key, then it's trusted if the master key
      // and self-signing key match
      if (this.userId === userCrossSigning.userId && this.getId() && this.getId() === userCrossSigning.getId() && this.getId("self_signing") && this.getId("self_signing") === userCrossSigning.getId("self_signing")) {
        return new UserTrustLevel(true, true, this.firstUse);
      }

      if (!this.keys.user_signing) {
        // If there's no user signing key, they can't possibly be verified.
        // They may be TOFU trusted though.
        return new UserTrustLevel(false, false, userCrossSigning.firstUse);
      }

      var userTrusted;
      var userMaster = userCrossSigning.keys.master;
      var uskId = this.getId('user_signing');

      try {
        (0, _olmlib.pkVerify)(userMaster, uskId, this.userId);
        userTrusted = true;
      } catch (e) {
        userTrusted = false;
      }

      return new UserTrustLevel(userTrusted, userCrossSigning.crossSigningVerifiedBefore, userCrossSigning.firstUse);
    }
    /**
     * Check whether a given device is trusted.
     *
     * @param {CrossSigningInfo} userCrossSigning Cross signing info for user
     * @param {module:crypto/deviceinfo} device The device to check
     * @param {bool} localTrust Whether the device is trusted locally
     * @param {bool} trustCrossSignedDevices Whether we trust cross signed devices
     *
     * @returns {DeviceTrustLevel}
     */

  }, {
    key: "checkDeviceTrust",
    value: function checkDeviceTrust(userCrossSigning, device, localTrust, trustCrossSignedDevices) {
      var userTrust = this.checkUserTrust(userCrossSigning);
      var userSSK = userCrossSigning.keys.self_signing;

      if (!userSSK) {
        // if the user has no self-signing key then we cannot make any
        // trust assertions about this device from cross-signing
        return new DeviceTrustLevel(false, false, localTrust, trustCrossSignedDevices);
      }

      var deviceObj = deviceToObject(device, userCrossSigning.userId);

      try {
        // if we can verify the user's SSK from their master key...
        (0, _olmlib.pkVerify)(userSSK, userCrossSigning.getId(), userCrossSigning.userId); // ...and this device's key from their SSK...

        (0, _olmlib.pkVerify)(deviceObj, publicKeyFromKeyInfo(userSSK), userCrossSigning.userId); // ...then we trust this device as much as far as we trust the user

        return DeviceTrustLevel.fromUserTrustLevel(userTrust, localTrust, trustCrossSignedDevices);
      } catch (e) {
        return new DeviceTrustLevel(false, false, localTrust, trustCrossSignedDevices);
      }
    }
    /**
     * @returns {object} Cache callbacks
     */

  }, {
    key: "getCacheCallbacks",
    value: function getCacheCallbacks() {
      return this._cacheCallbacks;
    }
  }], [{
    key: "fromStorage",
    value: function fromStorage(obj, userId) {
      var res = new CrossSigningInfo(userId);

      for (var prop in obj) {
        if (obj.hasOwnProperty(prop)) {
          res[prop] = obj[prop];
        }
      }

      return res;
    }
  }, {
    key: "storeInSecretStorage",
    value: function () {
      var _storeInSecretStorage = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee9(keys, secretStorage) {
        var _iterator, _step, _step$value, type, privateKey, encodedKey;

        return _regenerator["default"].wrap(function _callee9$(_context9) {
          while (1) {
            switch (_context9.prev = _context9.next) {
              case 0:
                _iterator = _createForOfIteratorHelper(keys);
                _context9.prev = 1;

                _iterator.s();

              case 3:
                if ((_step = _iterator.n()).done) {
                  _context9.next = 10;
                  break;
                }

                _step$value = (0, _slicedToArray2["default"])(_step.value, 2), type = _step$value[0], privateKey = _step$value[1];
                encodedKey = (0, _olmlib.encodeBase64)(privateKey);
                _context9.next = 8;
                return secretStorage.store("m.cross_signing.".concat(type), encodedKey);

              case 8:
                _context9.next = 3;
                break;

              case 10:
                _context9.next = 15;
                break;

              case 12:
                _context9.prev = 12;
                _context9.t0 = _context9["catch"](1);

                _iterator.e(_context9.t0);

              case 15:
                _context9.prev = 15;

                _iterator.f();

                return _context9.finish(15);

              case 18:
              case "end":
                return _context9.stop();
            }
          }
        }, _callee9, null, [[1, 12, 15, 18]]);
      }));

      function storeInSecretStorage(_x11, _x12) {
        return _storeInSecretStorage.apply(this, arguments);
      }

      return storeInSecretStorage;
    }()
    /**
     * Get private keys from secret storage created by some other device. This
     * also passes the private keys to the app-specific callback.
     *
     * @param {string} type The type of key to get.  One of "master",
     * "self_signing", or "user_signing".
     * @param {SecretStorage} secretStorage The secret store using account data
     * @return {Uint8Array} The private key
     */

  }, {
    key: "getFromSecretStorage",
    value: function () {
      var _getFromSecretStorage = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee10(type, secretStorage) {
        var encodedKey;
        return _regenerator["default"].wrap(function _callee10$(_context10) {
          while (1) {
            switch (_context10.prev = _context10.next) {
              case 0:
                _context10.next = 2;
                return secretStorage.get("m.cross_signing.".concat(type));

              case 2:
                encodedKey = _context10.sent;

                if (encodedKey) {
                  _context10.next = 5;
                  break;
                }

                return _context10.abrupt("return", null);

              case 5:
                return _context10.abrupt("return", (0, _olmlib.decodeBase64)(encodedKey));

              case 6:
              case "end":
                return _context10.stop();
            }
          }
        }, _callee10);
      }));

      function getFromSecretStorage(_x13, _x14) {
        return _getFromSecretStorage.apply(this, arguments);
      }

      return getFromSecretStorage;
    }()
  }]);
  return CrossSigningInfo;
}(_events.EventEmitter);

exports.CrossSigningInfo = CrossSigningInfo;

function deviceToObject(device, userId) {
  return {
    algorithms: device.algorithms,
    keys: device.keys,
    device_id: device.deviceId,
    user_id: userId,
    signatures: device.signatures
  };
}

var CrossSigningLevel = {
  MASTER: 4,
  USER_SIGNING: 2,
  SELF_SIGNING: 1
};
/**
 * Represents the ways in which we trust a user
 */

exports.CrossSigningLevel = CrossSigningLevel;

var UserTrustLevel = /*#__PURE__*/function () {
  function UserTrustLevel(crossSigningVerified, crossSigningVerifiedBefore, tofu) {
    (0, _classCallCheck2["default"])(this, UserTrustLevel);
    this._crossSigningVerified = crossSigningVerified;
    this._crossSigningVerifiedBefore = crossSigningVerifiedBefore;
    this._tofu = tofu;
  }
  /**
   * @returns {bool} true if this user is verified via any means
   */


  (0, _createClass2["default"])(UserTrustLevel, [{
    key: "isVerified",
    value: function isVerified() {
      return this.isCrossSigningVerified();
    }
    /**
     * @returns {bool} true if this user is verified via cross signing
     */

  }, {
    key: "isCrossSigningVerified",
    value: function isCrossSigningVerified() {
      return this._crossSigningVerified;
    }
    /**
     * @returns {bool} true if we ever verified this user before (at least for
     * the history of verifications observed by this device).
     */

  }, {
    key: "wasCrossSigningVerified",
    value: function wasCrossSigningVerified() {
      return this._crossSigningVerifiedBefore;
    }
    /**
     * @returns {bool} true if this user's key is trusted on first use
     */

  }, {
    key: "isTofu",
    value: function isTofu() {
      return this._tofu;
    }
  }]);
  return UserTrustLevel;
}();
/**
 * Represents the ways in which we trust a device
 */


exports.UserTrustLevel = UserTrustLevel;

var DeviceTrustLevel = /*#__PURE__*/function () {
  function DeviceTrustLevel(crossSigningVerified, tofu, localVerified, trustCrossSignedDevices) {
    (0, _classCallCheck2["default"])(this, DeviceTrustLevel);
    this._crossSigningVerified = crossSigningVerified;
    this._tofu = tofu;
    this._localVerified = localVerified;
    this._trustCrossSignedDevices = trustCrossSignedDevices;
  }

  (0, _createClass2["default"])(DeviceTrustLevel, [{
    key: "isVerified",

    /**
     * @returns {bool} true if this device is verified via any means
     */
    value: function isVerified() {
      return Boolean(this.isLocallyVerified() || this._trustCrossSignedDevices && this.isCrossSigningVerified());
    }
    /**
     * @returns {bool} true if this device is verified via cross signing
     */

  }, {
    key: "isCrossSigningVerified",
    value: function isCrossSigningVerified() {
      return this._crossSigningVerified;
    }
    /**
     * @returns {bool} true if this device is verified locally
     */

  }, {
    key: "isLocallyVerified",
    value: function isLocallyVerified() {
      return this._localVerified;
    }
    /**
     * @returns {bool} true if this device is trusted from a user's key
     * that is trusted on first use
     */

  }, {
    key: "isTofu",
    value: function isTofu() {
      return this._tofu;
    }
  }], [{
    key: "fromUserTrustLevel",
    value: function fromUserTrustLevel(userTrustLevel, localVerified, trustCrossSignedDevices) {
      return new DeviceTrustLevel(userTrustLevel._crossSigningVerified, userTrustLevel._tofu, localVerified, trustCrossSignedDevices);
    }
  }]);
  return DeviceTrustLevel;
}();

exports.DeviceTrustLevel = DeviceTrustLevel;

function createCryptoStoreCacheCallbacks(store, olmdevice) {
  return {
    getCrossSigningKeyCache: function () {
      var _getCrossSigningKeyCache = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee11(type, _expectedPublicKey) {
        var key, pickleKey, decrypted;
        return _regenerator["default"].wrap(function _callee11$(_context11) {
          while (1) {
            switch (_context11.prev = _context11.next) {
              case 0:
                _context11.next = 2;
                return new Promise(function (resolve) {
                  return store.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
                    store.getSecretStorePrivateKey(txn, resolve, type);
                  });
                });

              case 2:
                key = _context11.sent;

                if (!(key && key.ciphertext)) {
                  _context11.next = 11;
                  break;
                }

                pickleKey = Buffer.from(olmdevice._pickleKey);
                _context11.next = 7;
                return (0, _aes.decryptAES)(key, pickleKey, type);

              case 7:
                decrypted = _context11.sent;
                return _context11.abrupt("return", (0, _olmlib.decodeBase64)(decrypted));

              case 11:
                return _context11.abrupt("return", key);

              case 12:
              case "end":
                return _context11.stop();
            }
          }
        }, _callee11);
      }));

      function getCrossSigningKeyCache(_x15, _x16) {
        return _getCrossSigningKeyCache.apply(this, arguments);
      }

      return getCrossSigningKeyCache;
    }(),
    storeCrossSigningKeyCache: function () {
      var _storeCrossSigningKeyCache = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee12(type, key) {
        var pickleKey;
        return _regenerator["default"].wrap(function _callee12$(_context12) {
          while (1) {
            switch (_context12.prev = _context12.next) {
              case 0:
                if (key instanceof Uint8Array) {
                  _context12.next = 2;
                  break;
                }

                throw new Error("storeCrossSigningKeyCache expects Uint8Array, got ".concat(key));

              case 2:
                pickleKey = Buffer.from(olmdevice._pickleKey);
                _context12.next = 5;
                return (0, _aes.encryptAES)((0, _olmlib.encodeBase64)(key), pickleKey, type);

              case 5:
                key = _context12.sent;
                return _context12.abrupt("return", store.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
                  store.storeSecretStorePrivateKey(txn, type, key);
                }));

              case 7:
              case "end":
                return _context12.stop();
            }
          }
        }, _callee12);
      }));

      function storeCrossSigningKeyCache(_x17, _x18) {
        return _storeCrossSigningKeyCache.apply(this, arguments);
      }

      return storeCrossSigningKeyCache;
    }()
  };
}
/**
 * Request cross-signing keys from another device during verification.
 *
 * @param {module:base-apis~MatrixBaseApis} baseApis base Matrix API interface
 * @param {string} userId The user ID being verified
 * @param {string} deviceId The device ID being verified
 */


function requestKeysDuringVerification(_x19, _x20, _x21) {
  return _requestKeysDuringVerification.apply(this, arguments);
}

function _requestKeysDuringVerification() {
  _requestKeysDuringVerification = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee15(baseApis, userId, deviceId) {
    return _regenerator["default"].wrap(function _callee15$(_context15) {
      while (1) {
        switch (_context15.prev = _context15.next) {
          case 0:
            if (!(baseApis.getUserId() !== userId)) {
              _context15.next = 2;
              break;
            }

            return _context15.abrupt("return");

          case 2:
            _logger.logger.log("Cross-signing: Self-verification done; requesting keys"); // This happens asynchronously, and we're not concerned about waiting for
            // it.  We return here in order to test.


            return _context15.abrupt("return", new Promise(function (resolve, reject) {
              var client = baseApis;
              var original = client._crypto._crossSigningInfo; // We already have all of the infrastructure we need to validate and
              // cache cross-signing keys, so instead of replicating that, here we set
              // up callbacks that request them from the other device and call
              // CrossSigningInfo.getCrossSigningKey() to validate/cache

              var crossSigning = new CrossSigningInfo(original.userId, {
                getCrossSigningKey: function () {
                  var _getCrossSigningKey2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee13(type) {
                    var _client$requestSecret, promise, result, decoded;

                    return _regenerator["default"].wrap(function _callee13$(_context13) {
                      while (1) {
                        switch (_context13.prev = _context13.next) {
                          case 0:
                            _logger.logger.debug("Cross-signing: requesting secret", type, deviceId);

                            _client$requestSecret = client.requestSecret("m.cross_signing.".concat(type), [deviceId]), promise = _client$requestSecret.promise;
                            _context13.next = 4;
                            return promise;

                          case 4:
                            result = _context13.sent;
                            decoded = (0, _olmlib.decodeBase64)(result);
                            return _context13.abrupt("return", Uint8Array.from(decoded));

                          case 7:
                          case "end":
                            return _context13.stop();
                        }
                      }
                    }, _callee13);
                  }));

                  function getCrossSigningKey(_x22) {
                    return _getCrossSigningKey2.apply(this, arguments);
                  }

                  return getCrossSigningKey;
                }()
              }, original._cacheCallbacks);
              crossSigning.keys = original.keys; // XXX: get all keys out if we get one key out
              // https://github.com/vector-im/element-web/issues/12604
              // then change here to reject on the timeout
              // Requests can be ignored, so don't wait around forever

              var timeout = new Promise(function (resolve, reject) {
                setTimeout(resolve, KEY_REQUEST_TIMEOUT_MS, new Error("Timeout"));
              }); // also request and cache the key backup key

              var backupKeyPromise = new Promise( /*#__PURE__*/function () {
                var _ref = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee14(resolve) {
                  var cachedKey, secretReq, base64Key, decodedKey, backupInfo;
                  return _regenerator["default"].wrap(function _callee14$(_context14) {
                    while (1) {
                      switch (_context14.prev = _context14.next) {
                        case 0:
                          _context14.next = 2;
                          return client._crypto.getSessionBackupPrivateKey();

                        case 2:
                          cachedKey = _context14.sent;

                          if (cachedKey) {
                            _context14.next = 18;
                            break;
                          }

                          _logger.logger.info("No cached backup key found. Requesting...");

                          secretReq = client.requestSecret('m.megolm_backup.v1', [deviceId]);
                          _context14.next = 8;
                          return secretReq.promise;

                        case 8:
                          base64Key = _context14.sent;

                          _logger.logger.info("Got key backup key, decoding...");

                          decodedKey = (0, _olmlib.decodeBase64)(base64Key);

                          _logger.logger.info("Decoded backup key, storing...");

                          client._crypto.storeSessionBackupPrivateKey(Uint8Array.from(decodedKey));

                          _logger.logger.info("Backup key stored. Starting backup restore...");

                          _context14.next = 16;
                          return client.getKeyBackupVersion();

                        case 16:
                          backupInfo = _context14.sent;
                          // no need to await for this - just let it go in the bg
                          client.restoreKeyBackupWithCache(undefined, undefined, backupInfo).then(function () {
                            _logger.logger.info("Backup restored.");
                          });

                        case 18:
                          resolve();

                        case 19:
                        case "end":
                          return _context14.stop();
                      }
                    }
                  }, _callee14);
                }));

                return function (_x23) {
                  return _ref.apply(this, arguments);
                };
              }()); // We call getCrossSigningKey() for its side-effects

              return Promise.race([Promise.all([crossSigning.getCrossSigningKey("master"), crossSigning.getCrossSigningKey("self_signing"), crossSigning.getCrossSigningKey("user_signing"), backupKeyPromise]), timeout]).then(resolve, reject);
            })["catch"](function (e) {
              _logger.logger.warn("Cross-signing: failure while requesting keys:", e);
            }));

          case 4:
          case "end":
            return _context15.stop();
        }
      }
    }, _callee15);
  }));
  return _requestKeysDuringVerification.apply(this, arguments);
}