"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.DeviceList = void 0;

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _assertThisInitialized2 = _interopRequireDefault(require("@babel/runtime/helpers/assertThisInitialized"));

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _events = require("events");

var _logger = require("../logger");

var _deviceinfo = require("./deviceinfo");

var _CrossSigning = require("./CrossSigning");

var olmlib = _interopRequireWildcard(require("./olmlib"));

var _indexeddbCryptoStore = require("./store/indexeddb-crypto-store");

var _utils = require("../utils");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

/* State transition diagram for DeviceList._deviceTrackingStatus
 *
 *                                |
 *     stopTrackingDeviceList     V
 *   +---------------------> NOT_TRACKED
 *   |                            |
 *   +<--------------------+      | startTrackingDeviceList
 *   |                     |      V
 *   |   +-------------> PENDING_DOWNLOAD <--------------------+-+
 *   |   |                      ^ |                            | |
 *   |   | restart     download | |  start download            | | invalidateUserDeviceList
 *   |   | client        failed | |                            | |
 *   |   |                      | V                            | |
 *   |   +------------ DOWNLOAD_IN_PROGRESS -------------------+ |
 *   |                    |       |                              |
 *   +<-------------------+       |  download successful         |
 *   ^                            V                              |
 *   +----------------------- UP_TO_DATE ------------------------+
 */
// constants for DeviceList._deviceTrackingStatus
var TRACKING_STATUS_NOT_TRACKED = 0;
var TRACKING_STATUS_PENDING_DOWNLOAD = 1;
var TRACKING_STATUS_DOWNLOAD_IN_PROGRESS = 2;
var TRACKING_STATUS_UP_TO_DATE = 3;
/**
 * @alias module:crypto/DeviceList
 */

var DeviceList = /*#__PURE__*/function (_EventEmitter) {
  (0, _inherits2["default"])(DeviceList, _EventEmitter);

  var _super = _createSuper(DeviceList);

  function DeviceList(baseApis, cryptoStore, olmDevice) {
    var _this;

    (0, _classCallCheck2["default"])(this, DeviceList);
    _this = _super.call(this);
    _this._cryptoStore = cryptoStore; // userId -> {
    //     deviceId -> {
    //         [device info]
    //     }
    // }

    _this._devices = {}; // userId -> {
    //     [key info]
    // }

    _this._crossSigningInfo = {}; // map of identity keys to the user who owns it

    _this._userByIdentityKey = {}; // which users we are tracking device status for.
    // userId -> TRACKING_STATUS_*

    _this._deviceTrackingStatus = {}; // loaded from storage in load()
    // The 'next_batch' sync token at the point the data was writen,
    // ie. a token representing the point immediately after the
    // moment represented by the snapshot in the db.

    _this._syncToken = null;
    _this._serialiser = new DeviceListUpdateSerialiser(baseApis, olmDevice, (0, _assertThisInitialized2["default"])(_this)); // userId -> promise

    _this._keyDownloadsInProgressByUser = {}; // Set whenever changes are made other than setting the sync token

    _this._dirty = false; // Promise resolved when device data is saved

    _this._savePromise = null; // Function that resolves the save promise

    _this._resolveSavePromise = null; // The time the save is scheduled for

    _this._savePromiseTime = null; // The timer used to delay the save

    _this._saveTimer = null; // True if we have fetched data from the server or loaded a non-empty
    // set of device data from the store

    _this._hasFetched = null;
    return _this;
  }
  /**
   * Load the device tracking state from storage
   */


  (0, _createClass2["default"])(DeviceList, [{
    key: "load",
    value: function () {
      var _load = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee() {
        var _this2 = this;

        var _i3, _Object$keys3, u;

        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_DEVICE_DATA], function (txn) {
                  _this2._cryptoStore.getEndToEndDeviceData(txn, function (deviceData) {
                    _this2._hasFetched = Boolean(deviceData && deviceData.devices);
                    _this2._devices = deviceData ? deviceData.devices : {}, _this2._crossSigningInfo = deviceData ? deviceData.crossSigningInfo || {} : {};
                    _this2._deviceTrackingStatus = deviceData ? deviceData.trackingStatus : {};
                    _this2._syncToken = deviceData ? deviceData.syncToken : null;
                    _this2._userByIdentityKey = {};

                    for (var _i = 0, _Object$keys = Object.keys(_this2._devices); _i < _Object$keys.length; _i++) {
                      var user = _Object$keys[_i];
                      var userDevices = _this2._devices[user];

                      for (var _i2 = 0, _Object$keys2 = Object.keys(userDevices); _i2 < _Object$keys2.length; _i2++) {
                        var device = _Object$keys2[_i2];
                        var idKey = userDevices[device].keys['curve25519:' + device];

                        if (idKey !== undefined) {
                          _this2._userByIdentityKey[idKey] = user;
                        }
                      }
                    }
                  });
                });

              case 2:
                for (_i3 = 0, _Object$keys3 = Object.keys(this._deviceTrackingStatus); _i3 < _Object$keys3.length; _i3++) {
                  u = _Object$keys3[_i3];

                  // if a download was in progress when we got shut down, it isn't any more.
                  if (this._deviceTrackingStatus[u] == TRACKING_STATUS_DOWNLOAD_IN_PROGRESS) {
                    this._deviceTrackingStatus[u] = TRACKING_STATUS_PENDING_DOWNLOAD;
                  }
                }

              case 3:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function load() {
        return _load.apply(this, arguments);
      }

      return load;
    }()
  }, {
    key: "stop",
    value: function stop() {
      if (this._saveTimer !== null) {
        clearTimeout(this._saveTimer);
      }
    }
    /**
     * Save the device tracking state to storage, if any changes are
     * pending other than updating the sync token
     *
     * The actual save will be delayed by a short amount of time to
     * aggregate multiple writes to the database.
     *
     * @param {integer} delay Time in ms before which the save actually happens.
     *     By default, the save is delayed for a short period in order to batch
     *     multiple writes, but this behaviour can be disabled by passing 0.
     *
     * @return {Promise<bool>} true if the data was saved, false if
     *     it was not (eg. because no changes were pending). The promise
     *     will only resolve once the data is saved, so may take some time
     *     to resolve.
     */

  }, {
    key: "saveIfDirty",
    value: function () {
      var _saveIfDirty = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(delay) {
        var _this3 = this;

        var targetTime, savePromise, resolveSavePromise;
        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                if (this._dirty) {
                  _context2.next = 2;
                  break;
                }

                return _context2.abrupt("return", Promise.resolve(false));

              case 2:
                // Delay saves for a bit so we can aggregate multiple saves that happen
                // in quick succession (eg. when a whole room's devices are marked as known)
                if (delay === undefined) delay = 500;
                targetTime = Date.now + delay;

                if (this._savePromiseTime && targetTime < this._savePromiseTime) {
                  // There's a save scheduled but for after we would like: cancel
                  // it & schedule one for the time we want
                  clearTimeout(this._saveTimer);
                  this._saveTimer = null;
                  this._savePromiseTime = null; // (but keep the save promise since whatever called save before
                  // will still want to know when the save is done)
                }

                savePromise = this._savePromise;

                if (savePromise === null) {
                  savePromise = new Promise(function (resolve, reject) {
                    _this3._resolveSavePromise = resolve;
                  });
                  this._savePromise = savePromise;
                }

                if (this._saveTimer === null) {
                  resolveSavePromise = this._resolveSavePromise;
                  this._savePromiseTime = targetTime;
                  this._saveTimer = setTimeout(function () {
                    _logger.logger.log('Saving device tracking data', _this3._syncToken); // null out savePromise now (after the delay but before the write),
                    // otherwise we could return the existing promise when the save has
                    // actually already happened.


                    _this3._savePromiseTime = null;
                    _this3._saveTimer = null;
                    _this3._savePromise = null;
                    _this3._resolveSavePromise = null;

                    _this3._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_DEVICE_DATA], function (txn) {
                      _this3._cryptoStore.storeEndToEndDeviceData({
                        devices: _this3._devices,
                        crossSigningInfo: _this3._crossSigningInfo,
                        trackingStatus: _this3._deviceTrackingStatus,
                        syncToken: _this3._syncToken
                      }, txn);
                    }).then(function () {
                      // The device list is considered dirty until the write
                      // completes.
                      _this3._dirty = false;
                      resolveSavePromise();
                    }, function (err) {
                      _logger.logger.error('Failed to save device tracking data', _this3._syncToken);

                      _logger.logger.error(err);
                    });
                  }, delay);
                }

                return _context2.abrupt("return", savePromise);

              case 9:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function saveIfDirty(_x) {
        return _saveIfDirty.apply(this, arguments);
      }

      return saveIfDirty;
    }()
    /**
     * Gets the sync token last set with setSyncToken
     *
     * @return {string} The sync token
     */

  }, {
    key: "getSyncToken",
    value: function getSyncToken() {
      return this._syncToken;
    }
    /**
     * Sets the sync token that the app will pass as the 'since' to the /sync
     * endpoint next time it syncs.
     * The sync token must always be set after any changes made as a result of
     * data in that sync since setting the sync token to a newer one will mean
     * those changed will not be synced from the server if a new client starts
     * up with that data.
     *
     * @param {string} st The sync token
     */

  }, {
    key: "setSyncToken",
    value: function setSyncToken(st) {
      this._syncToken = st;
    }
    /**
     * Ensures up to date keys for a list of users are stored in the session store,
     * downloading and storing them if they're not (or if forceDownload is
     * true).
     * @param {Array} userIds The users to fetch.
     * @param {bool} forceDownload Always download the keys even if cached.
     *
     * @return {Promise} A promise which resolves to a map userId->deviceId->{@link
     * module:crypto/deviceinfo|DeviceInfo}.
     */

  }, {
    key: "downloadKeys",
    value: function downloadKeys(userIds, forceDownload) {
      var _this4 = this;

      var usersToDownload = [];
      var promises = [];
      userIds.forEach(function (u) {
        var trackingStatus = _this4._deviceTrackingStatus[u];

        if (_this4._keyDownloadsInProgressByUser[u]) {
          // already a key download in progress/queued for this user; its results
          // will be good enough for us.
          _logger.logger.log("downloadKeys: already have a download in progress for " + "".concat(u, ": awaiting its result"));

          promises.push(_this4._keyDownloadsInProgressByUser[u]);
        } else if (forceDownload || trackingStatus != TRACKING_STATUS_UP_TO_DATE) {
          usersToDownload.push(u);
        }
      });

      if (usersToDownload.length != 0) {
        _logger.logger.log("downloadKeys: downloading for", usersToDownload);

        var downloadPromise = this._doKeyDownload(usersToDownload);

        promises.push(downloadPromise);
      }

      if (promises.length === 0) {
        _logger.logger.log("downloadKeys: already have all necessary keys");
      }

      return Promise.all(promises).then(function () {
        return _this4._getDevicesFromStore(userIds);
      });
    }
    /**
     * Get the stored device keys for a list of user ids
     *
     * @param {string[]} userIds the list of users to list keys for.
     *
     * @return {Object} userId->deviceId->{@link module:crypto/deviceinfo|DeviceInfo}.
     */

  }, {
    key: "_getDevicesFromStore",
    value: function _getDevicesFromStore(userIds) {
      var stored = {};
      var self = this;
      userIds.map(function (u) {
        stored[u] = {};
        var devices = self.getStoredDevicesForUser(u) || [];
        devices.map(function (dev) {
          stored[u][dev.deviceId] = dev;
        });
      });
      return stored;
    }
    /**
     * Returns a list of all user IDs the DeviceList knows about
     *
     * @return {array} All known user IDs
     */

  }, {
    key: "getKnownUserIds",
    value: function getKnownUserIds() {
      return Object.keys(this._devices);
    }
    /**
     * Get the stored device keys for a user id
     *
     * @param {string} userId the user to list keys for.
     *
     * @return {module:crypto/deviceinfo[]|null} list of devices, or null if we haven't
     * managed to get a list of devices for this user yet.
     */

  }, {
    key: "getStoredDevicesForUser",
    value: function getStoredDevicesForUser(userId) {
      var devs = this._devices[userId];

      if (!devs) {
        return null;
      }

      var res = [];

      for (var deviceId in devs) {
        if (devs.hasOwnProperty(deviceId)) {
          res.push(_deviceinfo.DeviceInfo.fromStorage(devs[deviceId], deviceId));
        }
      }

      return res;
    }
    /**
     * Get the stored device data for a user, in raw object form
     *
     * @param {string} userId the user to get data for
     *
     * @return {Object} deviceId->{object} devices, or undefined if
     * there is no data for this user.
     */

  }, {
    key: "getRawStoredDevicesForUser",
    value: function getRawStoredDevicesForUser(userId) {
      return this._devices[userId];
    }
  }, {
    key: "getStoredCrossSigningForUser",
    value: function getStoredCrossSigningForUser(userId) {
      if (!this._crossSigningInfo[userId]) return null;
      return _CrossSigning.CrossSigningInfo.fromStorage(this._crossSigningInfo[userId], userId);
    }
  }, {
    key: "storeCrossSigningForUser",
    value: function storeCrossSigningForUser(userId, info) {
      this._crossSigningInfo[userId] = info;
      this._dirty = true;
    }
    /**
     * Get the stored keys for a single device
     *
     * @param {string} userId
     * @param {string} deviceId
     *
     * @return {module:crypto/deviceinfo?} device, or undefined
     * if we don't know about this device
     */

  }, {
    key: "getStoredDevice",
    value: function getStoredDevice(userId, deviceId) {
      var devs = this._devices[userId];

      if (!devs || !devs[deviceId]) {
        return undefined;
      }

      return _deviceinfo.DeviceInfo.fromStorage(devs[deviceId], deviceId);
    }
    /**
     * Get a user ID by one of their device's curve25519 identity key
     *
     * @param {string} algorithm  encryption algorithm
     * @param {string} senderKey  curve25519 key to match
     *
     * @return {string} user ID
     */

  }, {
    key: "getUserByIdentityKey",
    value: function getUserByIdentityKey(algorithm, senderKey) {
      if (algorithm !== olmlib.OLM_ALGORITHM && algorithm !== olmlib.MEGOLM_ALGORITHM) {
        // we only deal in olm keys
        return null;
      }

      return this._userByIdentityKey[senderKey];
    }
    /**
     * Find a device by curve25519 identity key
     *
     * @param {string} algorithm  encryption algorithm
     * @param {string} senderKey  curve25519 key to match
     *
     * @return {module:crypto/deviceinfo?}
     */

  }, {
    key: "getDeviceByIdentityKey",
    value: function getDeviceByIdentityKey(algorithm, senderKey) {
      var userId = this.getUserByIdentityKey(algorithm, senderKey);

      if (!userId) {
        return null;
      }

      var devices = this._devices[userId];

      if (!devices) {
        return null;
      }

      for (var deviceId in devices) {
        if (!devices.hasOwnProperty(deviceId)) {
          continue;
        }

        var device = devices[deviceId];

        for (var keyId in device.keys) {
          if (!device.keys.hasOwnProperty(keyId)) {
            continue;
          }

          if (keyId.indexOf("curve25519:") !== 0) {
            continue;
          }

          var deviceKey = device.keys[keyId];

          if (deviceKey == senderKey) {
            return _deviceinfo.DeviceInfo.fromStorage(device, deviceId);
          }
        }
      } // doesn't match a known device


      return null;
    }
    /**
     * Replaces the list of devices for a user with the given device list
     *
     * @param {string} u The user ID
     * @param {Object} devs New device info for user
     */

  }, {
    key: "storeDevicesForUser",
    value: function storeDevicesForUser(u, devs) {
      // remove previous devices from _userByIdentityKey
      if (this._devices[u] !== undefined) {
        for (var _i4 = 0, _Object$entries = Object.entries(this._devices[u]); _i4 < _Object$entries.length; _i4++) {
          var _Object$entries$_i = (0, _slicedToArray2["default"])(_Object$entries[_i4], 2),
              deviceId = _Object$entries$_i[0],
              dev = _Object$entries$_i[1];

          var identityKey = dev.keys['curve25519:' + deviceId];
          delete this._userByIdentityKey[identityKey];
        }
      }

      this._devices[u] = devs; // add new ones

      for (var _i5 = 0, _Object$entries2 = Object.entries(devs); _i5 < _Object$entries2.length; _i5++) {
        var _Object$entries2$_i = (0, _slicedToArray2["default"])(_Object$entries2[_i5], 2),
            _deviceId = _Object$entries2$_i[0],
            _dev = _Object$entries2$_i[1];

        var _identityKey = _dev.keys['curve25519:' + _deviceId];
        this._userByIdentityKey[_identityKey] = u;
      }

      this._dirty = true;
    }
    /**
     * flag the given user for device-list tracking, if they are not already.
     *
     * This will mean that a subsequent call to refreshOutdatedDeviceLists()
     * will download the device list for the user, and that subsequent calls to
     * invalidateUserDeviceList will trigger more updates.
     *
     * @param {String} userId
     */

  }, {
    key: "startTrackingDeviceList",
    value: function startTrackingDeviceList(userId) {
      // sanity-check the userId. This is mostly paranoia, but if synapse
      // can't parse the userId we give it as an mxid, it 500s the whole
      // request and we can never update the device lists again (because
      // the broken userId is always 'invalid' and always included in any
      // refresh request).
      // By checking it is at least a string, we can eliminate a class of
      // silly errors.
      if (typeof userId !== 'string') {
        throw new Error('userId must be a string; was ' + userId);
      }

      if (!this._deviceTrackingStatus[userId]) {
        _logger.logger.log('Now tracking device list for ' + userId);

        this._deviceTrackingStatus[userId] = TRACKING_STATUS_PENDING_DOWNLOAD; // we don't yet persist the tracking status, since there may be a lot
        // of calls; we save all data together once the sync is done

        this._dirty = true;
      }
    }
    /**
     * Mark the given user as no longer being tracked for device-list updates.
     *
     * This won't affect any in-progress downloads, which will still go on to
     * complete; it will just mean that we don't think that we have an up-to-date
     * list for future calls to downloadKeys.
     *
     * @param {String} userId
     */

  }, {
    key: "stopTrackingDeviceList",
    value: function stopTrackingDeviceList(userId) {
      if (this._deviceTrackingStatus[userId]) {
        _logger.logger.log('No longer tracking device list for ' + userId);

        this._deviceTrackingStatus[userId] = TRACKING_STATUS_NOT_TRACKED; // we don't yet persist the tracking status, since there may be a lot
        // of calls; we save all data together once the sync is done

        this._dirty = true;
      }
    }
    /**
     * Set all users we're currently tracking to untracked
     *
     * This will flag each user whose devices we are tracking as in need of an
     * update.
     */

  }, {
    key: "stopTrackingAllDeviceLists",
    value: function stopTrackingAllDeviceLists() {
      for (var _i6 = 0, _Object$keys4 = Object.keys(this._deviceTrackingStatus); _i6 < _Object$keys4.length; _i6++) {
        var userId = _Object$keys4[_i6];
        this._deviceTrackingStatus[userId] = TRACKING_STATUS_NOT_TRACKED;
      }

      this._dirty = true;
    }
    /**
     * Mark the cached device list for the given user outdated.
     *
     * If we are not tracking this user's devices, we'll do nothing. Otherwise
     * we flag the user as needing an update.
     *
     * This doesn't actually set off an update, so that several users can be
     * batched together. Call refreshOutdatedDeviceLists() for that.
     *
     * @param {String} userId
     */

  }, {
    key: "invalidateUserDeviceList",
    value: function invalidateUserDeviceList(userId) {
      if (this._deviceTrackingStatus[userId]) {
        _logger.logger.log("Marking device list outdated for", userId);

        this._deviceTrackingStatus[userId] = TRACKING_STATUS_PENDING_DOWNLOAD; // we don't yet persist the tracking status, since there may be a lot
        // of calls; we save all data together once the sync is done

        this._dirty = true;
      }
    }
    /**
     * If we have users who have outdated device lists, start key downloads for them
     *
     * @returns {Promise} which completes when the download completes; normally there
     *    is no need to wait for this (it's mostly for the unit tests).
     */

  }, {
    key: "refreshOutdatedDeviceLists",
    value: function refreshOutdatedDeviceLists() {
      this.saveIfDirty();
      var usersToDownload = [];

      for (var _i7 = 0, _Object$keys5 = Object.keys(this._deviceTrackingStatus); _i7 < _Object$keys5.length; _i7++) {
        var userId = _Object$keys5[_i7];
        var stat = this._deviceTrackingStatus[userId];

        if (stat == TRACKING_STATUS_PENDING_DOWNLOAD) {
          usersToDownload.push(userId);
        }
      }

      return this._doKeyDownload(usersToDownload);
    }
    /**
     * Set the stored device data for a user, in raw object form
     * Used only by internal class DeviceListUpdateSerialiser
     *
     * @param {string} userId the user to get data for
     *
     * @param {Object} devices deviceId->{object} the new devices
     */

  }, {
    key: "_setRawStoredDevicesForUser",
    value: function _setRawStoredDevicesForUser(userId, devices) {
      // remove old devices from _userByIdentityKey
      if (this._devices[userId] !== undefined) {
        for (var _i8 = 0, _Object$entries3 = Object.entries(this._devices[userId]); _i8 < _Object$entries3.length; _i8++) {
          var _Object$entries3$_i = (0, _slicedToArray2["default"])(_Object$entries3[_i8], 2),
              deviceId = _Object$entries3$_i[0],
              dev = _Object$entries3$_i[1];

          var identityKey = dev.keys['curve25519:' + deviceId];
          delete this._userByIdentityKey[identityKey];
        }
      }

      this._devices[userId] = devices; // add new devices into _userByIdentityKey

      for (var _i9 = 0, _Object$entries4 = Object.entries(devices); _i9 < _Object$entries4.length; _i9++) {
        var _Object$entries4$_i = (0, _slicedToArray2["default"])(_Object$entries4[_i9], 2),
            _deviceId2 = _Object$entries4$_i[0],
            _dev2 = _Object$entries4$_i[1];

        var _identityKey2 = _dev2.keys['curve25519:' + _deviceId2];
        this._userByIdentityKey[_identityKey2] = userId;
      }
    }
  }, {
    key: "setRawStoredCrossSigningForUser",
    value: function setRawStoredCrossSigningForUser(userId, info) {
      this._crossSigningInfo[userId] = info;
    }
    /**
     * Fire off download update requests for the given users, and update the
     * device list tracking status for them, and the
     * _keyDownloadsInProgressByUser map for them.
     *
     * @param {String[]} users  list of userIds
     *
     * @return {Promise} resolves when all the users listed have
     *     been updated. rejects if there was a problem updating any of the
     *     users.
     */

  }, {
    key: "_doKeyDownload",
    value: function _doKeyDownload(users) {
      var _this5 = this;

      if (users.length === 0) {
        // nothing to do
        return Promise.resolve();
      }

      var prom = this._serialiser.updateDevicesForUsers(users, this._syncToken).then(function () {
        finished(true);
      }, function (e) {
        _logger.logger.error('Error downloading keys for ' + users + ":", e);

        finished(false);
        throw e;
      });

      users.forEach(function (u) {
        _this5._keyDownloadsInProgressByUser[u] = prom;
        var stat = _this5._deviceTrackingStatus[u];

        if (stat == TRACKING_STATUS_PENDING_DOWNLOAD) {
          _this5._deviceTrackingStatus[u] = TRACKING_STATUS_DOWNLOAD_IN_PROGRESS;
        }
      });

      var finished = function finished(success) {
        _this5.emit("crypto.willUpdateDevices", users, !_this5._hasFetched);

        users.forEach(function (u) {
          _this5._dirty = true; // we may have queued up another download request for this user
          // since we started this request. If that happens, we should
          // ignore the completion of the first one.

          if (_this5._keyDownloadsInProgressByUser[u] !== prom) {
            _logger.logger.log('Another update in the queue for', u, '- not marking up-to-date');

            return;
          }

          delete _this5._keyDownloadsInProgressByUser[u];
          var stat = _this5._deviceTrackingStatus[u];

          if (stat == TRACKING_STATUS_DOWNLOAD_IN_PROGRESS) {
            if (success) {
              // we didn't get any new invalidations since this download started:
              // this user's device list is now up to date.
              _this5._deviceTrackingStatus[u] = TRACKING_STATUS_UP_TO_DATE;

              _logger.logger.log("Device list for", u, "now up to date");
            } else {
              _this5._deviceTrackingStatus[u] = TRACKING_STATUS_PENDING_DOWNLOAD;
            }
          }
        });

        _this5.saveIfDirty();

        _this5.emit("crypto.devicesUpdated", users, !_this5._hasFetched);

        _this5._hasFetched = true;
      };

      return prom;
    }
  }]);
  return DeviceList;
}(_events.EventEmitter);
/**
 * Serialises updates to device lists
 *
 * Ensures that results from /keys/query are not overwritten if a second call
 * completes *before* an earlier one.
 *
 * It currently does this by ensuring only one call to /keys/query happens at a
 * time (and queuing other requests up).
 */


exports.DeviceList = DeviceList;

var DeviceListUpdateSerialiser = /*#__PURE__*/function () {
  /*
   * @param {object} baseApis Base API object
   * @param {object} olmDevice The Olm Device
   * @param {object} deviceList The device list object
   */
  function DeviceListUpdateSerialiser(baseApis, olmDevice, deviceList) {
    (0, _classCallCheck2["default"])(this, DeviceListUpdateSerialiser);
    this._baseApis = baseApis;
    this._olmDevice = olmDevice;
    this._deviceList = deviceList; // the device list to be updated

    this._downloadInProgress = false; // users which are queued for download
    // userId -> true

    this._keyDownloadsQueuedByUser = {}; // deferred which is resolved when the queued users are downloaded.
    //
    // non-null indicates that we have users queued for download.

    this._queuedQueryDeferred = null;
    this._syncToken = null; // The sync token we send with the requests
  }
  /**
   * Make a key query request for the given users
   *
   * @param {String[]} users list of user ids
   *
   * @param {String} syncToken sync token to pass in the query request, to
   *     help the HS give the most recent results
   *
   * @return {Promise} resolves when all the users listed have
   *     been updated. rejects if there was a problem updating any of the
   *     users.
   */


  (0, _createClass2["default"])(DeviceListUpdateSerialiser, [{
    key: "updateDevicesForUsers",
    value: function updateDevicesForUsers(users, syncToken) {
      var _this6 = this;

      users.forEach(function (u) {
        _this6._keyDownloadsQueuedByUser[u] = true;
      });

      if (!this._queuedQueryDeferred) {
        this._queuedQueryDeferred = (0, _utils.defer)();
      } // We always take the new sync token and just use the latest one we've
      // been given, since it just needs to be at least as recent as the
      // sync response the device invalidation message arrived in


      this._syncToken = syncToken;

      if (this._downloadInProgress) {
        // just queue up these users
        _logger.logger.log('Queued key download for', users);

        return this._queuedQueryDeferred.promise;
      } // start a new download.


      return this._doQueuedQueries();
    }
  }, {
    key: "_doQueuedQueries",
    value: function _doQueuedQueries() {
      var _this7 = this;

      if (this._downloadInProgress) {
        throw new Error("DeviceListUpdateSerialiser._doQueuedQueries called with request active");
      }

      var downloadUsers = Object.keys(this._keyDownloadsQueuedByUser);
      this._keyDownloadsQueuedByUser = {};
      var deferred = this._queuedQueryDeferred;
      this._queuedQueryDeferred = null;

      _logger.logger.log('Starting key download for', downloadUsers);

      this._downloadInProgress = true;
      var opts = {};

      if (this._syncToken) {
        opts.token = this._syncToken;
      }

      this._baseApis.downloadKeysForUsers(downloadUsers, opts).then( /*#__PURE__*/function () {
        var _ref = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(res) {
          var dk, masterKeys, ssks, usks, _iterator, _step, userId;

          return _regenerator["default"].wrap(function _callee3$(_context3) {
            while (1) {
              switch (_context3.prev = _context3.next) {
                case 0:
                  dk = res.device_keys || {};
                  masterKeys = res.master_keys || {};
                  ssks = res.self_signing_keys || {};
                  usks = res.user_signing_keys || {}; // yield to other things that want to execute in between users, to
                  // avoid wedging the CPU
                  // (https://github.com/vector-im/element-web/issues/3158)
                  //
                  // of course we ought to do this in a web worker or similar, but
                  // this serves as an easy solution for now.

                  _iterator = _createForOfIteratorHelper(downloadUsers);
                  _context3.prev = 5;

                  _iterator.s();

                case 7:
                  if ((_step = _iterator.n()).done) {
                    _context3.next = 21;
                    break;
                  }

                  userId = _step.value;
                  _context3.next = 11;
                  return (0, _utils.sleep)(5);

                case 11:
                  _context3.prev = 11;
                  _context3.next = 14;
                  return _this7._processQueryResponseForUser(userId, dk[userId], {
                    master: masterKeys[userId],
                    self_signing: ssks[userId],
                    user_signing: usks[userId]
                  });

                case 14:
                  _context3.next = 19;
                  break;

                case 16:
                  _context3.prev = 16;
                  _context3.t0 = _context3["catch"](11);

                  // log the error but continue, so that one bad key
                  // doesn't kill the whole process
                  _logger.logger.error("Error processing keys for ".concat(userId, ":"), _context3.t0);

                case 19:
                  _context3.next = 7;
                  break;

                case 21:
                  _context3.next = 26;
                  break;

                case 23:
                  _context3.prev = 23;
                  _context3.t1 = _context3["catch"](5);

                  _iterator.e(_context3.t1);

                case 26:
                  _context3.prev = 26;

                  _iterator.f();

                  return _context3.finish(26);

                case 29:
                case "end":
                  return _context3.stop();
              }
            }
          }, _callee3, null, [[5, 23, 26, 29], [11, 16]]);
        }));

        return function (_x2) {
          return _ref.apply(this, arguments);
        };
      }()).then(function () {
        _logger.logger.log('Completed key download for ' + downloadUsers);

        _this7._downloadInProgress = false;
        deferred.resolve(); // if we have queued users, fire off another request.

        if (_this7._queuedQueryDeferred) {
          _this7._doQueuedQueries();
        }
      }, function (e) {
        _logger.logger.warn('Error downloading keys for ' + downloadUsers + ':', e);

        _this7._downloadInProgress = false;
        deferred.reject(e);
      });

      return deferred.promise;
    }
  }, {
    key: "_processQueryResponseForUser",
    value: function () {
      var _processQueryResponseForUser2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4(userId, dkResponse, crossSigningResponse) {
        var userStore, devs, storage, crossSigning;
        return _regenerator["default"].wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                _logger.logger.log('got device keys for ' + userId + ':', dkResponse);

                _logger.logger.log('got cross-signing keys for ' + userId + ':', crossSigningResponse);

                // map from deviceid -> deviceinfo for this user
                userStore = {};
                devs = this._deviceList.getRawStoredDevicesForUser(userId);

                if (devs) {
                  Object.keys(devs).forEach(function (deviceId) {
                    var d = _deviceinfo.DeviceInfo.fromStorage(devs[deviceId], deviceId);

                    userStore[deviceId] = d;
                  });
                }

                _context4.next = 7;
                return _updateStoredDeviceKeysForUser(this._olmDevice, userId, userStore, dkResponse || {}, this._baseApis.getUserId(), this._baseApis.deviceId);

              case 7:
                // put the updates into the object that will be returned as our results
                storage = {};
                Object.keys(userStore).forEach(function (deviceId) {
                  storage[deviceId] = userStore[deviceId].toStorage();
                });

                this._deviceList._setRawStoredDevicesForUser(userId, storage);

                // FIXME: should we be ignoring empty cross-signing responses, or
                // should we be dropping the keys?
                if (crossSigningResponse && (crossSigningResponse.master || crossSigningResponse.self_signing || crossSigningResponse.user_signing)) {
                  crossSigning = this._deviceList.getStoredCrossSigningForUser(userId) || new _CrossSigning.CrossSigningInfo(userId);
                  crossSigning.setKeys(crossSigningResponse);

                  this._deviceList.setRawStoredCrossSigningForUser(userId, crossSigning.toStorage()); // NB. Unlike most events in the js-sdk, this one is internal to the
                  // js-sdk and is not re-emitted


                  this._deviceList.emit('userCrossSigningUpdated', userId);
                }

              case 11:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function _processQueryResponseForUser(_x3, _x4, _x5) {
        return _processQueryResponseForUser2.apply(this, arguments);
      }

      return _processQueryResponseForUser;
    }()
  }]);
  return DeviceListUpdateSerialiser;
}();

function _updateStoredDeviceKeysForUser(_x6, _x7, _x8, _x9, _x10, _x11) {
  return _updateStoredDeviceKeysForUser2.apply(this, arguments);
}
/*
 * Process a device in a /query response, and add it to the userStore
 *
 * returns (a promise for) true if a change was made, else false
 */


function _updateStoredDeviceKeysForUser2() {
  _updateStoredDeviceKeysForUser2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5(_olmDevice, userId, userStore, userResult, localUserId, localDeviceId) {
    var updated, deviceId, _deviceId3, deviceResult;

    return _regenerator["default"].wrap(function _callee5$(_context5) {
      while (1) {
        switch (_context5.prev = _context5.next) {
          case 0:
            updated = false; // remove any devices in the store which aren't in the response

            _context5.t0 = _regenerator["default"].keys(userStore);

          case 2:
            if ((_context5.t1 = _context5.t0()).done) {
              _context5.next = 15;
              break;
            }

            deviceId = _context5.t1.value;

            if (userStore.hasOwnProperty(deviceId)) {
              _context5.next = 6;
              break;
            }

            return _context5.abrupt("continue", 2);

          case 6:
            if (deviceId in userResult) {
              _context5.next = 13;
              break;
            }

            if (!(userId === localUserId && deviceId === localDeviceId)) {
              _context5.next = 10;
              break;
            }

            _logger.logger.warn("Local device ".concat(deviceId, " missing from sync, skipping removal"));

            return _context5.abrupt("continue", 2);

          case 10:
            _logger.logger.log("Device " + userId + ":" + deviceId + " has been removed");

            delete userStore[deviceId];
            updated = true;

          case 13:
            _context5.next = 2;
            break;

          case 15:
            _context5.t2 = _regenerator["default"].keys(userResult);

          case 16:
            if ((_context5.t3 = _context5.t2()).done) {
              _context5.next = 33;
              break;
            }

            _deviceId3 = _context5.t3.value;

            if (userResult.hasOwnProperty(_deviceId3)) {
              _context5.next = 20;
              break;
            }

            return _context5.abrupt("continue", 16);

          case 20:
            deviceResult = userResult[_deviceId3]; // check that the user_id and device_id in the response object are
            // correct

            if (!(deviceResult.user_id !== userId)) {
              _context5.next = 24;
              break;
            }

            _logger.logger.warn("Mismatched user_id " + deviceResult.user_id + " in keys from " + userId + ":" + _deviceId3);

            return _context5.abrupt("continue", 16);

          case 24:
            if (!(deviceResult.device_id !== _deviceId3)) {
              _context5.next = 27;
              break;
            }

            _logger.logger.warn("Mismatched device_id " + deviceResult.device_id + " in keys from " + userId + ":" + _deviceId3);

            return _context5.abrupt("continue", 16);

          case 27:
            _context5.next = 29;
            return _storeDeviceKeys(_olmDevice, userStore, deviceResult);

          case 29:
            if (!_context5.sent) {
              _context5.next = 31;
              break;
            }

            updated = true;

          case 31:
            _context5.next = 16;
            break;

          case 33:
            return _context5.abrupt("return", updated);

          case 34:
          case "end":
            return _context5.stop();
        }
      }
    }, _callee5);
  }));
  return _updateStoredDeviceKeysForUser2.apply(this, arguments);
}

function _storeDeviceKeys(_x12, _x13, _x14) {
  return _storeDeviceKeys2.apply(this, arguments);
}

function _storeDeviceKeys2() {
  _storeDeviceKeys2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6(_olmDevice, userStore, deviceResult) {
    var deviceId, userId, signKeyId, signKey, unsigned, signatures, deviceStore;
    return _regenerator["default"].wrap(function _callee6$(_context6) {
      while (1) {
        switch (_context6.prev = _context6.next) {
          case 0:
            if (deviceResult.keys) {
              _context6.next = 2;
              break;
            }

            return _context6.abrupt("return", false);

          case 2:
            deviceId = deviceResult.device_id;
            userId = deviceResult.user_id;
            signKeyId = "ed25519:" + deviceId;
            signKey = deviceResult.keys[signKeyId];

            if (signKey) {
              _context6.next = 9;
              break;
            }

            _logger.logger.warn("Device " + userId + ":" + deviceId + " has no ed25519 key");

            return _context6.abrupt("return", false);

          case 9:
            unsigned = deviceResult.unsigned || {};
            signatures = deviceResult.signatures || {};
            _context6.prev = 11;
            _context6.next = 14;
            return olmlib.verifySignature(_olmDevice, deviceResult, userId, deviceId, signKey);

          case 14:
            _context6.next = 20;
            break;

          case 16:
            _context6.prev = 16;
            _context6.t0 = _context6["catch"](11);

            _logger.logger.warn("Unable to verify signature on device " + userId + ":" + deviceId + ":" + _context6.t0);

            return _context6.abrupt("return", false);

          case 20:
            if (!(deviceId in userStore)) {
              _context6.next = 27;
              break;
            }

            // already have this device.
            deviceStore = userStore[deviceId];

            if (!(deviceStore.getFingerprint() != signKey)) {
              _context6.next = 25;
              break;
            }

            // this should only happen if the list has been MITMed; we are
            // best off sticking with the original keys.
            //
            // Should we warn the user about it somehow?
            _logger.logger.warn("Ed25519 key for device " + userId + ":" + deviceId + " has changed");

            return _context6.abrupt("return", false);

          case 25:
            _context6.next = 28;
            break;

          case 27:
            userStore[deviceId] = deviceStore = new _deviceinfo.DeviceInfo(deviceId);

          case 28:
            deviceStore.keys = deviceResult.keys || {};
            deviceStore.algorithms = deviceResult.algorithms || [];
            deviceStore.unsigned = unsigned;
            deviceStore.signatures = signatures;
            return _context6.abrupt("return", true);

          case 33:
          case "end":
            return _context6.stop();
        }
      }
    }, _callee6, null, [[11, 16]]);
  }));
  return _storeDeviceKeys2.apply(this, arguments);
}