"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.EncryptionSetupOperation = exports.EncryptionSetupBuilder = void 0;

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _logger = require("../logger");

var _event = require("../models/event");

var _events = require("events");

var _CrossSigning = require("./CrossSigning");

var _indexeddbCryptoStore = require("./store/indexeddb-crypto-store");

var _httpApi = require("../http-api");

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

/**
 * Builds an EncryptionSetupOperation by calling any of the add.. methods.
 * Once done, `buildOperation()` can be called which allows to apply to operation.
 *
 * This is used as a helper by Crypto to keep track of all the network requests
 * and other side-effects of bootstrapping, so it can be applied in one go (and retried in the future)
 * Also keeps track of all the private keys created during bootstrapping, so we don't need to prompt for them
 * more than once.
 */
var EncryptionSetupBuilder = /*#__PURE__*/function () {
  /**
   * @param {Object.<String, MatrixEvent>} accountData pre-existing account data, will only be read, not written.
   * @param {CryptoCallbacks} delegateCryptoCallbacks crypto callbacks to delegate to if the key isn't in cache yet
   */
  function EncryptionSetupBuilder(accountData, delegateCryptoCallbacks) {
    (0, _classCallCheck2["default"])(this, EncryptionSetupBuilder);
    this.accountDataClientAdapter = new AccountDataClientAdapter(accountData);
    this.crossSigningCallbacks = new CrossSigningCallbacks();
    this.ssssCryptoCallbacks = new SSSSCryptoCallbacks(delegateCryptoCallbacks);
    this._crossSigningKeys = null;
    this._keySignatures = null;
    this._keyBackupInfo = null;
  }
  /**
   * Adds new cross-signing public keys
   *
   * @param {function} authUpload Function called to await an interactive auth
   * flow when uploading device signing keys.
   * Args:
   *     {function} A function that makes the request requiring auth. Receives
   *     the auth data as an object. Can be called multiple times, first with
   *     an empty authDict, to obtain the flows.
   * @param {Object} keys the new keys
   */


  (0, _createClass2["default"])(EncryptionSetupBuilder, [{
    key: "addCrossSigningKeys",
    value: function addCrossSigningKeys(authUpload, keys) {
      this._crossSigningKeys = {
        authUpload: authUpload,
        keys: keys
      };
    }
    /**
     * Adds the key backup info to be updated on the server
     *
     * Used either to create a new key backup, or add signatures
     * from the new MSK.
     *
     * @param {Object} keyBackupInfo as received from/sent to the server
     */

  }, {
    key: "addSessionBackup",
    value: function addSessionBackup(keyBackupInfo) {
      this._keyBackupInfo = keyBackupInfo;
    }
    /**
     * Adds the session backup private key to be updated in the local cache
     *
     * Used after fixing the format of the key
     *
     * @param {Uint8Array} privateKey
     */

  }, {
    key: "addSessionBackupPrivateKeyToCache",
    value: function addSessionBackupPrivateKeyToCache(privateKey) {
      this._sessionBackupPrivateKey = privateKey;
    }
    /**
     * Add signatures from a given user and device/x-sign key
     * Used to sign the new cross-signing key with the device key
     *
     * @param {String} userId
     * @param {String} deviceId
     * @param {String} signature
     */

  }, {
    key: "addKeySignature",
    value: function addKeySignature(userId, deviceId, signature) {
      if (!this._keySignatures) {
        this._keySignatures = {};
      }

      var userSignatures = this._keySignatures[userId] || {};
      this._keySignatures[userId] = userSignatures;
      userSignatures[deviceId] = signature;
    }
    /**
     * @param {String} type
     * @param {Object} content
     * @return {Promise}
     */

  }, {
    key: "setAccountData",
    value: function setAccountData(type, content) {
      return this.accountDataClientAdapter.setAccountData(type, content);
    }
    /**
     * builds the operation containing all the parts that have been added to the builder
     * @return {EncryptionSetupOperation}
     */

  }, {
    key: "buildOperation",
    value: function buildOperation() {
      var accountData = this.accountDataClientAdapter._values;
      return new EncryptionSetupOperation(accountData, this._crossSigningKeys, this._keyBackupInfo, this._keySignatures);
    }
    /**
     * Stores the created keys locally.
     *
     * This does not yet store the operation in a way that it can be restored,
     * but that is the idea in the future.
     *
     * @param  {Crypto} crypto
     * @return {Promise}
     */

  }, {
    key: "persist",
    value: function () {
      var _persist = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(crypto) {
        var _this = this;

        var cacheCallbacks, _i, _arr, type, privateKey;

        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                if (!this._crossSigningKeys) {
                  _context.next = 14;
                  break;
                }

                cacheCallbacks = (0, _CrossSigning.createCryptoStoreCacheCallbacks)(crypto._cryptoStore, crypto._olmDevice);
                _i = 0, _arr = ["master", "self_signing", "user_signing"];

              case 3:
                if (!(_i < _arr.length)) {
                  _context.next = 12;
                  break;
                }

                type = _arr[_i];

                _logger.logger.log("Cache ".concat(type, " cross-signing private key locally"));

                privateKey = this.crossSigningCallbacks.privateKeys.get(type);
                _context.next = 9;
                return cacheCallbacks.storeCrossSigningKeyCache(type, privateKey);

              case 9:
                _i++;
                _context.next = 3;
                break;

              case 12:
                _context.next = 14;
                return crypto._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
                  crypto._cryptoStore.storeCrossSigningKeys(txn, _this._crossSigningKeys.keys);
                });

              case 14:
                if (!this._sessionBackupPrivateKey) {
                  _context.next = 17;
                  break;
                }

                _context.next = 17;
                return crypto.storeSessionBackupPrivateKey(this._sessionBackupPrivateKey);

              case 17:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function persist(_x) {
        return _persist.apply(this, arguments);
      }

      return persist;
    }()
  }]);
  return EncryptionSetupBuilder;
}();
/**
 * Can be created from EncryptionSetupBuilder, or
 * (in a follow-up PR, not implemented yet) restored from storage, to retry.
 *
 * It does not have knowledge of any private keys, unlike the builder.
 */


exports.EncryptionSetupBuilder = EncryptionSetupBuilder;

var EncryptionSetupOperation = /*#__PURE__*/function () {
  /**
   * @param  {Map<String, Object>} accountData
   * @param  {Object} crossSigningKeys
   * @param  {Object} keyBackupInfo
   * @param  {Object} keySignatures
   */
  function EncryptionSetupOperation(accountData, crossSigningKeys, keyBackupInfo, keySignatures) {
    (0, _classCallCheck2["default"])(this, EncryptionSetupOperation);
    this._accountData = accountData;
    this._crossSigningKeys = crossSigningKeys;
    this._keyBackupInfo = keyBackupInfo;
    this._keySignatures = keySignatures;
  }
  /**
   * Runs the (remaining part of, in the future) operation by sending requests to the server.
   * @param  {Crypto} crypto
   */


  (0, _createClass2["default"])(EncryptionSetupOperation, [{
    key: "apply",
    value: function () {
      var _apply = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(crypto) {
        var baseApis, keys, _i2, _Object$entries, _Object$entries$_i, name, key, _iterator, _step, _step$value, type, content;

        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                baseApis = crypto._baseApis; // upload cross-signing keys

                if (!this._crossSigningKeys) {
                  _context2.next = 7;
                  break;
                }

                keys = {};

                for (_i2 = 0, _Object$entries = Object.entries(this._crossSigningKeys.keys); _i2 < _Object$entries.length; _i2++) {
                  _Object$entries$_i = (0, _slicedToArray2["default"])(_Object$entries[_i2], 2), name = _Object$entries$_i[0], key = _Object$entries$_i[1];
                  keys[name + "_key"] = key;
                } // We must only call `uploadDeviceSigningKeys` from inside this auth
                // helper to ensure we properly handle auth errors.


                _context2.next = 6;
                return this._crossSigningKeys.authUpload(function (authDict) {
                  return baseApis.uploadDeviceSigningKeys(authDict, keys);
                });

              case 6:
                // pass the new keys to the main instance of our own CrossSigningInfo.
                crypto._crossSigningInfo.setKeys(this._crossSigningKeys.keys);

              case 7:
                if (!this._accountData) {
                  _context2.next = 25;
                  break;
                }

                _iterator = _createForOfIteratorHelper(this._accountData);
                _context2.prev = 9;

                _iterator.s();

              case 11:
                if ((_step = _iterator.n()).done) {
                  _context2.next = 17;
                  break;
                }

                _step$value = (0, _slicedToArray2["default"])(_step.value, 2), type = _step$value[0], content = _step$value[1];
                _context2.next = 15;
                return baseApis.setAccountData(type, content);

              case 15:
                _context2.next = 11;
                break;

              case 17:
                _context2.next = 22;
                break;

              case 19:
                _context2.prev = 19;
                _context2.t0 = _context2["catch"](9);

                _iterator.e(_context2.t0);

              case 22:
                _context2.prev = 22;

                _iterator.f();

                return _context2.finish(22);

              case 25:
                if (!this._keySignatures) {
                  _context2.next = 28;
                  break;
                }

                _context2.next = 28;
                return baseApis.uploadKeySignatures(this._keySignatures);

              case 28:
                if (!this._keyBackupInfo) {
                  _context2.next = 36;
                  break;
                }

                if (!this._keyBackupInfo.version) {
                  _context2.next = 34;
                  break;
                }

                _context2.next = 32;
                return baseApis._http.authedRequest(undefined, "PUT", "/room_keys/version/" + this._keyBackupInfo.version, undefined, {
                  algorithm: this._keyBackupInfo.algorithm,
                  auth_data: this._keyBackupInfo.auth_data
                }, {
                  prefix: _httpApi.PREFIX_UNSTABLE
                });

              case 32:
                _context2.next = 36;
                break;

              case 34:
                _context2.next = 36;
                return baseApis._http.authedRequest(undefined, "POST", "/room_keys/version", undefined, this._keyBackupInfo, {
                  prefix: _httpApi.PREFIX_UNSTABLE
                });

              case 36:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this, [[9, 19, 22, 25]]);
      }));

      function apply(_x2) {
        return _apply.apply(this, arguments);
      }

      return apply;
    }()
  }]);
  return EncryptionSetupOperation;
}();
/**
 * Catches account data set by SecretStorage during bootstrapping by
 * implementing the methods related to account data in MatrixClient
 */


exports.EncryptionSetupOperation = EncryptionSetupOperation;

var AccountDataClientAdapter = /*#__PURE__*/function (_EventEmitter) {
  (0, _inherits2["default"])(AccountDataClientAdapter, _EventEmitter);

  var _super = _createSuper(AccountDataClientAdapter);

  /**
   * @param  {Object.<String, MatrixEvent>} accountData existing account data
   */
  function AccountDataClientAdapter(accountData) {
    var _this2;

    (0, _classCallCheck2["default"])(this, AccountDataClientAdapter);
    _this2 = _super.call(this);
    _this2._existingValues = accountData;
    _this2._values = new Map();
    return _this2;
  }
  /**
   * @param  {String} type
   * @return {Promise<Object>} the content of the account data
   */


  (0, _createClass2["default"])(AccountDataClientAdapter, [{
    key: "getAccountDataFromServer",
    value: function getAccountDataFromServer(type) {
      return Promise.resolve(this.getAccountData(type));
    }
    /**
     * @param  {String} type
     * @return {Object} the content of the account data
     */

  }, {
    key: "getAccountData",
    value: function getAccountData(type) {
      var modifiedValue = this._values.get(type);

      if (modifiedValue) {
        return modifiedValue;
      }

      var existingValue = this._existingValues[type];

      if (existingValue) {
        return existingValue.getContent();
      }

      return null;
    }
    /**
     * @param {String} type
     * @param {Object} content
     * @return {Promise}
     */

  }, {
    key: "setAccountData",
    value: function setAccountData(type, content) {
      var _this3 = this;

      var lastEvent = this._values.get(type);

      this._values.set(type, content); // ensure accountData is emitted on the next tick,
      // as SecretStorage listens for it while calling this method
      // and it seems to rely on this.


      return Promise.resolve().then(function () {
        var event = new _event.MatrixEvent({
          type: type,
          content: content
        });

        _this3.emit("accountData", event, lastEvent);
      });
    }
  }]);
  return AccountDataClientAdapter;
}(_events.EventEmitter);
/**
 * Catches the private cross-signing keys set during bootstrapping
 * by both cache callbacks (see createCryptoStoreCacheCallbacks) as non-cache callbacks.
 * See CrossSigningInfo constructor
 */


var CrossSigningCallbacks = /*#__PURE__*/function () {
  function CrossSigningCallbacks() {
    (0, _classCallCheck2["default"])(this, CrossSigningCallbacks);
    this.privateKeys = new Map();
  } // cache callbacks


  (0, _createClass2["default"])(CrossSigningCallbacks, [{
    key: "getCrossSigningKeyCache",
    value: function getCrossSigningKeyCache(type, expectedPublicKey) {
      return this.getCrossSigningKey(type, expectedPublicKey);
    }
  }, {
    key: "storeCrossSigningKeyCache",
    value: function storeCrossSigningKeyCache(type, key) {
      this.privateKeys.set(type, key);
      return Promise.resolve();
    } // non-cache callbacks

  }, {
    key: "getCrossSigningKey",
    value: function getCrossSigningKey(type, _expectedPubkey) {
      return Promise.resolve(this.privateKeys.get(type));
    }
  }, {
    key: "saveCrossSigningKeys",
    value: function saveCrossSigningKeys(privateKeys) {
      for (var _i3 = 0, _Object$entries2 = Object.entries(privateKeys); _i3 < _Object$entries2.length; _i3++) {
        var _Object$entries2$_i = (0, _slicedToArray2["default"])(_Object$entries2[_i3], 2),
            type = _Object$entries2$_i[0],
            privateKey = _Object$entries2$_i[1];

        this.privateKeys.set(type, privateKey);
      }
    }
  }]);
  return CrossSigningCallbacks;
}();
/**
 * Catches the 4S private key set during bootstrapping by implementing
 * the SecretStorage crypto callbacks
 */


var SSSSCryptoCallbacks = /*#__PURE__*/function () {
  function SSSSCryptoCallbacks(delegateCryptoCallbacks) {
    (0, _classCallCheck2["default"])(this, SSSSCryptoCallbacks);
    this._privateKeys = new Map();
    this._delegateCryptoCallbacks = delegateCryptoCallbacks;
  }

  (0, _createClass2["default"])(SSSSCryptoCallbacks, [{
    key: "getSecretStorageKey",
    value: function () {
      var _getSecretStorageKey = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(_ref, name) {
        var keys, _i4, _Object$keys, keyId, privateKey, result, _result, _keyId, _privateKey;

        return _regenerator["default"].wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                keys = _ref.keys;
                _i4 = 0, _Object$keys = Object.keys(keys);

              case 2:
                if (!(_i4 < _Object$keys.length)) {
                  _context3.next = 10;
                  break;
                }

                keyId = _Object$keys[_i4];
                privateKey = this._privateKeys.get(keyId);

                if (!privateKey) {
                  _context3.next = 7;
                  break;
                }

                return _context3.abrupt("return", [keyId, privateKey]);

              case 7:
                _i4++;
                _context3.next = 2;
                break;

              case 10:
                if (!this._delegateCryptoCallbacks) {
                  _context3.next = 16;
                  break;
                }

                _context3.next = 13;
                return this._delegateCryptoCallbacks.getSecretStorageKey({
                  keys: keys
                }, name);

              case 13:
                result = _context3.sent;

                if (result) {
                  _result = (0, _slicedToArray2["default"])(result, 2), _keyId = _result[0], _privateKey = _result[1];

                  this._privateKeys.set(_keyId, _privateKey);
                }

                return _context3.abrupt("return", result);

              case 16:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function getSecretStorageKey(_x3, _x4) {
        return _getSecretStorageKey.apply(this, arguments);
      }

      return getSecretStorageKey;
    }()
  }, {
    key: "addPrivateKey",
    value: function addPrivateKey(keyId, keyInfo, privKey) {
      this._privateKeys.set(keyId, privKey); // Also pass along to application to cache if it wishes


      if (this._delegateCryptoCallbacks && this._delegateCryptoCallbacks.cacheSecretStorageKey) {
        this._delegateCryptoCallbacks.cacheSecretStorageKey(keyId, keyInfo, privKey);
      }
    }
  }]);
  return SSSSCryptoCallbacks;
}();