"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.OutgoingRoomKeyRequestManager = exports.ROOM_KEY_REQUEST_STATES = void 0;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _logger = require("../logger");

var utils = _interopRequireWildcard(require("../utils"));

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

/**
 * Internal module. Management of outgoing room key requests.
 *
 * See https://docs.google.com/document/d/1m4gQkcnJkxNuBmb5NoFCIadIY-DyqqNAS3lloE73BlQ
 * for draft documentation on what we're supposed to be implementing here.
 *
 * @module
 */
// delay between deciding we want some keys, and sending out the request, to
// allow for (a) it turning up anyway, (b) grouping requests together
var SEND_KEY_REQUESTS_DELAY_MS = 500;
/** possible states for a room key request
 *
 * The state machine looks like:
 *
 *     |         (cancellation sent)
 *     | .-------------------------------------------------.
 *     | |                                                 |
 *     V V       (cancellation requested)                  |
 *   UNSENT  -----------------------------+                |
 *     |                                  |                |
 *     |                                  |                |
 *     | (send successful)                |  CANCELLATION_PENDING_AND_WILL_RESEND
 *     V                                  |                Λ
 *    SENT                                |                |
 *     |--------------------------------  |  --------------'
 *     |                                  |  (cancellation requested with intent
 *     |                                  |   to resend the original request)
 *     |                                  |
 *     | (cancellation requested)         |
 *     V                                  |
 * CANCELLATION_PENDING                   |
 *     |                                  |
 *     | (cancellation sent)              |
 *     V                                  |
 * (deleted)  <---------------------------+
 *
 * @enum {number}
 */

var ROOM_KEY_REQUEST_STATES = {
  /** request not yet sent */
  UNSENT: 0,

  /** request sent, awaiting reply */
  SENT: 1,

  /** reply received, cancellation not yet sent */
  CANCELLATION_PENDING: 2,

  /**
   * Cancellation not yet sent and will transition to UNSENT instead of
   * being deleted once the cancellation has been sent.
   */
  CANCELLATION_PENDING_AND_WILL_RESEND: 3
};
exports.ROOM_KEY_REQUEST_STATES = ROOM_KEY_REQUEST_STATES;

var OutgoingRoomKeyRequestManager = /*#__PURE__*/function () {
  function OutgoingRoomKeyRequestManager(baseApis, deviceId, cryptoStore) {
    (0, _classCallCheck2["default"])(this, OutgoingRoomKeyRequestManager);
    this._baseApis = baseApis;
    this._deviceId = deviceId;
    this._cryptoStore = cryptoStore; // handle for the delayed call to _sendOutgoingRoomKeyRequests. Non-null
    // if the callback has been set, or if it is still running.

    this._sendOutgoingRoomKeyRequestsTimer = null; // sanity check to ensure that we don't end up with two concurrent runs
    // of _sendOutgoingRoomKeyRequests

    this._sendOutgoingRoomKeyRequestsRunning = false;
    this._clientRunning = false;
  }
  /**
   * Called when the client is started. Sets background processes running.
   */


  (0, _createClass2["default"])(OutgoingRoomKeyRequestManager, [{
    key: "start",
    value: function start() {
      this._clientRunning = true;
    }
    /**
     * Called when the client is stopped. Stops any running background processes.
     */

  }, {
    key: "stop",
    value: function stop() {
      _logger.logger.log('stopping OutgoingRoomKeyRequestManager'); // stop the timer on the next run


      this._clientRunning = false;
    }
    /**
     * Send any requests that have been queued
     */

  }, {
    key: "sendQueuedRequests",
    value: function sendQueuedRequests() {
      this._startTimer();
    }
    /**
     * Queue up a room key request, if we haven't already queued or sent one.
     *
     * The `requestBody` is compared (with a deep-equality check) against
     * previous queued or sent requests and if it matches, no change is made.
     * Otherwise, a request is added to the pending list, and a job is started
     * in the background to send it.
     *
     * @param {module:crypto~RoomKeyRequestBody} requestBody
     * @param {Array<{userId: string, deviceId: string}>} recipients
     * @param {boolean} resend whether to resend the key request if there is
     *    already one
     *
     * @returns {Promise} resolves when the request has been added to the
     *    pending list (or we have established that a similar request already
     *    exists)
     */

  }, {
    key: "queueRoomKeyRequest",
    value: function () {
      var _queueRoomKeyRequest = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(requestBody, recipients) {
        var resend,
            req,
            state,
            _state,
            updatedReq,
            _args = arguments;

        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                resend = _args.length > 2 && _args[2] !== undefined ? _args[2] : false;
                _context.next = 3;
                return this._cryptoStore.getOutgoingRoomKeyRequest(requestBody);

              case 3:
                req = _context.sent;

                if (req) {
                  _context.next = 9;
                  break;
                }

                _context.next = 7;
                return this._cryptoStore.getOrAddOutgoingRoomKeyRequest({
                  requestBody: requestBody,
                  recipients: recipients,
                  requestId: this._baseApis.makeTxnId(),
                  state: ROOM_KEY_REQUEST_STATES.UNSENT
                });

              case 7:
                _context.next = 36;
                break;

              case 9:
                _context.t0 = req.state;
                _context.next = _context.t0 === ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING_AND_WILL_RESEND ? 12 : _context.t0 === ROOM_KEY_REQUEST_STATES.UNSENT ? 12 : _context.t0 === ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING ? 13 : _context.t0 === ROOM_KEY_REQUEST_STATES.SENT ? 17 : 35;
                break;

              case 12:
                return _context.abrupt("return");

              case 13:
                // existing request is about to be cancelled.  If we want to
                // resend, then change the state so that it resends after
                // cancelling.  Otherwise, just cancel the cancellation.
                state = resend ? ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING_AND_WILL_RESEND : ROOM_KEY_REQUEST_STATES.SENT;
                _context.next = 16;
                return this._cryptoStore.updateOutgoingRoomKeyRequest(req.requestId, ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING, {
                  state: state,
                  cancellationTxnId: this._baseApis.makeTxnId()
                });

              case 16:
                return _context.abrupt("break", 36);

              case 17:
                if (!resend) {
                  _context.next = 34;
                  break;
                }

                _state = ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING_AND_WILL_RESEND;
                _context.next = 21;
                return this._cryptoStore.updateOutgoingRoomKeyRequest(req.requestId, ROOM_KEY_REQUEST_STATES.SENT, {
                  state: _state,
                  cancellationTxnId: this._baseApis.makeTxnId(),
                  // need to use a new transaction ID so that
                  // the request gets sent
                  requestTxnId: this._baseApis.makeTxnId()
                });

              case 21:
                updatedReq = _context.sent;

                if (updatedReq) {
                  _context.next = 26;
                  break;
                }

                _context.next = 25;
                return this.queueRoomKeyRequest(requestBody, recipients, resend);

              case 25:
                return _context.abrupt("return", _context.sent);

              case 26:
                _context.prev = 26;
                _context.next = 29;
                return this._sendOutgoingRoomKeyRequestCancellation(updatedReq, true);

              case 29:
                _context.next = 34;
                break;

              case 31:
                _context.prev = 31;
                _context.t1 = _context["catch"](26);

                _logger.logger.error("Error sending room key request cancellation;" + " will retry later.", _context.t1);

              case 34:
                return _context.abrupt("break", 36);

              case 35:
                throw new Error('unhandled state: ' + req.state);

              case 36:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this, [[26, 31]]);
      }));

      function queueRoomKeyRequest(_x, _x2) {
        return _queueRoomKeyRequest.apply(this, arguments);
      }

      return queueRoomKeyRequest;
    }()
    /**
     * Cancel room key requests, if any match the given requestBody
     *
     * @param {module:crypto~RoomKeyRequestBody} requestBody
     *
     * @returns {Promise} resolves when the request has been updated in our
     *    pending list.
     */

  }, {
    key: "cancelRoomKeyRequest",
    value: function cancelRoomKeyRequest(requestBody) {
      var _this = this;

      return this._cryptoStore.getOutgoingRoomKeyRequest(requestBody).then(function (req) {
        if (!req) {
          // no request was made for this key
          return;
        }

        switch (req.state) {
          case ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING:
          case ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING_AND_WILL_RESEND:
            // nothing to do here
            return;

          case ROOM_KEY_REQUEST_STATES.UNSENT:
            // just delete it
            // FIXME: ghahah we may have attempted to send it, and
            // not yet got a successful response. So the server
            // may have seen it, so we still need to send a cancellation
            // in that case :/
            _logger.logger.log('deleting unnecessary room key request for ' + stringifyRequestBody(requestBody));

            return _this._cryptoStore.deleteOutgoingRoomKeyRequest(req.requestId, ROOM_KEY_REQUEST_STATES.UNSENT);

          case ROOM_KEY_REQUEST_STATES.SENT:
            {
              // send a cancellation.
              return _this._cryptoStore.updateOutgoingRoomKeyRequest(req.requestId, ROOM_KEY_REQUEST_STATES.SENT, {
                state: ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING,
                cancellationTxnId: _this._baseApis.makeTxnId()
              }).then(function (updatedReq) {
                if (!updatedReq) {
                  // updateOutgoingRoomKeyRequest couldn't find the
                  // request in state ROOM_KEY_REQUEST_STATES.SENT,
                  // so we must have raced with another tab to mark
                  // the request cancelled. There is no point in
                  // sending another cancellation since the other tab
                  // will do it.
                  _logger.logger.log('Tried to cancel room key request for ' + stringifyRequestBody(requestBody) + ' but it was already cancelled in another tab');

                  return;
                } // We don't want to wait for the timer, so we send it
                // immediately. (We might actually end up racing with the timer,
                // but that's ok: even if we make the request twice, we'll do it
                // with the same transaction_id, so only one message will get
                // sent).
                //
                // (We also don't want to wait for the response from the server
                // here, as it will slow down processing of received keys if we
                // do.)


                _this._sendOutgoingRoomKeyRequestCancellation(updatedReq)["catch"](function (e) {
                  _logger.logger.error("Error sending room key request cancellation;" + " will retry later.", e);

                  _this._startTimer();
                });
              });
            }

          default:
            throw new Error('unhandled state: ' + req.state);
        }
      });
    }
    /**
     * Look for room key requests by target device and state
     *
     * @param {string} userId Target user ID
     * @param {string} deviceId Target device ID
     *
     * @return {Promise} resolves to a list of all the
     *    {@link module:crypto/store/base~OutgoingRoomKeyRequest}
     */

  }, {
    key: "getOutgoingSentRoomKeyRequest",
    value: function getOutgoingSentRoomKeyRequest(userId, deviceId) {
      return this._cryptoStore.getOutgoingRoomKeyRequestsByTarget(userId, deviceId, [ROOM_KEY_REQUEST_STATES.SENT]);
    }
    /**
     * Find anything in `sent` state, and kick it around the loop again.
     * This is intended for situations where something substantial has changed, and we
     * don't really expect the other end to even care about the cancellation.
     * For example, after initialization or self-verification.
     * @return {Promise} An array of `queueRoomKeyRequest` outputs.
     */

  }, {
    key: "cancelAndResendAllOutgoingRequests",
    value: function () {
      var _cancelAndResendAllOutgoingRequests = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
        var _this2 = this;

        var outgoings;
        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return this._cryptoStore.getAllOutgoingRoomKeyRequestsByState(ROOM_KEY_REQUEST_STATES.SENT);

              case 2:
                outgoings = _context2.sent;
                return _context2.abrupt("return", Promise.all(outgoings.map(function (_ref) {
                  var requestBody = _ref.requestBody,
                      recipients = _ref.recipients;
                  return _this2.queueRoomKeyRequest(requestBody, recipients, true);
                })));

              case 4:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function cancelAndResendAllOutgoingRequests() {
        return _cancelAndResendAllOutgoingRequests.apply(this, arguments);
      }

      return cancelAndResendAllOutgoingRequests;
    }() // start the background timer to send queued requests, if the timer isn't
    // already running

  }, {
    key: "_startTimer",
    value: function _startTimer() {
      var _this3 = this;

      if (this._sendOutgoingRoomKeyRequestsTimer) {
        return;
      }

      var startSendingOutgoingRoomKeyRequests = function startSendingOutgoingRoomKeyRequests() {
        if (_this3._sendOutgoingRoomKeyRequestsRunning) {
          throw new Error("RoomKeyRequestSend already in progress!");
        }

        _this3._sendOutgoingRoomKeyRequestsRunning = true;

        _this3._sendOutgoingRoomKeyRequests()["finally"](function () {
          _this3._sendOutgoingRoomKeyRequestsRunning = false;
        })["catch"](function (e) {
          // this should only happen if there is an indexeddb error,
          // in which case we're a bit stuffed anyway.
          _logger.logger.warn("error in OutgoingRoomKeyRequestManager: ".concat(e));
        });
      };

      this._sendOutgoingRoomKeyRequestsTimer = global.setTimeout(startSendingOutgoingRoomKeyRequests, SEND_KEY_REQUESTS_DELAY_MS);
    } // look for and send any queued requests. Runs itself recursively until
    // there are no more requests, or there is an error (in which case, the
    // timer will be restarted before the promise resolves).

  }, {
    key: "_sendOutgoingRoomKeyRequests",
    value: function _sendOutgoingRoomKeyRequests() {
      var _this4 = this;

      if (!this._clientRunning) {
        this._sendOutgoingRoomKeyRequestsTimer = null;
        return Promise.resolve();
      }

      return this._cryptoStore.getOutgoingRoomKeyRequestByState([ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING, ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING_AND_WILL_RESEND, ROOM_KEY_REQUEST_STATES.UNSENT]).then(function (req) {
        if (!req) {
          _this4._sendOutgoingRoomKeyRequestsTimer = null;
          return;
        }

        var prom;

        switch (req.state) {
          case ROOM_KEY_REQUEST_STATES.UNSENT:
            prom = _this4._sendOutgoingRoomKeyRequest(req);
            break;

          case ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING:
            prom = _this4._sendOutgoingRoomKeyRequestCancellation(req);
            break;

          case ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING_AND_WILL_RESEND:
            prom = _this4._sendOutgoingRoomKeyRequestCancellation(req, true);
            break;
        }

        return prom.then(function () {
          // go around the loop again
          return _this4._sendOutgoingRoomKeyRequests();
        })["catch"](function (e) {
          _logger.logger.error("Error sending room key request; will retry later.", e);

          _this4._sendOutgoingRoomKeyRequestsTimer = null;
        });
      });
    } // given a RoomKeyRequest, send it and update the request record

  }, {
    key: "_sendOutgoingRoomKeyRequest",
    value: function _sendOutgoingRoomKeyRequest(req) {
      var _this5 = this;

      _logger.logger.log("Requesting keys for ".concat(stringifyRequestBody(req.requestBody)) + " from ".concat(stringifyRecipientList(req.recipients)) + "(id ".concat(req.requestId, ")"));

      var requestMessage = {
        action: "request",
        requesting_device_id: this._deviceId,
        request_id: req.requestId,
        body: req.requestBody
      };
      return this._sendMessageToDevices(requestMessage, req.recipients, req.requestTxnId || req.requestId).then(function () {
        return _this5._cryptoStore.updateOutgoingRoomKeyRequest(req.requestId, ROOM_KEY_REQUEST_STATES.UNSENT, {
          state: ROOM_KEY_REQUEST_STATES.SENT
        });
      });
    } // Given a RoomKeyRequest, cancel it and delete the request record unless
    // andResend is set, in which case transition to UNSENT.

  }, {
    key: "_sendOutgoingRoomKeyRequestCancellation",
    value: function _sendOutgoingRoomKeyRequestCancellation(req, andResend) {
      var _this6 = this;

      _logger.logger.log("Sending cancellation for key request for " + "".concat(stringifyRequestBody(req.requestBody), " to ") + "".concat(stringifyRecipientList(req.recipients), " ") + "(cancellation id ".concat(req.cancellationTxnId, ")"));

      var requestMessage = {
        action: "request_cancellation",
        requesting_device_id: this._deviceId,
        request_id: req.requestId
      };
      return this._sendMessageToDevices(requestMessage, req.recipients, req.cancellationTxnId).then(function () {
        if (andResend) {
          // We want to resend, so transition to UNSENT
          return _this6._cryptoStore.updateOutgoingRoomKeyRequest(req.requestId, ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING_AND_WILL_RESEND, {
            state: ROOM_KEY_REQUEST_STATES.UNSENT
          });
        }

        return _this6._cryptoStore.deleteOutgoingRoomKeyRequest(req.requestId, ROOM_KEY_REQUEST_STATES.CANCELLATION_PENDING);
      });
    } // send a RoomKeyRequest to a list of recipients

  }, {
    key: "_sendMessageToDevices",
    value: function _sendMessageToDevices(message, recipients, txnId) {
      var contentMap = {};

      var _iterator = _createForOfIteratorHelper(recipients),
          _step;

      try {
        for (_iterator.s(); !(_step = _iterator.n()).done;) {
          var recip = _step.value;

          if (!contentMap[recip.userId]) {
            contentMap[recip.userId] = {};
          }

          contentMap[recip.userId][recip.deviceId] = message;
        }
      } catch (err) {
        _iterator.e(err);
      } finally {
        _iterator.f();
      }

      return this._baseApis.sendToDevice('m.room_key_request', contentMap, txnId);
    }
  }]);
  return OutgoingRoomKeyRequestManager;
}();

exports.OutgoingRoomKeyRequestManager = OutgoingRoomKeyRequestManager;

function stringifyRequestBody(requestBody) {
  // we assume that the request is for megolm keys, which are identified by
  // room id and session id
  return requestBody.room_id + " / " + requestBody.session_id;
}

function stringifyRecipientList(recipients) {
  return '[' + utils.map(recipients, function (r) {
    return "".concat(r.userId, ":").concat(r.deviceId);
  }).join(",") + ']';
}