"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.encryptAES = encryptAES;
exports.decryptAES = decryptAES;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _utils = require("../utils");

var _olmlib = require("./olmlib");

/*
Copyright 2020 The Matrix.org Foundation C.I.C.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
var subtleCrypto = typeof window !== "undefined" && window.crypto ? window.crypto.subtle || window.crypto.webkitSubtle : null; // salt for HKDF, with 8 bytes of zeros

var zerosalt = new Uint8Array(8);
/**
 * encrypt a string in Node.js
 *
 * @param {string} data the plaintext to encrypt
 * @param {Uint8Array} key the encryption key to use
 * @param {string} name the name of the secret
 * @param {string} ivStr the initialization vector to use
 */

function encryptNode(_x, _x2, _x3, _x4) {
  return _encryptNode.apply(this, arguments);
}
/**
 * decrypt a string in Node.js
 *
 * @param {object} data the encrypted data
 * @param {string} data.ciphertext the ciphertext in base64
 * @param {string} data.iv the initialization vector in base64
 * @param {string} data.mac the HMAC in base64
 * @param {Uint8Array} key the encryption key to use
 * @param {string} name the name of the secret
 */


function _encryptNode() {
  _encryptNode = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(data, key, name, ivStr) {
    var crypto, iv, _deriveKeysNode, _deriveKeysNode2, aesKey, hmacKey, cipher, ciphertext, hmac;

    return _regenerator["default"].wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            crypto = (0, _utils.getCrypto)();

            if (crypto) {
              _context.next = 3;
              break;
            }

            throw new Error("No usable crypto implementation");

          case 3:
            if (ivStr) {
              iv = (0, _olmlib.decodeBase64)(ivStr);
            } else {
              iv = crypto.randomBytes(16);
            } // clear bit 63 of the IV to stop us hitting the 64-bit counter boundary
            // (which would mean we wouldn't be able to decrypt on Android). The loss
            // of a single bit of iv is a price we have to pay.


            iv[8] &= 0x7f;
            _deriveKeysNode = deriveKeysNode(key, name), _deriveKeysNode2 = (0, _slicedToArray2["default"])(_deriveKeysNode, 2), aesKey = _deriveKeysNode2[0], hmacKey = _deriveKeysNode2[1];
            cipher = crypto.createCipheriv("aes-256-ctr", aesKey, iv);
            ciphertext = cipher.update(data, "utf-8", "base64") + cipher["final"]("base64");
            hmac = crypto.createHmac("sha256", hmacKey).update(ciphertext, "base64").digest("base64");
            return _context.abrupt("return", {
              iv: (0, _olmlib.encodeBase64)(iv),
              ciphertext: ciphertext,
              mac: hmac
            });

          case 10:
          case "end":
            return _context.stop();
        }
      }
    }, _callee);
  }));
  return _encryptNode.apply(this, arguments);
}

function decryptNode(_x5, _x6, _x7) {
  return _decryptNode.apply(this, arguments);
}

function _decryptNode() {
  _decryptNode = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(data, key, name) {
    var crypto, _deriveKeysNode3, _deriveKeysNode4, aesKey, hmacKey, hmac, decipher;

    return _regenerator["default"].wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            crypto = (0, _utils.getCrypto)();

            if (crypto) {
              _context2.next = 3;
              break;
            }

            throw new Error("No usable crypto implementation");

          case 3:
            _deriveKeysNode3 = deriveKeysNode(key, name), _deriveKeysNode4 = (0, _slicedToArray2["default"])(_deriveKeysNode3, 2), aesKey = _deriveKeysNode4[0], hmacKey = _deriveKeysNode4[1];
            hmac = crypto.createHmac("sha256", hmacKey).update(data.ciphertext, "base64").digest("base64").replace(/=+$/g, '');

            if (!(hmac !== data.mac.replace(/=+$/g, ''))) {
              _context2.next = 7;
              break;
            }

            throw new Error("Error decrypting secret ".concat(name, ": bad MAC"));

          case 7:
            decipher = crypto.createDecipheriv("aes-256-ctr", aesKey, (0, _olmlib.decodeBase64)(data.iv));
            return _context2.abrupt("return", decipher.update(data.ciphertext, "base64", "utf-8") + decipher["final"]("utf-8"));

          case 9:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2);
  }));
  return _decryptNode.apply(this, arguments);
}

function deriveKeysNode(key, name) {
  var crypto = (0, _utils.getCrypto)();
  var prk = crypto.createHmac("sha256", zerosalt).update(key).digest();
  var b = Buffer.alloc(1, 1);
  var aesKey = crypto.createHmac("sha256", prk).update(name, "utf-8").update(b).digest();
  b[0] = 2;
  var hmacKey = crypto.createHmac("sha256", prk).update(aesKey).update(name, "utf-8").update(b).digest();
  return [aesKey, hmacKey];
}
/**
 * encrypt a string in Node.js
 *
 * @param {string} data the plaintext to encrypt
 * @param {Uint8Array} key the encryption key to use
 * @param {string} name the name of the secret
 * @param {string} ivStr the initialization vector to use
 */


function encryptBrowser(_x8, _x9, _x10, _x11) {
  return _encryptBrowser.apply(this, arguments);
}
/**
 * decrypt a string in the browser
 *
 * @param {object} data the encrypted data
 * @param {string} data.ciphertext the ciphertext in base64
 * @param {string} data.iv the initialization vector in base64
 * @param {string} data.mac the HMAC in base64
 * @param {Uint8Array} key the encryption key to use
 * @param {string} name the name of the secret
 */


function _encryptBrowser() {
  _encryptBrowser = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(data, key, name, ivStr) {
    var iv, _yield$deriveKeysBrow, _yield$deriveKeysBrow2, aesKey, hmacKey, encodedData, ciphertext, hmac;

    return _regenerator["default"].wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            if (ivStr) {
              iv = (0, _olmlib.decodeBase64)(ivStr);
            } else {
              iv = new Uint8Array(16);
              window.crypto.getRandomValues(iv);
            } // clear bit 63 of the IV to stop us hitting the 64-bit counter boundary
            // (which would mean we wouldn't be able to decrypt on Android). The loss
            // of a single bit of iv is a price we have to pay.


            iv[8] &= 0x7f;
            _context3.next = 4;
            return deriveKeysBrowser(key, name);

          case 4:
            _yield$deriveKeysBrow = _context3.sent;
            _yield$deriveKeysBrow2 = (0, _slicedToArray2["default"])(_yield$deriveKeysBrow, 2);
            aesKey = _yield$deriveKeysBrow2[0];
            hmacKey = _yield$deriveKeysBrow2[1];
            encodedData = new TextEncoder().encode(data);
            _context3.next = 11;
            return subtleCrypto.encrypt({
              name: "AES-CTR",
              counter: iv,
              length: 64
            }, aesKey, encodedData);

          case 11:
            ciphertext = _context3.sent;
            _context3.next = 14;
            return subtleCrypto.sign({
              name: 'HMAC'
            }, hmacKey, ciphertext);

          case 14:
            hmac = _context3.sent;
            return _context3.abrupt("return", {
              iv: (0, _olmlib.encodeBase64)(iv),
              ciphertext: (0, _olmlib.encodeBase64)(ciphertext),
              mac: (0, _olmlib.encodeBase64)(hmac)
            });

          case 16:
          case "end":
            return _context3.stop();
        }
      }
    }, _callee3);
  }));
  return _encryptBrowser.apply(this, arguments);
}

function decryptBrowser(_x12, _x13, _x14) {
  return _decryptBrowser.apply(this, arguments);
}

function _decryptBrowser() {
  _decryptBrowser = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4(data, key, name) {
    var _yield$deriveKeysBrow3, _yield$deriveKeysBrow4, aesKey, hmacKey, ciphertext, plaintext;

    return _regenerator["default"].wrap(function _callee4$(_context4) {
      while (1) {
        switch (_context4.prev = _context4.next) {
          case 0:
            _context4.next = 2;
            return deriveKeysBrowser(key, name);

          case 2:
            _yield$deriveKeysBrow3 = _context4.sent;
            _yield$deriveKeysBrow4 = (0, _slicedToArray2["default"])(_yield$deriveKeysBrow3, 2);
            aesKey = _yield$deriveKeysBrow4[0];
            hmacKey = _yield$deriveKeysBrow4[1];
            ciphertext = (0, _olmlib.decodeBase64)(data.ciphertext);
            _context4.next = 9;
            return subtleCrypto.verify({
              name: "HMAC"
            }, hmacKey, (0, _olmlib.decodeBase64)(data.mac), ciphertext);

          case 9:
            if (_context4.sent) {
              _context4.next = 11;
              break;
            }

            throw new Error("Error decrypting secret ".concat(name, ": bad MAC"));

          case 11:
            _context4.next = 13;
            return subtleCrypto.decrypt({
              name: "AES-CTR",
              counter: (0, _olmlib.decodeBase64)(data.iv),
              length: 64
            }, aesKey, ciphertext);

          case 13:
            plaintext = _context4.sent;
            return _context4.abrupt("return", new TextDecoder().decode(new Uint8Array(plaintext)));

          case 15:
          case "end":
            return _context4.stop();
        }
      }
    }, _callee4);
  }));
  return _decryptBrowser.apply(this, arguments);
}

function deriveKeysBrowser(_x15, _x16) {
  return _deriveKeysBrowser.apply(this, arguments);
}

function _deriveKeysBrowser() {
  _deriveKeysBrowser = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5(key, name) {
    var hkdfkey, keybits, aesKey, hmacKey, aesProm, hmacProm;
    return _regenerator["default"].wrap(function _callee5$(_context5) {
      while (1) {
        switch (_context5.prev = _context5.next) {
          case 0:
            _context5.next = 2;
            return subtleCrypto.importKey('raw', key, {
              name: "HKDF"
            }, false, ["deriveBits"]);

          case 2:
            hkdfkey = _context5.sent;
            _context5.next = 5;
            return subtleCrypto.deriveBits({
              name: "HKDF",
              salt: zerosalt,
              info: new TextEncoder().encode(name),
              hash: "SHA-256"
            }, hkdfkey, 512);

          case 5:
            keybits = _context5.sent;
            aesKey = keybits.slice(0, 32);
            hmacKey = keybits.slice(32);
            aesProm = subtleCrypto.importKey('raw', aesKey, {
              name: 'AES-CTR'
            }, false, ['encrypt', 'decrypt']);
            hmacProm = subtleCrypto.importKey('raw', hmacKey, {
              name: 'HMAC',
              hash: {
                name: 'SHA-256'
              }
            }, false, ['sign', 'verify']);
            _context5.next = 12;
            return Promise.all([aesProm, hmacProm]);

          case 12:
            return _context5.abrupt("return", _context5.sent);

          case 13:
          case "end":
            return _context5.stop();
        }
      }
    }, _callee5);
  }));
  return _deriveKeysBrowser.apply(this, arguments);
}

function encryptAES() {
  return subtleCrypto ? encryptBrowser.apply(void 0, arguments) : encryptNode.apply(void 0, arguments);
}

function decryptAES() {
  return subtleCrypto ? decryptBrowser.apply(void 0, arguments) : decryptNode.apply(void 0, arguments);
}