"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.registerAlgorithm = registerAlgorithm;
exports.UnknownDeviceError = exports.DecryptionError = exports.DecryptionAlgorithm = exports.EncryptionAlgorithm = exports.DECRYPTION_CLASSES = exports.ENCRYPTION_CLASSES = void 0;

var _assertThisInitialized2 = _interopRequireDefault(require("@babel/runtime/helpers/assertThisInitialized"));

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _wrapNativeSuper2 = _interopRequireDefault(require("@babel/runtime/helpers/wrapNativeSuper"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

/*
Copyright 2016 OpenMarket Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/**
 * Internal module. Defines the base classes of the encryption implementations
 *
 * @module
 */

/**
 * map of registered encryption algorithm classes. A map from string to {@link
 * module:crypto/algorithms/base.EncryptionAlgorithm|EncryptionAlgorithm} class
 *
 * @type {Object.<string, function(new: module:crypto/algorithms/base.EncryptionAlgorithm)>}
 */
var ENCRYPTION_CLASSES = {};
/**
 * map of registered encryption algorithm classes. Map from string to {@link
 * module:crypto/algorithms/base.DecryptionAlgorithm|DecryptionAlgorithm} class
 *
 * @type {Object.<string, function(new: module:crypto/algorithms/base.DecryptionAlgorithm)>}
 */

exports.ENCRYPTION_CLASSES = ENCRYPTION_CLASSES;
var DECRYPTION_CLASSES = {};
/**
 * base type for encryption implementations
 *
 * @alias module:crypto/algorithms/base.EncryptionAlgorithm
 *
 * @param {object} params parameters
 * @param {string} params.userId  The UserID for the local user
 * @param {string} params.deviceId The identifier for this device.
 * @param {module:crypto} params.crypto crypto core
 * @param {module:crypto/OlmDevice} params.olmDevice olm.js wrapper
 * @param {module:base-apis~MatrixBaseApis} baseApis base matrix api interface
 * @param {string} params.roomId  The ID of the room we will be sending to
 * @param {object} params.config  The body of the m.room.encryption event
 */

exports.DECRYPTION_CLASSES = DECRYPTION_CLASSES;

var EncryptionAlgorithm = /*#__PURE__*/function () {
  function EncryptionAlgorithm(params) {
    (0, _classCallCheck2["default"])(this, EncryptionAlgorithm);
    this._userId = params.userId;
    this._deviceId = params.deviceId;
    this._crypto = params.crypto;
    this._olmDevice = params.olmDevice;
    this._baseApis = params.baseApis;
    this._roomId = params.roomId;
  }
  /**
   * Perform any background tasks that can be done before a message is ready to
   * send, in order to speed up sending of the message.
   *
   * @param {module:models/room} room the room the event is in
   */


  (0, _createClass2["default"])(EncryptionAlgorithm, [{
    key: "prepareToEncrypt",
    value: function prepareToEncrypt(room) {}
    /**
     * Encrypt a message event
     *
     * @method module:crypto/algorithms/base.EncryptionAlgorithm.encryptMessage
     * @abstract
     *
     * @param {module:models/room} room
     * @param {string} eventType
     * @param {object} plaintext event content
     *
     * @return {Promise} Promise which resolves to the new event body
     */

    /**
     * Called when the membership of a member of the room changes.
     *
     * @param {module:models/event.MatrixEvent} event  event causing the change
     * @param {module:models/room-member} member  user whose membership changed
     * @param {string=} oldMembership  previous membership
     * @public
     */

  }, {
    key: "onRoomMembership",
    value: function onRoomMembership(event, member, oldMembership) {}
  }]);
  return EncryptionAlgorithm;
}();
/**
 * base type for decryption implementations
 *
 * @alias module:crypto/algorithms/base.DecryptionAlgorithm
 * @param {object} params parameters
 * @param {string} params.userId  The UserID for the local user
 * @param {module:crypto} params.crypto crypto core
 * @param {module:crypto/OlmDevice} params.olmDevice olm.js wrapper
 * @param {module:base-apis~MatrixBaseApis} baseApis base matrix api interface
 * @param {string=} params.roomId The ID of the room we will be receiving
 *     from. Null for to-device events.
 */


exports.EncryptionAlgorithm = EncryptionAlgorithm;

var DecryptionAlgorithm = /*#__PURE__*/function () {
  function DecryptionAlgorithm(params) {
    (0, _classCallCheck2["default"])(this, DecryptionAlgorithm);
    this._userId = params.userId;
    this._crypto = params.crypto;
    this._olmDevice = params.olmDevice;
    this._baseApis = params.baseApis;
    this._roomId = params.roomId;
  }
  /**
   * Decrypt an event
   *
   * @method module:crypto/algorithms/base.DecryptionAlgorithm#decryptEvent
   * @abstract
   *
   * @param {MatrixEvent} event undecrypted event
   *
   * @return {Promise<module:crypto~EventDecryptionResult>} promise which
   * resolves once we have finished decrypting. Rejects with an
   * `algorithms.DecryptionError` if there is a problem decrypting the event.
   */

  /**
   * Handle a key event
   *
   * @method module:crypto/algorithms/base.DecryptionAlgorithm#onRoomKeyEvent
   *
   * @param {module:models/event.MatrixEvent} params event key event
   */


  (0, _createClass2["default"])(DecryptionAlgorithm, [{
    key: "onRoomKeyEvent",
    value: function onRoomKeyEvent(params) {// ignore by default
    }
    /**
     * Import a room key
     *
     * @param {module:crypto/OlmDevice.MegolmSessionData} session
     */

  }, {
    key: "importRoomKey",
    value: function importRoomKey(session) {// ignore by default
    }
    /**
     * Determine if we have the keys necessary to respond to a room key request
     *
     * @param {module:crypto~IncomingRoomKeyRequest} keyRequest
     * @return {Promise<boolean>} true if we have the keys and could (theoretically) share
     *  them; else false.
     */

  }, {
    key: "hasKeysForKeyRequest",
    value: function hasKeysForKeyRequest(keyRequest) {
      return Promise.resolve(false);
    }
    /**
     * Send the response to a room key request
     *
     * @param {module:crypto~IncomingRoomKeyRequest} keyRequest
     */

  }, {
    key: "shareKeysWithDevice",
    value: function shareKeysWithDevice(keyRequest) {
      throw new Error("shareKeysWithDevice not supported for this DecryptionAlgorithm");
    }
    /**
     * Retry decrypting all the events from a sender that haven't been
     * decrypted yet.
     *
     * @param {string} senderKey the sender's key
     */

  }, {
    key: "retryDecryptionFromSender",
    value: function () {
      var _retryDecryptionFromSender = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(senderKey) {
        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }));

      function retryDecryptionFromSender(_x) {
        return _retryDecryptionFromSender.apply(this, arguments);
      }

      return retryDecryptionFromSender;
    }()
  }]);
  return DecryptionAlgorithm;
}();
/**
 * Exception thrown when decryption fails
 *
 * @alias module:crypto/algorithms/base.DecryptionError
 * @param {string} msg user-visible message describing the problem
 *
 * @param {Object=} details key/value pairs reported in the logs but not shown
 *   to the user.
 *
 * @extends Error
 */


exports.DecryptionAlgorithm = DecryptionAlgorithm;

var DecryptionError = /*#__PURE__*/function (_Error) {
  (0, _inherits2["default"])(DecryptionError, _Error);

  var _super = _createSuper(DecryptionError);

  function DecryptionError(code, msg, details) {
    var _this;

    (0, _classCallCheck2["default"])(this, DecryptionError);
    _this = _super.call(this, msg);
    _this.code = code;
    _this.name = 'DecryptionError';
    _this.detailedString = _detailedStringForDecryptionError((0, _assertThisInitialized2["default"])(_this), details);
    return _this;
  }

  return DecryptionError;
}( /*#__PURE__*/(0, _wrapNativeSuper2["default"])(Error));

exports.DecryptionError = DecryptionError;

function _detailedStringForDecryptionError(err, details) {
  var result = err.name + '[msg: ' + err.message;

  if (details) {
    result += ', ' + Object.keys(details).map(function (k) {
      return k + ': ' + details[k];
    }).join(', ');
  }

  result += ']';
  return result;
}
/**
 * Exception thrown specifically when we want to warn the user to consider
 * the security of their conversation before continuing
 *
 * @param {string} msg message describing the problem
 * @param {Object} devices userId -> {deviceId -> object}
 *      set of unknown devices per user we're warning about
 * @extends Error
 */


var UnknownDeviceError = /*#__PURE__*/function (_Error2) {
  (0, _inherits2["default"])(UnknownDeviceError, _Error2);

  var _super2 = _createSuper(UnknownDeviceError);

  function UnknownDeviceError(msg, devices) {
    var _this2;

    (0, _classCallCheck2["default"])(this, UnknownDeviceError);
    _this2 = _super2.call(this, msg);
    _this2.name = "UnknownDeviceError";
    _this2.devices = devices;
    return _this2;
  }

  return UnknownDeviceError;
}( /*#__PURE__*/(0, _wrapNativeSuper2["default"])(Error));
/**
 * Registers an encryption/decryption class for a particular algorithm
 *
 * @param {string} algorithm algorithm tag to register for
 *
 * @param {class} encryptor {@link
 *     module:crypto/algorithms/base.EncryptionAlgorithm|EncryptionAlgorithm}
 *     implementation
 *
 * @param {class} decryptor {@link
 *     module:crypto/algorithms/base.DecryptionAlgorithm|DecryptionAlgorithm}
 *     implementation
 */


exports.UnknownDeviceError = UnknownDeviceError;

function registerAlgorithm(algorithm, encryptor, decryptor) {
  ENCRYPTION_CLASSES[algorithm] = encryptor;
  DECRYPTION_CLASSES[algorithm] = decryptor;
}