"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.DehydrationManager = exports.DEHYDRATION_ALGORITHM = void 0;

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _olmlib = require("./olmlib");

var _indexeddbCryptoStore = require("../crypto/store/indexeddb-crypto-store");

var _aes = require("./aes");

var _anotherJson = _interopRequireDefault(require("another-json"));

var _logger = require("../logger");

/*
Copyright 2020 The Matrix.org Foundation C.I.C.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
var DEHYDRATION_ALGORITHM = "org.matrix.msc2697.v1.olm.libolm_pickle";
exports.DEHYDRATION_ALGORITHM = DEHYDRATION_ALGORITHM;
var oneweek = 7 * 24 * 60 * 60 * 1000;

var DehydrationManager = /*#__PURE__*/function () {
  function DehydrationManager(crypto) {
    (0, _classCallCheck2["default"])(this, DehydrationManager);
    this.crypto = crypto;
    (0, _defineProperty2["default"])(this, "inProgress", false);
    (0, _defineProperty2["default"])(this, "timeoutId", void 0);
    (0, _defineProperty2["default"])(this, "key", void 0);
    (0, _defineProperty2["default"])(this, "keyInfo", void 0);
    (0, _defineProperty2["default"])(this, "deviceDisplayName", void 0);
    this.getDehydrationKeyFromCache();
  }

  (0, _createClass2["default"])(DehydrationManager, [{
    key: "getDehydrationKeyFromCache",
    value: function () {
      var _getDehydrationKeyFromCache = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
        var _this = this;

        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _context2.next = 2;
                return this.crypto._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
                  _this.crypto._cryptoStore.getSecretStorePrivateKey(txn, /*#__PURE__*/function () {
                    var _ref = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(result) {
                      var key, keyInfo, deviceDisplayName, time, pickleKey, decrypted, now, delay;
                      return _regenerator["default"].wrap(function _callee$(_context) {
                        while (1) {
                          switch (_context.prev = _context.next) {
                            case 0:
                              if (!result) {
                                _context.next = 12;
                                break;
                              }

                              key = result.key, keyInfo = result.keyInfo, deviceDisplayName = result.deviceDisplayName, time = result.time;
                              pickleKey = Buffer.from(_this.crypto._olmDevice._pickleKey);
                              _context.next = 5;
                              return (0, _aes.decryptAES)(key, pickleKey, DEHYDRATION_ALGORITHM);

                            case 5:
                              decrypted = _context.sent;
                              _this.key = (0, _olmlib.decodeBase64)(decrypted);
                              _this.keyInfo = keyInfo;
                              _this.deviceDisplayName = deviceDisplayName;
                              now = Date.now();
                              delay = Math.max(1, time + oneweek - now);
                              _this.timeoutId = global.setTimeout(_this.dehydrateDevice.bind(_this), delay);

                            case 12:
                            case "end":
                              return _context.stop();
                          }
                        }
                      }, _callee);
                    }));

                    return function (_x) {
                      return _ref.apply(this, arguments);
                    };
                  }(), "dehydration");
                });

              case 2:
                return _context2.abrupt("return", _context2.sent);

              case 3:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function getDehydrationKeyFromCache() {
        return _getDehydrationKeyFromCache.apply(this, arguments);
      }

      return getDehydrationKeyFromCache;
    }()
    /** set the key, and queue periodic dehydration to the server in the background */

  }, {
    key: "setKeyAndQueueDehydration",
    value: function () {
      var _setKeyAndQueueDehydration = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(key) {
        var keyInfo,
            deviceDisplayName,
            matches,
            _args3 = arguments;
        return _regenerator["default"].wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                keyInfo = _args3.length > 1 && _args3[1] !== undefined ? _args3[1] : {};
                deviceDisplayName = _args3.length > 2 && _args3[2] !== undefined ? _args3[2] : undefined;
                _context3.next = 4;
                return this.setKey(key, keyInfo, deviceDisplayName);

              case 4:
                matches = _context3.sent;

                if (!matches) {
                  // start dehydration in the background
                  this.dehydrateDevice();
                }

              case 6:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function setKeyAndQueueDehydration(_x2) {
        return _setKeyAndQueueDehydration.apply(this, arguments);
      }

      return setKeyAndQueueDehydration;
    }()
  }, {
    key: "setKey",
    value: function () {
      var _setKey = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4(key) {
        var _this2 = this;

        var keyInfo,
            deviceDisplayName,
            matches,
            i,
            _args4 = arguments;
        return _regenerator["default"].wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                keyInfo = _args4.length > 1 && _args4[1] !== undefined ? _args4[1] : {};
                deviceDisplayName = _args4.length > 2 && _args4[2] !== undefined ? _args4[2] : undefined;

                if (key) {
                  _context4.next = 9;
                  break;
                }

                // unsetting the key -- cancel any pending dehydration task
                if (this.timeoutId) {
                  global.clearTimeout(this.timeoutId);
                  this.timeoutId = undefined;
                } // clear storage


                _context4.next = 6;
                return this.crypto._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
                  _this2.crypto._cryptoStore.storeSecretStorePrivateKey(txn, "dehydration", null);
                });

              case 6:
                this.key = undefined;
                this.keyInfo = undefined;
                return _context4.abrupt("return");

              case 9:
                // Check to see if it's the same key as before.  If it's different,
                // dehydrate a new device.  If it's the same, we can keep the same
                // device.  (Assume that keyInfo and deviceDisplayName will be the
                // same if the key is the same.)
                matches = this.key && key.length == this.key.length;

                for (i = 0; matches && i < key.length; i++) {
                  if (key[i] != this.key[i]) {
                    matches = false;
                  }
                }

                if (!matches) {
                  this.key = key;
                  this.keyInfo = keyInfo;
                  this.deviceDisplayName = deviceDisplayName;
                }

                return _context4.abrupt("return", matches);

              case 13:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this);
      }));

      function setKey(_x3) {
        return _setKey.apply(this, arguments);
      }

      return setKey;
    }()
    /** returns the device id of the newly created dehydrated device */

  }, {
    key: "dehydrateDevice",
    value: function () {
      var _dehydrateDevice = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5() {
        var _this3 = this;

        var _keys, pickleKey, key, account, e2eKeys, maxKeys, otks, fallbacks, pickledAccount, deviceData, dehydrateResult, deviceId, deviceKeys, deviceSignature, oneTimeKeys, _i, _Object$entries, _Object$entries$_i, keyId, _key, k, signature, fallbackKeys, _i2, _Object$entries2, _Object$entries2$_i, _keyId, _key2, _k, _signature;

        return _regenerator["default"].wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                if (!this.inProgress) {
                  _context5.next = 3;
                  break;
                }

                _logger.logger.log("Dehydration already in progress -- not starting new dehydration");

                return _context5.abrupt("return");

              case 3:
                this.inProgress = true;

                if (this.timeoutId) {
                  global.clearTimeout(this.timeoutId);
                  this.timeoutId = undefined;
                }

                _context5.prev = 5;
                pickleKey = Buffer.from(this.crypto._olmDevice._pickleKey); // update the crypto store with the timestamp

                _context5.next = 9;
                return (0, _aes.encryptAES)((0, _olmlib.encodeBase64)(this.key), pickleKey, DEHYDRATION_ALGORITHM);

              case 9:
                key = _context5.sent;
                _context5.next = 12;
                return this.crypto._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
                  _this3.crypto._cryptoStore.storeSecretStorePrivateKey(txn, "dehydration", {
                    keyInfo: _this3.keyInfo,
                    key: key,
                    deviceDisplayName: _this3.deviceDisplayName,
                    time: Date.now()
                  });
                });

              case 12:
                _logger.logger.log("Attempting to dehydrate device");

                _logger.logger.log("Creating account"); // create the account and all the necessary keys


                account = new global.Olm.Account();
                account.create();
                e2eKeys = JSON.parse(account.identity_keys());
                maxKeys = account.max_number_of_one_time_keys(); // FIXME: generate in small batches?

                account.generate_one_time_keys(maxKeys / 2);
                account.generate_fallback_key();
                otks = JSON.parse(account.one_time_keys());
                fallbacks = JSON.parse(account.fallback_key());
                account.mark_keys_as_published(); // dehydrate the account and store it on the server

                pickledAccount = account.pickle(new Uint8Array(this.key));
                deviceData = {
                  algorithm: DEHYDRATION_ALGORITHM,
                  account: pickledAccount
                };

                if (this.keyInfo.passphrase) {
                  deviceData.passphrase = this.keyInfo.passphrase;
                }

                _logger.logger.log("Uploading account to server");

                _context5.next = 29;
                return this.crypto._baseApis._http.authedRequest(undefined, "PUT", "/dehydrated_device", undefined, {
                  device_data: deviceData,
                  initial_device_display_name: this.deviceDisplayName
                }, {
                  prefix: "/_matrix/client/unstable/org.matrix.msc2697.v2"
                });

              case 29:
                dehydrateResult = _context5.sent;
                // send the keys to the server
                deviceId = dehydrateResult.device_id;

                _logger.logger.log("Preparing device keys", deviceId);

                deviceKeys = {
                  algorithms: this.crypto._supportedAlgorithms,
                  device_id: deviceId,
                  user_id: this.crypto._userId,
                  keys: (_keys = {}, (0, _defineProperty2["default"])(_keys, "ed25519:".concat(deviceId), e2eKeys.ed25519), (0, _defineProperty2["default"])(_keys, "curve25519:".concat(deviceId), e2eKeys.curve25519), _keys)
                };
                deviceSignature = account.sign(_anotherJson["default"].stringify(deviceKeys));
                deviceKeys.signatures = (0, _defineProperty2["default"])({}, this.crypto._userId, (0, _defineProperty2["default"])({}, "ed25519:".concat(deviceId), deviceSignature));

                if (!this.crypto._crossSigningInfo.getId("self_signing")) {
                  _context5.next = 38;
                  break;
                }

                _context5.next = 38;
                return this.crypto._crossSigningInfo.signObject(deviceKeys, "self_signing");

              case 38:
                _logger.logger.log("Preparing one-time keys");

                oneTimeKeys = {};

                for (_i = 0, _Object$entries = Object.entries(otks.curve25519); _i < _Object$entries.length; _i++) {
                  _Object$entries$_i = (0, _slicedToArray2["default"])(_Object$entries[_i], 2), keyId = _Object$entries$_i[0], _key = _Object$entries$_i[1];
                  k = {
                    key: _key
                  };
                  signature = account.sign(_anotherJson["default"].stringify(k));
                  k.signatures = (0, _defineProperty2["default"])({}, this.crypto._userId, (0, _defineProperty2["default"])({}, "ed25519:".concat(deviceId), signature));
                  oneTimeKeys["signed_curve25519:".concat(keyId)] = k;
                }

                _logger.logger.log("Preparing fallback keys");

                fallbackKeys = {};

                for (_i2 = 0, _Object$entries2 = Object.entries(fallbacks.curve25519); _i2 < _Object$entries2.length; _i2++) {
                  _Object$entries2$_i = (0, _slicedToArray2["default"])(_Object$entries2[_i2], 2), _keyId = _Object$entries2$_i[0], _key2 = _Object$entries2$_i[1];
                  _k = {
                    key: _key2,
                    fallback: true
                  };
                  _signature = account.sign(_anotherJson["default"].stringify(_k));
                  _k.signatures = (0, _defineProperty2["default"])({}, this.crypto._userId, (0, _defineProperty2["default"])({}, "ed25519:".concat(deviceId), _signature));
                  fallbackKeys["signed_curve25519:".concat(_keyId)] = _k;
                }

                _logger.logger.log("Uploading keys to server");

                _context5.next = 47;
                return this.crypto._baseApis._http.authedRequest(undefined, "POST", "/keys/upload/" + encodeURI(deviceId), undefined, {
                  "device_keys": deviceKeys,
                  "one_time_keys": oneTimeKeys,
                  "org.matrix.msc2732.fallback_keys": fallbackKeys
                });

              case 47:
                _logger.logger.log("Done dehydrating"); // dehydrate again in a week


                this.timeoutId = global.setTimeout(this.dehydrateDevice.bind(this), oneweek);
                return _context5.abrupt("return", deviceId);

              case 50:
                _context5.prev = 50;
                this.inProgress = false;
                return _context5.finish(50);

              case 53:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this, [[5,, 50, 53]]);
      }));

      function dehydrateDevice() {
        return _dehydrateDevice.apply(this, arguments);
      }

      return dehydrateDevice;
    }()
  }, {
    key: "stop",
    value: function stop() {
      if (this.timeoutId) {
        global.clearTimeout(this.timeoutId);
        this.timeoutId = undefined;
      }
    }
  }]);
  return DehydrationManager;
}();

exports.DehydrationManager = DehydrationManager;