"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.isCryptoAvailable = isCryptoAvailable;
exports.Crypto = Crypto;
exports.fixBackupKey = fixBackupKey;
exports.verificationMethods = void 0;

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _anotherJson = _interopRequireDefault(require("another-json"));

var _events = require("events");

var _ReEmitter = require("../ReEmitter");

var _logger = require("../logger");

var utils = _interopRequireWildcard(require("../utils"));

var _OlmDevice = require("./OlmDevice");

var olmlib = _interopRequireWildcard(require("./olmlib"));

var _DeviceList = require("./DeviceList");

var _deviceinfo = require("./deviceinfo");

var algorithms = _interopRequireWildcard(require("./algorithms"));

var _CrossSigning = require("./CrossSigning");

var _EncryptionSetup = require("./EncryptionSetup");

var _SecretStorage = require("./SecretStorage");

var _OutgoingRoomKeyRequestManager = require("./OutgoingRoomKeyRequestManager");

var _indexeddbCryptoStore = require("./store/indexeddb-crypto-store");

var _QRCode = require("./verification/QRCode");

var _SAS = require("./verification/SAS");

var _key_passphrase = require("./key_passphrase");

var _recoverykey = require("./recoverykey");

var _VerificationRequest = require("./verification/request/VerificationRequest");

var _InRoomChannel = require("./verification/request/InRoomChannel");

var _ToDeviceChannel = require("./verification/request/ToDeviceChannel");

var _IllegalMethod = require("./verification/IllegalMethod");

var _errors = require("../errors");

var _aes = require("./aes");

var _dehydration = require("./dehydration");

var _defaultVerificationM;

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var DeviceVerification = _deviceinfo.DeviceInfo.DeviceVerification;
var defaultVerificationMethods = (_defaultVerificationM = {}, (0, _defineProperty2["default"])(_defaultVerificationM, _QRCode.ReciprocateQRCode.NAME, _QRCode.ReciprocateQRCode), (0, _defineProperty2["default"])(_defaultVerificationM, _SAS.SAS.NAME, _SAS.SAS), (0, _defineProperty2["default"])(_defaultVerificationM, _QRCode.SHOW_QR_CODE_METHOD, _IllegalMethod.IllegalMethod), (0, _defineProperty2["default"])(_defaultVerificationM, _QRCode.SCAN_QR_CODE_METHOD, _IllegalMethod.IllegalMethod), _defaultVerificationM);
/**
 * verification method names
 */

var verificationMethods = {
  RECIPROCATE_QR_CODE: _QRCode.ReciprocateQRCode.NAME,
  SAS: _SAS.SAS.NAME
};
exports.verificationMethods = verificationMethods;

function isCryptoAvailable() {
  return Boolean(global.Olm);
}

var MIN_FORCE_SESSION_INTERVAL_MS = 60 * 60 * 1000;
var KEY_BACKUP_KEYS_PER_REQUEST = 200;
/**
 * Cryptography bits
 *
 * This module is internal to the js-sdk; the public API is via MatrixClient.
 *
 * @constructor
 * @alias module:crypto
 *
 * @internal
 *
 * @param {module:base-apis~MatrixBaseApis} baseApis base matrix api interface
 *
 * @param {module:store/session/webstorage~WebStorageSessionStore} sessionStore
 *    Store to be used for end-to-end crypto session data
 *
 * @param {string} userId The user ID for the local user
 *
 * @param {string} deviceId The identifier for this device.
 *
 * @param {Object} clientStore the MatrixClient data store.
 *
 * @param {module:crypto/store/base~CryptoStore} cryptoStore
 *    storage for the crypto layer.
 *
 * @param {RoomList} roomList An initialised RoomList object
 *
 * @param {Array} verificationMethods Array of verification methods to use.
 *    Each element can either be a string from MatrixClient.verificationMethods
 *    or a class that implements a verification method.
 */

function Crypto(baseApis, sessionStore, userId, deviceId, clientStore, cryptoStore, roomList, verificationMethods) {
  var _this = this;

  this._onDeviceListUserCrossSigningUpdated = this._onDeviceListUserCrossSigningUpdated.bind(this);
  this._trustCrossSignedDevices = true;
  this._reEmitter = new _ReEmitter.ReEmitter(this);
  this._baseApis = baseApis;
  this._sessionStore = sessionStore;
  this._userId = userId;
  this._deviceId = deviceId;
  this._clientStore = clientStore;
  this._cryptoStore = cryptoStore;
  this._roomList = roomList;

  if (verificationMethods) {
    this._verificationMethods = new Map();

    var _iterator = _createForOfIteratorHelper(verificationMethods),
        _step;

    try {
      for (_iterator.s(); !(_step = _iterator.n()).done;) {
        var method = _step.value;

        if (typeof method === "string") {
          if (defaultVerificationMethods[method]) {
            this._verificationMethods.set(method, defaultVerificationMethods[method]);
          }
        } else if (method.NAME) {
          this._verificationMethods.set(method.NAME, method);
        } else {
          _logger.logger.warn("Excluding unknown verification method ".concat(method));
        }
      }
    } catch (err) {
      _iterator.e(err);
    } finally {
      _iterator.f();
    }
  } else {
    this._verificationMethods = defaultVerificationMethods;
  } // track whether this device's megolm keys are being backed up incrementally
  // to the server or not.
  // XXX: this should probably have a single source of truth from OlmAccount


  this.backupInfo = null; // The info dict from /room_keys/version

  this.backupKey = null; // The encryption key object

  this._checkedForBackup = false; // Have we checked the server for a backup we can use?

  this._sendingBackups = false; // Are we currently sending backups?

  this._olmDevice = new _OlmDevice.OlmDevice(cryptoStore);
  this._deviceList = new _DeviceList.DeviceList(baseApis, cryptoStore, this._olmDevice); // XXX: This isn't removed at any point, but then none of the event listeners
  // this class sets seem to be removed at any point... :/

  this._deviceList.on('userCrossSigningUpdated', this._onDeviceListUserCrossSigningUpdated);

  this._reEmitter.reEmit(this._deviceList, ["crypto.devicesUpdated", "crypto.willUpdateDevices"]); // the last time we did a check for the number of one-time-keys on the
  // server.


  this._lastOneTimeKeyCheck = null;
  this._oneTimeKeyCheckInProgress = false; // EncryptionAlgorithm instance for each room

  this._roomEncryptors = {}; // map from algorithm to DecryptionAlgorithm instance, for each room

  this._roomDecryptors = {};
  this._supportedAlgorithms = utils.keys(algorithms.DECRYPTION_CLASSES);
  this._deviceKeys = {};
  this._globalBlacklistUnverifiedDevices = false;
  this._globalErrorOnUnknownDevices = true;
  this._outgoingRoomKeyRequestManager = new _OutgoingRoomKeyRequestManager.OutgoingRoomKeyRequestManager(baseApis, this._deviceId, this._cryptoStore); // list of IncomingRoomKeyRequests/IncomingRoomKeyRequestCancellations
  // we received in the current sync.

  this._receivedRoomKeyRequests = [];
  this._receivedRoomKeyRequestCancellations = []; // true if we are currently processing received room key requests

  this._processingRoomKeyRequests = false; // controls whether device tracking is delayed
  // until calling encryptEvent or trackRoomDevices,
  // or done immediately upon enabling room encryption.

  this._lazyLoadMembers = false; // in case _lazyLoadMembers is true,
  // track if an initial tracking of all the room members
  // has happened for a given room. This is delayed
  // to avoid loading room members as long as possible.

  this._roomDeviceTrackingState = {}; // The timestamp of the last time we forced establishment
  // of a new session for each device, in milliseconds.
  // {
  //     userId: {
  //         deviceId: 1234567890000,
  //     },
  // }

  this._lastNewSessionForced = {};
  this._toDeviceVerificationRequests = new _ToDeviceChannel.ToDeviceRequests();
  this._inRoomVerificationRequests = new _InRoomChannel.InRoomRequests(); // This flag will be unset whilst the client processes a sync response
  // so that we don't start requesting keys until we've actually finished
  // processing the response.

  this._sendKeyRequestsImmediately = false;
  var cryptoCallbacks = this._baseApis._cryptoCallbacks || {};
  var cacheCallbacks = (0, _CrossSigning.createCryptoStoreCacheCallbacks)(cryptoStore, this._olmDevice);
  this._crossSigningInfo = new _CrossSigning.CrossSigningInfo(userId, cryptoCallbacks, cacheCallbacks);
  this._secretStorage = new _SecretStorage.SecretStorage(baseApis, cryptoCallbacks);
  this._dehydrationManager = new _dehydration.DehydrationManager(this); // Assuming no app-supplied callback, default to getting from SSSS.

  if (!cryptoCallbacks.getCrossSigningKey && cryptoCallbacks.getSecretStorageKey) {
    cryptoCallbacks.getCrossSigningKey = /*#__PURE__*/function () {
      var _ref = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(type) {
        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                return _context.abrupt("return", _CrossSigning.CrossSigningInfo.getFromSecretStorage(type, _this._secretStorage));

              case 1:
              case "end":
                return _context.stop();
            }
          }
        }, _callee);
      }));

      return function (_x) {
        return _ref.apply(this, arguments);
      };
    }();
  }
}

utils.inherits(Crypto, _events.EventEmitter);
/**
 * Initialise the crypto module so that it is ready for use
 *
 * Returns a promise which resolves once the crypto module is ready for use.
 *
 * @param {Object} opts keyword arguments.
 * @param {string} opts.exportedOlmDevice (Optional) data from exported device
 *     that must be re-created.
 */

Crypto.prototype.init = /*#__PURE__*/function () {
  var _ref2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(opts) {
    var _this2 = this;

    var _ref3, exportedOlmDevice, pickleKey, myDevices, deviceInfo;

    return _regenerator["default"].wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            _ref3 = opts || {}, exportedOlmDevice = _ref3.exportedOlmDevice, pickleKey = _ref3.pickleKey;

            _logger.logger.log("Crypto: initialising Olm...");

            _context2.next = 4;
            return global.Olm.init();

          case 4:
            _logger.logger.log(exportedOlmDevice ? "Crypto: initialising Olm device from exported device..." : "Crypto: initialising Olm device...");

            _context2.next = 7;
            return this._olmDevice.init({
              fromExportedDevice: exportedOlmDevice,
              pickleKey: pickleKey
            });

          case 7:
            _logger.logger.log("Crypto: loading device list...");

            _context2.next = 10;
            return this._deviceList.load();

          case 10:
            // build our device keys: these will later be uploaded
            this._deviceKeys["ed25519:" + this._deviceId] = this._olmDevice.deviceEd25519Key;
            this._deviceKeys["curve25519:" + this._deviceId] = this._olmDevice.deviceCurve25519Key;

            _logger.logger.log("Crypto: fetching own devices...");

            myDevices = this._deviceList.getRawStoredDevicesForUser(this._userId);

            if (!myDevices) {
              myDevices = {};
            }

            if (!myDevices[this._deviceId]) {
              // add our own deviceinfo to the cryptoStore
              _logger.logger.log("Crypto: adding this device to the store...");

              deviceInfo = {
                keys: this._deviceKeys,
                algorithms: this._supportedAlgorithms,
                verified: DeviceVerification.VERIFIED,
                known: true
              };
              myDevices[this._deviceId] = deviceInfo;

              this._deviceList.storeDevicesForUser(this._userId, myDevices);

              this._deviceList.saveIfDirty();
            }

            _context2.next = 18;
            return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
              _this2._cryptoStore.getCrossSigningKeys(txn, function (keys) {
                // can be an empty object after resetting cross-signing keys, see _storeTrustedSelfKeys
                if (keys && Object.keys(keys).length !== 0) {
                  _logger.logger.log("Loaded cross-signing public keys from crypto store");

                  _this2._crossSigningInfo.setKeys(keys);
                }
              });
            });

          case 18:
            // make sure we are keeping track of our own devices
            // (this is important for key backups & things)
            this._deviceList.startTrackingDeviceList(this._userId);

            _logger.logger.log("Crypto: checking for key backup...");

            this._checkAndStartKeyBackup();

          case 21:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2, this);
  }));

  return function (_x2) {
    return _ref2.apply(this, arguments);
  };
}();
/**
 * Whether to trust a others users signatures of their devices.
 * If false, devices will only be considered 'verified' if we have
 * verified that device individually (effectively disabling cross-signing).
 *
 * Default: true
 *
 * @return {bool} True if trusting cross-signed devices
 */


Crypto.prototype.getCryptoTrustCrossSignedDevices = function () {
  return this._trustCrossSignedDevices;
};
/**
 * See getCryptoTrustCrossSignedDevices

 * This may be set before initCrypto() is called to ensure no races occur.
 *
 * @param {bool} val True to trust cross-signed devices
 */


Crypto.prototype.setCryptoTrustCrossSignedDevices = function (val) {
  this._trustCrossSignedDevices = val;

  var _iterator2 = _createForOfIteratorHelper(this._deviceList.getKnownUserIds()),
      _step2;

  try {
    for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
      var userId = _step2.value;

      var devices = this._deviceList.getRawStoredDevicesForUser(userId);

      for (var _i = 0, _Object$keys = Object.keys(devices); _i < _Object$keys.length; _i++) {
        var deviceId = _Object$keys[_i];
        var deviceTrust = this.checkDeviceTrust(userId, deviceId); // If the device is locally verified then isVerified() is always true,
        // so this will only have caused the value to change if the device is
        // cross-signing verified but not locally verified

        if (!deviceTrust.isLocallyVerified() && deviceTrust.isCrossSigningVerified()) {
          var deviceObj = this._deviceList.getStoredDevice(userId, deviceId);

          this.emit("deviceVerificationChanged", userId, deviceId, deviceObj);
        }
      }
    }
  } catch (err) {
    _iterator2.e(err);
  } finally {
    _iterator2.f();
  }
};
/**
 * Create a recovery key from a user-supplied passphrase.
 *
 * @param {string} password Passphrase string that can be entered by the user
 *     when restoring the backup as an alternative to entering the recovery key.
 *     Optional.
 * @returns {Promise<Object>} Object with public key metadata, encoded private
 *     recovery key which should be disposed of after displaying to the user,
 *     and raw private key to avoid round tripping if needed.
 */


Crypto.prototype.createRecoveryKeyFromPassphrase = /*#__PURE__*/function () {
  var _ref4 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(password) {
    var decryption, keyInfo, derivation, privateKey, encodedPrivateKey;
    return _regenerator["default"].wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            decryption = new global.Olm.PkDecryption();
            _context3.prev = 1;
            keyInfo = {};

            if (!password) {
              _context3.next = 11;
              break;
            }

            _context3.next = 6;
            return (0, _key_passphrase.keyFromPassphrase)(password);

          case 6:
            derivation = _context3.sent;
            keyInfo.passphrase = {
              algorithm: "m.pbkdf2",
              iterations: derivation.iterations,
              salt: derivation.salt
            };
            keyInfo.pubkey = decryption.init_with_private_key(derivation.key);
            _context3.next = 12;
            break;

          case 11:
            keyInfo.pubkey = decryption.generate_key();

          case 12:
            privateKey = decryption.get_private_key();
            encodedPrivateKey = (0, _recoverykey.encodeRecoveryKey)(privateKey);
            return _context3.abrupt("return", {
              keyInfo: keyInfo,
              encodedPrivateKey: encodedPrivateKey,
              privateKey: privateKey
            });

          case 15:
            _context3.prev = 15;
            if (decryption) decryption.free();
            return _context3.finish(15);

          case 18:
          case "end":
            return _context3.stop();
        }
      }
    }, _callee3, null, [[1,, 15, 18]]);
  }));

  return function (_x3) {
    return _ref4.apply(this, arguments);
  };
}();
/**
 * Checks whether cross signing:
 * - is enabled on this account and trusted by this device
 * - has private keys either cached locally or stored in secret storage
 *
 * If this function returns false, bootstrapCrossSigning() can be used
 * to fix things such that it returns true. That is to say, after
 * bootstrapCrossSigning() completes successfully, this function should
 * return true.
 *
 * The cross-signing API is currently UNSTABLE and may change without notice.
 *
 * @return {bool} True if cross-signing is ready to be used on this device
 */


Crypto.prototype.isCrossSigningReady = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4() {
  var publicKeysOnDevice, privateKeysExistSomewhere;
  return _regenerator["default"].wrap(function _callee4$(_context4) {
    while (1) {
      switch (_context4.prev = _context4.next) {
        case 0:
          publicKeysOnDevice = this._crossSigningInfo.getId();
          _context4.next = 3;
          return this._crossSigningInfo.isStoredInKeyCache();

        case 3:
          _context4.t0 = _context4.sent;

          if (_context4.t0) {
            _context4.next = 8;
            break;
          }

          _context4.next = 7;
          return this._crossSigningInfo.isStoredInSecretStorage(this._secretStorage);

        case 7:
          _context4.t0 = _context4.sent;

        case 8:
          privateKeysExistSomewhere = _context4.t0;
          return _context4.abrupt("return", !!(publicKeysOnDevice && privateKeysExistSomewhere));

        case 10:
        case "end":
          return _context4.stop();
      }
    }
  }, _callee4, this);
}));
/**
 * Checks whether secret storage:
 * - is enabled on this account
 * - is storing cross-signing private keys
 * - is storing session backup key (if enabled)
 *
 * If this function returns false, bootstrapSecretStorage() can be used
 * to fix things such that it returns true. That is to say, after
 * bootstrapSecretStorage() completes successfully, this function should
 * return true.
 *
 * The Secure Secret Storage API is currently UNSTABLE and may change without notice.
 *
 * @return {bool} True if secret storage is ready to be used on this device
 */

Crypto.prototype.isSecretStorageReady = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5() {
  var secretStorageKeyInAccount, privateKeysInStorage, sessionBackupInStorage;
  return _regenerator["default"].wrap(function _callee5$(_context5) {
    while (1) {
      switch (_context5.prev = _context5.next) {
        case 0:
          _context5.next = 2;
          return this._secretStorage.hasKey();

        case 2:
          secretStorageKeyInAccount = _context5.sent;
          _context5.next = 5;
          return this._crossSigningInfo.isStoredInSecretStorage(this._secretStorage);

        case 5:
          privateKeysInStorage = _context5.sent;
          sessionBackupInStorage = !this._baseApis.getKeyBackupEnabled() || this._baseApis.isKeyBackupKeyStored();
          return _context5.abrupt("return", !!(secretStorageKeyInAccount && privateKeysInStorage && sessionBackupInStorage));

        case 8:
        case "end":
          return _context5.stop();
      }
    }
  }, _callee5, this);
}));
/**
 * Bootstrap cross-signing by creating keys if needed. If everything is already
 * set up, then no changes are made, so this is safe to run to ensure
 * cross-signing is ready for use.
 *
 * This function:
 * - creates new cross-signing keys if they are not found locally cached nor in
 *   secret storage (if it has been setup)
 *
 * The cross-signing API is currently UNSTABLE and may change without notice.
 *
 * @param {function} opts.authUploadDeviceSigningKeys Function
 * called to await an interactive auth flow when uploading device signing keys.
 * @param {bool} [opts.setupNewCrossSigning] Optional. Reset even if keys
 * already exist.
 * Args:
 *     {function} A function that makes the request requiring auth. Receives the
 *     auth data as an object. Can be called multiple times, first with an empty
 *     authDict, to obtain the flows.
 */

Crypto.prototype.bootstrapCrossSigning = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee7() {
  var _this3 = this;

  var _ref8,
      authUploadDeviceSigningKeys,
      setupNewCrossSigning,
      delegateCryptoCallbacks,
      builder,
      crossSigningInfo,
      resetCrossSigning,
      publicKeysOnDevice,
      privateKeysInCache,
      privateKeysInStorage,
      privateKeysExistSomewhere,
      crossSigningPrivateKeys,
      secretStorage,
      operation,
      _args7 = arguments;

  return _regenerator["default"].wrap(function _callee7$(_context7) {
    while (1) {
      switch (_context7.prev = _context7.next) {
        case 0:
          _ref8 = _args7.length > 0 && _args7[0] !== undefined ? _args7[0] : {}, authUploadDeviceSigningKeys = _ref8.authUploadDeviceSigningKeys, setupNewCrossSigning = _ref8.setupNewCrossSigning;

          _logger.logger.log("Bootstrapping cross-signing");

          delegateCryptoCallbacks = this._baseApis._cryptoCallbacks;
          builder = new _EncryptionSetup.EncryptionSetupBuilder(this._baseApis.store.accountData, delegateCryptoCallbacks);
          crossSigningInfo = new _CrossSigning.CrossSigningInfo(this._userId, builder.crossSigningCallbacks, builder.crossSigningCallbacks); // Reset the cross-signing keys

          resetCrossSigning = /*#__PURE__*/function () {
            var _ref9 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6() {
              var device, deviceSignature;
              return _regenerator["default"].wrap(function _callee6$(_context6) {
                while (1) {
                  switch (_context6.prev = _context6.next) {
                    case 0:
                      crossSigningInfo.resetKeys(); // Sign master key with device key

                      _context6.next = 3;
                      return _this3._signObject(crossSigningInfo.keys.master);

                    case 3:
                      // Store auth flow helper function, as we need to call it when uploading
                      // to ensure we handle auth errors properly.
                      builder.addCrossSigningKeys(authUploadDeviceSigningKeys, crossSigningInfo.keys); // Cross-sign own device

                      device = _this3._deviceList.getStoredDevice(_this3._userId, _this3._deviceId);
                      _context6.next = 7;
                      return crossSigningInfo.signDevice(_this3._userId, device);

                    case 7:
                      deviceSignature = _context6.sent;
                      builder.addKeySignature(_this3._userId, _this3._deviceId, deviceSignature); // Sign message key backup with cross-signing master key

                      if (!_this3.backupInfo) {
                        _context6.next = 13;
                        break;
                      }

                      _context6.next = 12;
                      return crossSigningInfo.signObject(_this3.backupInfo.auth_data, "master");

                    case 12:
                      builder.addSessionBackup(_this3.backupInfo);

                    case 13:
                    case "end":
                      return _context6.stop();
                  }
                }
              }, _callee6);
            }));

            return function resetCrossSigning() {
              return _ref9.apply(this, arguments);
            };
          }();

          publicKeysOnDevice = this._crossSigningInfo.getId();
          _context7.next = 9;
          return this._crossSigningInfo.isStoredInKeyCache();

        case 9:
          privateKeysInCache = _context7.sent;
          _context7.next = 12;
          return this._crossSigningInfo.isStoredInSecretStorage(this._secretStorage);

        case 12:
          privateKeysInStorage = _context7.sent;
          privateKeysExistSomewhere = privateKeysInCache || privateKeysInStorage; // Log all relevant state for easier parsing of debug logs.

          _logger.logger.log({
            setupNewCrossSigning: setupNewCrossSigning,
            publicKeysOnDevice: publicKeysOnDevice,
            privateKeysInCache: privateKeysInCache,
            privateKeysInStorage: privateKeysInStorage,
            privateKeysExistSomewhere: privateKeysExistSomewhere
          });

          if (!(!privateKeysExistSomewhere || setupNewCrossSigning)) {
            _context7.next = 21;
            break;
          }

          _logger.logger.log("Cross-signing private keys not found locally or in secret storage, " + "creating new keys"); // If a user has multiple devices, it important to only call bootstrap
          // as part of some UI flow (and not silently during startup), as they
          // may have setup cross-signing on a platform which has not saved keys
          // to secret storage, and this would reset them. In such a case, you
          // should prompt the user to verify any existing devices first (and
          // request private keys from those devices) before calling bootstrap.


          _context7.next = 19;
          return resetCrossSigning();

        case 19:
          _context7.next = 29;
          break;

        case 21:
          if (!(publicKeysOnDevice && privateKeysInCache)) {
            _context7.next = 25;
            break;
          }

          _logger.logger.log("Cross-signing public keys trusted and private keys found locally");

          _context7.next = 29;
          break;

        case 25:
          if (!privateKeysInStorage) {
            _context7.next = 29;
            break;
          }

          _logger.logger.log("Cross-signing private keys not found locally, but they are available " + "in secret storage, reading storage and caching locally");

          _context7.next = 29;
          return this.checkOwnCrossSigningTrust();

        case 29:
          // Assuming no app-supplied callback, default to storing new private keys in
          // secret storage if it exists. If it does not, it is assumed this will be
          // done as part of setting up secret storage later.
          crossSigningPrivateKeys = builder.crossSigningCallbacks.privateKeys;

          if (!(crossSigningPrivateKeys.size && !this._baseApis._cryptoCallbacks.saveCrossSigningKeys)) {
            _context7.next = 38;
            break;
          }

          secretStorage = new _SecretStorage.SecretStorage(builder.accountDataClientAdapter, builder.ssssCryptoCallbacks);
          _context7.next = 34;
          return secretStorage.hasKey();

        case 34:
          if (!_context7.sent) {
            _context7.next = 38;
            break;
          }

          _logger.logger.log("Storing new cross-signing private keys in secret storage"); // This is writing to in-memory account data in
          // builder.accountDataClientAdapter so won't fail


          _context7.next = 38;
          return _CrossSigning.CrossSigningInfo.storeInSecretStorage(crossSigningPrivateKeys, secretStorage);

        case 38:
          operation = builder.buildOperation();
          _context7.next = 41;
          return operation.apply(this);

        case 41:
          _context7.next = 43;
          return builder.persist(this);

        case 43:
          _logger.logger.log("Cross-signing ready");

        case 44:
        case "end":
          return _context7.stop();
      }
    }
  }, _callee7, this);
}));
/**
 * Bootstrap Secure Secret Storage if needed by creating a default key. If everything is
 * already set up, then no changes are made, so this is safe to run to ensure secret
 * storage is ready for use.
 *
 * This function
 * - creates a new Secure Secret Storage key if no default key exists
 *   - if a key backup exists, it is migrated to store the key in the Secret
 *     Storage
 * - creates a backup if none exists, and one is requested
 * - migrates Secure Secret Storage to use the latest algorithm, if an outdated
 *   algorithm is found
 *
 * The Secure Secret Storage API is currently UNSTABLE and may change without notice.
 *
 * @param {function} [opts.createSecretStorageKey] Optional. Function
 * called to await a secret storage key creation flow.
 * Returns:
 *     {Promise<Object>} Object with public key metadata, encoded private
 *     recovery key which should be disposed of after displaying to the user,
 *     and raw private key to avoid round tripping if needed.
 * @param {object} [opts.keyBackupInfo] The current key backup object. If passed,
 * the passphrase and recovery key from this backup will be used.
 * @param {bool} [opts.setupNewKeyBackup] If true, a new key backup version will be
 * created and the private key stored in the new SSSS store. Ignored if keyBackupInfo
 * is supplied.
 * @param {bool} [opts.setupNewSecretStorage] Optional. Reset even if keys already exist.
 * @param {func} [opts.getKeyBackupPassphrase] Optional. Function called to get the user's
 *     current key backup passphrase. Should return a promise that resolves with a Buffer
 *     containing the key, or rejects if the key cannot be obtained.
 * Returns:
 *     {Promise} A promise which resolves to key creation data for
 *     SecretStorage#addKey: an object with `passphrase` etc fields.
 */

Crypto.prototype.bootstrapSecretStorage = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee12() {
  var _this4 = this;

  var _ref11,
      _ref11$createSecretSt,
      createSecretStorageKey,
      keyBackupInfo,
      setupNewKeyBackup,
      setupNewSecretStorage,
      getKeyBackupPassphrase,
      delegateCryptoCallbacks,
      builder,
      secretStorage,
      newKeyId,
      createSSSS,
      ensureCanCheckPassphrase,
      signKeyBackupWithCrossSigning,
      oldSSSSKey,
      _ref16,
      _ref17,
      oldKeyId,
      oldKeyInfo,
      storageExists,
      _yield$createSecretSt,
      keyInfo,
      privateKey,
      backupKey,
      opts,
      crossSigningPrivateKeys,
      info,
      _privateKey,
      data,
      sessionBackupKey,
      fixedBackupKey,
      decodedBackupKey,
      operation,
      _args12 = arguments;

  return _regenerator["default"].wrap(function _callee12$(_context12) {
    while (1) {
      switch (_context12.prev = _context12.next) {
        case 0:
          _ref11 = _args12.length > 0 && _args12[0] !== undefined ? _args12[0] : {}, _ref11$createSecretSt = _ref11.createSecretStorageKey, createSecretStorageKey = _ref11$createSecretSt === void 0 ? /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee8() {
            return _regenerator["default"].wrap(function _callee8$(_context8) {
              while (1) {
                switch (_context8.prev = _context8.next) {
                  case 0:
                    return _context8.abrupt("return", {});

                  case 1:
                  case "end":
                    return _context8.stop();
                }
              }
            }, _callee8);
          })) : _ref11$createSecretSt, keyBackupInfo = _ref11.keyBackupInfo, setupNewKeyBackup = _ref11.setupNewKeyBackup, setupNewSecretStorage = _ref11.setupNewSecretStorage, getKeyBackupPassphrase = _ref11.getKeyBackupPassphrase;

          _logger.logger.log("Bootstrapping Secure Secret Storage");

          delegateCryptoCallbacks = this._baseApis._cryptoCallbacks;
          builder = new _EncryptionSetup.EncryptionSetupBuilder(this._baseApis.store.accountData, delegateCryptoCallbacks);
          secretStorage = new _SecretStorage.SecretStorage(builder.accountDataClientAdapter, builder.ssssCryptoCallbacks); // the ID of the new SSSS key, if we create one

          newKeyId = null; // create a new SSSS key and set it as default

          createSSSS = /*#__PURE__*/function () {
            var _ref13 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee9(opts, privateKey) {
              var _yield$secretStorage$, keyId, keyInfo;

              return _regenerator["default"].wrap(function _callee9$(_context9) {
                while (1) {
                  switch (_context9.prev = _context9.next) {
                    case 0:
                      opts = opts || {};

                      if (privateKey) {
                        opts.key = privateKey;
                      }

                      _context9.next = 4;
                      return secretStorage.addKey(_SecretStorage.SECRET_STORAGE_ALGORITHM_V1_AES, opts);

                    case 4:
                      _yield$secretStorage$ = _context9.sent;
                      keyId = _yield$secretStorage$.keyId;
                      keyInfo = _yield$secretStorage$.keyInfo;

                      if (privateKey) {
                        // make the private key available to encrypt 4S secrets
                        builder.ssssCryptoCallbacks.addPrivateKey(keyId, keyInfo, privateKey);
                      }

                      _context9.next = 10;
                      return secretStorage.setDefaultKeyId(keyId);

                    case 10:
                      return _context9.abrupt("return", keyId);

                    case 11:
                    case "end":
                      return _context9.stop();
                  }
                }
              }, _callee9);
            }));

            return function createSSSS(_x4, _x5) {
              return _ref13.apply(this, arguments);
            };
          }();

          ensureCanCheckPassphrase = /*#__PURE__*/function () {
            var _ref14 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee10(keyId, keyInfo) {
              var key, privateKey, _yield$SecretStorage$, iv, mac;

              return _regenerator["default"].wrap(function _callee10$(_context10) {
                while (1) {
                  switch (_context10.prev = _context10.next) {
                    case 0:
                      if (keyInfo.mac) {
                        _context10.next = 16;
                        break;
                      }

                      _context10.next = 3;
                      return _this4._baseApis._cryptoCallbacks.getSecretStorageKey({
                        keys: (0, _defineProperty2["default"])({}, keyId, keyInfo)
                      }, "");

                    case 3:
                      key = _context10.sent;

                      if (!key) {
                        _context10.next = 16;
                        break;
                      }

                      privateKey = key[1];
                      builder.ssssCryptoCallbacks.addPrivateKey(keyId, keyInfo, privateKey);
                      _context10.next = 9;
                      return _SecretStorage.SecretStorage._calculateKeyCheck(privateKey);

                    case 9:
                      _yield$SecretStorage$ = _context10.sent;
                      iv = _yield$SecretStorage$.iv;
                      mac = _yield$SecretStorage$.mac;
                      keyInfo.iv = iv;
                      keyInfo.mac = mac;
                      _context10.next = 16;
                      return builder.setAccountData("m.secret_storage.key.".concat(keyId), keyInfo);

                    case 16:
                    case "end":
                      return _context10.stop();
                  }
                }
              }, _callee10);
            }));

            return function ensureCanCheckPassphrase(_x6, _x7) {
              return _ref14.apply(this, arguments);
            };
          }();

          signKeyBackupWithCrossSigning = /*#__PURE__*/function () {
            var _ref15 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee11(keyBackupAuthData) {
              return _regenerator["default"].wrap(function _callee11$(_context11) {
                while (1) {
                  switch (_context11.prev = _context11.next) {
                    case 0:
                      _context11.t0 = _this4._crossSigningInfo.getId();

                      if (!_context11.t0) {
                        _context11.next = 5;
                        break;
                      }

                      _context11.next = 4;
                      return _this4._crossSigningInfo.isStoredInKeyCache("master");

                    case 4:
                      _context11.t0 = _context11.sent;

                    case 5:
                      if (!_context11.t0) {
                        _context11.next = 17;
                        break;
                      }

                      _context11.prev = 6;

                      _logger.logger.log("Adding cross-signing signature to key backup");

                      _context11.next = 10;
                      return _this4._crossSigningInfo.signObject(keyBackupAuthData, "master");

                    case 10:
                      _context11.next = 15;
                      break;

                    case 12:
                      _context11.prev = 12;
                      _context11.t1 = _context11["catch"](6);

                      // This step is not critical (just helpful), so we catch here
                      // and continue if it fails.
                      _logger.logger.error("Signing key backup with cross-signing keys failed", _context11.t1);

                    case 15:
                      _context11.next = 18;
                      break;

                    case 17:
                      _logger.logger.warn("Cross-signing keys not available, skipping signature on key backup");

                    case 18:
                    case "end":
                      return _context11.stop();
                  }
                }
              }, _callee11, null, [[6, 12]]);
            }));

            return function signKeyBackupWithCrossSigning(_x8) {
              return _ref15.apply(this, arguments);
            };
          }();

          _context12.next = 11;
          return this.getSecretStorageKey();

        case 11:
          oldSSSSKey = _context12.sent;
          _ref16 = oldSSSSKey || [null, null], _ref17 = (0, _slicedToArray2["default"])(_ref16, 2), oldKeyId = _ref17[0], oldKeyInfo = _ref17[1];
          storageExists = !setupNewSecretStorage && oldKeyInfo && oldKeyInfo.algorithm === _SecretStorage.SECRET_STORAGE_ALGORITHM_V1_AES; // Log all relevant state for easier parsing of debug logs.

          _logger.logger.log({
            keyBackupInfo: keyBackupInfo,
            setupNewKeyBackup: setupNewKeyBackup,
            setupNewSecretStorage: setupNewSecretStorage,
            storageExists: storageExists,
            oldKeyInfo: oldKeyInfo
          });

          if (!(!storageExists && !keyBackupInfo)) {
            _context12.next = 27;
            break;
          }

          // either we don't have anything, or we've been asked to restart
          // from scratch
          _logger.logger.log("Secret storage does not exist, creating new storage key"); // if we already have a usable default SSSS key and aren't resetting
          // SSSS just use it. otherwise, create a new one
          // Note: we leave the old SSSS key in place: there could be other
          // secrets using it, in theory. We could move them to the new key but a)
          // that would mean we'd need to prompt for the old passphrase, and b)
          // it's not clear that would be the right thing to do anyway.


          _context12.next = 19;
          return createSecretStorageKey();

        case 19:
          _yield$createSecretSt = _context12.sent;
          keyInfo = _yield$createSecretSt.keyInfo;
          privateKey = _yield$createSecretSt.privateKey;
          _context12.next = 24;
          return createSSSS(keyInfo, privateKey);

        case 24:
          newKeyId = _context12.sent;
          _context12.next = 53;
          break;

        case 27:
          if (!(!storageExists && keyBackupInfo)) {
            _context12.next = 49;
            break;
          }

          // we have an existing backup, but no SSSS
          _logger.logger.log("Secret storage does not exist, using key backup key"); // if we have the backup key already cached, use it; otherwise use the
          // callback to prompt for the key


          _context12.next = 31;
          return this.getSessionBackupPrivateKey();

        case 31:
          _context12.t0 = _context12.sent;

          if (_context12.t0) {
            _context12.next = 36;
            break;
          }

          _context12.next = 35;
          return getKeyBackupPassphrase();

        case 35:
          _context12.t0 = _context12.sent;

        case 36:
          backupKey = _context12.t0;
          // create a new SSSS key and use the backup key as the new SSSS key
          opts = {};

          if (keyBackupInfo.auth_data.private_key_salt && keyBackupInfo.auth_data.private_key_iterations) {
            opts.passphrase = {
              algorithm: "m.pbkdf2",
              iterations: keyBackupInfo.auth_data.private_key_iterations,
              salt: keyBackupInfo.auth_data.private_key_salt,
              bits: 256
            };
          }

          _context12.next = 41;
          return createSSSS(opts, backupKey);

        case 41:
          newKeyId = _context12.sent;
          _context12.next = 44;
          return secretStorage.store("m.megolm_backup.v1", olmlib.encodeBase64(backupKey), [newKeyId]);

        case 44:
          _context12.next = 46;
          return signKeyBackupWithCrossSigning(keyBackupInfo.auth_data);

        case 46:
          builder.addSessionBackup(keyBackupInfo);
          _context12.next = 53;
          break;

        case 49:
          // 4S is already set up
          _logger.logger.log("Secret storage exists");

          if (!(oldKeyInfo && oldKeyInfo.algorithm === _SecretStorage.SECRET_STORAGE_ALGORITHM_V1_AES)) {
            _context12.next = 53;
            break;
          }

          _context12.next = 53;
          return ensureCanCheckPassphrase(oldKeyId, oldKeyInfo);

        case 53:
          _context12.t2 = !this._baseApis._cryptoCallbacks.saveCrossSigningKeys;

          if (!_context12.t2) {
            _context12.next = 58;
            break;
          }

          _context12.next = 57;
          return this.isCrossSigningReady();

        case 57:
          _context12.t2 = _context12.sent;

        case 58:
          _context12.t1 = _context12.t2;

          if (!_context12.t1) {
            _context12.next = 66;
            break;
          }

          _context12.t3 = newKeyId;

          if (_context12.t3) {
            _context12.next = 65;
            break;
          }

          _context12.next = 64;
          return this._crossSigningInfo.isStoredInSecretStorage(secretStorage);

        case 64:
          _context12.t3 = !_context12.sent;

        case 65:
          _context12.t1 = _context12.t3;

        case 66:
          if (!_context12.t1) {
            _context12.next = 73;
            break;
          }

          _logger.logger.log("Copying cross-signing private keys from cache to secret storage");

          _context12.next = 70;
          return this._crossSigningInfo.getCrossSigningKeysFromCache();

        case 70:
          crossSigningPrivateKeys = _context12.sent;
          _context12.next = 73;
          return _CrossSigning.CrossSigningInfo.storeInSecretStorage(crossSigningPrivateKeys, secretStorage);

        case 73:
          if (!(setupNewKeyBackup && !keyBackupInfo)) {
            _context12.next = 87;
            break;
          }

          _logger.logger.log("Creating new message key backup version");

          _context12.next = 77;
          return this._baseApis.prepareKeyBackupVersion(null
          /* random key */
          , // don't write to secret storage, as it will write to this._secretStorage.
          // Here, we want to capture all the side-effects of bootstrapping,
          // and want to write to the local secretStorage object
          {
            secureSecretStorage: false
          });

        case 77:
          info = _context12.sent;
          // write the key ourselves to 4S
          _privateKey = (0, _recoverykey.decodeRecoveryKey)(info.recovery_key);
          _context12.next = 81;
          return secretStorage.store("m.megolm_backup.v1", olmlib.encodeBase64(_privateKey));

        case 81:
          // create keyBackupInfo object to add to builder
          data = {
            algorithm: info.algorithm,
            auth_data: info.auth_data
          }; // Sign with cross-signing master key

          _context12.next = 84;
          return signKeyBackupWithCrossSigning(data.auth_data);

        case 84:
          _context12.next = 86;
          return this._signObject(data.auth_data);

        case 86:
          builder.addSessionBackup(data);

        case 87:
          _context12.next = 89;
          return secretStorage.get('m.megolm_backup.v1');

        case 89:
          sessionBackupKey = _context12.sent;

          if (!sessionBackupKey) {
            _context12.next = 99;
            break;
          }

          _logger.logger.info("Got session backup key from secret storage: caching"); // fix up the backup key if it's in the wrong format, and replace
          // in secret storage


          fixedBackupKey = fixBackupKey(sessionBackupKey);

          if (!fixedBackupKey) {
            _context12.next = 96;
            break;
          }

          _context12.next = 96;
          return secretStorage.store("m.megolm_backup.v1", fixedBackupKey, [newKeyId || oldKeyId]);

        case 96:
          decodedBackupKey = new Uint8Array(olmlib.decodeBase64(fixedBackupKey || sessionBackupKey));
          _context12.next = 99;
          return builder.addSessionBackupPrivateKeyToCache(decodedBackupKey);

        case 99:
          operation = builder.buildOperation();
          _context12.next = 102;
          return operation.apply(this);

        case 102:
          _context12.next = 104;
          return builder.persist(this);

        case 104:
          _logger.logger.log("Secure Secret Storage ready");

        case 105:
        case "end":
          return _context12.stop();
      }
    }
  }, _callee12, this);
}));
/**
 * Fix up the backup key, that may be in the wrong format due to a bug in a
 * migration step.  Some backup keys were stored as a comma-separated list of
 * integers, rather than a base64-encoded byte array.  If this function is
 * passed a string that looks like a list of integers rather than a base64
 * string, it will attempt to convert it to the right format.
 *
 * @param {string} key the key to check
 * @returns {null | string} If the key is in the wrong format, then the fixed
 * key will be returned. Otherwise null will be returned.
 *
 */

function fixBackupKey(key) {
  if (typeof key !== "string" || key.indexOf(",") < 0) {
    return null;
  }

  var fixedKey = Uint8Array.from(key.split(","), function (x) {
    return parseInt(x);
  });
  return olmlib.encodeBase64(fixedKey);
}

Crypto.prototype.addSecretStorageKey = function (algorithm, opts, keyID) {
  return this._secretStorage.addKey(algorithm, opts, keyID);
};

Crypto.prototype.hasSecretStorageKey = function (keyID) {
  return this._secretStorage.hasKey(keyID);
};

Crypto.prototype.getSecretStorageKey = function (keyID) {
  return this._secretStorage.getKey(keyID);
};

Crypto.prototype.storeSecret = function (name, secret, keys) {
  return this._secretStorage.store(name, secret, keys);
};

Crypto.prototype.getSecret = function (name) {
  return this._secretStorage.get(name);
};

Crypto.prototype.isSecretStored = function (name, checkKey) {
  return this._secretStorage.isStored(name, checkKey);
};

Crypto.prototype.requestSecret = function (name, devices) {
  if (!devices) {
    devices = Object.keys(this._deviceList.getRawStoredDevicesForUser(this._userId));
  }

  return this._secretStorage.request(name, devices);
};

Crypto.prototype.getDefaultSecretStorageKeyId = function () {
  return this._secretStorage.getDefaultKeyId();
};

Crypto.prototype.setDefaultSecretStorageKeyId = function (k) {
  return this._secretStorage.setDefaultKeyId(k);
};

Crypto.prototype.checkSecretStorageKey = function (key, info) {
  return this._secretStorage.checkKey(key, info);
};
/**
 * Checks that a given secret storage private key matches a given public key.
 * This can be used by the getSecretStorageKey callback to verify that the
 * private key it is about to supply is the one that was requested.
 *
 * @param {Uint8Array} privateKey The private key
 * @param {string} expectedPublicKey The public key
 * @returns {boolean} true if the key matches, otherwise false
 */


Crypto.prototype.checkSecretStoragePrivateKey = function (privateKey, expectedPublicKey) {
  var decryption = null;

  try {
    decryption = new global.Olm.PkDecryption();
    var gotPubkey = decryption.init_with_private_key(privateKey); // make sure it agrees with the given pubkey

    return gotPubkey === expectedPublicKey;
  } finally {
    if (decryption) decryption.free();
  }
};
/**
 * Fetches the backup private key, if cached
 * @returns {Promise} the key, if any, or null
 */


Crypto.prototype.getSessionBackupPrivateKey = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee13() {
  var _this5 = this;

  var key, pickleKey, decrypted;
  return _regenerator["default"].wrap(function _callee13$(_context13) {
    while (1) {
      switch (_context13.prev = _context13.next) {
        case 0:
          _context13.next = 2;
          return new Promise(function (resolve) {
            _this5._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
              _this5._cryptoStore.getSecretStorePrivateKey(txn, resolve, "m.megolm_backup.v1");
            });
          });

        case 2:
          key = _context13.sent;

          if (!(key && typeof key === "string")) {
            _context13.next = 7;
            break;
          }

          key = new Uint8Array(olmlib.decodeBase64(fixBackupKey(key) || key));
          _context13.next = 7;
          return this.storeSessionBackupPrivateKey(key);

        case 7:
          if (!(key && key.ciphertext)) {
            _context13.next = 13;
            break;
          }

          pickleKey = Buffer.from(this._olmDevice._pickleKey);
          _context13.next = 11;
          return (0, _aes.decryptAES)(key, pickleKey, "m.megolm_backup.v1");

        case 11:
          decrypted = _context13.sent;
          key = olmlib.decodeBase64(decrypted);

        case 13:
          return _context13.abrupt("return", key);

        case 14:
        case "end":
          return _context13.stop();
      }
    }
  }, _callee13, this);
}));
/**
 * Stores the session backup key to the cache
 * @param {Uint8Array} key the private key
 * @returns {Promise} so you can catch failures
 */

Crypto.prototype.storeSessionBackupPrivateKey = /*#__PURE__*/function () {
  var _ref19 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee14(key) {
    var _this6 = this;

    var pickleKey;
    return _regenerator["default"].wrap(function _callee14$(_context14) {
      while (1) {
        switch (_context14.prev = _context14.next) {
          case 0:
            if (key instanceof Uint8Array) {
              _context14.next = 2;
              break;
            }

            throw new Error("storeSessionBackupPrivateKey expects Uint8Array, got ".concat(key));

          case 2:
            pickleKey = Buffer.from(this._olmDevice._pickleKey);
            _context14.next = 5;
            return (0, _aes.encryptAES)(olmlib.encodeBase64(key), pickleKey, "m.megolm_backup.v1");

          case 5:
            key = _context14.sent;
            return _context14.abrupt("return", this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
              _this6._cryptoStore.storeSecretStorePrivateKey(txn, "m.megolm_backup.v1", key);
            }));

          case 7:
          case "end":
            return _context14.stop();
        }
      }
    }, _callee14, this);
  }));

  return function (_x9) {
    return _ref19.apply(this, arguments);
  };
}();
/**
 * Checks that a given cross-signing private key matches a given public key.
 * This can be used by the getCrossSigningKey callback to verify that the
 * private key it is about to supply is the one that was requested.
 *
 * @param {Uint8Array} privateKey The private key
 * @param {string} expectedPublicKey The public key
 * @returns {boolean} true if the key matches, otherwise false
 */


Crypto.prototype.checkCrossSigningPrivateKey = function (privateKey, expectedPublicKey) {
  var signing = null;

  try {
    signing = new global.Olm.PkSigning();
    var gotPubkey = signing.init_with_seed(privateKey); // make sure it agrees with the given pubkey

    return gotPubkey === expectedPublicKey;
  } finally {
    if (signing) signing.free();
  }
};
/**
 * Run various follow-up actions after cross-signing keys have changed locally
 * (either by resetting the keys for the account or by getting them from secret
 * storage), such as signing the current device, upgrading device
 * verifications, etc.
 */


Crypto.prototype._afterCrossSigningLocalKeyChange = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee15() {
  var _this7 = this;

  var device, signedDevice, upload, shouldUpgradeCb, users, _i2, _Object$entries, _Object$entries$_i, userId, crossSigningInfo, upgradeInfo, usersToUpgrade, _iterator3, _step3, _userId;

  return _regenerator["default"].wrap(function _callee15$(_context15) {
    while (1) {
      switch (_context15.prev = _context15.next) {
        case 0:
          _logger.logger.info("Starting cross-signing key change post-processing"); // sign the current device with the new key, and upload to the server


          device = this._deviceList.getStoredDevice(this._userId, this._deviceId);
          _context15.next = 4;
          return this._crossSigningInfo.signDevice(this._userId, device);

        case 4:
          signedDevice = _context15.sent;

          _logger.logger.info("Starting background key sig upload for ".concat(this._deviceId));

          upload = function upload(_ref21) {
            var shouldEmit = _ref21.shouldEmit;
            return _this7._baseApis.uploadKeySignatures((0, _defineProperty2["default"])({}, _this7._userId, (0, _defineProperty2["default"])({}, _this7._deviceId, signedDevice))).then(function (response) {
              var _ref22 = response || {},
                  failures = _ref22.failures;

              if (Object.keys(failures || []).length > 0) {
                if (shouldEmit) {
                  _this7._baseApis.emit("crypto.keySignatureUploadFailure", failures, "_afterCrossSigningLocalKeyChange", upload // continuation
                  );
                }

                throw new _errors.KeySignatureUploadError("Key upload failed", {
                  failures: failures
                });
              }

              _logger.logger.info("Finished background key sig upload for ".concat(_this7._deviceId));
            })["catch"](function (e) {
              _logger.logger.error("Error during background key sig upload for ".concat(_this7._deviceId), e);
            });
          };

          upload({
            shouldEmit: true
          });
          shouldUpgradeCb = this._baseApis._cryptoCallbacks.shouldUpgradeDeviceVerifications;

          if (!shouldUpgradeCb) {
            _context15.next = 53;
            break;
          }

          _logger.logger.info("Starting device verification upgrade"); // Check all users for signatures if upgrade callback present
          // FIXME: do this in batches


          users = {};
          _i2 = 0, _Object$entries = Object.entries(this._deviceList._crossSigningInfo);

        case 13:
          if (!(_i2 < _Object$entries.length)) {
            _context15.next = 22;
            break;
          }

          _Object$entries$_i = (0, _slicedToArray2["default"])(_Object$entries[_i2], 2), userId = _Object$entries$_i[0], crossSigningInfo = _Object$entries$_i[1];
          _context15.next = 17;
          return this._checkForDeviceVerificationUpgrade(userId, _CrossSigning.CrossSigningInfo.fromStorage(crossSigningInfo, userId));

        case 17:
          upgradeInfo = _context15.sent;

          if (upgradeInfo) {
            users[userId] = upgradeInfo;
          }

        case 19:
          _i2++;
          _context15.next = 13;
          break;

        case 22:
          if (!(Object.keys(users).length > 0)) {
            _context15.next = 52;
            break;
          }

          _logger.logger.info("Found ".concat(Object.keys(users).length, " verif users to upgrade"));

          _context15.prev = 24;
          _context15.next = 27;
          return shouldUpgradeCb({
            users: users
          });

        case 27:
          usersToUpgrade = _context15.sent;

          if (!usersToUpgrade) {
            _context15.next = 47;
            break;
          }

          _iterator3 = _createForOfIteratorHelper(usersToUpgrade);
          _context15.prev = 30;

          _iterator3.s();

        case 32:
          if ((_step3 = _iterator3.n()).done) {
            _context15.next = 39;
            break;
          }

          _userId = _step3.value;

          if (!(_userId in users)) {
            _context15.next = 37;
            break;
          }

          _context15.next = 37;
          return this._baseApis.setDeviceVerified(_userId, users[_userId].crossSigningInfo.getId());

        case 37:
          _context15.next = 32;
          break;

        case 39:
          _context15.next = 44;
          break;

        case 41:
          _context15.prev = 41;
          _context15.t0 = _context15["catch"](30);

          _iterator3.e(_context15.t0);

        case 44:
          _context15.prev = 44;

          _iterator3.f();

          return _context15.finish(44);

        case 47:
          _context15.next = 52;
          break;

        case 49:
          _context15.prev = 49;
          _context15.t1 = _context15["catch"](24);

          _logger.logger.log("shouldUpgradeDeviceVerifications threw an error: not upgrading", _context15.t1);

        case 52:
          _logger.logger.info("Finished device verification upgrade");

        case 53:
          _logger.logger.info("Finished cross-signing key change post-processing");

        case 54:
        case "end":
          return _context15.stop();
      }
    }
  }, _callee15, this, [[24, 49], [30, 41, 44, 47]]);
}));
/**
 * Check if a user's cross-signing key is a candidate for upgrading from device
 * verification.
 *
 * @param {string} userId the user whose cross-signing information is to be checked
 * @param {object} crossSigningInfo the cross-signing information to check
 */

Crypto.prototype._checkForDeviceVerificationUpgrade = /*#__PURE__*/function () {
  var _ref23 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee16(userId, crossSigningInfo) {
    var trustLevel, devices, deviceIds;
    return _regenerator["default"].wrap(function _callee16$(_context16) {
      while (1) {
        switch (_context16.prev = _context16.next) {
          case 0:
            // only upgrade if this is the first cross-signing key that we've seen for
            // them, and if their cross-signing key isn't already verified
            trustLevel = this._crossSigningInfo.checkUserTrust(crossSigningInfo);

            if (!(crossSigningInfo.firstUse && !trustLevel.verified)) {
              _context16.next = 8;
              break;
            }

            devices = this._deviceList.getRawStoredDevicesForUser(userId);
            _context16.next = 5;
            return this._checkForValidDeviceSignature(userId, crossSigningInfo.keys.master, devices);

          case 5:
            deviceIds = _context16.sent;

            if (!deviceIds.length) {
              _context16.next = 8;
              break;
            }

            return _context16.abrupt("return", {
              devices: deviceIds.map(function (deviceId) {
                return _deviceinfo.DeviceInfo.fromStorage(devices[deviceId], deviceId);
              }),
              crossSigningInfo: crossSigningInfo
            });

          case 8:
          case "end":
            return _context16.stop();
        }
      }
    }, _callee16, this);
  }));

  return function (_x10, _x11) {
    return _ref23.apply(this, arguments);
  };
}();
/**
 * Check if the cross-signing key is signed by a verified device.
 *
 * @param {string} userId the user ID whose key is being checked
 * @param {object} key the key that is being checked
 * @param {object} devices the user's devices.  Should be a map from device ID
 *     to device info
 */


Crypto.prototype._checkForValidDeviceSignature = /*#__PURE__*/function () {
  var _ref24 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee17(userId, key, devices) {
    var deviceIds, _i3, _Object$keys2, signame, _signame$split, _signame$split2, deviceId;

    return _regenerator["default"].wrap(function _callee17$(_context17) {
      while (1) {
        switch (_context17.prev = _context17.next) {
          case 0:
            deviceIds = [];

            if (!(devices && key.signatures && key.signatures[userId])) {
              _context17.next = 18;
              break;
            }

            _i3 = 0, _Object$keys2 = Object.keys(key.signatures[userId]);

          case 3:
            if (!(_i3 < _Object$keys2.length)) {
              _context17.next = 18;
              break;
            }

            signame = _Object$keys2[_i3];
            _signame$split = signame.split(':', 2), _signame$split2 = (0, _slicedToArray2["default"])(_signame$split, 2), deviceId = _signame$split2[1];

            if (!(deviceId in devices && devices[deviceId].verified === DeviceVerification.VERIFIED)) {
              _context17.next = 15;
              break;
            }

            _context17.prev = 7;
            _context17.next = 10;
            return olmlib.verifySignature(this._olmDevice, key, userId, deviceId, devices[deviceId].keys[signame]);

          case 10:
            deviceIds.push(deviceId);
            _context17.next = 15;
            break;

          case 13:
            _context17.prev = 13;
            _context17.t0 = _context17["catch"](7);

          case 15:
            _i3++;
            _context17.next = 3;
            break;

          case 18:
            return _context17.abrupt("return", deviceIds);

          case 19:
          case "end":
            return _context17.stop();
        }
      }
    }, _callee17, this, [[7, 13]]);
  }));

  return function (_x12, _x13, _x14) {
    return _ref24.apply(this, arguments);
  };
}();
/**
 * Get the user's cross-signing key ID.
 *
 * @param {string} [type=master] The type of key to get the ID of.  One of
 *     "master", "self_signing", or "user_signing".  Defaults to "master".
 *
 * @returns {string} the key ID
 */


Crypto.prototype.getCrossSigningId = function (type) {
  return this._crossSigningInfo.getId(type);
};
/**
 * Get the cross signing information for a given user.
 *
 * @param {string} userId the user ID to get the cross-signing info for.
 *
 * @returns {CrossSigningInfo} the cross signing informmation for the user.
 */


Crypto.prototype.getStoredCrossSigningForUser = function (userId) {
  return this._deviceList.getStoredCrossSigningForUser(userId);
};
/**
 * Check whether a given user is trusted.
 *
 * @param {string} userId The ID of the user to check.
 *
 * @returns {UserTrustLevel}
 */


Crypto.prototype.checkUserTrust = function (userId) {
  var userCrossSigning = this._deviceList.getStoredCrossSigningForUser(userId);

  if (!userCrossSigning) {
    return new _CrossSigning.UserTrustLevel(false, false, false);
  }

  return this._crossSigningInfo.checkUserTrust(userCrossSigning);
};
/**
 * Check whether a given device is trusted.
 *
 * @param {string} userId The ID of the user whose devices is to be checked.
 * @param {string} deviceId The ID of the device to check
 *
 * @returns {DeviceTrustLevel}
 */


Crypto.prototype.checkDeviceTrust = function (userId, deviceId) {
  var device = this._deviceList.getStoredDevice(userId, deviceId);

  return this._checkDeviceInfoTrust(userId, device);
};
/**
 * Check whether a given deviceinfo is trusted.
 *
 * @param {string} userId The ID of the user whose devices is to be checked.
 * @param {module:crypto/deviceinfo?} device The device info object to check
 *
 * @returns {DeviceTrustLevel}
 */


Crypto.prototype._checkDeviceInfoTrust = function (userId, device) {
  var trustedLocally = !!(device && device.isVerified());

  var userCrossSigning = this._deviceList.getStoredCrossSigningForUser(userId);

  if (device && userCrossSigning) {
    // The _trustCrossSignedDevices only affects trust of other people's cross-signing
    // signatures
    var trustCrossSig = this._trustCrossSignedDevices || userId === this._userId;
    return this._crossSigningInfo.checkDeviceTrust(userCrossSigning, device, trustedLocally, trustCrossSig);
  } else {
    return new _CrossSigning.DeviceTrustLevel(false, false, trustedLocally, false);
  }
};
/*
 * Event handler for DeviceList's userNewDevices event
 */


Crypto.prototype._onDeviceListUserCrossSigningUpdated = /*#__PURE__*/function () {
  var _ref25 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee18(userId) {
    var newCrossSigning, seenPubkey, currentPubkey, changed, crossSigning;
    return _regenerator["default"].wrap(function _callee18$(_context18) {
      while (1) {
        switch (_context18.prev = _context18.next) {
          case 0:
            if (!(userId === this._userId)) {
              _context18.next = 15;
              break;
            }

            // An update to our own cross-signing key.
            // Get the new key first:
            newCrossSigning = this._deviceList.getStoredCrossSigningForUser(userId);
            seenPubkey = newCrossSigning ? newCrossSigning.getId() : null;
            currentPubkey = this._crossSigningInfo.getId();
            changed = currentPubkey !== seenPubkey;

            if (!(currentPubkey && seenPubkey && !changed)) {
              _context18.next = 10;
              break;
            }

            _context18.next = 8;
            return this.checkOwnCrossSigningTrust();

          case 8:
            _context18.next = 13;
            break;

          case 10:
            // We'll now be in a state where cross-signing on the account is not trusted
            // because our locally stored cross-signing keys will not match the ones
            // on the server for our account. So we clear our own stored cross-signing keys,
            // effectively disabling cross-signing until the user gets verified by the device
            // that reset the keys
            this._storeTrustedSelfKeys(null); // emit cross-signing has been disabled


            this.emit("crossSigning.keysChanged", {}); // as the trust for our own user has changed,
            // also emit an event for this

            this.emit("userTrustStatusChanged", this._userId, this.checkUserTrust(userId));

          case 13:
            _context18.next = 20;
            break;

          case 15:
            _context18.next = 17;
            return this._checkDeviceVerifications(userId);

          case 17:
            // Update verified before latch using the current state and save the new
            // latch value in the device list store.
            crossSigning = this._deviceList.getStoredCrossSigningForUser(userId);

            if (crossSigning) {
              crossSigning.updateCrossSigningVerifiedBefore(this.checkUserTrust(userId).isCrossSigningVerified());

              this._deviceList.setRawStoredCrossSigningForUser(userId, crossSigning.toStorage());
            }

            this.emit("userTrustStatusChanged", userId, this.checkUserTrust(userId));

          case 20:
          case "end":
            return _context18.stop();
        }
      }
    }, _callee18, this);
  }));

  return function (_x15) {
    return _ref25.apply(this, arguments);
  };
}();
/**
 * Check the copy of our cross-signing key that we have in the device list and
 * see if we can get the private key. If so, mark it as trusted.
 */


Crypto.prototype.checkOwnCrossSigningTrust = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee19() {
  var _this8 = this;

  var userId, newCrossSigning, seenPubkey, masterChanged, signing, ret, oldSelfSigningId, oldUserSigningId, selfSigningChanged, userSigningChanged, keySignatures, _signing, _ret, device, _signedDevice, _signing2, _ret2, masterKey, deviceSig, keysToUpload, _upload;

  return _regenerator["default"].wrap(function _callee19$(_context19) {
    while (1) {
      switch (_context19.prev = _context19.next) {
        case 0:
          userId = this._userId; // Before proceeding, ensure our cross-signing public keys have been
          // downloaded via the device list.

          _context19.next = 3;
          return this.downloadKeys([this._userId]);

        case 3:
          // If we see an update to our own master key, check it against the master
          // key we have and, if it matches, mark it as verified
          // First, get the new cross-signing info
          newCrossSigning = this._deviceList.getStoredCrossSigningForUser(userId);

          if (newCrossSigning) {
            _context19.next = 7;
            break;
          }

          _logger.logger.error("Got cross-signing update event for user " + userId + " but no new cross-signing information found!");

          return _context19.abrupt("return");

        case 7:
          seenPubkey = newCrossSigning.getId();
          masterChanged = this._crossSigningInfo.getId() !== seenPubkey;

          if (!masterChanged) {
            _context19.next = 27;
            break;
          }

          _logger.logger.info("Got new master public key", seenPubkey);

          _logger.logger.info("Attempting to retrieve cross-signing master private key");

          signing = null;
          _context19.prev = 13;
          _context19.next = 16;
          return this._crossSigningInfo.getCrossSigningKey('master', seenPubkey);

        case 16:
          ret = _context19.sent;
          signing = ret[1];

          _logger.logger.info("Got cross-signing master private key");

          _context19.next = 24;
          break;

        case 21:
          _context19.prev = 21;
          _context19.t0 = _context19["catch"](13);

          _logger.logger.error("Cross-signing master private key not available", _context19.t0);

        case 24:
          _context19.prev = 24;
          if (signing) signing.free();
          return _context19.finish(24);

        case 27:
          oldSelfSigningId = this._crossSigningInfo.getId("self_signing");
          oldUserSigningId = this._crossSigningInfo.getId("user_signing"); // Update the version of our keys in our cross-signing object and the local store

          this._storeTrustedSelfKeys(newCrossSigning.keys);

          selfSigningChanged = oldSelfSigningId !== newCrossSigning.getId("self_signing");
          userSigningChanged = oldUserSigningId !== newCrossSigning.getId("user_signing");
          keySignatures = {};

          if (!selfSigningChanged) {
            _context19.next = 56;
            break;
          }

          _logger.logger.info("Got new self-signing key", newCrossSigning.getId("self_signing"));

          _logger.logger.info("Attempting to retrieve cross-signing self-signing private key");

          _signing = null;
          _context19.prev = 37;
          _context19.next = 40;
          return this._crossSigningInfo.getCrossSigningKey("self_signing", newCrossSigning.getId("self_signing"));

        case 40:
          _ret = _context19.sent;
          _signing = _ret[1];

          _logger.logger.info("Got cross-signing self-signing private key");

          _context19.next = 48;
          break;

        case 45:
          _context19.prev = 45;
          _context19.t1 = _context19["catch"](37);

          _logger.logger.error("Cross-signing self-signing private key not available", _context19.t1);

        case 48:
          _context19.prev = 48;
          if (_signing) _signing.free();
          return _context19.finish(48);

        case 51:
          device = this._deviceList.getStoredDevice(this._userId, this._deviceId);
          _context19.next = 54;
          return this._crossSigningInfo.signDevice(this._userId, device);

        case 54:
          _signedDevice = _context19.sent;
          keySignatures[this._deviceId] = _signedDevice;

        case 56:
          if (!userSigningChanged) {
            _context19.next = 74;
            break;
          }

          _logger.logger.info("Got new user-signing key", newCrossSigning.getId("user_signing"));

          _logger.logger.info("Attempting to retrieve cross-signing user-signing private key");

          _signing2 = null;
          _context19.prev = 60;
          _context19.next = 63;
          return this._crossSigningInfo.getCrossSigningKey("user_signing", newCrossSigning.getId("user_signing"));

        case 63:
          _ret2 = _context19.sent;
          _signing2 = _ret2[1];

          _logger.logger.info("Got cross-signing user-signing private key");

          _context19.next = 71;
          break;

        case 68:
          _context19.prev = 68;
          _context19.t2 = _context19["catch"](60);

          _logger.logger.error("Cross-signing user-signing private key not available", _context19.t2);

        case 71:
          _context19.prev = 71;
          if (_signing2) _signing2.free();
          return _context19.finish(71);

        case 74:
          if (!masterChanged) {
            _context19.next = 80;
            break;
          }

          masterKey = this._crossSigningInfo.keys.master;
          _context19.next = 78;
          return this._signObject(masterKey);

        case 78:
          deviceSig = masterKey.signatures[this._userId]["ed25519:" + this._deviceId]; // Include only the _new_ device signature in the upload.
          // We may have existing signatures from deleted devices, which will cause
          // the entire upload to fail.

          keySignatures[this._crossSigningInfo.getId()] = Object.assign({}, masterKey, {
            signatures: (0, _defineProperty2["default"])({}, this._userId, (0, _defineProperty2["default"])({}, "ed25519:" + this._deviceId, deviceSig))
          });

        case 80:
          keysToUpload = Object.keys(keySignatures);

          if (keysToUpload.length) {
            _upload = function _upload(_ref27) {
              var shouldEmit = _ref27.shouldEmit;

              _logger.logger.info("Starting background key sig upload for ".concat(keysToUpload));

              return _this8._baseApis.uploadKeySignatures((0, _defineProperty2["default"])({}, _this8._userId, keySignatures)).then(function (response) {
                var _ref28 = response || {},
                    failures = _ref28.failures;

                _logger.logger.info("Finished background key sig upload for ".concat(keysToUpload));

                if (Object.keys(failures || []).length > 0) {
                  if (shouldEmit) {
                    _this8._baseApis.emit("crypto.keySignatureUploadFailure", failures, "checkOwnCrossSigningTrust", _upload);
                  }

                  throw new _errors.KeySignatureUploadError("Key upload failed", {
                    failures: failures
                  });
                }
              })["catch"](function (e) {
                _logger.logger.error("Error during background key sig upload for ".concat(keysToUpload), e);
              });
            };

            _upload({
              shouldEmit: true
            });
          }

          this.emit("userTrustStatusChanged", userId, this.checkUserTrust(userId));

          if (!masterChanged) {
            _context19.next = 87;
            break;
          }

          this._baseApis.emit("crossSigning.keysChanged", {});

          _context19.next = 87;
          return this._afterCrossSigningLocalKeyChange();

        case 87:
          _context19.next = 89;
          return this.checkKeyBackup();

        case 89:
        case "end":
          return _context19.stop();
      }
    }
  }, _callee19, this, [[13, 21, 24, 27], [37, 45, 48, 51], [60, 68, 71, 74]]);
}));
/**
 * Store a set of keys as our own, trusted, cross-signing keys.
 *
 * @param {object} keys The new trusted set of keys
 */

Crypto.prototype._storeTrustedSelfKeys = /*#__PURE__*/function () {
  var _ref29 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee20(keys) {
    var _this9 = this;

    return _regenerator["default"].wrap(function _callee20$(_context20) {
      while (1) {
        switch (_context20.prev = _context20.next) {
          case 0:
            if (keys) {
              this._crossSigningInfo.setKeys(keys);
            } else {
              this._crossSigningInfo.clearKeys();
            }

            _context20.next = 3;
            return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_ACCOUNT], function (txn) {
              _this9._cryptoStore.storeCrossSigningKeys(txn, _this9._crossSigningInfo.keys);
            });

          case 3:
          case "end":
            return _context20.stop();
        }
      }
    }, _callee20, this);
  }));

  return function (_x16) {
    return _ref29.apply(this, arguments);
  };
}();
/**
 * Check if the master key is signed by a verified device, and if so, prompt
 * the application to mark it as verified.
 *
 * @param {string} userId the user ID whose key should be checked
 */


Crypto.prototype._checkDeviceVerifications = /*#__PURE__*/function () {
  var _ref30 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee21(userId) {
    var shouldUpgradeCb, crossSigningInfo, upgradeInfo, usersToUpgrade;
    return _regenerator["default"].wrap(function _callee21$(_context21) {
      while (1) {
        switch (_context21.prev = _context21.next) {
          case 0:
            shouldUpgradeCb = this._baseApis._cryptoCallbacks.shouldUpgradeDeviceVerifications;

            if (shouldUpgradeCb) {
              _context21.next = 3;
              break;
            }

            return _context21.abrupt("return");

          case 3:
            _logger.logger.info("Starting device verification upgrade for ".concat(userId));

            if (!this._crossSigningInfo.keys.user_signing) {
              _context21.next = 17;
              break;
            }

            crossSigningInfo = this._deviceList.getStoredCrossSigningForUser(userId);

            if (!crossSigningInfo) {
              _context21.next = 17;
              break;
            }

            _context21.next = 9;
            return this._checkForDeviceVerificationUpgrade(userId, crossSigningInfo);

          case 9:
            upgradeInfo = _context21.sent;

            if (!upgradeInfo) {
              _context21.next = 17;
              break;
            }

            _context21.next = 13;
            return shouldUpgradeCb({
              users: (0, _defineProperty2["default"])({}, userId, upgradeInfo)
            });

          case 13:
            usersToUpgrade = _context21.sent;

            if (!usersToUpgrade.includes(userId)) {
              _context21.next = 17;
              break;
            }

            _context21.next = 17;
            return this._baseApis.setDeviceVerified(userId, crossSigningInfo.getId());

          case 17:
            _logger.logger.info("Finished device verification upgrade for ".concat(userId));

          case 18:
          case "end":
            return _context21.stop();
        }
      }
    }, _callee21, this);
  }));

  return function (_x17) {
    return _ref30.apply(this, arguments);
  };
}();
/**
 * Check the server for an active key backup and
 * if one is present and has a valid signature from
 * one of the user's verified devices, start backing up
 * to it.
 */


Crypto.prototype._checkAndStartKeyBackup = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee22() {
  var backupInfo, trustInfo;
  return _regenerator["default"].wrap(function _callee22$(_context22) {
    while (1) {
      switch (_context22.prev = _context22.next) {
        case 0:
          _logger.logger.log("Checking key backup status...");

          if (!this._baseApis.isGuest()) {
            _context22.next = 5;
            break;
          }

          _logger.logger.log("Skipping key backup check since user is guest");

          this._checkedForBackup = true;
          return _context22.abrupt("return", null);

        case 5:
          _context22.prev = 5;
          _context22.next = 8;
          return this._baseApis.getKeyBackupVersion();

        case 8:
          backupInfo = _context22.sent;
          _context22.next = 16;
          break;

        case 11:
          _context22.prev = 11;
          _context22.t0 = _context22["catch"](5);

          _logger.logger.log("Error checking for active key backup", _context22.t0);

          if (_context22.t0.httpStatus === 404) {
            // 404 is returned when the key backup does not exist, so that
            // counts as successfully checking.
            this._checkedForBackup = true;
          }

          return _context22.abrupt("return", null);

        case 16:
          this._checkedForBackup = true;
          _context22.next = 19;
          return this.isKeyBackupTrusted(backupInfo);

        case 19:
          trustInfo = _context22.sent;

          if (!(trustInfo.usable && !this.backupInfo)) {
            _context22.next = 25;
            break;
          }

          _logger.logger.log("Found usable key backup v" + backupInfo.version + ": enabling key backups");

          this._baseApis.enableKeyBackup(backupInfo);

          _context22.next = 44;
          break;

        case 25:
          if (!(!trustInfo.usable && this.backupInfo)) {
            _context22.next = 30;
            break;
          }

          _logger.logger.log("No usable key backup: disabling key backup");

          this._baseApis.disableKeyBackup();

          _context22.next = 44;
          break;

        case 30:
          if (!(!trustInfo.usable && !this.backupInfo)) {
            _context22.next = 34;
            break;
          }

          _logger.logger.log("No usable key backup: not enabling key backup");

          _context22.next = 44;
          break;

        case 34:
          if (!(trustInfo.usable && this.backupInfo)) {
            _context22.next = 44;
            break;
          }

          if (!(backupInfo.version !== this.backupInfo.version)) {
            _context22.next = 43;
            break;
          }

          _logger.logger.log("On backup version " + this.backupInfo.version + " but found " + "version " + backupInfo.version + ": switching.");

          this._baseApis.disableKeyBackup();

          this._baseApis.enableKeyBackup(backupInfo); // We're now using a new backup, so schedule all the keys we have to be
          // uploaded to the new backup. This is a bit of a workaround to upload
          // keys to a new backup in *most* cases, but it won't cover all cases
          // because we don't remember what backup version we uploaded keys to:
          // see https://github.com/vector-im/element-web/issues/14833


          _context22.next = 41;
          return this.scheduleAllGroupSessionsForBackup();

        case 41:
          _context22.next = 44;
          break;

        case 43:
          _logger.logger.log("Backup version " + backupInfo.version + " still current");

        case 44:
          return _context22.abrupt("return", {
            backupInfo: backupInfo,
            trustInfo: trustInfo
          });

        case 45:
        case "end":
          return _context22.stop();
      }
    }
  }, _callee22, this, [[5, 11]]);
}));

Crypto.prototype.setTrustedBackupPubKey = /*#__PURE__*/function () {
  var _ref32 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee23(trustedPubKey) {
    return _regenerator["default"].wrap(function _callee23$(_context23) {
      while (1) {
        switch (_context23.prev = _context23.next) {
          case 0:
            // This should be redundant post cross-signing is a thing, so just
            // plonk it in localStorage for now.
            this._sessionStore.setLocalTrustedBackupPubKey(trustedPubKey);

            _context23.next = 3;
            return this.checkKeyBackup();

          case 3:
          case "end":
            return _context23.stop();
        }
      }
    }, _callee23, this);
  }));

  return function (_x18) {
    return _ref32.apply(this, arguments);
  };
}();
/**
 * Forces a re-check of the key backup and enables/disables it
 * as appropriate.
 *
 * @return {Object} Object with backup info (as returned by
 *     getKeyBackupVersion) in backupInfo and
 *     trust information (as returned by isKeyBackupTrusted)
 *     in trustInfo.
 */


Crypto.prototype.checkKeyBackup = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee24() {
  return _regenerator["default"].wrap(function _callee24$(_context24) {
    while (1) {
      switch (_context24.prev = _context24.next) {
        case 0:
          this._checkedForBackup = false;
          return _context24.abrupt("return", this._checkAndStartKeyBackup());

        case 2:
        case "end":
          return _context24.stop();
      }
    }
  }, _callee24, this);
}));
/**
 * @param {object} backupInfo key backup info dict from /room_keys/version
 * @return {object} {
 *     usable: [bool], // is the backup trusted, true iff there is a sig that is valid & from a trusted device
 *     sigs: [
 *         valid: [bool || null], // true: valid, false: invalid, null: cannot attempt validation
 *         deviceId: [string],
 *         device: [DeviceInfo || null],
 *     ]
 * }
 */

Crypto.prototype.isKeyBackupTrusted = /*#__PURE__*/function () {
  var _ref34 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee25(backupInfo) {
    var ret, trustedPubkey, mySigs, _i4, _Object$keys3, keyId, keyIdParts, sigInfo, crossSigningId, device;

    return _regenerator["default"].wrap(function _callee25$(_context25) {
      while (1) {
        switch (_context25.prev = _context25.next) {
          case 0:
            ret = {
              usable: false,
              trusted_locally: false,
              sigs: []
            };

            if (!(!backupInfo || !backupInfo.algorithm || !backupInfo.auth_data || !backupInfo.auth_data.public_key || !backupInfo.auth_data.signatures)) {
              _context25.next = 4;
              break;
            }

            _logger.logger.info("Key backup is absent or missing required data");

            return _context25.abrupt("return", ret);

          case 4:
            trustedPubkey = this._sessionStore.getLocalTrustedBackupPubKey();

            if (backupInfo.auth_data.public_key === trustedPubkey) {
              _logger.logger.info("Backup public key " + trustedPubkey + " is trusted locally");

              ret.trusted_locally = true;
            }

            mySigs = backupInfo.auth_data.signatures[this._userId] || [];
            _i4 = 0, _Object$keys3 = Object.keys(mySigs);

          case 8:
            if (!(_i4 < _Object$keys3.length)) {
              _context25.next = 54;
              break;
            }

            keyId = _Object$keys3[_i4];
            keyIdParts = keyId.split(':');

            if (!(keyIdParts[0] !== 'ed25519')) {
              _context25.next = 14;
              break;
            }

            _logger.logger.log("Ignoring unknown signature type: " + keyIdParts[0]);

            return _context25.abrupt("continue", 51);

          case 14:
            // Could be a cross-signing master key, but just say this is the device
            // ID for backwards compat
            sigInfo = {
              deviceId: keyIdParts[1]
            }; // first check to see if it's from our cross-signing key

            crossSigningId = this._crossSigningInfo.getId();

            if (!(crossSigningId === sigInfo.deviceId)) {
              _context25.next = 30;
              break;
            }

            sigInfo.crossSigningId = true;
            _context25.prev = 18;
            _context25.next = 21;
            return olmlib.verifySignature(this._olmDevice, backupInfo.auth_data, this._userId, sigInfo.deviceId, crossSigningId);

          case 21:
            sigInfo.valid = true;
            _context25.next = 28;
            break;

          case 24:
            _context25.prev = 24;
            _context25.t0 = _context25["catch"](18);

            _logger.logger.warning("Bad signature from cross signing key " + crossSigningId, _context25.t0);

            sigInfo.valid = false;

          case 28:
            ret.sigs.push(sigInfo);
            return _context25.abrupt("continue", 51);

          case 30:
            // Now look for a sig from a device
            // At some point this can probably go away and we'll just support
            // it being signed by the cross-signing master key
            device = this._deviceList.getStoredDevice(this._userId, sigInfo.deviceId);

            if (!device) {
              _context25.next = 48;
              break;
            }

            sigInfo.device = device;
            _context25.next = 35;
            return this.checkDeviceTrust(this._userId, sigInfo.deviceId);

          case 35:
            sigInfo.deviceTrust = _context25.sent;
            _context25.prev = 36;
            _context25.next = 39;
            return olmlib.verifySignature(this._olmDevice, backupInfo.auth_data, this._userId, device.deviceId, device.getFingerprint());

          case 39:
            sigInfo.valid = true;
            _context25.next = 46;
            break;

          case 42:
            _context25.prev = 42;
            _context25.t1 = _context25["catch"](36);

            _logger.logger.info("Bad signature from key ID " + keyId + " userID " + this._userId + " device ID " + device.deviceId + " fingerprint: " + device.getFingerprint(), backupInfo.auth_data, _context25.t1);

            sigInfo.valid = false;

          case 46:
            _context25.next = 50;
            break;

          case 48:
            sigInfo.valid = null; // Can't determine validity because we don't have the signing device

            _logger.logger.info("Ignoring signature from unknown key " + keyId);

          case 50:
            ret.sigs.push(sigInfo);

          case 51:
            _i4++;
            _context25.next = 8;
            break;

          case 54:
            ret.usable = ret.sigs.some(function (s) {
              return s.valid && (s.device && s.deviceTrust.isVerified() || s.crossSigningId);
            });
            ret.usable |= ret.trusted_locally;
            return _context25.abrupt("return", ret);

          case 57:
          case "end":
            return _context25.stop();
        }
      }
    }, _callee25, this, [[18, 24], [36, 42]]);
  }));

  return function (_x19) {
    return _ref34.apply(this, arguments);
  };
}();
/**
 */


Crypto.prototype.enableLazyLoading = function () {
  this._lazyLoadMembers = true;
};
/**
 * Tell the crypto module to register for MatrixClient events which it needs to
 * listen for
 *
 * @param {external:EventEmitter} eventEmitter event source where we can register
 *    for event notifications
 */


Crypto.prototype.registerEventHandlers = function (eventEmitter) {
  var crypto = this;
  eventEmitter.on("RoomMember.membership", function (event, member, oldMembership) {
    try {
      crypto._onRoomMembership(event, member, oldMembership);
    } catch (e) {
      _logger.logger.error("Error handling membership change:", e);
    }
  });
  eventEmitter.on("toDeviceEvent", crypto._onToDeviceEvent.bind(crypto));

  var timelineHandler = crypto._onTimelineEvent.bind(crypto);

  eventEmitter.on("Room.timeline", timelineHandler);
  eventEmitter.on("Event.decrypted", timelineHandler);
};
/** Start background processes related to crypto */


Crypto.prototype.start = function () {
  this._outgoingRoomKeyRequestManager.start();
};
/** Stop background processes related to crypto */


Crypto.prototype.stop = function () {
  this._outgoingRoomKeyRequestManager.stop();

  this._deviceList.stop();

  this._dehydrationManager.stop();
};
/**
 * @return {string} The version of Olm.
 */


Crypto.getOlmVersion = function () {
  return _OlmDevice.OlmDevice.getOlmVersion();
};
/**
 * Get the Ed25519 key for this device
 *
 * @return {string} base64-encoded ed25519 key.
 */


Crypto.prototype.getDeviceEd25519Key = function () {
  return this._olmDevice.deviceEd25519Key;
};
/**
 * Get the Curve25519 key for this device
 *
 * @return {string} base64-encoded curve25519 key.
 */


Crypto.prototype.getDeviceCurve25519Key = function () {
  return this._olmDevice.deviceCurve25519Key;
};
/**
 * Set the global override for whether the client should ever send encrypted
 * messages to unverified devices.  This provides the default for rooms which
 * do not specify a value.
 *
 * @param {boolean} value whether to blacklist all unverified devices by default
 */


Crypto.prototype.setGlobalBlacklistUnverifiedDevices = function (value) {
  this._globalBlacklistUnverifiedDevices = value;
};
/**
 * @return {boolean} whether to blacklist all unverified devices by default
 */


Crypto.prototype.getGlobalBlacklistUnverifiedDevices = function () {
  return this._globalBlacklistUnverifiedDevices;
};
/**
 * Set whether sendMessage in a room with unknown and unverified devices
 * should throw an error and not send them message. This has 'Global' for
 * symmertry with setGlobalBlacklistUnverifiedDevices but there is currently
 * no room-level equivalent for this setting.
 *
 * This API is currently UNSTABLE and may change or be removed without notice.
 *
 * @param {boolean} value whether error on unknown devices
 */


Crypto.prototype.setGlobalErrorOnUnknownDevices = function (value) {
  this._globalErrorOnUnknownDevices = value;
};
/**
 * @return {boolean} whether to error on unknown devices
 *
 * This API is currently UNSTABLE and may change or be removed without notice.
 */


Crypto.prototype.getGlobalErrorOnUnknownDevices = function () {
  return this._globalErrorOnUnknownDevices;
};
/**
 * Upload the device keys to the homeserver.
 * @return {object} A promise that will resolve when the keys are uploaded.
 */


Crypto.prototype.uploadDeviceKeys = function () {
  var crypto = this;
  var userId = crypto._userId;
  var deviceId = crypto._deviceId;
  var deviceKeys = {
    algorithms: crypto._supportedAlgorithms,
    device_id: deviceId,
    keys: crypto._deviceKeys,
    user_id: userId
  };
  return crypto._signObject(deviceKeys).then(function () {
    return crypto._baseApis.uploadKeysRequest({
      device_keys: deviceKeys
    });
  });
};
/**
 * Stores the current one_time_key count which will be handled later (in a call of
 * onSyncCompleted). The count is e.g. coming from a /sync response.
 *
 * @param {Number} currentCount The current count of one_time_keys to be stored
 */


Crypto.prototype.updateOneTimeKeyCount = function (currentCount) {
  if (isFinite(currentCount)) {
    this._oneTimeKeyCount = currentCount;
  } else {
    throw new TypeError("Parameter for updateOneTimeKeyCount has to be a number");
  }
};

Crypto.prototype.setNeedsNewFallback = function (needsNewFallback) {
  this._needsNewFallback = !!needsNewFallback;
};

Crypto.prototype.getNeedsNewFallback = function () {
  return this._needsNewFallback;
}; // check if it's time to upload one-time keys, and do so if so.


function _maybeUploadOneTimeKeys(crypto) {
  // frequency with which to check & upload one-time keys
  var uploadPeriod = 1000 * 60; // one minute
  // max number of keys to upload at once
  // Creating keys can be an expensive operation so we limit the
  // number we generate in one go to avoid blocking the application
  // for too long.

  var maxKeysPerCycle = 5;

  if (crypto._oneTimeKeyCheckInProgress) {
    return;
  }

  var now = Date.now();

  if (crypto._lastOneTimeKeyCheck !== null && now - crypto._lastOneTimeKeyCheck < uploadPeriod) {
    // we've done a key upload recently.
    return;
  }

  crypto._lastOneTimeKeyCheck = now; // We need to keep a pool of one time public keys on the server so that
  // other devices can start conversations with us. But we can only store
  // a finite number of private keys in the olm Account object.
  // To complicate things further then can be a delay between a device
  // claiming a public one time key from the server and it sending us a
  // message. We need to keep the corresponding private key locally until
  // we receive the message.
  // But that message might never arrive leaving us stuck with duff
  // private keys clogging up our local storage.
  // So we need some kind of enginering compromise to balance all of
  // these factors.
  // Check how many keys we can store in the Account object.

  var maxOneTimeKeys = crypto._olmDevice.maxNumberOfOneTimeKeys(); // Try to keep at most half that number on the server. This leaves the
  // rest of the slots free to hold keys that have been claimed from the
  // server but we haven't recevied a message for.
  // If we run out of slots when generating new keys then olm will
  // discard the oldest private keys first. This will eventually clean
  // out stale private keys that won't receive a message.


  var keyLimit = Math.floor(maxOneTimeKeys / 2);

  function uploadLoop(_x20) {
    return _uploadLoop.apply(this, arguments);
  }

  function _uploadLoop() {
    _uploadLoop = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee26(keyCount) {
      var keysThisLoop, res;
      return _regenerator["default"].wrap(function _callee26$(_context26) {
        while (1) {
          switch (_context26.prev = _context26.next) {
            case 0:
              if (!(keyLimit > keyCount || crypto.getNeedsNewFallback())) {
                _context26.next = 21;
                break;
              }

              if (!(keyLimit > keyCount)) {
                _context26.next = 6;
                break;
              }

              _logger.logger.info("generating oneTimeKeys");

              keysThisLoop = Math.min(keyLimit - keyCount, maxKeysPerCycle);
              _context26.next = 6;
              return crypto._olmDevice.generateOneTimeKeys(keysThisLoop);

            case 6:
              if (!crypto.getNeedsNewFallback()) {
                _context26.next = 10;
                break;
              }

              _logger.logger.info("generating fallback key");

              _context26.next = 10;
              return crypto._olmDevice.generateFallbackKey();

            case 10:
              _logger.logger.info("calling _uploadOneTimeKeys");

              _context26.next = 13;
              return _uploadOneTimeKeys(crypto);

            case 13:
              res = _context26.sent;

              if (!(res.one_time_key_counts && res.one_time_key_counts.signed_curve25519)) {
                _context26.next = 18;
                break;
              }

              // if the response contains a more up to date value use this
              // for the next loop
              keyCount = res.one_time_key_counts.signed_curve25519;
              _context26.next = 19;
              break;

            case 18:
              throw new Error("response for uploading keys does not contain " + "one_time_key_counts.signed_curve25519");

            case 19:
              _context26.next = 0;
              break;

            case 21:
            case "end":
              return _context26.stop();
          }
        }
      }, _callee26);
    }));
    return _uploadLoop.apply(this, arguments);
  }

  crypto._oneTimeKeyCheckInProgress = true;
  Promise.resolve().then(function () {
    if (crypto._oneTimeKeyCount !== undefined) {
      // We already have the current one_time_key count from a /sync response.
      // Use this value instead of asking the server for the current key count.
      return Promise.resolve(crypto._oneTimeKeyCount);
    } // ask the server how many keys we have


    return crypto._baseApis.uploadKeysRequest({}).then(function (res) {
      return res.one_time_key_counts.signed_curve25519 || 0;
    });
  }).then(function (keyCount) {
    // Start the uploadLoop with the current keyCount. The function checks if
    // we need to upload new keys or not.
    // If there are too many keys on the server then we don't need to
    // create any more keys.
    return uploadLoop(keyCount);
  })["catch"](function (e) {
    _logger.logger.error("Error uploading one-time keys", e.stack || e);
  })["finally"](function () {
    // reset _oneTimeKeyCount to prevent start uploading based on old data.
    // it will be set again on the next /sync-response
    crypto._oneTimeKeyCount = undefined;
    crypto._oneTimeKeyCheckInProgress = false;
  });
} // returns a promise which resolves to the response


function _uploadOneTimeKeys(_x21) {
  return _uploadOneTimeKeys2.apply(this, arguments);
}
/**
 * Download the keys for a list of users and stores the keys in the session
 * store.
 * @param {Array} userIds The users to fetch.
 * @param {bool} forceDownload Always download the keys even if cached.
 *
 * @return {Promise} A promise which resolves to a map userId->deviceId->{@link
 * module:crypto/deviceinfo|DeviceInfo}.
 */


function _uploadOneTimeKeys2() {
  _uploadOneTimeKeys2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee54(crypto) {
    var promises, fallbackJson, fallbackKeys, _i6, _Object$entries2, _Object$entries2$_i, keyId, key, k, oneTimeKeys, oneTimeJson, _keyId, _k, res;

    return _regenerator["default"].wrap(function _callee54$(_context54) {
      while (1) {
        switch (_context54.prev = _context54.next) {
          case 0:
            promises = [];
            fallbackJson = {};

            if (!crypto.getNeedsNewFallback()) {
              _context54.next = 8;
              break;
            }

            _context54.next = 5;
            return crypto._olmDevice.getFallbackKey();

          case 5:
            fallbackKeys = _context54.sent;

            for (_i6 = 0, _Object$entries2 = Object.entries(fallbackKeys.curve25519); _i6 < _Object$entries2.length; _i6++) {
              _Object$entries2$_i = (0, _slicedToArray2["default"])(_Object$entries2[_i6], 2), keyId = _Object$entries2$_i[0], key = _Object$entries2$_i[1];
              k = {
                key: key,
                fallback: true
              };
              fallbackJson["signed_curve25519:" + keyId] = k;
              promises.push(crypto._signObject(k));
            }

            crypto.setNeedsNewFallback(false);

          case 8:
            _context54.next = 10;
            return crypto._olmDevice.getOneTimeKeys();

          case 10:
            oneTimeKeys = _context54.sent;
            oneTimeJson = {};

            for (_keyId in oneTimeKeys.curve25519) {
              if (oneTimeKeys.curve25519.hasOwnProperty(_keyId)) {
                _k = {
                  key: oneTimeKeys.curve25519[_keyId]
                };
                oneTimeJson["signed_curve25519:" + _keyId] = _k;
                promises.push(crypto._signObject(_k));
              }
            }

            _context54.next = 15;
            return Promise.all(promises);

          case 15:
            _context54.next = 17;
            return crypto._baseApis.uploadKeysRequest({
              "one_time_keys": oneTimeJson,
              "org.matrix.msc2732.fallback_keys": fallbackJson
            });

          case 17:
            res = _context54.sent;
            _context54.next = 20;
            return crypto._olmDevice.markKeysAsPublished();

          case 20:
            return _context54.abrupt("return", res);

          case 21:
          case "end":
            return _context54.stop();
        }
      }
    }, _callee54);
  }));
  return _uploadOneTimeKeys2.apply(this, arguments);
}

Crypto.prototype.downloadKeys = function (userIds, forceDownload) {
  return this._deviceList.downloadKeys(userIds, forceDownload);
};
/**
 * Get the stored device keys for a user id
 *
 * @param {string} userId the user to list keys for.
 *
 * @return {module:crypto/deviceinfo[]|null} list of devices, or null if we haven't
 * managed to get a list of devices for this user yet.
 */


Crypto.prototype.getStoredDevicesForUser = function (userId) {
  return this._deviceList.getStoredDevicesForUser(userId);
};
/**
 * Get the stored keys for a single device
 *
 * @param {string} userId
 * @param {string} deviceId
 *
 * @return {module:crypto/deviceinfo?} device, or undefined
 * if we don't know about this device
 */


Crypto.prototype.getStoredDevice = function (userId, deviceId) {
  return this._deviceList.getStoredDevice(userId, deviceId);
};
/**
 * Save the device list, if necessary
 *
 * @param {integer} delay Time in ms before which the save actually happens.
 *     By default, the save is delayed for a short period in order to batch
 *     multiple writes, but this behaviour can be disabled by passing 0.
 *
 * @return {Promise<bool>} true if the data was saved, false if
 *     it was not (eg. because no changes were pending). The promise
 *     will only resolve once the data is saved, so may take some time
 *     to resolve.
 */


Crypto.prototype.saveDeviceList = function (delay) {
  return this._deviceList.saveIfDirty(delay);
};
/**
 * Update the blocked/verified state of the given device
 *
 * @param {string} userId owner of the device
 * @param {string} deviceId unique identifier for the device or user's
 * cross-signing public key ID.
 *
 * @param {?boolean} verified whether to mark the device as verified. Null to
 *     leave unchanged.
 *
 * @param {?boolean} blocked whether to mark the device as blocked. Null to
 *      leave unchanged.
 *
 * @param {?boolean} known whether to mark that the user has been made aware of
 *      the existence of this device. Null to leave unchanged
 *
 * @return {Promise<module:crypto/deviceinfo>} updated DeviceInfo
 */


Crypto.prototype.setDeviceVerification = /*#__PURE__*/function () {
  var _ref35 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee29(userId, deviceId, verified, blocked, known) {
    var _this10 = this;

    var xsk, device, _upload2, devices, dev, verificationStatus, knownStatus, _device, deviceTrust, _upload3, deviceObj;

    return _regenerator["default"].wrap(function _callee29$(_context29) {
      while (1) {
        switch (_context29.prev = _context29.next) {
          case 0:
            // get rid of any `undefined`s here so we can just check
            // for null rather than null or undefined
            if (verified === undefined) verified = null;
            if (blocked === undefined) blocked = null;
            if (known === undefined) known = null; // Check if the 'device' is actually a cross signing key
            // The js-sdk's verification treats cross-signing keys as devices
            // and so uses this method to mark them verified.

            xsk = this._deviceList.getStoredCrossSigningForUser(userId);

            if (!(xsk && xsk.getId() === deviceId)) {
              _context29.next = 23;
              break;
            }

            if (!(blocked !== null || known !== null)) {
              _context29.next = 7;
              break;
            }

            throw new Error("Cannot set blocked or known for a cross-signing key");

          case 7:
            if (verified) {
              _context29.next = 9;
              break;
            }

            throw new Error("Cannot set a cross-signing key as unverified");

          case 9:
            if (!this._crossSigningInfo.getId() && userId === this._crossSigningInfo.userId) {
              this._storeTrustedSelfKeys(xsk.keys); // This will cause our own user trust to change, so emit the event


              this.emit("userTrustStatusChanged", this._userId, this.checkUserTrust(userId));
            } // Now sign the master key with our user signing key (unless it's ourself)


            if (!(userId !== this._userId)) {
              _context29.next = 22;
              break;
            }

            _logger.logger.info("Master key " + xsk.getId() + " for " + userId + " marked verified. Signing...");

            _context29.next = 14;
            return this._crossSigningInfo.signUser(xsk);

          case 14:
            device = _context29.sent;

            if (!device) {
              _context29.next = 19;
              break;
            }

            _upload2 = /*#__PURE__*/function () {
              var _ref37 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee27(_ref36) {
                var shouldEmit, response, _ref38, failures;

                return _regenerator["default"].wrap(function _callee27$(_context27) {
                  while (1) {
                    switch (_context27.prev = _context27.next) {
                      case 0:
                        shouldEmit = _ref36.shouldEmit;

                        _logger.logger.info("Uploading signature for " + userId + "...");

                        _context27.next = 4;
                        return _this10._baseApis.uploadKeySignatures((0, _defineProperty2["default"])({}, userId, (0, _defineProperty2["default"])({}, deviceId, device)));

                      case 4:
                        response = _context27.sent;
                        _ref38 = response || {}, failures = _ref38.failures;

                        if (!(Object.keys(failures || []).length > 0)) {
                          _context27.next = 9;
                          break;
                        }

                        if (shouldEmit) {
                          _this10._baseApis.emit("crypto.keySignatureUploadFailure", failures, "setDeviceVerification", _upload2);
                        }
                        /* Throwing here causes the process to be cancelled and the other
                        * user to be notified */


                        throw new _errors.KeySignatureUploadError("Key upload failed", {
                          failures: failures
                        });

                      case 9:
                      case "end":
                        return _context27.stop();
                    }
                  }
                }, _callee27);
              }));

              return function _upload2(_x27) {
                return _ref37.apply(this, arguments);
              };
            }();

            _context29.next = 19;
            return _upload2({
              shouldEmit: true
            });

          case 19:
            return _context29.abrupt("return", device);

          case 22:
            return _context29.abrupt("return", xsk);

          case 23:
            devices = this._deviceList.getRawStoredDevicesForUser(userId);

            if (!(!devices || !devices[deviceId])) {
              _context29.next = 26;
              break;
            }

            throw new Error("Unknown device " + userId + ":" + deviceId);

          case 26:
            dev = devices[deviceId];
            verificationStatus = dev.verified;

            if (verified) {
              verificationStatus = DeviceVerification.VERIFIED;
            } else if (verified !== null && verificationStatus == DeviceVerification.VERIFIED) {
              verificationStatus = DeviceVerification.UNVERIFIED;
            }

            if (blocked) {
              verificationStatus = DeviceVerification.BLOCKED;
            } else if (blocked !== null && verificationStatus == DeviceVerification.BLOCKED) {
              verificationStatus = DeviceVerification.UNVERIFIED;
            }

            knownStatus = dev.known;

            if (known !== null) {
              knownStatus = known;
            }

            if (dev.verified !== verificationStatus || dev.known !== knownStatus) {
              dev.verified = verificationStatus;
              dev.known = knownStatus;

              this._deviceList.storeDevicesForUser(userId, devices);

              this._deviceList.saveIfDirty();
            } // do cross-signing


            if (!(verified && userId === this._userId)) {
              _context29.next = 47;
              break;
            }

            _logger.logger.info("Own device " + deviceId + " marked verified: signing"); // Signing only needed if other device not already signed


            deviceTrust = this.checkDeviceTrust(userId, deviceId);

            if (!deviceTrust.isCrossSigningVerified()) {
              _context29.next = 40;
              break;
            }

            _logger.logger.log("Own device ".concat(deviceId, " already cross-signing verified"));

            _context29.next = 43;
            break;

          case 40:
            _context29.next = 42;
            return this._crossSigningInfo.signDevice(userId, _deviceinfo.DeviceInfo.fromStorage(dev, deviceId));

          case 42:
            _device = _context29.sent;

          case 43:
            if (!_device) {
              _context29.next = 47;
              break;
            }

            _upload3 = /*#__PURE__*/function () {
              var _ref40 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee28(_ref39) {
                var shouldEmit, response, _ref41, failures;

                return _regenerator["default"].wrap(function _callee28$(_context28) {
                  while (1) {
                    switch (_context28.prev = _context28.next) {
                      case 0:
                        shouldEmit = _ref39.shouldEmit;

                        _logger.logger.info("Uploading signature for " + deviceId);

                        _context28.next = 4;
                        return _this10._baseApis.uploadKeySignatures((0, _defineProperty2["default"])({}, userId, (0, _defineProperty2["default"])({}, deviceId, _device)));

                      case 4:
                        response = _context28.sent;
                        _ref41 = response || {}, failures = _ref41.failures;

                        if (!(Object.keys(failures || []).length > 0)) {
                          _context28.next = 9;
                          break;
                        }

                        if (shouldEmit) {
                          _this10._baseApis.emit("crypto.keySignatureUploadFailure", failures, "setDeviceVerification", _upload3 // continuation
                          );
                        }

                        throw new _errors.KeySignatureUploadError("Key upload failed", {
                          failures: failures
                        });

                      case 9:
                      case "end":
                        return _context28.stop();
                    }
                  }
                }, _callee28);
              }));

              return function _upload3(_x28) {
                return _ref40.apply(this, arguments);
              };
            }();

            _context29.next = 47;
            return _upload3({
              shouldEmit: true
            });

          case 47:
            deviceObj = _deviceinfo.DeviceInfo.fromStorage(dev, deviceId);
            this.emit("deviceVerificationChanged", userId, deviceId, deviceObj);
            return _context29.abrupt("return", deviceObj);

          case 50:
          case "end":
            return _context29.stop();
        }
      }
    }, _callee29, this);
  }));

  return function (_x22, _x23, _x24, _x25, _x26) {
    return _ref35.apply(this, arguments);
  };
}();

Crypto.prototype.findVerificationRequestDMInProgress = function (roomId) {
  return this._inRoomVerificationRequests.findRequestInProgress(roomId);
};

Crypto.prototype.getVerificationRequestsToDeviceInProgress = function (userId) {
  return this._toDeviceVerificationRequests.getRequestsInProgress(userId);
};

Crypto.prototype.requestVerificationDM = function (userId, roomId) {
  var existingRequest = this._inRoomVerificationRequests.findRequestInProgress(roomId);

  if (existingRequest) {
    return Promise.resolve(existingRequest);
  }

  var channel = new _InRoomChannel.InRoomChannel(this._baseApis, roomId, userId);
  return this._requestVerificationWithChannel(userId, channel, this._inRoomVerificationRequests);
};

Crypto.prototype.requestVerification = function (userId, devices) {
  if (!devices) {
    devices = Object.keys(this._deviceList.getRawStoredDevicesForUser(userId));
  }

  var existingRequest = this._toDeviceVerificationRequests.findRequestInProgress(userId, devices);

  if (existingRequest) {
    return Promise.resolve(existingRequest);
  }

  var channel = new _ToDeviceChannel.ToDeviceChannel(this._baseApis, userId, devices, _ToDeviceChannel.ToDeviceChannel.makeTransactionId());
  return this._requestVerificationWithChannel(userId, channel, this._toDeviceVerificationRequests);
};

Crypto.prototype._requestVerificationWithChannel = /*#__PURE__*/function () {
  var _ref42 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee30(userId, channel, requestsMap) {
    var request, racingRequest;
    return _regenerator["default"].wrap(function _callee30$(_context30) {
      while (1) {
        switch (_context30.prev = _context30.next) {
          case 0:
            request = new _VerificationRequest.VerificationRequest(channel, this._verificationMethods, this._baseApis); // if transaction id is already known, add request

            if (channel.transactionId) {
              requestsMap.setRequestByChannel(channel, request);
            }

            _context30.next = 4;
            return request.sendRequest();

          case 4:
            // don't replace the request created by a racing remote echo
            racingRequest = requestsMap.getRequestByChannel(channel);

            if (racingRequest) {
              request = racingRequest;
            } else {
              _logger.logger.log("Crypto: adding new request to " + "requestsByTxnId with id ".concat(channel.transactionId, " ").concat(channel.roomId));

              requestsMap.setRequestByChannel(channel, request);
            }

            return _context30.abrupt("return", request);

          case 7:
          case "end":
            return _context30.stop();
        }
      }
    }, _callee30, this);
  }));

  return function (_x29, _x30, _x31) {
    return _ref42.apply(this, arguments);
  };
}();

Crypto.prototype.beginKeyVerification = function (method, userId, deviceId) {
  var transactionId = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
  var request;

  if (transactionId) {
    request = this._toDeviceVerificationRequests.getRequestBySenderAndTxnId(userId, transactionId);

    if (!request) {
      throw new Error("No request found for user ".concat(userId, " with ") + "transactionId ".concat(transactionId));
    }
  } else {
    transactionId = _ToDeviceChannel.ToDeviceChannel.makeTransactionId();
    var channel = new _ToDeviceChannel.ToDeviceChannel(this._baseApis, userId, [deviceId], transactionId, deviceId);
    request = new _VerificationRequest.VerificationRequest(channel, this._verificationMethods, this._baseApis);

    this._toDeviceVerificationRequests.setRequestBySenderAndTxnId(userId, transactionId, request);
  }

  return request.beginKeyVerification(method, {
    userId: userId,
    deviceId: deviceId
  });
};

Crypto.prototype.legacyDeviceVerification = /*#__PURE__*/function () {
  var _ref43 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee31(userId, deviceId, method) {
    var transactionId, channel, request, verifier;
    return _regenerator["default"].wrap(function _callee31$(_context31) {
      while (1) {
        switch (_context31.prev = _context31.next) {
          case 0:
            transactionId = _ToDeviceChannel.ToDeviceChannel.makeTransactionId();
            channel = new _ToDeviceChannel.ToDeviceChannel(this._baseApis, userId, [deviceId], transactionId, deviceId);
            request = new _VerificationRequest.VerificationRequest(channel, this._verificationMethods, this._baseApis);

            this._toDeviceVerificationRequests.setRequestBySenderAndTxnId(userId, transactionId, request);

            verifier = request.beginKeyVerification(method, {
              userId: userId,
              deviceId: deviceId
            }); // either reject by an error from verify() while sending .start
            // or resolve when the request receives the
            // local (fake remote) echo for sending the .start event

            _context31.next = 7;
            return Promise.race([verifier.verify(), request.waitFor(function (r) {
              return r.started;
            })]);

          case 7:
            return _context31.abrupt("return", request);

          case 8:
          case "end":
            return _context31.stop();
        }
      }
    }, _callee31, this);
  }));

  return function (_x32, _x33, _x34) {
    return _ref43.apply(this, arguments);
  };
}();
/**
 * Get information on the active olm sessions with a user
 * <p>
 * Returns a map from device id to an object with keys 'deviceIdKey' (the
 * device's curve25519 identity key) and 'sessions' (an array of objects in the
 * same format as that returned by
 * {@link module:crypto/OlmDevice#getSessionInfoForDevice}).
 * <p>
 * This method is provided for debugging purposes.
 *
 * @param {string} userId id of user to inspect
 *
 * @return {Promise<Object.<string, {deviceIdKey: string, sessions: object[]}>>}
 */


Crypto.prototype.getOlmSessionsForUser = /*#__PURE__*/function () {
  var _ref44 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee32(userId) {
    var devices, result, j, device, deviceKey, sessions;
    return _regenerator["default"].wrap(function _callee32$(_context32) {
      while (1) {
        switch (_context32.prev = _context32.next) {
          case 0:
            devices = this.getStoredDevicesForUser(userId) || [];
            result = {};
            j = 0;

          case 3:
            if (!(j < devices.length)) {
              _context32.next = 13;
              break;
            }

            device = devices[j];
            deviceKey = device.getIdentityKey();
            _context32.next = 8;
            return this._olmDevice.getSessionInfoForDevice(deviceKey);

          case 8:
            sessions = _context32.sent;
            result[device.deviceId] = {
              deviceIdKey: deviceKey,
              sessions: sessions
            };

          case 10:
            ++j;
            _context32.next = 3;
            break;

          case 13:
            return _context32.abrupt("return", result);

          case 14:
          case "end":
            return _context32.stop();
        }
      }
    }, _callee32, this);
  }));

  return function (_x35) {
    return _ref44.apply(this, arguments);
  };
}();
/**
 * Get the device which sent an event
 *
 * @param {module:models/event.MatrixEvent} event event to be checked
 *
 * @return {module:crypto/deviceinfo?}
 */


Crypto.prototype.getEventSenderDeviceInfo = function (event) {
  var senderKey = event.getSenderKey();
  var algorithm = event.getWireContent().algorithm;

  if (!senderKey || !algorithm) {
    return null;
  }

  var forwardingChain = event.getForwardingCurve25519KeyChain();

  if (forwardingChain.length > 0) {
    // we got the key this event from somewhere else
    // TODO: check if we can trust the forwarders.
    return null;
  }

  if (event.isKeySourceUntrusted()) {
    // we got the key for this event from a source that we consider untrusted
    return null;
  } // senderKey is the Curve25519 identity key of the device which the event
  // was sent from. In the case of Megolm, it's actually the Curve25519
  // identity key of the device which set up the Megolm session.


  var device = this._deviceList.getDeviceByIdentityKey(algorithm, senderKey);

  if (device === null) {
    // we haven't downloaded the details of this device yet.
    return null;
  } // so far so good, but now we need to check that the sender of this event
  // hadn't advertised someone else's Curve25519 key as their own. We do that
  // by checking the Ed25519 claimed by the event (or, in the case of megolm,
  // the event which set up the megolm session), to check that it matches the
  // fingerprint of the purported sending device.
  //
  // (see https://github.com/vector-im/vector-web/issues/2215)


  var claimedKey = event.getClaimedEd25519Key();

  if (!claimedKey) {
    _logger.logger.warn("Event " + event.getId() + " claims no ed25519 key: " + "cannot verify sending device");

    return null;
  }

  if (claimedKey !== device.getFingerprint()) {
    _logger.logger.warn("Event " + event.getId() + " claims ed25519 key " + claimedKey + " but sender device has key " + device.getFingerprint());

    return null;
  }

  return device;
};
/**
 * Get information about the encryption of an event
 *
 * @param {module:models/event.MatrixEvent} event event to be checked
 *
 * @return {object} An object with the fields:
 *    - encrypted: whether the event is encrypted (if not encrypted, some of the
 *      other properties may not be set)
 *    - senderKey: the sender's key
 *    - algorithm: the algorithm used to encrypt the event
 *    - authenticated: whether we can be sure that the owner of the senderKey
 *      sent the event
 *    - sender: the sender's device information, if available
 *    - mismatchedSender: if the event's ed25519 and curve25519 keys don't match
 *      (only meaningful if `sender` is set)
 */


Crypto.prototype.getEventEncryptionInfo = function (event) {
  var ret = {};
  ret.senderKey = event.getSenderKey();
  ret.algorithm = event.getWireContent().algorithm;

  if (!ret.senderKey || !ret.algorithm) {
    ret.encrypted = false;
    return ret;
  }

  ret.encrypted = true;
  var forwardingChain = event.getForwardingCurve25519KeyChain();

  if (forwardingChain.length > 0 || event.isKeySourceUntrusted()) {
    // we got the key this event from somewhere else
    // TODO: check if we can trust the forwarders.
    ret.authenticated = false;
  } else {
    ret.authenticated = true;
  } // senderKey is the Curve25519 identity key of the device which the event
  // was sent from. In the case of Megolm, it's actually the Curve25519
  // identity key of the device which set up the Megolm session.


  ret.sender = this._deviceList.getDeviceByIdentityKey(ret.algorithm, ret.senderKey); // so far so good, but now we need to check that the sender of this event
  // hadn't advertised someone else's Curve25519 key as their own. We do that
  // by checking the Ed25519 claimed by the event (or, in the case of megolm,
  // the event which set up the megolm session), to check that it matches the
  // fingerprint of the purported sending device.
  //
  // (see https://github.com/vector-im/vector-web/issues/2215)

  var claimedKey = event.getClaimedEd25519Key();

  if (!claimedKey) {
    _logger.logger.warn("Event " + event.getId() + " claims no ed25519 key: " + "cannot verify sending device");

    ret.mismatchedSender = true;
  }

  if (ret.sender && claimedKey !== ret.sender.getFingerprint()) {
    _logger.logger.warn("Event " + event.getId() + " claims ed25519 key " + claimedKey + "but sender device has key " + ret.sender.getFingerprint());

    ret.mismatchedSender = true;
  }

  return ret;
};
/**
 * Forces the current outbound group session to be discarded such
 * that another one will be created next time an event is sent.
 *
 * @param {string} roomId The ID of the room to discard the session for
 *
 * This should not normally be necessary.
 */


Crypto.prototype.forceDiscardSession = function (roomId) {
  var alg = this._roomEncryptors[roomId];
  if (alg === undefined) throw new Error("Room not encrypted");

  if (alg.forceDiscardSession === undefined) {
    throw new Error("Room encryption algorithm doesn't support session discarding");
  }

  alg.forceDiscardSession();
};
/**
 * Configure a room to use encryption (ie, save a flag in the cryptoStore).
 *
 * @param {string} roomId The room ID to enable encryption in.
 *
 * @param {object} config The encryption config for the room.
 *
 * @param {boolean=} inhibitDeviceQuery true to suppress device list query for
 *   users in the room (for now). In case lazy loading is enabled,
 *   the device query is always inhibited as the members are not tracked.
 */


Crypto.prototype.setRoomEncryption = /*#__PURE__*/function () {
  var _ref45 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee33(roomId, config, inhibitDeviceQuery) {
    var existingConfig, existingAlg, storeConfigPromise, AlgClass, alg;
    return _regenerator["default"].wrap(function _callee33$(_context33) {
      while (1) {
        switch (_context33.prev = _context33.next) {
          case 0:
            if (config.algorithm) {
              _context33.next = 3;
              break;
            }

            _logger.logger.log("Ignoring setRoomEncryption with no algorithm");

            return _context33.abrupt("return");

          case 3:
            // if state is being replayed from storage, we might already have a configuration
            // for this room as they are persisted as well.
            // We just need to make sure the algorithm is initialized in this case.
            // However, if the new config is different,
            // we should bail out as room encryption can't be changed once set.
            existingConfig = this._roomList.getRoomEncryption(roomId);

            if (!existingConfig) {
              _context33.next = 8;
              break;
            }

            if (!(JSON.stringify(existingConfig) != JSON.stringify(config))) {
              _context33.next = 8;
              break;
            }

            _logger.logger.error("Ignoring m.room.encryption event which requests " + "a change of config in " + roomId);

            return _context33.abrupt("return");

          case 8:
            // if we already have encryption in this room, we should ignore this event,
            // as it would reset the encryption algorithm.
            // This is at least expected to be called twice, as sync calls onCryptoEvent
            // for both the timeline and state sections in the /sync response,
            // the encryption event would appear in both.
            // If it's called more than twice though,
            // it signals a bug on client or server.
            existingAlg = this._roomEncryptors[roomId];

            if (!existingAlg) {
              _context33.next = 11;
              break;
            }

            return _context33.abrupt("return");

          case 11:
            // _roomList.getRoomEncryption will not race with _roomList.setRoomEncryption
            // because it first stores in memory. We should await the promise only
            // after all the in-memory state (_roomEncryptors and _roomList) has been updated
            // to avoid races when calling this method multiple times. Hence keep a hold of the promise.
            storeConfigPromise = null;

            if (!existingConfig) {
              storeConfigPromise = this._roomList.setRoomEncryption(roomId, config);
            }

            AlgClass = algorithms.ENCRYPTION_CLASSES[config.algorithm];

            if (AlgClass) {
              _context33.next = 16;
              break;
            }

            throw new Error("Unable to encrypt with " + config.algorithm);

          case 16:
            alg = new AlgClass({
              userId: this._userId,
              deviceId: this._deviceId,
              crypto: this,
              olmDevice: this._olmDevice,
              baseApis: this._baseApis,
              roomId: roomId,
              config: config
            });
            this._roomEncryptors[roomId] = alg;

            if (!storeConfigPromise) {
              _context33.next = 21;
              break;
            }

            _context33.next = 21;
            return storeConfigPromise;

          case 21:
            if (this._lazyLoadMembers) {
              _context33.next = 28;
              break;
            }

            _logger.logger.log("Enabling encryption in " + roomId + "; " + "starting to track device lists for all users therein");

            _context33.next = 25;
            return this.trackRoomDevices(roomId);

          case 25:
            // TODO: this flag is only not used from MatrixClient::setRoomEncryption
            // which is never used (inside Element at least)
            // but didn't want to remove it as it technically would
            // be a breaking change.
            if (!this.inhibitDeviceQuery) {
              this._deviceList.refreshOutdatedDeviceLists();
            }

            _context33.next = 29;
            break;

          case 28:
            _logger.logger.log("Enabling encryption in " + roomId);

          case 29:
          case "end":
            return _context33.stop();
        }
      }
    }, _callee33, this);
  }));

  return function (_x36, _x37, _x38) {
    return _ref45.apply(this, arguments);
  };
}();
/**
 * Make sure we are tracking the device lists for all users in this room.
 *
 * @param {string} roomId The room ID to start tracking devices in.
 * @returns {Promise} when all devices for the room have been fetched and marked to track
 */


Crypto.prototype.trackRoomDevices = function (roomId) {
  var _this11 = this;

  var trackMembers = /*#__PURE__*/function () {
    var _ref46 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee34() {
      var room, members;
      return _regenerator["default"].wrap(function _callee34$(_context34) {
        while (1) {
          switch (_context34.prev = _context34.next) {
            case 0:
              if (_this11._roomEncryptors[roomId]) {
                _context34.next = 2;
                break;
              }

              return _context34.abrupt("return");

            case 2:
              room = _this11._clientStore.getRoom(roomId);

              if (room) {
                _context34.next = 5;
                break;
              }

              throw new Error("Unable to start tracking devices in unknown room ".concat(roomId));

            case 5:
              _logger.logger.log("Starting to track devices for room ".concat(roomId, " ..."));

              _context34.next = 8;
              return room.getEncryptionTargetMembers();

            case 8:
              members = _context34.sent;
              members.forEach(function (m) {
                _this11._deviceList.startTrackingDeviceList(m.userId);
              });

            case 10:
            case "end":
              return _context34.stop();
          }
        }
      }, _callee34);
    }));

    return function trackMembers() {
      return _ref46.apply(this, arguments);
    };
  }();

  var promise = this._roomDeviceTrackingState[roomId];

  if (!promise) {
    promise = trackMembers();
    this._roomDeviceTrackingState[roomId] = promise["catch"](function (err) {
      _this11._roomDeviceTrackingState[roomId] = null;
      throw err;
    });
  }

  return promise;
};
/**
 * @typedef {Object} module:crypto~OlmSessionResult
 * @property {module:crypto/deviceinfo} device  device info
 * @property {string?} sessionId base64 olm session id; null if no session
 *    could be established
 */

/**
 * Try to make sure we have established olm sessions for all known devices for
 * the given users.
 *
 * @param {string[]} users list of user ids
 *
 * @return {Promise} resolves once the sessions are complete, to
 *    an Object mapping from userId to deviceId to
 *    {@link module:crypto~OlmSessionResult}
 */


Crypto.prototype.ensureOlmSessionsForUsers = function (users) {
  var devicesByUser = {};

  for (var i = 0; i < users.length; ++i) {
    var userId = users[i];
    devicesByUser[userId] = [];
    var devices = this.getStoredDevicesForUser(userId) || [];

    for (var j = 0; j < devices.length; ++j) {
      var deviceInfo = devices[j];
      var key = deviceInfo.getIdentityKey();

      if (key == this._olmDevice.deviceCurve25519Key) {
        // don't bother setting up session to ourself
        continue;
      }

      if (deviceInfo.verified == DeviceVerification.BLOCKED) {
        // don't bother setting up sessions with blocked users
        continue;
      }

      devicesByUser[userId].push(deviceInfo);
    }
  }

  return olmlib.ensureOlmSessionsForDevices(this._olmDevice, this._baseApis, devicesByUser);
};
/**
 * Get a list containing all of the room keys
 *
 * @return {module:crypto/OlmDevice.MegolmSessionData[]} a list of session export objects
 */


Crypto.prototype.exportRoomKeys = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee35() {
  var _this12 = this;

  var exportedSessions;
  return _regenerator["default"].wrap(function _callee35$(_context35) {
    while (1) {
      switch (_context35.prev = _context35.next) {
        case 0:
          exportedSessions = [];
          _context35.next = 3;
          return this._cryptoStore.doTxn('readonly', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS], function (txn) {
            _this12._cryptoStore.getAllEndToEndInboundGroupSessions(txn, function (s) {
              if (s === null) return;

              var sess = _this12._olmDevice.exportInboundGroupSession(s.senderKey, s.sessionId, s.sessionData);

              delete sess.first_known_index;
              sess.algorithm = olmlib.MEGOLM_ALGORITHM;
              exportedSessions.push(sess);
            });
          });

        case 3:
          return _context35.abrupt("return", exportedSessions);

        case 4:
        case "end":
          return _context35.stop();
      }
    }
  }, _callee35, this);
}));
/**
 * Import a list of room keys previously exported by exportRoomKeys
 *
 * @param {Object[]} keys a list of session export objects
 * @param {Object} opts
 * @param {Function} opts.progressCallback called with an object which has a stage param
 * @return {Promise} a promise which resolves once the keys have been imported
 */

Crypto.prototype.importRoomKeys = function (keys) {
  var _this13 = this;

  var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  var successes = 0;
  var failures = 0;
  var total = keys.length;

  function updateProgress() {
    opts.progressCallback({
      stage: "load_keys",
      successes: successes,
      failures: failures,
      total: total
    });
  }

  return Promise.all(keys.map(function (key) {
    if (!key.room_id || !key.algorithm) {
      _logger.logger.warn("ignoring room key entry with missing fields", key);

      failures++;

      if (opts.progressCallback) {
        updateProgress();
      }

      return null;
    }

    var alg = _this13._getRoomDecryptor(key.room_id, key.algorithm);

    return alg.importRoomKey(key, opts)["finally"](function (r) {
      successes++;

      if (opts.progressCallback) {
        updateProgress();
      }
    });
  }));
};
/**
 * Schedules sending all keys waiting to be sent to the backup, if not already
 * scheduled. Retries if necessary.
 *
 * @param {number} maxDelay Maximum delay to wait in ms. 0 means no delay.
 */


Crypto.prototype.scheduleKeyBackupSend = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee36() {
  var maxDelay,
      delay,
      numFailures,
      numBackedUp,
      _args36 = arguments;
  return _regenerator["default"].wrap(function _callee36$(_context36) {
    while (1) {
      switch (_context36.prev = _context36.next) {
        case 0:
          maxDelay = _args36.length > 0 && _args36[0] !== undefined ? _args36[0] : 10000;

          if (!this._sendingBackups) {
            _context36.next = 3;
            break;
          }

          return _context36.abrupt("return");

        case 3:
          this._sendingBackups = true;
          _context36.prev = 4;
          // wait between 0 and `maxDelay` seconds, to avoid backup
          // requests from different clients hitting the server all at
          // the same time when a new key is sent
          delay = Math.random() * maxDelay;
          _context36.next = 8;
          return (0, utils.sleep)(delay);

        case 8:
          numFailures = 0; // number of consecutive failures

        case 9:
          if (!1) {
            _context36.next = 36;
            break;
          }

          if (this.backupKey) {
            _context36.next = 12;
            break;
          }

          return _context36.abrupt("return");

        case 12:
          _context36.prev = 12;
          _context36.next = 15;
          return this._backupPendingKeys(KEY_BACKUP_KEYS_PER_REQUEST);

        case 15:
          numBackedUp = _context36.sent;

          if (!(numBackedUp === 0)) {
            _context36.next = 18;
            break;
          }

          return _context36.abrupt("return");

        case 18:
          numFailures = 0;
          _context36.next = 31;
          break;

        case 21:
          _context36.prev = 21;
          _context36.t0 = _context36["catch"](12);
          numFailures++;

          _logger.logger.log("Key backup request failed", _context36.t0);

          if (!_context36.t0.data) {
            _context36.next = 31;
            break;
          }

          if (!(_context36.t0.data.errcode == 'M_NOT_FOUND' || _context36.t0.data.errcode == 'M_WRONG_ROOM_KEYS_VERSION')) {
            _context36.next = 31;
            break;
          }

          _context36.next = 29;
          return this.checkKeyBackup();

        case 29:
          // Backup version has changed or this backup version
          // has been deleted
          this.emit("crypto.keyBackupFailed", _context36.t0.data.errcode);
          throw _context36.t0;

        case 31:
          if (!numFailures) {
            _context36.next = 34;
            break;
          }

          _context36.next = 34;
          return (0, utils.sleep)(1000 * Math.pow(2, Math.min(numFailures - 1, 4)));

        case 34:
          _context36.next = 9;
          break;

        case 36:
          _context36.prev = 36;
          this._sendingBackups = false;
          return _context36.finish(36);

        case 39:
        case "end":
          return _context36.stop();
      }
    }
  }, _callee36, this, [[4,, 36, 39], [12, 21]]);
}));
/**
 * Take some e2e keys waiting to be backed up and send them
 * to the backup.
 *
 * @param {integer} limit Maximum number of keys to back up
 * @returns {integer} Number of sessions backed up
 */

Crypto.prototype._backupPendingKeys = /*#__PURE__*/function () {
  var _ref49 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee37(limit) {
    var sessions, remaining, data, _iterator4, _step4, session, roomId, sessionData, firstKnownIndex, encrypted, forwardedCount, userId, device, verified;

    return _regenerator["default"].wrap(function _callee37$(_context37) {
      while (1) {
        switch (_context37.prev = _context37.next) {
          case 0:
            _context37.next = 2;
            return this._cryptoStore.getSessionsNeedingBackup(limit);

          case 2:
            sessions = _context37.sent;

            if (sessions.length) {
              _context37.next = 5;
              break;
            }

            return _context37.abrupt("return", 0);

          case 5:
            _context37.next = 7;
            return this._cryptoStore.countSessionsNeedingBackup();

          case 7:
            remaining = _context37.sent;
            this.emit("crypto.keyBackupSessionsRemaining", remaining);
            data = {};
            _iterator4 = _createForOfIteratorHelper(sessions);
            _context37.prev = 11;

            _iterator4.s();

          case 13:
            if ((_step4 = _iterator4.n()).done) {
              _context37.next = 33;
              break;
            }

            session = _step4.value;
            roomId = session.sessionData.room_id;

            if (data[roomId] === undefined) {
              data[roomId] = {
                sessions: {}
              };
            }

            _context37.next = 19;
            return this._olmDevice.exportInboundGroupSession(session.senderKey, session.sessionId, session.sessionData);

          case 19:
            sessionData = _context37.sent;
            sessionData.algorithm = olmlib.MEGOLM_ALGORITHM;
            delete sessionData.session_id;
            delete sessionData.room_id;
            firstKnownIndex = sessionData.first_known_index;
            delete sessionData.first_known_index;
            encrypted = this.backupKey.encrypt(JSON.stringify(sessionData));
            forwardedCount = (sessionData.forwarding_curve25519_key_chain || []).length;
            userId = this._deviceList.getUserByIdentityKey(olmlib.MEGOLM_ALGORITHM, session.senderKey);
            device = this._deviceList.getDeviceByIdentityKey(olmlib.MEGOLM_ALGORITHM, session.senderKey);
            verified = this._checkDeviceInfoTrust(userId, device).isVerified();
            data[roomId]['sessions'][session.sessionId] = {
              first_message_index: firstKnownIndex,
              forwarded_count: forwardedCount,
              is_verified: verified,
              session_data: encrypted
            };

          case 31:
            _context37.next = 13;
            break;

          case 33:
            _context37.next = 38;
            break;

          case 35:
            _context37.prev = 35;
            _context37.t0 = _context37["catch"](11);

            _iterator4.e(_context37.t0);

          case 38:
            _context37.prev = 38;

            _iterator4.f();

            return _context37.finish(38);

          case 41:
            _context37.next = 43;
            return this._baseApis.sendKeyBackup(undefined, undefined, this.backupInfo.version, {
              rooms: data
            });

          case 43:
            _context37.next = 45;
            return this._cryptoStore.unmarkSessionsNeedingBackup(sessions);

          case 45:
            _context37.next = 47;
            return this._cryptoStore.countSessionsNeedingBackup();

          case 47:
            remaining = _context37.sent;
            this.emit("crypto.keyBackupSessionsRemaining", remaining);
            return _context37.abrupt("return", sessions.length);

          case 50:
          case "end":
            return _context37.stop();
        }
      }
    }, _callee37, this, [[11, 35, 38, 41]]);
  }));

  return function (_x39) {
    return _ref49.apply(this, arguments);
  };
}();

Crypto.prototype.backupGroupSession = /*#__PURE__*/function () {
  var _ref50 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee38(roomId, senderKey, forwardingCurve25519KeyChain, sessionId, sessionKey, keysClaimed, exportFormat) {
    return _regenerator["default"].wrap(function _callee38$(_context38) {
      while (1) {
        switch (_context38.prev = _context38.next) {
          case 0:
            if (this.backupInfo) {
              _context38.next = 2;
              break;
            }

            throw new Error("Key backups are not enabled");

          case 2:
            _context38.next = 4;
            return this._cryptoStore.markSessionsNeedingBackup([{
              senderKey: senderKey,
              sessionId: sessionId
            }]);

          case 4:
            // don't wait for this to complete: it will delay so
            // happens in the background
            this.scheduleKeyBackupSend();

          case 5:
          case "end":
            return _context38.stop();
        }
      }
    }, _callee38, this);
  }));

  return function (_x40, _x41, _x42, _x43, _x44, _x45, _x46) {
    return _ref50.apply(this, arguments);
  };
}();
/**
 * Marks all group sessions as needing to be backed up and schedules them to
 * upload in the background as soon as possible.
 */


Crypto.prototype.scheduleAllGroupSessionsForBackup = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee39() {
  return _regenerator["default"].wrap(function _callee39$(_context39) {
    while (1) {
      switch (_context39.prev = _context39.next) {
        case 0:
          _context39.next = 2;
          return this.flagAllGroupSessionsForBackup();

        case 2:
          // Schedule keys to upload in the background as soon as possible.
          this.scheduleKeyBackupSend(0
          /* maxDelay */
          );

        case 3:
        case "end":
          return _context39.stop();
      }
    }
  }, _callee39, this);
}));
/**
 * Marks all group sessions as needing to be backed up without scheduling
 * them to upload in the background.
 * @returns {Promise<int>} Resolves to the number of sessions now requiring a backup
 *     (which will be equal to the number of sessions in the store).
 */

Crypto.prototype.flagAllGroupSessionsForBackup = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee40() {
  var _this14 = this;

  var remaining;
  return _regenerator["default"].wrap(function _callee40$(_context40) {
    while (1) {
      switch (_context40.prev = _context40.next) {
        case 0:
          _context40.next = 2;
          return this._cryptoStore.doTxn('readwrite', [_indexeddbCryptoStore.IndexedDBCryptoStore.STORE_INBOUND_GROUP_SESSIONS, _indexeddbCryptoStore.IndexedDBCryptoStore.STORE_BACKUP], function (txn) {
            _this14._cryptoStore.getAllEndToEndInboundGroupSessions(txn, function (session) {
              if (session !== null) {
                _this14._cryptoStore.markSessionsNeedingBackup([session], txn);
              }
            });
          });

        case 2:
          _context40.next = 4;
          return this._cryptoStore.countSessionsNeedingBackup();

        case 4:
          remaining = _context40.sent;
          this.emit("crypto.keyBackupSessionsRemaining", remaining);
          return _context40.abrupt("return", remaining);

        case 7:
        case "end":
          return _context40.stop();
      }
    }
  }, _callee40, this);
}));
/**
 * Counts the number of end to end session keys that are waiting to be backed up
 * @returns {Promise<int>} Resolves to the number of sessions requiring backup
 */

Crypto.prototype.countSessionsNeedingBackup = function () {
  return this._cryptoStore.countSessionsNeedingBackup();
};
/**
 * Perform any background tasks that can be done before a message is ready to
 * send, in order to speed up sending of the message.
 *
 * @param {module:models/room} room the room the event is in
 */


Crypto.prototype.prepareToEncrypt = function (room) {
  var roomId = room.roomId;
  var alg = this._roomEncryptors[roomId];

  if (alg) {
    alg.prepareToEncrypt(room);
  }
};
/* eslint-disable valid-jsdoc */
//https://github.com/eslint/eslint/issues/7307

/**
 * Encrypt an event according to the configuration of the room.
 *
 * @param {module:models/event.MatrixEvent} event  event to be sent
 *
 * @param {module:models/room} room destination room.
 *
 * @return {Promise?} Promise which resolves when the event has been
 *     encrypted, or null if nothing was needed
 */

/* eslint-enable valid-jsdoc */


Crypto.prototype.encryptEvent = /*#__PURE__*/function () {
  var _ref53 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee41(event, room) {
    var roomId, alg, content, mRelatesTo, encryptedContent;
    return _regenerator["default"].wrap(function _callee41$(_context41) {
      while (1) {
        switch (_context41.prev = _context41.next) {
          case 0:
            if (room) {
              _context41.next = 2;
              break;
            }

            throw new Error("Cannot send encrypted messages in unknown rooms");

          case 2:
            roomId = event.getRoomId();
            alg = this._roomEncryptors[roomId];

            if (alg) {
              _context41.next = 6;
              break;
            }

            throw new Error("Room was previously configured to use encryption, but is " + "no longer. Perhaps the homeserver is hiding the " + "configuration event.");

          case 6:
            if (!this._roomDeviceTrackingState[roomId]) {
              this.trackRoomDevices(roomId);
            } // wait for all the room devices to be loaded


            _context41.next = 9;
            return this._roomDeviceTrackingState[roomId];

          case 9:
            content = event.getContent(); // If event has an m.relates_to then we need
            // to put this on the wrapping event instead

            mRelatesTo = content['m.relates_to'];

            if (mRelatesTo) {
              // Clone content here so we don't remove `m.relates_to` from the local-echo
              content = Object.assign({}, content);
              delete content['m.relates_to'];
            }

            _context41.next = 14;
            return alg.encryptMessage(room, event.getType(), content);

          case 14:
            encryptedContent = _context41.sent;

            if (mRelatesTo) {
              encryptedContent['m.relates_to'] = mRelatesTo;
            }

            event.makeEncrypted("m.room.encrypted", encryptedContent, this._olmDevice.deviceCurve25519Key, this._olmDevice.deviceEd25519Key);

          case 17:
          case "end":
            return _context41.stop();
        }
      }
    }, _callee41, this);
  }));

  return function (_x47, _x48) {
    return _ref53.apply(this, arguments);
  };
}();
/**
 * Decrypt a received event
 *
 * @param {MatrixEvent} event
 *
 * @return {Promise<module:crypto~EventDecryptionResult>} resolves once we have
 *  finished decrypting. Rejects with an `algorithms.DecryptionError` if there
 *  is a problem decrypting the event.
 */


Crypto.prototype.decryptEvent = function (event) {
  if (event.isRedacted()) {
    return Promise.resolve({
      clearEvent: {
        room_id: event.getRoomId(),
        type: "m.room.message",
        content: {}
      }
    });
  }

  var content = event.getWireContent();

  var alg = this._getRoomDecryptor(event.getRoomId(), content.algorithm);

  return alg.decryptEvent(event);
};
/**
 * Handle the notification from /sync or /keys/changes that device lists have
 * been changed.
 *
 * @param {Object} syncData Object containing sync tokens associated with this sync
 * @param {Object} syncDeviceLists device_lists field from /sync, or response from
 * /keys/changes
 */


Crypto.prototype.handleDeviceListChanges = /*#__PURE__*/function () {
  var _ref54 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee42(syncData, syncDeviceLists) {
    return _regenerator["default"].wrap(function _callee42$(_context42) {
      while (1) {
        switch (_context42.prev = _context42.next) {
          case 0:
            if (syncData.oldSyncToken) {
              _context42.next = 2;
              break;
            }

            return _context42.abrupt("return");

          case 2:
            _context42.next = 4;
            return this._evalDeviceListChanges(syncDeviceLists);

          case 4:
          case "end":
            return _context42.stop();
        }
      }
    }, _callee42, this);
  }));

  return function (_x49, _x50) {
    return _ref54.apply(this, arguments);
  };
}();
/**
 * Send a request for some room keys, if we have not already done so
 *
 * @param {module:crypto~RoomKeyRequestBody} requestBody
 * @param {Array<{userId: string, deviceId: string}>} recipients
 * @param {boolean} resend whether to resend the key request if there is
 *    already one
 *
 * @return {Promise} a promise that resolves when the key request is queued
 */


Crypto.prototype.requestRoomKey = function (requestBody, recipients) {
  var _this15 = this;

  var resend = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;
  return this._outgoingRoomKeyRequestManager.queueRoomKeyRequest(requestBody, recipients, resend).then(function () {
    if (_this15._sendKeyRequestsImmediately) {
      _this15._outgoingRoomKeyRequestManager.sendQueuedRequests();
    }
  })["catch"](function (e) {
    // this normally means we couldn't talk to the store
    _logger.logger.error('Error requesting key for event', e);
  });
};
/**
 * Cancel any earlier room key request
 *
 * @param {module:crypto~RoomKeyRequestBody} requestBody
 *    parameters to match for cancellation
 */


Crypto.prototype.cancelRoomKeyRequest = function (requestBody) {
  this._outgoingRoomKeyRequestManager.cancelRoomKeyRequest(requestBody)["catch"](function (e) {
    _logger.logger.warn("Error clearing pending room key requests", e);
  });
};
/**
 * Re-send any outgoing key requests, eg after verification
 * @returns {Promise}
 */


Crypto.prototype.cancelAndResendAllOutgoingKeyRequests = function () {
  return this._outgoingRoomKeyRequestManager.cancelAndResendAllOutgoingRequests();
};
/**
 * handle an m.room.encryption event
 *
 * @param {module:models/event.MatrixEvent} event encryption event
 */


Crypto.prototype.onCryptoEvent = /*#__PURE__*/function () {
  var _ref55 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee43(event) {
    var roomId, content;
    return _regenerator["default"].wrap(function _callee43$(_context43) {
      while (1) {
        switch (_context43.prev = _context43.next) {
          case 0:
            roomId = event.getRoomId();
            content = event.getContent();
            _context43.prev = 2;
            _context43.next = 5;
            return this.setRoomEncryption(roomId, content, true);

          case 5:
            _context43.next = 10;
            break;

          case 7:
            _context43.prev = 7;
            _context43.t0 = _context43["catch"](2);

            _logger.logger.error("Error configuring encryption in room " + roomId + ":", _context43.t0);

          case 10:
          case "end":
            return _context43.stop();
        }
      }
    }, _callee43, this, [[2, 7]]);
  }));

  return function (_x51) {
    return _ref55.apply(this, arguments);
  };
}();
/**
 * Called before the result of a sync is procesed
 *
 * @param {Object} syncData  the data from the 'MatrixClient.sync' event
 */


Crypto.prototype.onSyncWillProcess = /*#__PURE__*/function () {
  var _ref56 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee44(syncData) {
    return _regenerator["default"].wrap(function _callee44$(_context44) {
      while (1) {
        switch (_context44.prev = _context44.next) {
          case 0:
            if (!syncData.oldSyncToken) {
              // If there is no old sync token, we start all our tracking from
              // scratch, so mark everything as untracked. onCryptoEvent will
              // be called for all e2e rooms during the processing of the sync,
              // at which point we'll start tracking all the users of that room.
              _logger.logger.log("Initial sync performed - resetting device tracking state");

              this._deviceList.stopTrackingAllDeviceLists(); // we always track our own device list (for key backups etc)


              this._deviceList.startTrackingDeviceList(this._userId);

              this._roomDeviceTrackingState = {};
            }

            this._sendKeyRequestsImmediately = false;

          case 2:
          case "end":
            return _context44.stop();
        }
      }
    }, _callee44, this);
  }));

  return function (_x52) {
    return _ref56.apply(this, arguments);
  };
}();
/**
 * handle the completion of a /sync
 *
 * This is called after the processing of each successful /sync response.
 * It is an opportunity to do a batch process on the information received.
 *
 * @param {Object} syncData  the data from the 'MatrixClient.sync' event
 */


Crypto.prototype.onSyncCompleted = /*#__PURE__*/function () {
  var _ref57 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee45(syncData) {
    var nextSyncToken;
    return _regenerator["default"].wrap(function _callee45$(_context45) {
      while (1) {
        switch (_context45.prev = _context45.next) {
          case 0:
            nextSyncToken = syncData.nextSyncToken;

            this._deviceList.setSyncToken(syncData.nextSyncToken);

            this._deviceList.saveIfDirty(); // catch up on any new devices we got told about during the sync.


            this._deviceList.lastKnownSyncToken = nextSyncToken; // we always track our own device list (for key backups etc)

            this._deviceList.startTrackingDeviceList(this._userId);

            this._deviceList.refreshOutdatedDeviceLists(); // we don't start uploading one-time keys until we've caught up with
            // to-device messages, to help us avoid throwing away one-time-keys that we
            // are about to receive messages for
            // (https://github.com/vector-im/element-web/issues/2782).


            if (!syncData.catchingUp) {
              _maybeUploadOneTimeKeys(this);

              this._processReceivedRoomKeyRequests(); // likewise don't start requesting keys until we've caught up
              // on to_device messages, otherwise we'll request keys that we're
              // just about to get.


              this._outgoingRoomKeyRequestManager.sendQueuedRequests(); // Sync has finished so send key requests straight away.


              this._sendKeyRequestsImmediately = true;
            }

          case 7:
          case "end":
            return _context45.stop();
        }
      }
    }, _callee45, this);
  }));

  return function (_x53) {
    return _ref57.apply(this, arguments);
  };
}();
/**
 * Trigger the appropriate invalidations and removes for a given
 * device list
 *
 * @param {Object} deviceLists device_lists field from /sync, or response from
 * /keys/changes
 */


Crypto.prototype._evalDeviceListChanges = /*#__PURE__*/function () {
  var _ref58 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee46(deviceLists) {
    var _this16 = this;

    var e2eUserIds;
    return _regenerator["default"].wrap(function _callee46$(_context46) {
      while (1) {
        switch (_context46.prev = _context46.next) {
          case 0:
            if (deviceLists.changed && Array.isArray(deviceLists.changed)) {
              deviceLists.changed.forEach(function (u) {
                _this16._deviceList.invalidateUserDeviceList(u);
              });
            }

            if (!(deviceLists.left && Array.isArray(deviceLists.left) && deviceLists.left.length)) {
              _context46.next = 8;
              break;
            }

            _context46.t0 = Set;
            _context46.next = 5;
            return this._getTrackedE2eUsers();

          case 5:
            _context46.t1 = _context46.sent;
            e2eUserIds = new _context46.t0(_context46.t1);
            deviceLists.left.forEach(function (u) {
              if (!e2eUserIds.has(u)) {
                _this16._deviceList.stopTrackingDeviceList(u);
              }
            });

          case 8:
          case "end":
            return _context46.stop();
        }
      }
    }, _callee46, this);
  }));

  return function (_x54) {
    return _ref58.apply(this, arguments);
  };
}();
/**
 * Get a list of all the IDs of users we share an e2e room with
 * for which we are tracking devices already
 *
 * @returns {string[]} List of user IDs
 */


Crypto.prototype._getTrackedE2eUsers = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee47() {
  var e2eUserIds, _iterator5, _step5, room, members, _iterator6, _step6, member;

  return _regenerator["default"].wrap(function _callee47$(_context47) {
    while (1) {
      switch (_context47.prev = _context47.next) {
        case 0:
          e2eUserIds = [];
          _iterator5 = _createForOfIteratorHelper(this._getTrackedE2eRooms());
          _context47.prev = 2;

          _iterator5.s();

        case 4:
          if ((_step5 = _iterator5.n()).done) {
            _context47.next = 13;
            break;
          }

          room = _step5.value;
          _context47.next = 8;
          return room.getEncryptionTargetMembers();

        case 8:
          members = _context47.sent;
          _iterator6 = _createForOfIteratorHelper(members);

          try {
            for (_iterator6.s(); !(_step6 = _iterator6.n()).done;) {
              member = _step6.value;
              e2eUserIds.push(member.userId);
            }
          } catch (err) {
            _iterator6.e(err);
          } finally {
            _iterator6.f();
          }

        case 11:
          _context47.next = 4;
          break;

        case 13:
          _context47.next = 18;
          break;

        case 15:
          _context47.prev = 15;
          _context47.t0 = _context47["catch"](2);

          _iterator5.e(_context47.t0);

        case 18:
          _context47.prev = 18;

          _iterator5.f();

          return _context47.finish(18);

        case 21:
          return _context47.abrupt("return", e2eUserIds);

        case 22:
        case "end":
          return _context47.stop();
      }
    }
  }, _callee47, this, [[2, 15, 18, 21]]);
}));
/**
 * Get a list of the e2e-enabled rooms we are members of,
 * and for which we are already tracking the devices
 *
 * @returns {module:models.Room[]}
 */

Crypto.prototype._getTrackedE2eRooms = function () {
  var _this17 = this;

  return this._clientStore.getRooms().filter(function (room) {
    // check for rooms with encryption enabled
    var alg = _this17._roomEncryptors[room.roomId];

    if (!alg) {
      return false;
    }

    if (!_this17._roomDeviceTrackingState[room.roomId]) {
      return false;
    } // ignore any rooms which we have left


    var myMembership = room.getMyMembership();
    return myMembership === "join" || myMembership === "invite";
  });
};

Crypto.prototype._onToDeviceEvent = function (event) {
  var _this18 = this;

  try {
    _logger.logger.log("received to_device ".concat(event.getType(), " from: ") + "".concat(event.getSender(), " id: ").concat(event.getId()));

    if (event.getType() == "m.room_key" || event.getType() == "m.forwarded_room_key") {
      this._onRoomKeyEvent(event);
    } else if (event.getType() == "m.room_key_request") {
      this._onRoomKeyRequestEvent(event);
    } else if (event.getType() === "m.secret.request") {
      this._secretStorage._onRequestReceived(event);
    } else if (event.getType() === "m.secret.send") {
      this._secretStorage._onSecretReceived(event);
    } else if (event.getType() === "org.matrix.room_key.withheld") {
      this._onRoomKeyWithheldEvent(event);
    } else if (event.getContent().transaction_id) {
      this._onKeyVerificationMessage(event);
    } else if (event.getContent().msgtype === "m.bad.encrypted") {
      this._onToDeviceBadEncrypted(event);
    } else if (event.isBeingDecrypted()) {
      // once the event has been decrypted, try again
      event.once('Event.decrypted', function (ev) {
        _this18._onToDeviceEvent(ev);
      });
    }
  } catch (e) {
    _logger.logger.error("Error handling toDeviceEvent:", e);
  }
};
/**
 * Handle a key event
 *
 * @private
 * @param {module:models/event.MatrixEvent} event key event
 */


Crypto.prototype._onRoomKeyEvent = function (event) {
  var content = event.getContent();

  if (!content.room_id || !content.algorithm) {
    _logger.logger.error("key event is missing fields");

    return;
  }

  if (!this._checkedForBackup) {
    // don't bother awaiting on this - the important thing is that we retry if we
    // haven't managed to check before
    this._checkAndStartKeyBackup();
  }

  var alg = this._getRoomDecryptor(content.room_id, content.algorithm);

  alg.onRoomKeyEvent(event);
};
/**
 * Handle a key withheld event
 *
 * @private
 * @param {module:models/event.MatrixEvent} event key withheld event
 */


Crypto.prototype._onRoomKeyWithheldEvent = function (event) {
  var content = event.getContent();

  if (content.code !== "m.no_olm" && (!content.room_id || !content.session_id) || !content.algorithm || !content.sender_key) {
    _logger.logger.error("key withheld event is missing fields");

    return;
  }

  _logger.logger.info("Got room key withheld event from ".concat(event.getSender(), " (").concat(content.sender_key, ") ") + "for ".concat(content.algorithm, "/").concat(content.room_id, "/").concat(content.session_id, " ") + "with reason ".concat(content.code, " (").concat(content.reason, ")"));

  var alg = this._getRoomDecryptor(content.room_id, content.algorithm);

  if (alg.onRoomKeyWithheldEvent) {
    alg.onRoomKeyWithheldEvent(event);
  }

  if (!content.room_id) {
    // retry decryption for all events sent by the sender_key.  This will
    // update the events to show a message indicating that the olm session was
    // wedged.
    var roomDecryptors = this._getRoomDecryptors(content.algorithm);

    var _iterator7 = _createForOfIteratorHelper(roomDecryptors),
        _step7;

    try {
      for (_iterator7.s(); !(_step7 = _iterator7.n()).done;) {
        var decryptor = _step7.value;
        decryptor.retryDecryptionFromSender(content.sender_key);
      }
    } catch (err) {
      _iterator7.e(err);
    } finally {
      _iterator7.f();
    }
  }
};
/**
 * Handle a general key verification event.
 *
 * @private
 * @param {module:models/event.MatrixEvent} event verification start event
 */


Crypto.prototype._onKeyVerificationMessage = function (event) {
  var _this19 = this;

  if (!_ToDeviceChannel.ToDeviceChannel.validateEvent(event, this._baseApis)) {
    return;
  }

  var createRequest = function createRequest(event) {
    if (!_ToDeviceChannel.ToDeviceChannel.canCreateRequest(_ToDeviceChannel.ToDeviceChannel.getEventType(event))) {
      return;
    }

    var content = event.getContent();
    var deviceId = content && content.from_device;

    if (!deviceId) {
      return;
    }

    var userId = event.getSender();
    var channel = new _ToDeviceChannel.ToDeviceChannel(_this19._baseApis, userId, [deviceId]);
    return new _VerificationRequest.VerificationRequest(channel, _this19._verificationMethods, _this19._baseApis);
  };

  this._handleVerificationEvent(event, this._toDeviceVerificationRequests, createRequest);
};
/**
 * Handle key verification requests sent as timeline events
 *
 * @private
 * @param {module:models/event.MatrixEvent} event the timeline event
 * @param {module:models/Room} room not used
 * @param {bool} atStart not used
 * @param {bool} removed not used
 * @param {bool} data.liveEvent whether this is a live event
 */


Crypto.prototype._onTimelineEvent = function (event, room, atStart, removed) {
  var _this20 = this;

  var _ref60 = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {},
      liveEvent = _ref60.liveEvent;

  if (!_InRoomChannel.InRoomChannel.validateEvent(event, this._baseApis)) {
    return;
  }

  var createRequest = function createRequest(event) {
    var channel = new _InRoomChannel.InRoomChannel(_this20._baseApis, event.getRoomId());
    return new _VerificationRequest.VerificationRequest(channel, _this20._verificationMethods, _this20._baseApis);
  };

  this._handleVerificationEvent(event, this._inRoomVerificationRequests, createRequest, liveEvent);
};

Crypto.prototype._handleVerificationEvent = /*#__PURE__*/function () {
  var _ref61 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee48(event, requestsMap, createRequest) {
    var isLiveEvent,
        request,
        isNewRequest,
        shouldEmit,
        _args48 = arguments;
    return _regenerator["default"].wrap(function _callee48$(_context48) {
      while (1) {
        switch (_context48.prev = _context48.next) {
          case 0:
            isLiveEvent = _args48.length > 3 && _args48[3] !== undefined ? _args48[3] : true;
            request = requestsMap.getRequest(event);
            isNewRequest = false;

            if (request) {
              _context48.next = 10;
              break;
            }

            request = createRequest(event); // a request could not be made from this event, so ignore event

            if (request) {
              _context48.next = 8;
              break;
            }

            _logger.logger.log("Crypto: could not find VerificationRequest for " + "".concat(event.getType(), ", and could not create one, so ignoring."));

            return _context48.abrupt("return");

          case 8:
            isNewRequest = true;
            requestsMap.setRequest(event, request);

          case 10:
            event.setVerificationRequest(request);
            _context48.prev = 11;
            _context48.next = 14;
            return request.channel.handleEvent(event, request, isLiveEvent);

          case 14:
            _context48.next = 19;
            break;

          case 16:
            _context48.prev = 16;
            _context48.t0 = _context48["catch"](11);

            _logger.logger.error("error while handling verification event: " + _context48.t0.message);

          case 19:
            shouldEmit = isNewRequest && !request.initiatedByMe && !request.invalid && // check it has enough events to pass the UNSENT stage
            !request.observeOnly;

            if (shouldEmit) {
              this._baseApis.emit("crypto.verification.request", request);
            }

          case 21:
          case "end":
            return _context48.stop();
        }
      }
    }, _callee48, this, [[11, 16]]);
  }));

  return function (_x55, _x56, _x57) {
    return _ref61.apply(this, arguments);
  };
}();
/**
 * Handle a toDevice event that couldn't be decrypted
 *
 * @private
 * @param {module:models/event.MatrixEvent} event undecryptable event
 */


Crypto.prototype._onToDeviceBadEncrypted = /*#__PURE__*/function () {
  var _ref62 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee49(event) {
    var _this21 = this;

    var content, sender, algorithm, deviceKey, retryDecryption, lastNewSessionForced, device, devicesByUser, encryptedContent, requestsToResend, _iterator9, _step9, keyReq;

    return _regenerator["default"].wrap(function _callee49$(_context49) {
      while (1) {
        switch (_context49.prev = _context49.next) {
          case 0:
            content = event.getWireContent();
            sender = event.getSender();
            algorithm = content.algorithm;
            deviceKey = content.sender_key; // retry decryption for all events sent by the sender_key.  This will
            // update the events to show a message indicating that the olm session was
            // wedged.

            retryDecryption = function retryDecryption() {
              var roomDecryptors = _this21._getRoomDecryptors(olmlib.MEGOLM_ALGORITHM);

              var _iterator8 = _createForOfIteratorHelper(roomDecryptors),
                  _step8;

              try {
                for (_iterator8.s(); !(_step8 = _iterator8.n()).done;) {
                  var decryptor = _step8.value;
                  decryptor.retryDecryptionFromSender(deviceKey);
                }
              } catch (err) {
                _iterator8.e(err);
              } finally {
                _iterator8.f();
              }
            };

            if (!(sender === undefined || deviceKey === undefined || deviceKey === undefined)) {
              _context49.next = 7;
              break;
            }

            return _context49.abrupt("return");

          case 7:
            // check when we last forced a new session with this device: if we've already done so
            // recently, don't do it again.
            this._lastNewSessionForced[sender] = this._lastNewSessionForced[sender] || {};
            lastNewSessionForced = this._lastNewSessionForced[sender][deviceKey] || 0;

            if (!(lastNewSessionForced + MIN_FORCE_SESSION_INTERVAL_MS > Date.now())) {
              _context49.next = 15;
              break;
            }

            _logger.logger.debug("New session already forced with device " + sender + ":" + deviceKey + " at " + lastNewSessionForced + ": not forcing another");

            _context49.next = 13;
            return this._olmDevice.recordSessionProblem(deviceKey, "wedged", true);

          case 13:
            retryDecryption();
            return _context49.abrupt("return");

          case 15:
            // establish a new olm session with this device since we're failing to decrypt messages
            // on a current session.
            // Note that an undecryptable message from another device could easily be spoofed -
            // is there anything we can do to mitigate this?
            device = this._deviceList.getDeviceByIdentityKey(algorithm, deviceKey);

            if (device) {
              _context49.next = 26;
              break;
            }

            _context49.next = 19;
            return this.downloadKeys([sender], false);

          case 19:
            device = this._deviceList.getDeviceByIdentityKey(algorithm, deviceKey);

            if (device) {
              _context49.next = 26;
              break;
            }

            _logger.logger.info("Couldn't find device for identity key " + deviceKey + ": not re-establishing session");

            _context49.next = 24;
            return this._olmDevice.recordSessionProblem(deviceKey, "wedged", false);

          case 24:
            retryDecryption();
            return _context49.abrupt("return");

          case 26:
            devicesByUser = {};
            devicesByUser[sender] = [device];
            _context49.next = 30;
            return olmlib.ensureOlmSessionsForDevices(this._olmDevice, this._baseApis, devicesByUser, true);

          case 30:
            this._lastNewSessionForced[sender][deviceKey] = Date.now(); // Now send a blank message on that session so the other side knows about it.
            // (The keyshare request is sent in the clear so that won't do)
            // We send this first such that, as long as the toDevice messages arrive in the
            // same order we sent them, the other end will get this first, set up the new session,
            // then get the keyshare request and send the key over this new session (because it
            // is the session it has most recently received a message on).

            encryptedContent = {
              algorithm: olmlib.OLM_ALGORITHM,
              sender_key: this._olmDevice.deviceCurve25519Key,
              ciphertext: {}
            };
            _context49.next = 34;
            return olmlib.encryptMessageForDevice(encryptedContent.ciphertext, this._userId, this._deviceId, this._olmDevice, sender, device, {
              type: "m.dummy"
            });

          case 34:
            _context49.next = 36;
            return this._olmDevice.recordSessionProblem(deviceKey, "wedged", true);

          case 36:
            retryDecryption();
            _context49.next = 39;
            return this._baseApis.sendToDevice("m.room.encrypted", (0, _defineProperty2["default"])({}, sender, (0, _defineProperty2["default"])({}, device.deviceId, encryptedContent)));

          case 39:
            _context49.next = 41;
            return this._outgoingRoomKeyRequestManager.getOutgoingSentRoomKeyRequest(sender, device.deviceId);

          case 41:
            requestsToResend = _context49.sent;
            _iterator9 = _createForOfIteratorHelper(requestsToResend);

            try {
              for (_iterator9.s(); !(_step9 = _iterator9.n()).done;) {
                keyReq = _step9.value;
                this.requestRoomKey(keyReq.requestBody, keyReq.recipients, true);
              }
            } catch (err) {
              _iterator9.e(err);
            } finally {
              _iterator9.f();
            }

          case 44:
          case "end":
            return _context49.stop();
        }
      }
    }, _callee49, this);
  }));

  return function (_x58) {
    return _ref62.apply(this, arguments);
  };
}();
/**
 * Handle a change in the membership state of a member of a room
 *
 * @private
 * @param {module:models/event.MatrixEvent} event  event causing the change
 * @param {module:models/room-member} member  user whose membership changed
 * @param {string=} oldMembership  previous membership
 */


Crypto.prototype._onRoomMembership = function (event, member, oldMembership) {
  // this event handler is registered on the *client* (as opposed to the room
  // member itself), which means it is only called on changes to the *live*
  // membership state (ie, it is not called when we back-paginate, nor when
  // we load the state in the initialsync).
  //
  // Further, it is automatically registered and called when new members
  // arrive in the room.
  var roomId = member.roomId;
  var alg = this._roomEncryptors[roomId];

  if (!alg) {
    // not encrypting in this room
    return;
  } // only mark users in this room as tracked if we already started tracking in this room
  // this way we don't start device queries after sync on behalf of this room which we won't use
  // the result of anyway, as we'll need to do a query again once all the members are fetched
  // by calling _trackRoomDevices


  if (this._roomDeviceTrackingState[roomId]) {
    if (member.membership == 'join') {
      _logger.logger.log('Join event for ' + member.userId + ' in ' + roomId); // make sure we are tracking the deviceList for this user


      this._deviceList.startTrackingDeviceList(member.userId);
    } else if (member.membership == 'invite' && this._clientStore.getRoom(roomId).shouldEncryptForInvitedMembers()) {
      _logger.logger.log('Invite event for ' + member.userId + ' in ' + roomId);

      this._deviceList.startTrackingDeviceList(member.userId);
    }
  }

  alg.onRoomMembership(event, member, oldMembership);
};
/**
 * Called when we get an m.room_key_request event.
 *
 * @private
 * @param {module:models/event.MatrixEvent} event key request event
 */


Crypto.prototype._onRoomKeyRequestEvent = function (event) {
  var content = event.getContent();

  if (content.action === "request") {
    // Queue it up for now, because they tend to arrive before the room state
    // events at initial sync, and we want to see if we know anything about the
    // room before passing them on to the app.
    var req = new IncomingRoomKeyRequest(event);

    this._receivedRoomKeyRequests.push(req);
  } else if (content.action === "request_cancellation") {
    var _req = new IncomingRoomKeyRequestCancellation(event);

    this._receivedRoomKeyRequestCancellations.push(_req);
  }
};
/**
 * Process any m.room_key_request events which were queued up during the
 * current sync.
 *
 * @private
 */


Crypto.prototype._processReceivedRoomKeyRequests = /*#__PURE__*/(0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee50() {
  var _this22 = this;

  var requests, cancellations;
  return _regenerator["default"].wrap(function _callee50$(_context50) {
    while (1) {
      switch (_context50.prev = _context50.next) {
        case 0:
          if (!this._processingRoomKeyRequests) {
            _context50.next = 2;
            break;
          }

          return _context50.abrupt("return");

        case 2:
          this._processingRoomKeyRequests = true;
          _context50.prev = 3;
          // we need to grab and clear the queues in the synchronous bit of this method,
          // so that we don't end up racing with the next /sync.
          requests = this._receivedRoomKeyRequests;
          this._receivedRoomKeyRequests = [];
          cancellations = this._receivedRoomKeyRequestCancellations;
          this._receivedRoomKeyRequestCancellations = []; // Process all of the requests, *then* all of the cancellations.
          //
          // This makes sure that if we get a request and its cancellation in the
          // same /sync result, then we process the request before the
          // cancellation (and end up with a cancelled request), rather than the
          // cancellation before the request (and end up with an outstanding
          // request which should have been cancelled.)

          _context50.next = 10;
          return Promise.all(requests.map(function (req) {
            return _this22._processReceivedRoomKeyRequest(req);
          }));

        case 10:
          _context50.next = 12;
          return Promise.all(cancellations.map(function (cancellation) {
            return _this22._processReceivedRoomKeyRequestCancellation(cancellation);
          }));

        case 12:
          _context50.next = 17;
          break;

        case 14:
          _context50.prev = 14;
          _context50.t0 = _context50["catch"](3);

          _logger.logger.error("Error processing room key requsts: ".concat(_context50.t0));

        case 17:
          _context50.prev = 17;
          this._processingRoomKeyRequests = false;
          return _context50.finish(17);

        case 20:
        case "end":
          return _context50.stop();
      }
    }
  }, _callee50, this, [[3, 14, 17, 20]]);
}));
/**
 * Helper for processReceivedRoomKeyRequests
 *
 * @param {IncomingRoomKeyRequest} req
 */

Crypto.prototype._processReceivedRoomKeyRequest = /*#__PURE__*/function () {
  var _ref64 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee51(req) {
    var userId, deviceId, body, roomId, alg, encryptor, device, decryptor;
    return _regenerator["default"].wrap(function _callee51$(_context51) {
      while (1) {
        switch (_context51.prev = _context51.next) {
          case 0:
            userId = req.userId;
            deviceId = req.deviceId;
            body = req.requestBody;
            roomId = body.room_id;
            alg = body.algorithm;

            _logger.logger.log("m.room_key_request from ".concat(userId, ":").concat(deviceId) + " for ".concat(roomId, " / ").concat(body.session_id, " (id ").concat(req.requestId, ")"));

            if (!(userId !== this._userId)) {
              _context51.next = 24;
              break;
            }

            if (this._roomEncryptors[roomId]) {
              _context51.next = 10;
              break;
            }

            _logger.logger.debug("room key request for unencrypted room ".concat(roomId));

            return _context51.abrupt("return");

          case 10:
            encryptor = this._roomEncryptors[roomId];
            device = this._deviceList.getStoredDevice(userId, deviceId);

            if (device) {
              _context51.next = 15;
              break;
            }

            _logger.logger.debug("Ignoring keyshare for unknown device ".concat(userId, ":").concat(deviceId));

            return _context51.abrupt("return");

          case 15:
            _context51.prev = 15;
            _context51.next = 18;
            return encryptor.reshareKeyWithDevice(body.sender_key, body.session_id, userId, device);

          case 18:
            _context51.next = 23;
            break;

          case 20:
            _context51.prev = 20;
            _context51.t0 = _context51["catch"](15);

            _logger.logger.warn("Failed to re-share keys for session " + body.session_id + " with device " + userId + ":" + device.deviceId, _context51.t0);

          case 23:
            return _context51.abrupt("return");

          case 24:
            if (!(deviceId === this._deviceId)) {
              _context51.next = 27;
              break;
            }

            // We'll always get these because we send room key requests to
            // '*' (ie. 'all devices') which includes the sending device,
            // so ignore requests from ourself because apart from it being
            // very silly, it won't work because an Olm session cannot send
            // messages to itself.
            // The log here is probably superfluous since we know this will
            // always happen, but let's log anyway for now just in case it
            // causes issues.
            _logger.logger.log("Ignoring room key request from ourselves");

            return _context51.abrupt("return");

          case 27:
            if (this._roomDecryptors[roomId]) {
              _context51.next = 30;
              break;
            }

            _logger.logger.log("room key request for unencrypted room ".concat(roomId));

            return _context51.abrupt("return");

          case 30:
            decryptor = this._roomDecryptors[roomId][alg];

            if (decryptor) {
              _context51.next = 34;
              break;
            }

            _logger.logger.log("room key request for unknown alg ".concat(alg, " in room ").concat(roomId));

            return _context51.abrupt("return");

          case 34:
            _context51.next = 36;
            return decryptor.hasKeysForKeyRequest(req);

          case 36:
            if (_context51.sent) {
              _context51.next = 39;
              break;
            }

            _logger.logger.log("room key request for unknown session ".concat(roomId, " / ") + body.session_id);

            return _context51.abrupt("return");

          case 39:
            req.share = function () {
              decryptor.shareKeysWithDevice(req);
            }; // if the device is verified already, share the keys


            if (!this.checkDeviceTrust(userId, deviceId).isVerified()) {
              _context51.next = 44;
              break;
            }

            _logger.logger.log('device is already verified: sharing keys');

            req.share();
            return _context51.abrupt("return");

          case 44:
            this.emit("crypto.roomKeyRequest", req);

          case 45:
          case "end":
            return _context51.stop();
        }
      }
    }, _callee51, this, [[15, 20]]);
  }));

  return function (_x59) {
    return _ref64.apply(this, arguments);
  };
}();
/**
 * Helper for processReceivedRoomKeyRequests
 *
 * @param {IncomingRoomKeyRequestCancellation} cancellation
 */


Crypto.prototype._processReceivedRoomKeyRequestCancellation = /*#__PURE__*/function () {
  var _ref65 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee52(cancellation) {
    return _regenerator["default"].wrap(function _callee52$(_context52) {
      while (1) {
        switch (_context52.prev = _context52.next) {
          case 0:
            _logger.logger.log("m.room_key_request cancellation for ".concat(cancellation.userId, ":") + "".concat(cancellation.deviceId, " (id ").concat(cancellation.requestId, ")")); // we should probably only notify the app of cancellations we told it
            // about, but we don't currently have a record of that, so we just pass
            // everything through.


            this.emit("crypto.roomKeyRequestCancellation", cancellation);

          case 2:
          case "end":
            return _context52.stop();
        }
      }
    }, _callee52, this);
  }));

  return function (_x60) {
    return _ref65.apply(this, arguments);
  };
}();
/**
 * Get a decryptor for a given room and algorithm.
 *
 * If we already have a decryptor for the given room and algorithm, return
 * it. Otherwise try to instantiate it.
 *
 * @private
 *
 * @param {string?} roomId   room id for decryptor. If undefined, a temporary
 * decryptor is instantiated.
 *
 * @param {string} algorithm  crypto algorithm
 *
 * @return {module:crypto.algorithms.base.DecryptionAlgorithm}
 *
 * @raises {module:crypto.algorithms.DecryptionError} if the algorithm is
 * unknown
 */


Crypto.prototype._getRoomDecryptor = function (roomId, algorithm) {
  var decryptors;
  var alg;
  roomId = roomId || null;

  if (roomId) {
    decryptors = this._roomDecryptors[roomId];

    if (!decryptors) {
      this._roomDecryptors[roomId] = decryptors = {};
    }

    alg = decryptors[algorithm];

    if (alg) {
      return alg;
    }
  }

  var AlgClass = algorithms.DECRYPTION_CLASSES[algorithm];

  if (!AlgClass) {
    throw new algorithms.DecryptionError('UNKNOWN_ENCRYPTION_ALGORITHM', 'Unknown encryption algorithm "' + algorithm + '".');
  }

  alg = new AlgClass({
    userId: this._userId,
    crypto: this,
    olmDevice: this._olmDevice,
    baseApis: this._baseApis,
    roomId: roomId
  });

  if (decryptors) {
    decryptors[algorithm] = alg;
  }

  return alg;
};
/**
 * Get all the room decryptors for a given encryption algorithm.
 *
 * @param {string} algorithm The encryption algorithm
 *
 * @return {array} An array of room decryptors
 */


Crypto.prototype._getRoomDecryptors = function (algorithm) {
  var decryptors = [];

  for (var _i5 = 0, _Object$values = Object.values(this._roomDecryptors); _i5 < _Object$values.length; _i5++) {
    var d = _Object$values[_i5];

    if (algorithm in d) {
      decryptors.push(d[algorithm]);
    }
  }

  return decryptors;
};
/**
 * sign the given object with our ed25519 key
 *
 * @param {Object} obj  Object to which we will add a 'signatures' property
 */


Crypto.prototype._signObject = /*#__PURE__*/function () {
  var _ref66 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee53(obj) {
    var sigs, unsigned;
    return _regenerator["default"].wrap(function _callee53$(_context53) {
      while (1) {
        switch (_context53.prev = _context53.next) {
          case 0:
            sigs = obj.signatures || {};
            unsigned = obj.unsigned;
            delete obj.signatures;
            delete obj.unsigned;
            sigs[this._userId] = sigs[this._userId] || {};
            _context53.next = 7;
            return this._olmDevice.sign(_anotherJson["default"].stringify(obj));

          case 7:
            sigs[this._userId]["ed25519:" + this._deviceId] = _context53.sent;
            obj.signatures = sigs;
            if (unsigned !== undefined) obj.unsigned = unsigned;

          case 10:
          case "end":
            return _context53.stop();
        }
      }
    }, _callee53, this);
  }));

  return function (_x61) {
    return _ref66.apply(this, arguments);
  };
}();
/**
 * The parameters of a room key request. The details of the request may
 * vary with the crypto algorithm, but the management and storage layers for
 * outgoing requests expect it to have 'room_id' and 'session_id' properties.
 *
 * @typedef {Object} RoomKeyRequestBody
 */

/**
 * Represents a received m.room_key_request event
 *
 * @property {string} userId    user requesting the key
 * @property {string} deviceId  device requesting the key
 * @property {string} requestId unique id for the request
 * @property {module:crypto~RoomKeyRequestBody} requestBody
 * @property {function()} share  callback which, when called, will ask
 *    the relevant crypto algorithm implementation to share the keys for
 *    this request.
 */


var IncomingRoomKeyRequest = function IncomingRoomKeyRequest(event) {
  (0, _classCallCheck2["default"])(this, IncomingRoomKeyRequest);
  var content = event.getContent();
  this.userId = event.getSender();
  this.deviceId = content.requesting_device_id;
  this.requestId = content.request_id;
  this.requestBody = content.body || {};

  this.share = function () {
    throw new Error("don't know how to share keys for this request yet");
  };
};
/**
 * Represents a received m.room_key_request cancellation
 *
 * @property {string} userId    user requesting the cancellation
 * @property {string} deviceId  device requesting the cancellation
 * @property {string} requestId unique id for the request to be cancelled
 */


var IncomingRoomKeyRequestCancellation = function IncomingRoomKeyRequestCancellation(event) {
  (0, _classCallCheck2["default"])(this, IncomingRoomKeyRequestCancellation);
  var content = event.getContent();
  this.userId = event.getSender();
  this.deviceId = content.requesting_device_id;
  this.requestId = content.request_id;
};
/**
 * The result of a (successful) call to decryptEvent.
 *
 * @typedef {Object} EventDecryptionResult
 *
 * @property {Object} clearEvent The plaintext payload for the event
 *     (typically containing <tt>type</tt> and <tt>content</tt> fields).
 *
 * @property {?string} senderCurve25519Key Key owned by the sender of this
 *    event.  See {@link module:models/event.MatrixEvent#getSenderKey}.
 *
 * @property {?string} claimedEd25519Key ed25519 key claimed by the sender of
 *    this event. See
 *    {@link module:models/event.MatrixEvent#getClaimedEd25519Key}.
 *
 * @property {?Array<string>} forwardingCurve25519KeyChain list of curve25519
 *     keys involved in telling us about the senderCurve25519Key and
 *     claimedEd25519Key. See
 *     {@link module:models/event.MatrixEvent#getForwardingCurve25519KeyChain}.
 */

/**
 * Fires when we receive a room key request
 *
 * @event module:client~MatrixClient#"crypto.roomKeyRequest"
 * @param {module:crypto~IncomingRoomKeyRequest} req  request details
 */

/**
 * Fires when we receive a room key request cancellation
 *
 * @event module:client~MatrixClient#"crypto.roomKeyRequestCancellation"
 * @param {module:crypto~IncomingRoomKeyRequestCancellation} req
 */

/**
 * Fires when the app may wish to warn the user about something related
 * the end-to-end crypto.
 *
 * @event module:client~MatrixClient#"crypto.warning"
 * @param {string} type One of the strings listed above
 */