"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.keyFromAuthData = keyFromAuthData;
exports.keyFromPassphrase = keyFromPassphrase;
exports.deriveKey = deriveKey;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _randomstring = require("../randomstring");

/*
Copyright 2018 New Vector Ltd
Copyright 2019 The Matrix.org Foundation C.I.C.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/
var DEFAULT_ITERATIONS = 500000;
var DEFAULT_BITSIZE = 256;

function keyFromAuthData(_x, _x2) {
  return _keyFromAuthData.apply(this, arguments);
}

function _keyFromAuthData() {
  _keyFromAuthData = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(authData, password) {
    return _regenerator["default"].wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            if (global.Olm) {
              _context.next = 2;
              break;
            }

            throw new Error("Olm is not available");

          case 2:
            if (!(!authData.private_key_salt || !authData.private_key_iterations)) {
              _context.next = 4;
              break;
            }

            throw new Error("Salt and/or iterations not found: " + "this backup cannot be restored with a passphrase");

          case 4:
            _context.next = 6;
            return deriveKey(password, authData.private_key_salt, authData.private_key_iterations, authData.private_key_bits || DEFAULT_BITSIZE);

          case 6:
            return _context.abrupt("return", _context.sent);

          case 7:
          case "end":
            return _context.stop();
        }
      }
    }, _callee);
  }));
  return _keyFromAuthData.apply(this, arguments);
}

function keyFromPassphrase(_x3) {
  return _keyFromPassphrase.apply(this, arguments);
}

function _keyFromPassphrase() {
  _keyFromPassphrase = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(password) {
    var salt, key;
    return _regenerator["default"].wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            if (global.Olm) {
              _context2.next = 2;
              break;
            }

            throw new Error("Olm is not available");

          case 2:
            salt = (0, _randomstring.randomString)(32);
            _context2.next = 5;
            return deriveKey(password, salt, DEFAULT_ITERATIONS, DEFAULT_BITSIZE);

          case 5:
            key = _context2.sent;
            return _context2.abrupt("return", {
              key: key,
              salt: salt,
              iterations: DEFAULT_ITERATIONS
            });

          case 7:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2);
  }));
  return _keyFromPassphrase.apply(this, arguments);
}

function deriveKey(_x4, _x5, _x6) {
  return _deriveKey.apply(this, arguments);
}

function _deriveKey() {
  _deriveKey = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(password, salt, iterations) {
    var numBits,
        subtleCrypto,
        TextEncoder,
        key,
        keybits,
        _args3 = arguments;
    return _regenerator["default"].wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            numBits = _args3.length > 3 && _args3[3] !== undefined ? _args3[3] : DEFAULT_BITSIZE;
            subtleCrypto = global.crypto.subtle;
            TextEncoder = global.TextEncoder;

            if (!(!subtleCrypto || !TextEncoder)) {
              _context3.next = 5;
              break;
            }

            throw new Error("Password-based backup is not avaiable on this platform");

          case 5:
            _context3.next = 7;
            return subtleCrypto.importKey('raw', new TextEncoder().encode(password), {
              name: 'PBKDF2'
            }, false, ['deriveBits']);

          case 7:
            key = _context3.sent;
            _context3.next = 10;
            return subtleCrypto.deriveBits({
              name: 'PBKDF2',
              salt: new TextEncoder().encode(salt),
              iterations: iterations,
              hash: 'SHA-512'
            }, key, numBits);

          case 10:
            keybits = _context3.sent;
            return _context3.abrupt("return", new Uint8Array(keybits));

          case 12:
          case "end":
            return _context3.stop();
        }
      }
    }, _callee3);
  }));
  return _deriveKey.apply(this, arguments);
}