"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.encryptMessageForDevice = encryptMessageForDevice;
exports.getExistingOlmSessions = getExistingOlmSessions;
exports.ensureOlmSessionsForDevices = ensureOlmSessionsForDevices;
exports.verifySignature = verifySignature;
exports.pkSign = pkSign;
exports.pkVerify = pkVerify;
exports.encodeBase64 = encodeBase64;
exports.encodeUnpaddedBase64 = encodeUnpaddedBase64;
exports.decodeBase64 = decodeBase64;
exports.MEGOLM_BACKUP_ALGORITHM = exports.MEGOLM_ALGORITHM = exports.OLM_ALGORITHM = void 0;

var _toConsumableArray2 = _interopRequireDefault(require("@babel/runtime/helpers/toConsumableArray"));

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _logger = require("../logger");

var utils = _interopRequireWildcard(require("../utils"));

var _anotherJson = _interopRequireDefault(require("another-json"));

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

/**
 * matrix algorithm tag for olm
 */
var OLM_ALGORITHM = "m.olm.v1.curve25519-aes-sha2";
/**
 * matrix algorithm tag for megolm
 */

exports.OLM_ALGORITHM = OLM_ALGORITHM;
var MEGOLM_ALGORITHM = "m.megolm.v1.aes-sha2";
/**
 * matrix algorithm tag for megolm backups
 */

exports.MEGOLM_ALGORITHM = MEGOLM_ALGORITHM;
var MEGOLM_BACKUP_ALGORITHM = "m.megolm_backup.v1.curve25519-aes-sha2";
/**
 * Encrypt an event payload for an Olm device
 *
 * @param {Object<string, string>} resultsObject  The `ciphertext` property
 *   of the m.room.encrypted event to which to add our result
 *
 * @param {string} ourUserId
 * @param {string} ourDeviceId
 * @param {module:crypto/OlmDevice} olmDevice olm.js wrapper
 * @param {string} recipientUserId
 * @param {module:crypto/deviceinfo} recipientDevice
 * @param {object} payloadFields fields to include in the encrypted payload
 *
 * Returns a promise which resolves (to undefined) when the payload
 *    has been encrypted into `resultsObject`
 */

exports.MEGOLM_BACKUP_ALGORITHM = MEGOLM_BACKUP_ALGORITHM;

function encryptMessageForDevice(_x, _x2, _x3, _x4, _x5, _x6, _x7) {
  return _encryptMessageForDevice.apply(this, arguments);
}
/**
 * Get the existing olm sessions for the given devices, and the devices that
 * don't have olm sessions.
 *
 * @param {module:crypto/OlmDevice} olmDevice
 *
 * @param {module:base-apis~MatrixBaseApis} baseApis
 *
 * @param {object<string, module:crypto/deviceinfo[]>} devicesByUser
 *    map from userid to list of devices to ensure sessions for
 *
 * @return {Promise} resolves to an array.  The first element of the array is a
 *    a map of user IDs to arrays of deviceInfo, representing the devices that
 *    don't have established olm sessions.  The second element of the array is
 *    a map from userId to deviceId to {@link module:crypto~OlmSessionResult}
 */


function _encryptMessageForDevice() {
  _encryptMessageForDevice = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(resultsObject, ourUserId, ourDeviceId, olmDevice, recipientUserId, recipientDevice, payloadFields) {
    var deviceKey, sessionId, payload;
    return _regenerator["default"].wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            deviceKey = recipientDevice.getIdentityKey();
            _context.next = 3;
            return olmDevice.getSessionIdForDevice(deviceKey);

          case 3:
            sessionId = _context.sent;

            if (!(sessionId === null)) {
              _context.next = 6;
              break;
            }

            return _context.abrupt("return");

          case 6:
            _logger.logger.log("Using sessionid " + sessionId + " for device " + recipientUserId + ":" + recipientDevice.deviceId);

            payload = {
              sender: ourUserId,
              sender_device: ourDeviceId,
              // Include the Ed25519 key so that the recipient knows what
              // device this message came from.
              // We don't need to include the curve25519 key since the
              // recipient will already know this from the olm headers.
              // When combined with the device keys retrieved from the
              // homeserver signed by the ed25519 key this proves that
              // the curve25519 key and the ed25519 key are owned by
              // the same device.
              keys: {
                "ed25519": olmDevice.deviceEd25519Key
              },
              // include the recipient device details in the payload,
              // to avoid unknown key attacks, per
              // https://github.com/vector-im/vector-web/issues/2483
              recipient: recipientUserId,
              recipient_keys: {
                "ed25519": recipientDevice.getFingerprint()
              }
            }; // TODO: technically, a bunch of that stuff only needs to be included for
            // pre-key messages: after that, both sides know exactly which devices are
            // involved in the session. If we're looking to reduce data transfer in the
            // future, we could elide them for subsequent messages.

            utils.extend(payload, payloadFields);
            _context.next = 11;
            return olmDevice.encryptMessage(deviceKey, sessionId, JSON.stringify(payload));

          case 11:
            resultsObject[deviceKey] = _context.sent;

          case 12:
          case "end":
            return _context.stop();
        }
      }
    }, _callee);
  }));
  return _encryptMessageForDevice.apply(this, arguments);
}

function getExistingOlmSessions(_x8, _x9, _x10) {
  return _getExistingOlmSessions.apply(this, arguments);
}
/**
 * Try to make sure we have established olm sessions for the given devices.
 *
 * @param {module:crypto/OlmDevice} olmDevice
 *
 * @param {module:base-apis~MatrixBaseApis} baseApis
 *
 * @param {object<string, module:crypto/deviceinfo[]>} devicesByUser
 *    map from userid to list of devices to ensure sessions for
 *
 * @param {boolean} [force=false] If true, establish a new session even if one
 *     already exists.
 *
 * @param {Number} [otkTimeout] The timeout in milliseconds when requesting
 *     one-time keys for establishing new olm sessions.
 *
 * @param {Array} [failedServers] An array to fill with remote servers that
 *     failed to respond to one-time-key requests.
 *
 * @return {Promise} resolves once the sessions are complete, to
 *    an Object mapping from userId to deviceId to
 *    {@link module:crypto~OlmSessionResult}
 */


function _getExistingOlmSessions() {
  _getExistingOlmSessions = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(olmDevice, baseApis, devicesByUser) {
    var devicesWithoutSession, sessions, promises, _loop, _i, _Object$entries;

    return _regenerator["default"].wrap(function _callee3$(_context3) {
      while (1) {
        switch (_context3.prev = _context3.next) {
          case 0:
            devicesWithoutSession = {};
            sessions = {};
            promises = [];

            _loop = function _loop() {
              var _Object$entries$_i = (0, _slicedToArray2["default"])(_Object$entries[_i], 2),
                  userId = _Object$entries$_i[0],
                  devices = _Object$entries$_i[1];

              var _iterator = _createForOfIteratorHelper(devices),
                  _step;

              try {
                var _loop2 = function _loop2() {
                  var deviceInfo = _step.value;
                  var deviceId = deviceInfo.deviceId;
                  var key = deviceInfo.getIdentityKey();
                  promises.push((0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
                    var sessionId;
                    return _regenerator["default"].wrap(function _callee2$(_context2) {
                      while (1) {
                        switch (_context2.prev = _context2.next) {
                          case 0:
                            _context2.next = 2;
                            return olmDevice.getSessionIdForDevice(key, true);

                          case 2:
                            sessionId = _context2.sent;

                            if (sessionId === null) {
                              devicesWithoutSession[userId] = devicesWithoutSession[userId] || [];
                              devicesWithoutSession[userId].push(deviceInfo);
                            } else {
                              sessions[userId] = sessions[userId] || {};
                              sessions[userId][deviceId] = {
                                device: deviceInfo,
                                sessionId: sessionId
                              };
                            }

                          case 4:
                          case "end":
                            return _context2.stop();
                        }
                      }
                    }, _callee2);
                  }))());
                };

                for (_iterator.s(); !(_step = _iterator.n()).done;) {
                  _loop2();
                }
              } catch (err) {
                _iterator.e(err);
              } finally {
                _iterator.f();
              }
            };

            for (_i = 0, _Object$entries = Object.entries(devicesByUser); _i < _Object$entries.length; _i++) {
              _loop();
            }

            _context3.next = 7;
            return Promise.all(promises);

          case 7:
            return _context3.abrupt("return", [devicesWithoutSession, sessions]);

          case 8:
          case "end":
            return _context3.stop();
        }
      }
    }, _callee3);
  }));
  return _getExistingOlmSessions.apply(this, arguments);
}

function ensureOlmSessionsForDevices(_x11, _x12, _x13, _x14, _x15, _x16) {
  return _ensureOlmSessionsForDevices.apply(this, arguments);
}

function _ensureOlmSessionsForDevices() {
  _ensureOlmSessionsForDevices = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4(olmDevice, baseApis, devicesByUser, force, otkTimeout, failedServers) {
    var devicesWithoutSession, result, resolveSession, _i2, _Object$entries2, _Object$entries2$_i, userId, devices, _iterator2, _step2, _loop3, _ret, oneTimeKeyAlgorithm, res, _i3, _Object$values, resolver, _failedServers, otk_res, promises, _loop4, _i4, _Object$entries3;

    return _regenerator["default"].wrap(function _callee4$(_context5) {
      while (1) {
        switch (_context5.prev = _context5.next) {
          case 0:
            if (typeof force === "number") {
              failedServers = otkTimeout;
              otkTimeout = force;
              force = false;
            }

            devicesWithoutSession = [// [userId, deviceId], ...
            ];
            result = {};
            resolveSession = {};
            _i2 = 0, _Object$entries2 = Object.entries(devicesByUser);

          case 5:
            if (!(_i2 < _Object$entries2.length)) {
              _context5.next = 30;
              break;
            }

            _Object$entries2$_i = (0, _slicedToArray2["default"])(_Object$entries2[_i2], 2), userId = _Object$entries2$_i[0], devices = _Object$entries2$_i[1];
            result[userId] = {};
            _iterator2 = _createForOfIteratorHelper(devices);
            _context5.prev = 9;
            _loop3 = /*#__PURE__*/_regenerator["default"].mark(function _loop3() {
              var deviceInfo, deviceId, key, sessionId;
              return _regenerator["default"].wrap(function _loop3$(_context4) {
                while (1) {
                  switch (_context4.prev = _context4.next) {
                    case 0:
                      deviceInfo = _step2.value;
                      deviceId = deviceInfo.deviceId;
                      key = deviceInfo.getIdentityKey();

                      if (!(key === olmDevice.deviceCurve25519Key)) {
                        _context4.next = 7;
                        break;
                      }

                      // We should never be trying to start a session with ourself.
                      // Apart from talking to yourself being the first sign of madness,
                      // olm sessions can't do this because they get confused when
                      // they get a message and see that the 'other side' has started a
                      // new chain when this side has an active sender chain.
                      // If you see this message being logged in the wild, we should find
                      // the thing that is trying to send Olm messages to itself and fix it.
                      _logger.logger.info("Attempted to start session with ourself! Ignoring"); // We must fill in the section in the return value though, as callers
                      // expect it to be there.


                      result[userId][deviceId] = {
                        device: deviceInfo,
                        sessionId: null
                      };
                      return _context4.abrupt("return", "continue");

                    case 7:
                      if (!olmDevice._sessionsInProgress[key]) {
                        // pre-emptively mark the session as in-progress to avoid race
                        // conditions.  If we find that we already have a session, then
                        // we'll resolve
                        olmDevice._sessionsInProgress[key] = new Promise(function (_resolve, _reject) {
                          resolveSession[key] = {
                            resolve: function resolve() {
                              delete olmDevice._sessionsInProgress[key];

                              _resolve.apply(void 0, arguments);
                            },
                            reject: function reject() {
                              delete olmDevice._sessionsInProgress[key];

                              _reject.apply(void 0, arguments);
                            }
                          };
                        });
                      }

                      _context4.next = 10;
                      return olmDevice.getSessionIdForDevice(key, resolveSession[key]);

                    case 10:
                      sessionId = _context4.sent;

                      if (sessionId !== null && resolveSession[key]) {
                        // we found a session, but we had marked the session as
                        // in-progress, so unmark it and unblock anything that was
                        // waiting
                        delete olmDevice._sessionsInProgress[key];
                        resolveSession[key].resolve();
                        delete resolveSession[key];
                      }

                      if (sessionId === null || force) {
                        if (force) {
                          _logger.logger.info("Forcing new Olm session for " + userId + ":" + deviceId);
                        } else {
                          _logger.logger.info("Making new Olm session for " + userId + ":" + deviceId);
                        }

                        devicesWithoutSession.push([userId, deviceId]);
                      }

                      result[userId][deviceId] = {
                        device: deviceInfo,
                        sessionId: sessionId
                      };

                    case 14:
                    case "end":
                      return _context4.stop();
                  }
                }
              }, _loop3);
            });

            _iterator2.s();

          case 12:
            if ((_step2 = _iterator2.n()).done) {
              _context5.next = 19;
              break;
            }

            return _context5.delegateYield(_loop3(), "t0", 14);

          case 14:
            _ret = _context5.t0;

            if (!(_ret === "continue")) {
              _context5.next = 17;
              break;
            }

            return _context5.abrupt("continue", 17);

          case 17:
            _context5.next = 12;
            break;

          case 19:
            _context5.next = 24;
            break;

          case 21:
            _context5.prev = 21;
            _context5.t1 = _context5["catch"](9);

            _iterator2.e(_context5.t1);

          case 24:
            _context5.prev = 24;

            _iterator2.f();

            return _context5.finish(24);

          case 27:
            _i2++;
            _context5.next = 5;
            break;

          case 30:
            if (!(devicesWithoutSession.length === 0)) {
              _context5.next = 32;
              break;
            }

            return _context5.abrupt("return", result);

          case 32:
            oneTimeKeyAlgorithm = "signed_curve25519";
            _context5.prev = 33;
            _context5.next = 36;
            return baseApis.claimOneTimeKeys(devicesWithoutSession, oneTimeKeyAlgorithm, otkTimeout);

          case 36:
            res = _context5.sent;
            _context5.next = 44;
            break;

          case 39:
            _context5.prev = 39;
            _context5.t2 = _context5["catch"](33);

            for (_i3 = 0, _Object$values = Object.values(resolveSession); _i3 < _Object$values.length; _i3++) {
              resolver = _Object$values[_i3];
              resolver.resolve();
            }

            _logger.logger.log("failed to claim one-time keys", _context5.t2, devicesWithoutSession);

            throw _context5.t2;

          case 44:
            if (failedServers && "failures" in res) {
              (_failedServers = failedServers).push.apply(_failedServers, (0, _toConsumableArray2["default"])(Object.keys(res.failures)));
            }

            otk_res = res.one_time_keys || {};
            promises = [];

            _loop4 = function _loop4() {
              var _Object$entries3$_i = (0, _slicedToArray2["default"])(_Object$entries3[_i4], 2),
                  userId = _Object$entries3$_i[0],
                  devices = _Object$entries3$_i[1];

              var userRes = otk_res[userId] || {};

              var _loop5 = function _loop5(j) {
                var deviceInfo = devices[j];
                var deviceId = deviceInfo.deviceId;
                var key = deviceInfo.getIdentityKey();

                if (key === olmDevice.deviceCurve25519Key) {
                  // We've already logged about this above. Skip here too
                  // otherwise we'll log saying there are no one-time keys
                  // which will be confusing.
                  return "continue";
                }

                if (result[userId][deviceId].sessionId && !force) {
                  // we already have a result for this device
                  return "continue";
                }

                var deviceRes = userRes[deviceId] || {};
                var oneTimeKey = null;

                for (var keyId in deviceRes) {
                  if (keyId.indexOf(oneTimeKeyAlgorithm + ":") === 0) {
                    oneTimeKey = deviceRes[keyId];
                  }
                }

                if (!oneTimeKey) {
                  var msg = "No one-time keys (alg=" + oneTimeKeyAlgorithm + ") for device " + userId + ":" + deviceId;

                  _logger.logger.warn(msg);

                  if (resolveSession[key]) {
                    resolveSession[key].resolve();
                  }

                  return "continue";
                }

                promises.push(_verifyKeyAndStartSession(olmDevice, oneTimeKey, userId, deviceInfo).then(function (sid) {
                  if (resolveSession[key]) {
                    resolveSession[key].resolve(sid);
                  }

                  result[userId][deviceId].sessionId = sid;
                }, function (e) {
                  if (resolveSession[key]) {
                    resolveSession[key].resolve();
                  }

                  throw e;
                }));
              };

              for (var j = 0; j < devices.length; j++) {
                var _ret2 = _loop5(j);

                if (_ret2 === "continue") continue;
              }
            };

            for (_i4 = 0, _Object$entries3 = Object.entries(devicesByUser); _i4 < _Object$entries3.length; _i4++) {
              _loop4();
            }

            _context5.next = 51;
            return Promise.all(promises);

          case 51:
            return _context5.abrupt("return", result);

          case 52:
          case "end":
            return _context5.stop();
        }
      }
    }, _callee4, null, [[9, 21, 24, 27], [33, 39]]);
  }));
  return _ensureOlmSessionsForDevices.apply(this, arguments);
}

function _verifyKeyAndStartSession(_x17, _x18, _x19, _x20) {
  return _verifyKeyAndStartSession2.apply(this, arguments);
}
/**
 * Verify the signature on an object
 *
 * @param {module:crypto/OlmDevice} olmDevice olm wrapper to use for verify op
 *
 * @param {Object} obj object to check signature on.
 *
 * @param {string} signingUserId  ID of the user whose signature should be checked
 *
 * @param {string} signingDeviceId  ID of the device whose signature should be checked
 *
 * @param {string} signingKey   base64-ed ed25519 public key
 *
 * Returns a promise which resolves (to undefined) if the the signature is good,
 * or rejects with an Error if it is bad.
 */


function _verifyKeyAndStartSession2() {
  _verifyKeyAndStartSession2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5(olmDevice, oneTimeKey, userId, deviceInfo) {
    var deviceId, sid;
    return _regenerator["default"].wrap(function _callee5$(_context6) {
      while (1) {
        switch (_context6.prev = _context6.next) {
          case 0:
            deviceId = deviceInfo.deviceId;
            _context6.prev = 1;
            _context6.next = 4;
            return verifySignature(olmDevice, oneTimeKey, userId, deviceId, deviceInfo.getFingerprint());

          case 4:
            _context6.next = 10;
            break;

          case 6:
            _context6.prev = 6;
            _context6.t0 = _context6["catch"](1);

            _logger.logger.error("Unable to verify signature on one-time key for device " + userId + ":" + deviceId + ":", _context6.t0);

            return _context6.abrupt("return", null);

          case 10:
            _context6.prev = 10;
            _context6.next = 13;
            return olmDevice.createOutboundSession(deviceInfo.getIdentityKey(), oneTimeKey.key);

          case 13:
            sid = _context6.sent;
            _context6.next = 20;
            break;

          case 16:
            _context6.prev = 16;
            _context6.t1 = _context6["catch"](10);

            // possibly a bad key
            _logger.logger.error("Error starting olm session with device " + userId + ":" + deviceId + ": " + _context6.t1);

            return _context6.abrupt("return", null);

          case 20:
            _logger.logger.log("Started new olm sessionid " + sid + " for device " + userId + ":" + deviceId);

            return _context6.abrupt("return", sid);

          case 22:
          case "end":
            return _context6.stop();
        }
      }
    }, _callee5, null, [[1, 6], [10, 16]]);
  }));
  return _verifyKeyAndStartSession2.apply(this, arguments);
}

function verifySignature(_x21, _x22, _x23, _x24, _x25) {
  return _verifySignature.apply(this, arguments);
}
/**
 * Sign a JSON object using public key cryptography
 * @param {Object} obj Object to sign.  The object will be modified to include
 *     the new signature
 * @param {Olm.PkSigning|Uint8Array} key the signing object or the private key
 * seed
 * @param {string} userId The user ID who owns the signing key
 * @param {string} pubkey The public key (ignored if key is a seed)
 * @returns {string} the signature for the object
 */


function _verifySignature() {
  _verifySignature = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6(olmDevice, obj, signingUserId, signingDeviceId, signingKey) {
    var signKeyId, signatures, userSigs, signature, mangledObj, json;
    return _regenerator["default"].wrap(function _callee6$(_context7) {
      while (1) {
        switch (_context7.prev = _context7.next) {
          case 0:
            signKeyId = "ed25519:" + signingDeviceId;
            signatures = obj.signatures || {};
            userSigs = signatures[signingUserId] || {};
            signature = userSigs[signKeyId];

            if (signature) {
              _context7.next = 6;
              break;
            }

            throw Error("No signature");

          case 6:
            // prepare the canonical json: remove unsigned and signatures, and stringify with
            // anotherjson
            mangledObj = Object.assign({}, obj);
            delete mangledObj.unsigned;
            delete mangledObj.signatures;
            json = _anotherJson["default"].stringify(mangledObj);
            olmDevice.verifySignature(signingKey, json, signature);

          case 11:
          case "end":
            return _context7.stop();
        }
      }
    }, _callee6);
  }));
  return _verifySignature.apply(this, arguments);
}

function pkSign(obj, key, userId, pubkey) {
  var createdKey = false;

  if (key instanceof Uint8Array) {
    var keyObj = new global.Olm.PkSigning();
    pubkey = keyObj.init_with_seed(key);
    key = keyObj;
    createdKey = true;
  }

  var sigs = obj.signatures || {};
  delete obj.signatures;
  var unsigned = obj.unsigned;
  if (obj.unsigned) delete obj.unsigned;

  try {
    var mysigs = sigs[userId] || {};
    sigs[userId] = mysigs;
    return mysigs['ed25519:' + pubkey] = key.sign(_anotherJson["default"].stringify(obj));
  } finally {
    obj.signatures = sigs;
    if (unsigned) obj.unsigned = unsigned;

    if (createdKey) {
      key.free();
    }
  }
}
/**
 * Verify a signed JSON object
 * @param {Object} obj Object to verify
 * @param {string} pubkey The public key to use to verify
 * @param {string} userId The user ID who signed the object
 */


function pkVerify(obj, pubkey, userId) {
  var keyId = "ed25519:" + pubkey;

  if (!(obj.signatures && obj.signatures[userId] && obj.signatures[userId][keyId])) {
    throw new Error("No signature");
  }

  var signature = obj.signatures[userId][keyId];
  var util = new global.Olm.Utility();
  var sigs = obj.signatures;
  delete obj.signatures;
  var unsigned = obj.unsigned;
  if (obj.unsigned) delete obj.unsigned;

  try {
    util.ed25519_verify(pubkey, _anotherJson["default"].stringify(obj), signature);
  } finally {
    obj.signatures = sigs;
    if (unsigned) obj.unsigned = unsigned;
    util.free();
  }
}
/**
 * Encode a typed array of uint8 as base64.
 * @param {Uint8Array} uint8Array The data to encode.
 * @return {string} The base64.
 */


function encodeBase64(uint8Array) {
  return Buffer.from(uint8Array).toString("base64");
}
/**
 * Encode a typed array of uint8 as unpadded base64.
 * @param {Uint8Array} uint8Array The data to encode.
 * @return {string} The unpadded base64.
 */


function encodeUnpaddedBase64(uint8Array) {
  return encodeBase64(uint8Array).replace(/=+$/g, '');
}
/**
 * Decode a base64 string to a typed array of uint8.
 * @param {string} base64 The base64 to decode.
 * @return {Uint8Array} The decoded data.
 */


function decodeBase64(base64) {
  return Buffer.from(base64, "base64");
}