"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.upgradeDatabase = upgradeDatabase;
exports.Backend = exports.VERSION = void 0;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _logger = require("../../logger");

var utils = _interopRequireWildcard(require("../../utils"));

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var VERSION = 9;
/**
 * Implementation of a CryptoStore which is backed by an existing
 * IndexedDB connection. Generally you want IndexedDBCryptoStore
 * which connects to the database and defers to one of these.
 *
 * @implements {module:crypto/store/base~CryptoStore}
 */

exports.VERSION = VERSION;

var Backend = /*#__PURE__*/function () {
  /**
   * @param {IDBDatabase} db
   */
  function Backend(db) {
    var _this = this;

    (0, _classCallCheck2["default"])(this, Backend);
    this._db = db; // make sure we close the db on `onversionchange` - otherwise
    // attempts to delete the database will block (and subsequent
    // attempts to re-create it will also block).

    db.onversionchange = function (ev) {
      _logger.logger.log("versionchange for indexeddb ".concat(_this._dbName, ": closing"));

      db.close();
    };
  }
  /**
   * Look for an existing outgoing room key request, and if none is found,
   * add a new one
   *
   * @param {module:crypto/store/base~OutgoingRoomKeyRequest} request
   *
   * @returns {Promise} resolves to
   *    {@link module:crypto/store/base~OutgoingRoomKeyRequest}: either the
   *    same instance as passed in, or the existing one.
   */


  (0, _createClass2["default"])(Backend, [{
    key: "getOrAddOutgoingRoomKeyRequest",
    value: function getOrAddOutgoingRoomKeyRequest(request) {
      var _this2 = this;

      var requestBody = request.requestBody;
      return new Promise(function (resolve, reject) {
        var txn = _this2._db.transaction("outgoingRoomKeyRequests", "readwrite");

        txn.onerror = reject; // first see if we already have an entry for this request.

        _this2._getOutgoingRoomKeyRequest(txn, requestBody, function (existing) {
          if (existing) {
            // this entry matches the request - return it.
            _logger.logger.log("already have key request outstanding for " + "".concat(requestBody.room_id, " / ").concat(requestBody.session_id, ": ") + "not sending another");

            resolve(existing);
            return;
          } // we got to the end of the list without finding a match
          // - add the new request.


          _logger.logger.log("enqueueing key request for ".concat(requestBody.room_id, " / ") + requestBody.session_id);

          txn.oncomplete = function () {
            resolve(request);
          };

          var store = txn.objectStore("outgoingRoomKeyRequests");
          store.add(request);
        });
      });
    }
    /**
     * Look for an existing room key request
     *
     * @param {module:crypto~RoomKeyRequestBody} requestBody
     *    existing request to look for
     *
     * @return {Promise} resolves to the matching
     *    {@link module:crypto/store/base~OutgoingRoomKeyRequest}, or null if
     *    not found
     */

  }, {
    key: "getOutgoingRoomKeyRequest",
    value: function getOutgoingRoomKeyRequest(requestBody) {
      var _this3 = this;

      return new Promise(function (resolve, reject) {
        var txn = _this3._db.transaction("outgoingRoomKeyRequests", "readonly");

        txn.onerror = reject;

        _this3._getOutgoingRoomKeyRequest(txn, requestBody, function (existing) {
          resolve(existing);
        });
      });
    }
    /**
     * look for an existing room key request in the db
     *
     * @private
     * @param {IDBTransaction} txn  database transaction
     * @param {module:crypto~RoomKeyRequestBody} requestBody
     *    existing request to look for
     * @param {Function} callback  function to call with the results of the
     *    search. Either passed a matching
     *    {@link module:crypto/store/base~OutgoingRoomKeyRequest}, or null if
     *    not found.
     */

  }, {
    key: "_getOutgoingRoomKeyRequest",
    value: function _getOutgoingRoomKeyRequest(txn, requestBody, callback) {
      var store = txn.objectStore("outgoingRoomKeyRequests");
      var idx = store.index("session");
      var cursorReq = idx.openCursor([requestBody.room_id, requestBody.session_id]);

      cursorReq.onsuccess = function (ev) {
        var cursor = ev.target.result;

        if (!cursor) {
          // no match found
          callback(null);
          return;
        }

        var existing = cursor.value;

        if (utils.deepCompare(existing.requestBody, requestBody)) {
          // got a match
          callback(existing);
          return;
        } // look at the next entry in the index


        cursor["continue"]();
      };
    }
    /**
     * Look for room key requests by state
     *
     * @param {Array<Number>} wantedStates list of acceptable states
     *
     * @return {Promise} resolves to the a
     *    {@link module:crypto/store/base~OutgoingRoomKeyRequest}, or null if
     *    there are no pending requests in those states. If there are multiple
     *    requests in those states, an arbitrary one is chosen.
     */

  }, {
    key: "getOutgoingRoomKeyRequestByState",
    value: function getOutgoingRoomKeyRequestByState(wantedStates) {
      if (wantedStates.length === 0) {
        return Promise.resolve(null);
      } // this is a bit tortuous because we need to make sure we do the lookup
      // in a single transaction, to avoid having a race with the insertion
      // code.
      // index into the wantedStates array


      var stateIndex = 0;
      var result;

      function onsuccess(ev) {
        var cursor = ev.target.result;

        if (cursor) {
          // got a match
          result = cursor.value;
          return;
        } // try the next state in the list


        stateIndex++;

        if (stateIndex >= wantedStates.length) {
          // no matches
          return;
        }

        var wantedState = wantedStates[stateIndex];
        var cursorReq = ev.target.source.openCursor(wantedState);
        cursorReq.onsuccess = onsuccess;
      }

      var txn = this._db.transaction("outgoingRoomKeyRequests", "readonly");

      var store = txn.objectStore("outgoingRoomKeyRequests");
      var wantedState = wantedStates[stateIndex];
      var cursorReq = store.index("state").openCursor(wantedState);
      cursorReq.onsuccess = onsuccess;
      return promiseifyTxn(txn).then(function () {
        return result;
      });
    }
    /**
     *
     * @param {Number} wantedState
     * @return {Promise<Array<*>>} All elements in a given state
     */

  }, {
    key: "getAllOutgoingRoomKeyRequestsByState",
    value: function getAllOutgoingRoomKeyRequestsByState(wantedState) {
      var _this4 = this;

      return new Promise(function (resolve, reject) {
        var txn = _this4._db.transaction("outgoingRoomKeyRequests", "readonly");

        var store = txn.objectStore("outgoingRoomKeyRequests");
        var index = store.index("state");
        var request = index.getAll(wantedState);

        request.onsuccess = function (ev) {
          return resolve(ev.target.result);
        };

        request.onerror = function (ev) {
          return reject(ev.target.error);
        };
      });
    }
  }, {
    key: "getOutgoingRoomKeyRequestsByTarget",
    value: function getOutgoingRoomKeyRequestsByTarget(userId, deviceId, wantedStates) {
      var stateIndex = 0;
      var results = [];

      function onsuccess(ev) {
        var cursor = ev.target.result;

        if (cursor) {
          var keyReq = cursor.value;

          if (keyReq.recipients.includes({
            userId: userId,
            deviceId: deviceId
          })) {
            results.push(keyReq);
          }

          cursor["continue"]();
        } else {
          // try the next state in the list
          stateIndex++;

          if (stateIndex >= wantedStates.length) {
            // no matches
            return;
          }

          var _wantedState = wantedStates[stateIndex];

          var _cursorReq = ev.target.source.openCursor(_wantedState);

          _cursorReq.onsuccess = onsuccess;
        }
      }

      var txn = this._db.transaction("outgoingRoomKeyRequests", "readonly");

      var store = txn.objectStore("outgoingRoomKeyRequests");
      var wantedState = wantedStates[stateIndex];
      var cursorReq = store.index("state").openCursor(wantedState);
      cursorReq.onsuccess = onsuccess;
      return promiseifyTxn(txn).then(function () {
        return results;
      });
    }
    /**
     * Look for an existing room key request by id and state, and update it if
     * found
     *
     * @param {string} requestId      ID of request to update
     * @param {number} expectedState  state we expect to find the request in
     * @param {Object} updates        name/value map of updates to apply
     *
     * @returns {Promise} resolves to
     *    {@link module:crypto/store/base~OutgoingRoomKeyRequest}
     *    updated request, or null if no matching row was found
     */

  }, {
    key: "updateOutgoingRoomKeyRequest",
    value: function updateOutgoingRoomKeyRequest(requestId, expectedState, updates) {
      var result = null;

      function onsuccess(ev) {
        var cursor = ev.target.result;

        if (!cursor) {
          return;
        }

        var data = cursor.value;

        if (data.state != expectedState) {
          _logger.logger.warn("Cannot update room key request from ".concat(expectedState, " ") + "as it was already updated to ".concat(data.state));

          return;
        }

        Object.assign(data, updates);
        cursor.update(data);
        result = data;
      }

      var txn = this._db.transaction("outgoingRoomKeyRequests", "readwrite");

      var cursorReq = txn.objectStore("outgoingRoomKeyRequests").openCursor(requestId);
      cursorReq.onsuccess = onsuccess;
      return promiseifyTxn(txn).then(function () {
        return result;
      });
    }
    /**
     * Look for an existing room key request by id and state, and delete it if
     * found
     *
     * @param {string} requestId      ID of request to update
     * @param {number} expectedState  state we expect to find the request in
     *
     * @returns {Promise} resolves once the operation is completed
     */

  }, {
    key: "deleteOutgoingRoomKeyRequest",
    value: function deleteOutgoingRoomKeyRequest(requestId, expectedState) {
      var txn = this._db.transaction("outgoingRoomKeyRequests", "readwrite");

      var cursorReq = txn.objectStore("outgoingRoomKeyRequests").openCursor(requestId);

      cursorReq.onsuccess = function (ev) {
        var cursor = ev.target.result;

        if (!cursor) {
          return;
        }

        var data = cursor.value;

        if (data.state != expectedState) {
          _logger.logger.warn("Cannot delete room key request in state ".concat(data.state, " ") + "(expected ".concat(expectedState, ")"));

          return;
        }

        cursor["delete"]();
      };

      return promiseifyTxn(txn);
    } // Olm Account

  }, {
    key: "getAccount",
    value: function getAccount(txn, func) {
      var objectStore = txn.objectStore("account");
      var getReq = objectStore.get("-");

      getReq.onsuccess = function () {
        try {
          func(getReq.result || null);
        } catch (e) {
          abortWithException(txn, e);
        }
      };
    }
  }, {
    key: "storeAccount",
    value: function storeAccount(txn, newData) {
      var objectStore = txn.objectStore("account");
      objectStore.put(newData, "-");
    }
  }, {
    key: "getCrossSigningKeys",
    value: function getCrossSigningKeys(txn, func) {
      var objectStore = txn.objectStore("account");
      var getReq = objectStore.get("crossSigningKeys");

      getReq.onsuccess = function () {
        try {
          func(getReq.result || null);
        } catch (e) {
          abortWithException(txn, e);
        }
      };
    }
  }, {
    key: "getSecretStorePrivateKey",
    value: function getSecretStorePrivateKey(txn, func, type) {
      var objectStore = txn.objectStore("account");
      var getReq = objectStore.get("ssss_cache:".concat(type));

      getReq.onsuccess = function () {
        try {
          func(getReq.result || null);
        } catch (e) {
          abortWithException(txn, e);
        }
      };
    }
  }, {
    key: "storeCrossSigningKeys",
    value: function storeCrossSigningKeys(txn, keys) {
      var objectStore = txn.objectStore("account");
      objectStore.put(keys, "crossSigningKeys");
    }
  }, {
    key: "storeSecretStorePrivateKey",
    value: function storeSecretStorePrivateKey(txn, type, key) {
      var objectStore = txn.objectStore("account");
      objectStore.put(key, "ssss_cache:".concat(type));
    } // Olm Sessions

  }, {
    key: "countEndToEndSessions",
    value: function countEndToEndSessions(txn, func) {
      var objectStore = txn.objectStore("sessions");
      var countReq = objectStore.count();

      countReq.onsuccess = function () {
        try {
          func(countReq.result);
        } catch (e) {
          abortWithException(txn, e);
        }
      };
    }
  }, {
    key: "getEndToEndSessions",
    value: function getEndToEndSessions(deviceKey, txn, func) {
      var objectStore = txn.objectStore("sessions");
      var idx = objectStore.index("deviceKey");
      var getReq = idx.openCursor(deviceKey);
      var results = {};

      getReq.onsuccess = function () {
        var cursor = getReq.result;

        if (cursor) {
          results[cursor.value.sessionId] = {
            session: cursor.value.session,
            lastReceivedMessageTs: cursor.value.lastReceivedMessageTs
          };
          cursor["continue"]();
        } else {
          try {
            func(results);
          } catch (e) {
            abortWithException(txn, e);
          }
        }
      };
    }
  }, {
    key: "getEndToEndSession",
    value: function getEndToEndSession(deviceKey, sessionId, txn, func) {
      var objectStore = txn.objectStore("sessions");
      var getReq = objectStore.get([deviceKey, sessionId]);

      getReq.onsuccess = function () {
        try {
          if (getReq.result) {
            func({
              session: getReq.result.session,
              lastReceivedMessageTs: getReq.result.lastReceivedMessageTs
            });
          } else {
            func(null);
          }
        } catch (e) {
          abortWithException(txn, e);
        }
      };
    }
  }, {
    key: "getAllEndToEndSessions",
    value: function getAllEndToEndSessions(txn, func) {
      var objectStore = txn.objectStore("sessions");
      var getReq = objectStore.openCursor();

      getReq.onsuccess = function () {
        try {
          var cursor = getReq.result;

          if (cursor) {
            func(cursor.value);
            cursor["continue"]();
          } else {
            func(null);
          }
        } catch (e) {
          abortWithException(txn, e);
        }
      };
    }
  }, {
    key: "storeEndToEndSession",
    value: function storeEndToEndSession(deviceKey, sessionId, sessionInfo, txn) {
      var objectStore = txn.objectStore("sessions");
      objectStore.put({
        deviceKey: deviceKey,
        sessionId: sessionId,
        session: sessionInfo.session,
        lastReceivedMessageTs: sessionInfo.lastReceivedMessageTs
      });
    }
  }, {
    key: "storeEndToEndSessionProblem",
    value: function () {
      var _storeEndToEndSessionProblem = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(deviceKey, type, fixed) {
        var txn, objectStore;
        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                txn = this._db.transaction("session_problems", "readwrite");
                objectStore = txn.objectStore("session_problems");
                objectStore.put({
                  deviceKey: deviceKey,
                  type: type,
                  fixed: fixed,
                  time: Date.now()
                });
                return _context.abrupt("return", promiseifyTxn(txn));

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function storeEndToEndSessionProblem(_x, _x2, _x3) {
        return _storeEndToEndSessionProblem.apply(this, arguments);
      }

      return storeEndToEndSessionProblem;
    }()
  }, {
    key: "getEndToEndSessionProblem",
    value: function () {
      var _getEndToEndSessionProblem = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(deviceKey, timestamp) {
        var result, txn, objectStore, index, req;
        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                txn = this._db.transaction("session_problems", "readwrite");
                objectStore = txn.objectStore("session_problems");
                index = objectStore.index("deviceKey");
                req = index.getAll(deviceKey);

                req.onsuccess = function (event) {
                  var problems = req.result;

                  if (!problems.length) {
                    result = null;
                    return;
                  }

                  problems.sort(function (a, b) {
                    return a.time - b.time;
                  });
                  var lastProblem = problems[problems.length - 1];

                  var _iterator = _createForOfIteratorHelper(problems),
                      _step;

                  try {
                    for (_iterator.s(); !(_step = _iterator.n()).done;) {
                      var problem = _step.value;

                      if (problem.time > timestamp) {
                        result = Object.assign({}, problem, {
                          fixed: lastProblem.fixed
                        });
                        return;
                      }
                    }
                  } catch (err) {
                    _iterator.e(err);
                  } finally {
                    _iterator.f();
                  }

                  if (lastProblem.fixed) {
                    result = null;
                  } else {
                    result = lastProblem;
                  }
                };

                _context2.next = 7;
                return promiseifyTxn(txn);

              case 7:
                return _context2.abrupt("return", result);

              case 8:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function getEndToEndSessionProblem(_x4, _x5) {
        return _getEndToEndSessionProblem.apply(this, arguments);
      }

      return getEndToEndSessionProblem;
    }() // FIXME: we should probably prune this when devices get deleted

  }, {
    key: "filterOutNotifiedErrorDevices",
    value: function () {
      var _filterOutNotifiedErrorDevices = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3(devices) {
        var txn, objectStore, ret;
        return _regenerator["default"].wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                txn = this._db.transaction("notified_error_devices", "readwrite");
                objectStore = txn.objectStore("notified_error_devices");
                ret = [];
                _context3.next = 5;
                return Promise.all(devices.map(function (device) {
                  return new Promise(function (resolve) {
                    var userId = device.userId,
                        deviceInfo = device.deviceInfo;
                    var getReq = objectStore.get([userId, deviceInfo.deviceId]);

                    getReq.onsuccess = function () {
                      if (!getReq.result) {
                        objectStore.put({
                          userId: userId,
                          deviceId: deviceInfo.deviceId
                        });
                        ret.push(device);
                      }

                      resolve();
                    };
                  });
                }));

              case 5:
                return _context3.abrupt("return", ret);

              case 6:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function filterOutNotifiedErrorDevices(_x6) {
        return _filterOutNotifiedErrorDevices.apply(this, arguments);
      }

      return filterOutNotifiedErrorDevices;
    }() // Inbound group sessions

  }, {
    key: "getEndToEndInboundGroupSession",
    value: function getEndToEndInboundGroupSession(senderCurve25519Key, sessionId, txn, func) {
      var session = false;
      var withheld = false;
      var objectStore = txn.objectStore("inbound_group_sessions");
      var getReq = objectStore.get([senderCurve25519Key, sessionId]);

      getReq.onsuccess = function () {
        try {
          if (getReq.result) {
            session = getReq.result.session;
          } else {
            session = null;
          }

          if (withheld !== false) {
            func(session, withheld);
          }
        } catch (e) {
          abortWithException(txn, e);
        }
      };

      var withheldObjectStore = txn.objectStore("inbound_group_sessions_withheld");
      var withheldGetReq = withheldObjectStore.get([senderCurve25519Key, sessionId]);

      withheldGetReq.onsuccess = function () {
        try {
          if (withheldGetReq.result) {
            withheld = withheldGetReq.result.session;
          } else {
            withheld = null;
          }

          if (session !== false) {
            func(session, withheld);
          }
        } catch (e) {
          abortWithException(txn, e);
        }
      };
    }
  }, {
    key: "getAllEndToEndInboundGroupSessions",
    value: function getAllEndToEndInboundGroupSessions(txn, func) {
      var objectStore = txn.objectStore("inbound_group_sessions");
      var getReq = objectStore.openCursor();

      getReq.onsuccess = function () {
        var cursor = getReq.result;

        if (cursor) {
          try {
            func({
              senderKey: cursor.value.senderCurve25519Key,
              sessionId: cursor.value.sessionId,
              sessionData: cursor.value.session
            });
          } catch (e) {
            abortWithException(txn, e);
          }

          cursor["continue"]();
        } else {
          try {
            func(null);
          } catch (e) {
            abortWithException(txn, e);
          }
        }
      };
    }
  }, {
    key: "addEndToEndInboundGroupSession",
    value: function addEndToEndInboundGroupSession(senderCurve25519Key, sessionId, sessionData, txn) {
      var objectStore = txn.objectStore("inbound_group_sessions");
      var addReq = objectStore.add({
        senderCurve25519Key: senderCurve25519Key,
        sessionId: sessionId,
        session: sessionData
      });

      addReq.onerror = function (ev) {
        if (addReq.error.name === 'ConstraintError') {
          // This stops the error from triggering the txn's onerror
          ev.stopPropagation(); // ...and this stops it from aborting the transaction

          ev.preventDefault();

          _logger.logger.log("Ignoring duplicate inbound group session: " + senderCurve25519Key + " / " + sessionId);
        } else {
          abortWithException(txn, new Error("Failed to add inbound group session: " + addReq.error));
        }
      };
    }
  }, {
    key: "storeEndToEndInboundGroupSession",
    value: function storeEndToEndInboundGroupSession(senderCurve25519Key, sessionId, sessionData, txn) {
      var objectStore = txn.objectStore("inbound_group_sessions");
      objectStore.put({
        senderCurve25519Key: senderCurve25519Key,
        sessionId: sessionId,
        session: sessionData
      });
    }
  }, {
    key: "storeEndToEndInboundGroupSessionWithheld",
    value: function storeEndToEndInboundGroupSessionWithheld(senderCurve25519Key, sessionId, sessionData, txn) {
      var objectStore = txn.objectStore("inbound_group_sessions_withheld");
      objectStore.put({
        senderCurve25519Key: senderCurve25519Key,
        sessionId: sessionId,
        session: sessionData
      });
    }
  }, {
    key: "getEndToEndDeviceData",
    value: function getEndToEndDeviceData(txn, func) {
      var objectStore = txn.objectStore("device_data");
      var getReq = objectStore.get("-");

      getReq.onsuccess = function () {
        try {
          func(getReq.result || null);
        } catch (e) {
          abortWithException(txn, e);
        }
      };
    }
  }, {
    key: "storeEndToEndDeviceData",
    value: function storeEndToEndDeviceData(deviceData, txn) {
      var objectStore = txn.objectStore("device_data");
      objectStore.put(deviceData, "-");
    }
  }, {
    key: "storeEndToEndRoom",
    value: function storeEndToEndRoom(roomId, roomInfo, txn) {
      var objectStore = txn.objectStore("rooms");
      objectStore.put(roomInfo, roomId);
    }
  }, {
    key: "getEndToEndRooms",
    value: function getEndToEndRooms(txn, func) {
      var rooms = {};
      var objectStore = txn.objectStore("rooms");
      var getReq = objectStore.openCursor();

      getReq.onsuccess = function () {
        var cursor = getReq.result;

        if (cursor) {
          rooms[cursor.key] = cursor.value;
          cursor["continue"]();
        } else {
          try {
            func(rooms);
          } catch (e) {
            abortWithException(txn, e);
          }
        }
      };
    } // session backups

  }, {
    key: "getSessionsNeedingBackup",
    value: function getSessionsNeedingBackup(limit) {
      var _this5 = this;

      return new Promise(function (resolve, reject) {
        var sessions = [];

        var txn = _this5._db.transaction(["sessions_needing_backup", "inbound_group_sessions"], "readonly");

        txn.onerror = reject;

        txn.oncomplete = function () {
          resolve(sessions);
        };

        var objectStore = txn.objectStore("sessions_needing_backup");
        var sessionStore = txn.objectStore("inbound_group_sessions");
        var getReq = objectStore.openCursor();

        getReq.onsuccess = function () {
          var cursor = getReq.result;

          if (cursor) {
            var sessionGetReq = sessionStore.get(cursor.key);

            sessionGetReq.onsuccess = function () {
              sessions.push({
                senderKey: sessionGetReq.result.senderCurve25519Key,
                sessionId: sessionGetReq.result.sessionId,
                sessionData: sessionGetReq.result.session
              });
            };

            if (!limit || sessions.length < limit) {
              cursor["continue"]();
            }
          }
        };
      });
    }
  }, {
    key: "countSessionsNeedingBackup",
    value: function countSessionsNeedingBackup(txn) {
      if (!txn) {
        txn = this._db.transaction("sessions_needing_backup", "readonly");
      }

      var objectStore = txn.objectStore("sessions_needing_backup");
      return new Promise(function (resolve, reject) {
        var req = objectStore.count();
        req.onerror = reject;

        req.onsuccess = function () {
          return resolve(req.result);
        };
      });
    }
  }, {
    key: "unmarkSessionsNeedingBackup",
    value: function unmarkSessionsNeedingBackup(sessions, txn) {
      if (!txn) {
        txn = this._db.transaction("sessions_needing_backup", "readwrite");
      }

      var objectStore = txn.objectStore("sessions_needing_backup");
      return Promise.all(sessions.map(function (session) {
        return new Promise(function (resolve, reject) {
          var req = objectStore["delete"]([session.senderKey, session.sessionId]);
          req.onsuccess = resolve;
          req.onerror = reject;
        });
      }));
    }
  }, {
    key: "markSessionsNeedingBackup",
    value: function markSessionsNeedingBackup(sessions, txn) {
      if (!txn) {
        txn = this._db.transaction("sessions_needing_backup", "readwrite");
      }

      var objectStore = txn.objectStore("sessions_needing_backup");
      return Promise.all(sessions.map(function (session) {
        return new Promise(function (resolve, reject) {
          var req = objectStore.put({
            senderCurve25519Key: session.senderKey,
            sessionId: session.sessionId
          });
          req.onsuccess = resolve;
          req.onerror = reject;
        });
      }));
    }
  }, {
    key: "doTxn",
    value: function doTxn(mode, stores, func) {
      var txn = this._db.transaction(stores, mode);

      var promise = promiseifyTxn(txn);
      var result = func(txn);
      return promise.then(function () {
        return result;
      });
    }
  }]);
  return Backend;
}();

exports.Backend = Backend;

function upgradeDatabase(db, oldVersion) {
  _logger.logger.log("Upgrading IndexedDBCryptoStore from version ".concat(oldVersion) + " to ".concat(VERSION));

  if (oldVersion < 1) {
    // The database did not previously exist.
    createDatabase(db);
  }

  if (oldVersion < 2) {
    db.createObjectStore("account");
  }

  if (oldVersion < 3) {
    var sessionsStore = db.createObjectStore("sessions", {
      keyPath: ["deviceKey", "sessionId"]
    });
    sessionsStore.createIndex("deviceKey", "deviceKey");
  }

  if (oldVersion < 4) {
    db.createObjectStore("inbound_group_sessions", {
      keyPath: ["senderCurve25519Key", "sessionId"]
    });
  }

  if (oldVersion < 5) {
    db.createObjectStore("device_data");
  }

  if (oldVersion < 6) {
    db.createObjectStore("rooms");
  }

  if (oldVersion < 7) {
    db.createObjectStore("sessions_needing_backup", {
      keyPath: ["senderCurve25519Key", "sessionId"]
    });
  }

  if (oldVersion < 8) {
    db.createObjectStore("inbound_group_sessions_withheld", {
      keyPath: ["senderCurve25519Key", "sessionId"]
    });
  }

  if (oldVersion < 9) {
    var problemsStore = db.createObjectStore("session_problems", {
      keyPath: ["deviceKey", "time"]
    });
    problemsStore.createIndex("deviceKey", "deviceKey");
    db.createObjectStore("notified_error_devices", {
      keyPath: ["userId", "deviceId"]
    });
  } // Expand as needed.

}

function createDatabase(db) {
  var outgoingRoomKeyRequestsStore = db.createObjectStore("outgoingRoomKeyRequests", {
    keyPath: "requestId"
  }); // we assume that the RoomKeyRequestBody will have room_id and session_id
  // properties, to make the index efficient.

  outgoingRoomKeyRequestsStore.createIndex("session", ["requestBody.room_id", "requestBody.session_id"]);
  outgoingRoomKeyRequestsStore.createIndex("state", "state");
}
/*
 * Aborts a transaction with a given exception
 * The transaction promise will be rejected with this exception.
 */


function abortWithException(txn, e) {
  // We cheekily stick our exception onto the transaction object here
  // We could alternatively make the thing we pass back to the app
  // an object containing the transaction and exception.
  txn._mx_abortexception = e;

  try {
    txn.abort();
  } catch (e) {// sometimes we won't be able to abort the transaction
    // (ie. if it's aborted or completed)
  }
}

function promiseifyTxn(txn) {
  return new Promise(function (resolve, reject) {
    txn.oncomplete = function () {
      if (txn._mx_abortexception !== undefined) {
        reject(txn._mx_abortexception);
      }

      resolve();
    };

    txn.onerror = function (event) {
      if (txn._mx_abortexception !== undefined) {
        reject(txn._mx_abortexception);
      } else {
        _logger.logger.log("Error performing indexeddb txn", event);

        reject(event.target.error);
      }
    };

    txn.onabort = function (event) {
      if (txn._mx_abortexception !== undefined) {
        reject(txn._mx_abortexception);
      } else {
        _logger.logger.log("Error performing indexeddb txn", event);

        reject(event.target.error);
      }
    };
  });
}