"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.VerificationBase = exports.SwitchStartEventError = void 0;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _wrapNativeSuper2 = _interopRequireDefault(require("@babel/runtime/helpers/wrapNativeSuper"));

var _event = require("../../models/event");

var _events = require("events");

var _logger = require("../../logger");

var _deviceinfo = require("../deviceinfo");

var _Error2 = require("./Error");

var _CrossSigning = require("../CrossSigning");

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

var timeoutException = new Error("Verification timed out");

var SwitchStartEventError = /*#__PURE__*/function (_Error) {
  (0, _inherits2["default"])(SwitchStartEventError, _Error);

  var _super = _createSuper(SwitchStartEventError);

  function SwitchStartEventError(startEvent) {
    var _this;

    (0, _classCallCheck2["default"])(this, SwitchStartEventError);
    _this = _super.call(this);
    _this.startEvent = startEvent;
    return _this;
  }

  return SwitchStartEventError;
}( /*#__PURE__*/(0, _wrapNativeSuper2["default"])(Error));

exports.SwitchStartEventError = SwitchStartEventError;

var VerificationBase = /*#__PURE__*/function (_EventEmitter) {
  (0, _inherits2["default"])(VerificationBase, _EventEmitter);

  var _super2 = _createSuper(VerificationBase);

  /**
   * Base class for verification methods.
   *
   * <p>Once a verifier object is created, the verification can be started by
   * calling the verify() method, which will return a promise that will
   * resolve when the verification is completed, or reject if it could not
   * complete.</p>
   *
   * <p>Subclasses must have a NAME class property.</p>
   *
   * @class
   *
   * @param {module:base-apis~Channel} channel the verification channel to send verification messages over.
   *
   * @param {module:base-apis~MatrixBaseApis} baseApis base matrix api interface
   *
   * @param {string} userId the user ID that is being verified
   *
   * @param {string} deviceId the device ID that is being verified
   *
   * @param {object} [startEvent] the m.key.verification.start event that
   * initiated this verification, if any
   *
   * @param {object} [request] the key verification request object related to
   * this verification, if any
   */
  function VerificationBase(channel, baseApis, userId, deviceId, startEvent, request) {
    var _this2;

    (0, _classCallCheck2["default"])(this, VerificationBase);
    _this2 = _super2.call(this);
    _this2._channel = channel;
    _this2._baseApis = baseApis;
    _this2.userId = userId;
    _this2.deviceId = deviceId;
    _this2.startEvent = startEvent;
    _this2.request = request;
    _this2.cancelled = false;
    _this2._done = false;
    _this2._promise = null;
    _this2._transactionTimeoutTimer = null;
    return _this2;
  }

  (0, _createClass2["default"])(VerificationBase, [{
    key: "_resetTimer",
    value: function _resetTimer() {
      var _this3 = this;

      _logger.logger.info("Refreshing/starting the verification transaction timeout timer");

      if (this._transactionTimeoutTimer !== null) {
        clearTimeout(this._transactionTimeoutTimer);
      }

      this._transactionTimeoutTimer = setTimeout(function () {
        if (!_this3._done && !_this3.cancelled) {
          _logger.logger.info("Triggering verification timeout");

          _this3.cancel(timeoutException);
        }
      }, 10 * 60 * 1000); // 10 minutes
    }
  }, {
    key: "_endTimer",
    value: function _endTimer() {
      if (this._transactionTimeoutTimer !== null) {
        clearTimeout(this._transactionTimeoutTimer);
        this._transactionTimeoutTimer = null;
      }
    }
  }, {
    key: "_send",
    value: function _send(type, uncompletedContent) {
      return this._channel.send(type, uncompletedContent);
    }
  }, {
    key: "_waitForEvent",
    value: function _waitForEvent(type) {
      var _this4 = this;

      if (this._done) {
        return Promise.reject(new Error("Verification is already done"));
      }

      var existingEvent = this.request.getEventFromOtherParty(type);

      if (existingEvent) {
        return Promise.resolve(existingEvent);
      }

      this._expectedEvent = type;
      return new Promise(function (resolve, reject) {
        _this4._resolveEvent = resolve;
        _this4._rejectEvent = reject;
      });
    }
  }, {
    key: "canSwitchStartEvent",
    value: function canSwitchStartEvent() {
      return false;
    }
  }, {
    key: "switchStartEvent",
    value: function switchStartEvent(event) {
      if (this.canSwitchStartEvent(event)) {
        _logger.logger.log("Verification Base: switching verification start event", {
          restartingFlow: !!this._rejectEvent
        });

        if (this._rejectEvent) {
          var reject = this._rejectEvent;
          this._rejectEvent = undefined;
          reject(new SwitchStartEventError(event));
        } else {
          this.startEvent = event;
        }
      }
    }
  }, {
    key: "handleEvent",
    value: function handleEvent(e) {
      if (this._done) {
        return;
      } else if (e.getType() === this._expectedEvent) {
        // if we receive an expected m.key.verification.done, then just
        // ignore it, since we don't need to do anything about it
        if (this._expectedEvent !== "m.key.verification.done") {
          this._expectedEvent = undefined;
          this._rejectEvent = undefined;

          this._resetTimer();

          this._resolveEvent(e);
        }
      } else if (e.getType() === "m.key.verification.cancel") {
        var reject = this._reject;
        this._reject = undefined; // there is only promise to reject if verify has been called

        if (reject) {
          var content = e.getContent();
          var reason = content.reason,
              code = content.code;
          reject(new Error("Other side cancelled verification " + "because ".concat(reason, " (").concat(code, ")")));
        }
      } else if (this._expectedEvent) {
        // only cancel if there is an event expected.
        // if there is no event expected, it means verify() wasn't called
        // and we're just replaying the timeline events when syncing
        // after a refresh when the events haven't been stored in the cache yet.
        var exception = new Error("Unexpected message: expecting " + this._expectedEvent + " but got " + e.getType());
        this._expectedEvent = undefined;

        if (this._rejectEvent) {
          var _reject = this._rejectEvent;
          this._rejectEvent = undefined;

          _reject(exception);
        }

        this.cancel(exception);
      }
    }
  }, {
    key: "done",
    value: function done() {
      this._endTimer(); // always kill the activity timer


      if (!this._done) {
        this.request.onVerifierFinished();

        this._resolve();

        return (0, _CrossSigning.requestKeysDuringVerification)(this._baseApis, this.userId, this.deviceId);
      }
    }
  }, {
    key: "cancel",
    value: function cancel(e) {
      this._endTimer(); // always kill the activity timer


      if (!this._done) {
        this.cancelled = true;
        this.request.onVerifierCancelled();

        if (this.userId && this.deviceId) {
          // send a cancellation to the other user (if it wasn't
          // cancelled by the other user)
          if (e === timeoutException) {
            var timeoutEvent = (0, _Error2.newTimeoutError)();

            this._send(timeoutEvent.getType(), timeoutEvent.getContent());
          } else if (e instanceof _event.MatrixEvent) {
            var sender = e.getSender();

            if (sender !== this.userId) {
              var content = e.getContent();

              if (e.getType() === "m.key.verification.cancel") {
                content.code = content.code || "m.unknown";
                content.reason = content.reason || content.body || "Unknown reason";

                this._send("m.key.verification.cancel", content);
              } else {
                this._send("m.key.verification.cancel", {
                  code: "m.unknown",
                  reason: content.body || "Unknown reason"
                });
              }
            }
          } else {
            this._send("m.key.verification.cancel", {
              code: "m.unknown",
              reason: e.toString()
            });
          }
        }

        if (this._promise !== null) {
          // when we cancel without a promise, we end up with a promise
          // but no reject function. If cancel is called again, we'd error.
          if (this._reject) this._reject(e);
        } else {
          // FIXME: this causes an "Uncaught promise" console message
          // if nothing ends up chaining this promise.
          this._promise = Promise.reject(e);
        } // Also emit a 'cancel' event that the app can listen for to detect cancellation
        // before calling verify()


        this.emit('cancel', e);
      }
    }
    /**
     * Begin the key verification
     *
     * @returns {Promise} Promise which resolves when the verification has
     *     completed.
     */

  }, {
    key: "verify",
    value: function verify() {
      var _this5 = this;

      if (this._promise) return this._promise;
      this._promise = new Promise(function (resolve, reject) {
        _this5._resolve = function () {
          _this5._done = true;

          _this5._endTimer();

          resolve.apply(void 0, arguments);
        };

        _this5._reject = function () {
          _this5._done = true;

          _this5._endTimer();

          reject.apply(void 0, arguments);
        };
      });

      if (this._doVerification && !this._started) {
        this._started = true;

        this._resetTimer(); // restart the timeout


        Promise.resolve(this._doVerification()).then(this.done.bind(this), this.cancel.bind(this));
      }

      return this._promise;
    }
  }, {
    key: "_verifyKeys",
    value: function () {
      var _verifyKeys2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(userId, keys, verifier) {
        var verifiedDevices, _i, _Object$entries, _Object$entries$_i, keyId, keyInfo, deviceId, device, crossSigningInfo, _i2, _verifiedDevices, _deviceId;

        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                // we try to verify all the keys that we're told about, but we might
                // not know about all of them, so keep track of the keys that we know
                // about, and ignore the rest
                verifiedDevices = [];
                _i = 0, _Object$entries = Object.entries(keys);

              case 2:
                if (!(_i < _Object$entries.length)) {
                  _context.next = 23;
                  break;
                }

                _Object$entries$_i = (0, _slicedToArray2["default"])(_Object$entries[_i], 2), keyId = _Object$entries$_i[0], keyInfo = _Object$entries$_i[1];
                deviceId = keyId.split(':', 2)[1];
                device = this._baseApis.getStoredDevice(userId, deviceId);

                if (!device) {
                  _context.next = 12;
                  break;
                }

                _context.next = 9;
                return verifier(keyId, device, keyInfo);

              case 9:
                verifiedDevices.push(deviceId);
                _context.next = 20;
                break;

              case 12:
                crossSigningInfo = this._baseApis._crypto._deviceList.getStoredCrossSigningForUser(userId);

                if (!(crossSigningInfo && crossSigningInfo.getId() === deviceId)) {
                  _context.next = 19;
                  break;
                }

                _context.next = 16;
                return verifier(keyId, _deviceinfo.DeviceInfo.fromStorage({
                  keys: (0, _defineProperty2["default"])({}, keyId, deviceId)
                }, deviceId), keyInfo);

              case 16:
                verifiedDevices.push(deviceId);
                _context.next = 20;
                break;

              case 19:
                _logger.logger.warn("verification: Could not find device ".concat(deviceId, " to verify"));

              case 20:
                _i++;
                _context.next = 2;
                break;

              case 23:
                if (verifiedDevices.length) {
                  _context.next = 25;
                  break;
                }

                throw new Error("No devices could be verified");

              case 25:
                _logger.logger.info("Verification completed! Marking devices verified: ", verifiedDevices); // TODO: There should probably be a batch version of this, otherwise it's going
                // to upload each signature in a separate API call which is silly because the
                // API supports as many signatures as you like.


                _i2 = 0, _verifiedDevices = verifiedDevices;

              case 27:
                if (!(_i2 < _verifiedDevices.length)) {
                  _context.next = 34;
                  break;
                }

                _deviceId = _verifiedDevices[_i2];
                _context.next = 31;
                return this._baseApis.setDeviceVerified(userId, _deviceId);

              case 31:
                _i2++;
                _context.next = 27;
                break;

              case 34:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function _verifyKeys(_x, _x2, _x3) {
        return _verifyKeys2.apply(this, arguments);
      }

      return _verifyKeys;
    }()
  }, {
    key: "initiatedByMe",
    get: function get() {
      // if there is no start event yet,
      // we probably want to send it,
      // which happens if we initiate
      if (!this.startEvent) {
        return true;
      }

      var sender = this.startEvent.getSender();
      var content = this.startEvent.getContent();
      return sender === this._baseApis.getUserId() && content.from_device === this._baseApis.getDeviceId();
    }
  }]);
  return VerificationBase;
}(_events.EventEmitter);

exports.VerificationBase = VerificationBase;