"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.SAS = void 0;

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _Base2 = require("./Base");

var _anotherJson = _interopRequireDefault(require("another-json"));

var _Error = require("./Error");

var _logger = require("../../logger");

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var START_TYPE = "m.key.verification.start";
var EVENTS = ["m.key.verification.accept", "m.key.verification.key", "m.key.verification.mac"];
var olmutil;
var newMismatchedSASError = (0, _Error.errorFactory)("m.mismatched_sas", "Mismatched short authentication string");
var newMismatchedCommitmentError = (0, _Error.errorFactory)("m.mismatched_commitment", "Mismatched commitment");

function generateDecimalSas(sasBytes) {
  /**
   *      +--------+--------+--------+--------+--------+
   *      | Byte 0 | Byte 1 | Byte 2 | Byte 3 | Byte 4 |
   *      +--------+--------+--------+--------+--------+
   * bits: 87654321 87654321 87654321 87654321 87654321
   *       \____________/\_____________/\____________/
   *         1st number    2nd number     3rd number
   */
  return [(sasBytes[0] << 5 | sasBytes[1] >> 3) + 1000, ((sasBytes[1] & 0x7) << 10 | sasBytes[2] << 2 | sasBytes[3] >> 6) + 1000, ((sasBytes[3] & 0x3f) << 7 | sasBytes[4] >> 1) + 1000];
}

var emojiMapping = [["🐶", "dog"], //  0
["🐱", "cat"], //  1
["🦁", "lion"], //  2
["🐎", "horse"], //  3
["🦄", "unicorn"], //  4
["🐷", "pig"], //  5
["🐘", "elephant"], //  6
["🐰", "rabbit"], //  7
["🐼", "panda"], //  8
["🐓", "rooster"], //  9
["🐧", "penguin"], // 10
["🐢", "turtle"], // 11
["🐟", "fish"], // 12
["🐙", "octopus"], // 13
["🦋", "butterfly"], // 14
["🌷", "flower"], // 15
["🌳", "tree"], // 16
["🌵", "cactus"], // 17
["🍄", "mushroom"], // 18
["🌏", "globe"], // 19
["🌙", "moon"], // 20
["☁️", "cloud"], // 21
["🔥", "fire"], // 22
["🍌", "banana"], // 23
["🍎", "apple"], // 24
["🍓", "strawberry"], // 25
["🌽", "corn"], // 26
["🍕", "pizza"], // 27
["🎂", "cake"], // 28
["❤️", "heart"], // 29
["🙂", "smiley"], // 30
["🤖", "robot"], // 31
["🎩", "hat"], // 32
["👓", "glasses"], // 33
["🔧", "spanner"], // 34
["🎅", "santa"], // 35
["👍", "thumbs up"], // 36
["☂️", "umbrella"], // 37
["⌛", "hourglass"], // 38
["⏰", "clock"], // 39
["🎁", "gift"], // 40
["💡", "light bulb"], // 41
["📕", "book"], // 42
["✏️", "pencil"], // 43
["📎", "paperclip"], // 44
["✂️", "scissors"], // 45
["🔒", "lock"], // 46
["🔑", "key"], // 47
["🔨", "hammer"], // 48
["☎️", "telephone"], // 49
["🏁", "flag"], // 50
["🚂", "train"], // 51
["🚲", "bicycle"], // 52
["✈️", "aeroplane"], // 53
["🚀", "rocket"], // 54
["🏆", "trophy"], // 55
["⚽", "ball"], // 56
["🎸", "guitar"], // 57
["🎺", "trumpet"], // 58
["🔔", "bell"], // 59
["⚓️", "anchor"], // 60
["🎧", "headphones"], // 61
["📁", "folder"], // 62
["📌", "pin"] // 63
];

function generateEmojiSas(sasBytes) {
  var emojis = [// just like base64 encoding
  sasBytes[0] >> 2, (sasBytes[0] & 0x3) << 4 | sasBytes[1] >> 4, (sasBytes[1] & 0xf) << 2 | sasBytes[2] >> 6, sasBytes[2] & 0x3f, sasBytes[3] >> 2, (sasBytes[3] & 0x3) << 4 | sasBytes[4] >> 4, (sasBytes[4] & 0xf) << 2 | sasBytes[5] >> 6];
  return emojis.map(function (num) {
    return emojiMapping[num];
  });
}

var sasGenerators = {
  decimal: generateDecimalSas,
  emoji: generateEmojiSas
};

function generateSas(sasBytes, methods) {
  var sas = {};

  var _iterator = _createForOfIteratorHelper(methods),
      _step;

  try {
    for (_iterator.s(); !(_step = _iterator.n()).done;) {
      var method = _step.value;

      if (method in sasGenerators) {
        sas[method] = sasGenerators[method](sasBytes);
      }
    }
  } catch (err) {
    _iterator.e(err);
  } finally {
    _iterator.f();
  }

  return sas;
}

var macMethods = {
  "hkdf-hmac-sha256": "calculate_mac",
  "hmac-sha256": "calculate_mac_long_kdf"
};

function calculateMAC(olmSAS, method) {
  return function () {
    var macFunction = olmSAS[macMethods[method]];

    for (var _len = arguments.length, args = new Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    var mac = macFunction.apply(olmSAS, args);

    _logger.logger.log("SAS calculateMAC:", method, args, mac);

    return mac;
  };
}

var calculateKeyAgreement = {
  "curve25519-hkdf-sha256": function curve25519HkdfSha256(sas, olmSAS, bytes) {
    var ourInfo = "".concat(sas._baseApis.getUserId(), "|").concat(sas._baseApis.deviceId, "|") + "".concat(sas.ourSASPubKey, "|");
    var theirInfo = "".concat(sas.userId, "|").concat(sas.deviceId, "|").concat(sas.theirSASPubKey, "|");
    var sasInfo = "MATRIX_KEY_VERIFICATION_SAS|" + (sas.initiatedByMe ? ourInfo + theirInfo : theirInfo + ourInfo) + sas._channel.transactionId;
    return olmSAS.generate_bytes(sasInfo, bytes);
  },
  "curve25519": function curve25519(sas, olmSAS, bytes) {
    var ourInfo = "".concat(sas._baseApis.getUserId()).concat(sas._baseApis.deviceId);
    var theirInfo = "".concat(sas.userId).concat(sas.deviceId);
    var sasInfo = "MATRIX_KEY_VERIFICATION_SAS" + (sas.initiatedByMe ? ourInfo + theirInfo : theirInfo + ourInfo) + sas._channel.transactionId;
    return olmSAS.generate_bytes(sasInfo, bytes);
  }
};
/* lists of algorithms/methods that are supported.  The key agreement, hashes,
 * and MAC lists should be sorted in order of preference (most preferred
 * first).
 */

var KEY_AGREEMENT_LIST = ["curve25519-hkdf-sha256", "curve25519"];
var HASHES_LIST = ["sha256"];
var MAC_LIST = ["hkdf-hmac-sha256", "hmac-sha256"];
var SAS_LIST = Object.keys(sasGenerators);
var KEY_AGREEMENT_SET = new Set(KEY_AGREEMENT_LIST);
var HASHES_SET = new Set(HASHES_LIST);
var MAC_SET = new Set(MAC_LIST);
var SAS_SET = new Set(SAS_LIST);

function intersection(anArray, aSet) {
  return anArray instanceof Array ? anArray.filter(function (x) {
    return aSet.has(x);
  }) : [];
}
/**
 * @alias module:crypto/verification/SAS
 * @extends {module:crypto/verification/Base}
 */


var SAS = /*#__PURE__*/function (_Base) {
  (0, _inherits2["default"])(SAS, _Base);

  var _super = _createSuper(SAS);

  function SAS() {
    (0, _classCallCheck2["default"])(this, SAS);
    return _super.apply(this, arguments);
  }

  (0, _createClass2["default"])(SAS, [{
    key: "_doVerification",
    value: function () {
      var _doVerification2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee() {
        var retry;
        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                _context.next = 2;
                return global.Olm.init();

              case 2:
                olmutil = olmutil || new global.Olm.Utility(); // make sure user's keys are downloaded

                _context.next = 5;
                return this._baseApis.downloadKeys([this.userId]);

              case 5:
                retry = false;

              case 6:
                _context.prev = 6;

                if (!this.initiatedByMe) {
                  _context.next = 13;
                  break;
                }

                _context.next = 10;
                return this._doSendVerification();

              case 10:
                return _context.abrupt("return", _context.sent);

              case 13:
                _context.next = 15;
                return this._doRespondVerification();

              case 15:
                return _context.abrupt("return", _context.sent);

              case 16:
                _context.next = 26;
                break;

              case 18:
                _context.prev = 18;
                _context.t0 = _context["catch"](6);

                if (!(_context.t0 instanceof _Base2.SwitchStartEventError)) {
                  _context.next = 25;
                  break;
                }

                // this changes what initiatedByMe returns
                this.startEvent = _context.t0.startEvent;
                retry = true;
                _context.next = 26;
                break;

              case 25:
                throw _context.t0;

              case 26:
                if (retry) {
                  _context.next = 6;
                  break;
                }

              case 27:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this, [[6, 18]]);
      }));

      function _doVerification() {
        return _doVerification2.apply(this, arguments);
      }

      return _doVerification;
    }()
  }, {
    key: "canSwitchStartEvent",
    value: function canSwitchStartEvent(event) {
      if (event.getType() !== START_TYPE) {
        return false;
      }

      var content = event.getContent();
      return content && content.method === SAS.NAME && this._waitingForAccept;
    }
  }, {
    key: "_sendStart",
    value: function () {
      var _sendStart2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
        var startContent;
        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                startContent = this._channel.completeContent(START_TYPE, {
                  method: SAS.NAME,
                  from_device: this._baseApis.deviceId,
                  key_agreement_protocols: KEY_AGREEMENT_LIST,
                  hashes: HASHES_LIST,
                  message_authentication_codes: MAC_LIST,
                  // FIXME: allow app to specify what SAS methods can be used
                  short_authentication_string: SAS_LIST
                });
                _context2.next = 3;
                return this._channel.sendCompleted(START_TYPE, startContent);

              case 3:
                return _context2.abrupt("return", startContent);

              case 4:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function _sendStart() {
        return _sendStart2.apply(this, arguments);
      }

      return _sendStart;
    }()
  }, {
    key: "_doSendVerification",
    value: function () {
      var _doSendVerification2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4() {
        var _this = this;

        var startContent, e, content, sasMethods, keyAgreement, macMethod, hashCommitment, olmSAS, commitmentStr, sasBytes, verifySAS, _yield$Promise$all, _yield$Promise$all2;

        return _regenerator["default"].wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                this._waitingForAccept = true;

                if (!this.startEvent) {
                  _context4.next = 5;
                  break;
                }

                startContent = this._channel.completedContentFromEvent(this.startEvent);
                _context4.next = 8;
                break;

              case 5:
                _context4.next = 7;
                return this._sendStart();

              case 7:
                startContent = _context4.sent;

              case 8:
                if (this.initiatedByMe) {
                  _context4.next = 10;
                  break;
                }

                throw new _Base2.SwitchStartEventError(this.startEvent);

              case 10:
                _context4.prev = 10;
                _context4.next = 13;
                return this._waitForEvent("m.key.verification.accept");

              case 13:
                e = _context4.sent;

              case 14:
                _context4.prev = 14;
                this._waitingForAccept = false;
                return _context4.finish(14);

              case 17:
                content = e.getContent();
                sasMethods = intersection(content.short_authentication_string, SAS_SET);

                if (KEY_AGREEMENT_SET.has(content.key_agreement_protocol) && HASHES_SET.has(content.hash) && MAC_SET.has(content.message_authentication_code) && sasMethods.length) {
                  _context4.next = 21;
                  break;
                }

                throw (0, _Error.newUnknownMethodError)();

              case 21:
                if (!(typeof content.commitment !== "string")) {
                  _context4.next = 23;
                  break;
                }

                throw (0, _Error.newInvalidMessageError)();

              case 23:
                keyAgreement = content.key_agreement_protocol;
                macMethod = content.message_authentication_code;
                hashCommitment = content.commitment;
                olmSAS = new global.Olm.SAS();
                _context4.prev = 27;
                this.ourSASPubKey = olmSAS.get_pubkey();
                _context4.next = 31;
                return this._send("m.key.verification.key", {
                  key: this.ourSASPubKey
                });

              case 31:
                _context4.next = 33;
                return this._waitForEvent("m.key.verification.key");

              case 33:
                e = _context4.sent;
                // FIXME: make sure event is properly formed
                content = e.getContent();
                commitmentStr = content.key + _anotherJson["default"].stringify(startContent); // TODO: use selected hash function (when we support multiple)

                if (!(olmutil.sha256(commitmentStr) !== hashCommitment)) {
                  _context4.next = 38;
                  break;
                }

                throw newMismatchedCommitmentError();

              case 38:
                this.theirSASPubKey = content.key;
                olmSAS.set_their_key(content.key);
                sasBytes = calculateKeyAgreement[keyAgreement](this, olmSAS, 6);
                verifySAS = new Promise(function (resolve, reject) {
                  _this.sasEvent = {
                    sas: generateSas(sasBytes, sasMethods),
                    confirm: function () {
                      var _confirm = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3() {
                        return _regenerator["default"].wrap(function _callee3$(_context3) {
                          while (1) {
                            switch (_context3.prev = _context3.next) {
                              case 0:
                                _context3.prev = 0;
                                _context3.next = 3;
                                return _this._sendMAC(olmSAS, macMethod);

                              case 3:
                                resolve();
                                _context3.next = 9;
                                break;

                              case 6:
                                _context3.prev = 6;
                                _context3.t0 = _context3["catch"](0);
                                reject(_context3.t0);

                              case 9:
                              case "end":
                                return _context3.stop();
                            }
                          }
                        }, _callee3, null, [[0, 6]]);
                      }));

                      function confirm() {
                        return _confirm.apply(this, arguments);
                      }

                      return confirm;
                    }(),
                    cancel: function cancel() {
                      return reject((0, _Error.newUserCancelledError)());
                    },
                    mismatch: function mismatch() {
                      return reject(newMismatchedSASError());
                    }
                  };

                  _this.emit("show_sas", _this.sasEvent);
                });
                _context4.next = 44;
                return Promise.all([this._waitForEvent("m.key.verification.mac").then(function (e) {
                  // we don't expect any more messages from the other
                  // party, and they may send a m.key.verification.done
                  // when they're done on their end
                  _this._expectedEvent = "m.key.verification.done";
                  return e;
                }), verifySAS]);

              case 44:
                _yield$Promise$all = _context4.sent;
                _yield$Promise$all2 = (0, _slicedToArray2["default"])(_yield$Promise$all, 1);
                e = _yield$Promise$all2[0];
                content = e.getContent();
                _context4.next = 50;
                return this._checkMAC(olmSAS, content, macMethod);

              case 50:
                _context4.prev = 50;
                olmSAS.free();
                return _context4.finish(50);

              case 53:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this, [[10,, 14, 17], [27,, 50, 53]]);
      }));

      function _doSendVerification() {
        return _doSendVerification2.apply(this, arguments);
      }

      return _doSendVerification;
    }()
  }, {
    key: "_doRespondVerification",
    value: function () {
      var _doRespondVerification2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee6() {
        var _this2 = this;

        var content, keyAgreement, hashMethod, macMethod, sasMethods, olmSAS, commitmentStr, e, sasBytes, verifySAS, _yield$Promise$all3, _yield$Promise$all4;

        return _regenerator["default"].wrap(function _callee6$(_context6) {
          while (1) {
            switch (_context6.prev = _context6.next) {
              case 0:
                // as m.related_to is not included in the encrypted content in e2e rooms,
                // we need to make sure it is added
                content = this._channel.completedContentFromEvent(this.startEvent); // Note: we intersect using our pre-made lists, rather than the sets,
                // so that the result will be in our order of preference.  Then
                // fetching the first element from the array will give our preferred
                // method out of the ones offered by the other party.

                keyAgreement = intersection(KEY_AGREEMENT_LIST, new Set(content.key_agreement_protocols))[0];
                hashMethod = intersection(HASHES_LIST, new Set(content.hashes))[0];
                macMethod = intersection(MAC_LIST, new Set(content.message_authentication_codes))[0]; // FIXME: allow app to specify what SAS methods can be used

                sasMethods = intersection(content.short_authentication_string, SAS_SET);

                if (keyAgreement !== undefined && hashMethod !== undefined && macMethod !== undefined && sasMethods.length) {
                  _context6.next = 7;
                  break;
                }

                throw (0, _Error.newUnknownMethodError)();

              case 7:
                olmSAS = new global.Olm.SAS();
                _context6.prev = 8;
                commitmentStr = olmSAS.get_pubkey() + _anotherJson["default"].stringify(content);
                _context6.next = 12;
                return this._send("m.key.verification.accept", {
                  key_agreement_protocol: keyAgreement,
                  hash: hashMethod,
                  message_authentication_code: macMethod,
                  short_authentication_string: sasMethods,
                  // TODO: use selected hash function (when we support multiple)
                  commitment: olmutil.sha256(commitmentStr)
                });

              case 12:
                _context6.next = 14;
                return this._waitForEvent("m.key.verification.key");

              case 14:
                e = _context6.sent;
                // FIXME: make sure event is properly formed
                content = e.getContent();
                this.theirSASPubKey = content.key;
                olmSAS.set_their_key(content.key);
                this.ourSASPubKey = olmSAS.get_pubkey();
                _context6.next = 21;
                return this._send("m.key.verification.key", {
                  key: this.ourSASPubKey
                });

              case 21:
                sasBytes = calculateKeyAgreement[keyAgreement](this, olmSAS, 6);
                verifySAS = new Promise(function (resolve, reject) {
                  _this2.sasEvent = {
                    sas: generateSas(sasBytes, sasMethods),
                    confirm: function () {
                      var _confirm2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5() {
                        return _regenerator["default"].wrap(function _callee5$(_context5) {
                          while (1) {
                            switch (_context5.prev = _context5.next) {
                              case 0:
                                _context5.prev = 0;
                                _context5.next = 3;
                                return _this2._sendMAC(olmSAS, macMethod);

                              case 3:
                                resolve();
                                _context5.next = 9;
                                break;

                              case 6:
                                _context5.prev = 6;
                                _context5.t0 = _context5["catch"](0);
                                reject(_context5.t0);

                              case 9:
                              case "end":
                                return _context5.stop();
                            }
                          }
                        }, _callee5, null, [[0, 6]]);
                      }));

                      function confirm() {
                        return _confirm2.apply(this, arguments);
                      }

                      return confirm;
                    }(),
                    cancel: function cancel() {
                      return reject((0, _Error.newUserCancelledError)());
                    },
                    mismatch: function mismatch() {
                      return reject(newMismatchedSASError());
                    }
                  };

                  _this2.emit("show_sas", _this2.sasEvent);
                });
                _context6.next = 25;
                return Promise.all([this._waitForEvent("m.key.verification.mac").then(function (e) {
                  // we don't expect any more messages from the other
                  // party, and they may send a m.key.verification.done
                  // when they're done on their end
                  _this2._expectedEvent = "m.key.verification.done";
                  return e;
                }), verifySAS]);

              case 25:
                _yield$Promise$all3 = _context6.sent;
                _yield$Promise$all4 = (0, _slicedToArray2["default"])(_yield$Promise$all3, 1);
                e = _yield$Promise$all4[0];
                content = e.getContent();
                _context6.next = 31;
                return this._checkMAC(olmSAS, content, macMethod);

              case 31:
                _context6.prev = 31;
                olmSAS.free();
                return _context6.finish(31);

              case 34:
              case "end":
                return _context6.stop();
            }
          }
        }, _callee6, this, [[8,, 31, 34]]);
      }));

      function _doRespondVerification() {
        return _doRespondVerification2.apply(this, arguments);
      }

      return _doRespondVerification;
    }()
  }, {
    key: "_sendMAC",
    value: function _sendMAC(olmSAS, method) {
      var mac = {};
      var keyList = [];

      var baseInfo = "MATRIX_KEY_VERIFICATION_MAC" + this._baseApis.getUserId() + this._baseApis.deviceId + this.userId + this.deviceId + this._channel.transactionId;

      var deviceKeyId = "ed25519:".concat(this._baseApis.deviceId);
      mac[deviceKeyId] = calculateMAC(olmSAS, method)(this._baseApis.getDeviceEd25519Key(), baseInfo + deviceKeyId);
      keyList.push(deviceKeyId);

      var crossSigningId = this._baseApis.getCrossSigningId();

      if (crossSigningId) {
        var crossSigningKeyId = "ed25519:".concat(crossSigningId);
        mac[crossSigningKeyId] = calculateMAC(olmSAS, method)(crossSigningId, baseInfo + crossSigningKeyId);
        keyList.push(crossSigningKeyId);
      }

      var keys = calculateMAC(olmSAS, method)(keyList.sort().join(","), baseInfo + "KEY_IDS");
      return this._send("m.key.verification.mac", {
        mac: mac,
        keys: keys
      });
    }
  }, {
    key: "_checkMAC",
    value: function () {
      var _checkMAC2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee7(olmSAS, content, method) {
        var baseInfo;
        return _regenerator["default"].wrap(function _callee7$(_context7) {
          while (1) {
            switch (_context7.prev = _context7.next) {
              case 0:
                baseInfo = "MATRIX_KEY_VERIFICATION_MAC" + this.userId + this.deviceId + this._baseApis.getUserId() + this._baseApis.deviceId + this._channel.transactionId;

                if (!(content.keys !== calculateMAC(olmSAS, method)(Object.keys(content.mac).sort().join(","), baseInfo + "KEY_IDS"))) {
                  _context7.next = 3;
                  break;
                }

                throw (0, _Error.newKeyMismatchError)();

              case 3:
                _context7.next = 5;
                return this._verifyKeys(this.userId, content.mac, function (keyId, device, keyInfo) {
                  if (keyInfo !== calculateMAC(olmSAS, method)(device.keys[keyId], baseInfo + keyId)) {
                    throw (0, _Error.newKeyMismatchError)();
                  }
                });

              case 5:
              case "end":
                return _context7.stop();
            }
          }
        }, _callee7, this);
      }));

      function _checkMAC(_x, _x2, _x3) {
        return _checkMAC2.apply(this, arguments);
      }

      return _checkMAC;
    }()
  }, {
    key: "events",
    get: function get() {
      return EVENTS;
    }
  }], [{
    key: "NAME",
    get: function get() {
      return "m.sas.v1";
    }
  }]);
  return SAS;
}(_Base2.VerificationBase);

exports.SAS = SAS;