"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.InRoomRequests = exports.InRoomChannel = void 0;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _VerificationRequest = require("./VerificationRequest");

var _logger = require("../../../logger");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

var MESSAGE_TYPE = "m.room.message";
var M_REFERENCE = "m.reference";
var M_RELATES_TO = "m.relates_to";
/**
 * A key verification channel that sends verification events in the timeline of a room.
 * Uses the event id of the initial m.key.verification.request event as a transaction id.
 */

var InRoomChannel = /*#__PURE__*/function () {
  /**
   * @param {MatrixClient} client the matrix client, to send messages with and get current user & device from.
   * @param {string} roomId id of the room where verification events should be posted in, should be a DM with the given user.
   * @param {string} userId id of user that the verification request is directed at, should be present in the room.
   */
  function InRoomChannel(client, roomId) {
    var userId = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;
    (0, _classCallCheck2["default"])(this, InRoomChannel);
    this._client = client;
    this._roomId = roomId;
    this.userId = userId;
    this._requestEventId = null;
  }

  (0, _createClass2["default"])(InRoomChannel, [{
    key: "getTimestamp",

    /**
     * @param {MatrixEvent} event the event to get the timestamp of
     * @return {number} the timestamp when the event was sent
     */
    value: function getTimestamp(event) {
      return event.getTs();
    }
    /**
     * Checks whether the given event type should be allowed to initiate a new VerificationRequest over this channel
     * @param {string} type the event type to check
     * @returns {bool} boolean flag
     */

  }, {
    key: "handleEvent",

    /**
     * Changes the state of the channel, request, and verifier in response to a key verification event.
     * @param {MatrixEvent} event to handle
     * @param {VerificationRequest} request the request to forward handling to
     * @param {bool} isLiveEvent whether this is an even received through sync or not
     * @returns {Promise} a promise that resolves when any requests as an anwser to the passed-in event are sent.
     */
    value: function () {
      var _handleEvent = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(event, request, isLiveEvent) {
        var type, userId, ownUserId, sender, isRemoteEcho, isSentByUs;
        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                if (!request.hasEventId(event.getId())) {
                  _context.next = 2;
                  break;
                }

                return _context.abrupt("return");

              case 2:
                type = InRoomChannel.getEventType(event); // do validations that need state (roomId, userId),
                // ignore if invalid

                if (!(event.getRoomId() !== this._roomId)) {
                  _context.next = 5;
                  break;
                }

                return _context.abrupt("return");

              case 5:
                // set userId if not set already
                if (this.userId === null) {
                  userId = InRoomChannel.getOtherPartyUserId(event, this._client);

                  if (userId) {
                    this.userId = userId;
                  }
                } // ignore events not sent by us or the other party


                ownUserId = this._client.getUserId();
                sender = event.getSender();

                if (!(this.userId !== null)) {
                  _context.next = 12;
                  break;
                }

                if (!(sender !== ownUserId && sender !== this.userId)) {
                  _context.next = 12;
                  break;
                }

                _logger.logger.log("InRoomChannel: ignoring verification event from " + "non-participating sender ".concat(sender));

                return _context.abrupt("return");

              case 12:
                if (this._requestEventId === null) {
                  this._requestEventId = InRoomChannel.getTransactionId(event);
                }

                isRemoteEcho = !!event.getUnsigned().transaction_id;
                isSentByUs = event.getSender() === this._client.getUserId();
                _context.next = 17;
                return request.handleEvent(type, event, isLiveEvent, isRemoteEcho, isSentByUs);

              case 17:
                return _context.abrupt("return", _context.sent);

              case 18:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function handleEvent(_x, _x2, _x3) {
        return _handleEvent.apply(this, arguments);
      }

      return handleEvent;
    }()
    /**
     * Adds the transaction id (relation) back to a received event
     * so it has the same format as returned by `completeContent` before sending.
     * The relation can not appear on the event content because of encryption,
     * relations are excluded from encryption.
     * @param {MatrixEvent} event the received event
     * @returns {Object} the content object with the relation added again
     */

  }, {
    key: "completedContentFromEvent",
    value: function completedContentFromEvent(event) {
      // ensure m.related_to is included in e2ee rooms
      // as the field is excluded from encryption
      var content = Object.assign({}, event.getContent());
      content[M_RELATES_TO] = event.getRelation();
      return content;
    }
    /**
     * Add all the fields to content needed for sending it over this channel.
     * This is public so verification methods (SAS uses this) can get the exact
     * content that will be sent independent of the used channel,
     * as they need to calculate the hash of it.
     * @param {string} type the event type
     * @param {object} content the (incomplete) content
     * @returns {object} the complete content, as it will be sent.
     */

  }, {
    key: "completeContent",
    value: function completeContent(type, content) {
      content = Object.assign({}, content);

      if (type === _VerificationRequest.REQUEST_TYPE || type === _VerificationRequest.READY_TYPE || type === _VerificationRequest.START_TYPE) {
        content.from_device = this._client.getDeviceId();
      }

      if (type === _VerificationRequest.REQUEST_TYPE) {
        // type is mapped to m.room.message in the send method
        content = {
          body: this._client.getUserId() + " is requesting to verify " + "your key, but your client does not support in-chat key " + "verification.  You will need to use legacy key " + "verification to verify keys.",
          msgtype: _VerificationRequest.REQUEST_TYPE,
          to: this.userId,
          from_device: content.from_device,
          methods: content.methods
        };
      } else {
        content[M_RELATES_TO] = {
          rel_type: M_REFERENCE,
          event_id: this.transactionId
        };
      }

      return content;
    }
    /**
     * Send an event over the channel with the content not having gone through `completeContent`.
     * @param {string} type the event type
     * @param {object} uncompletedContent the (incomplete) content
     * @returns {Promise} the promise of the request
     */

  }, {
    key: "send",
    value: function send(type, uncompletedContent) {
      var content = this.completeContent(type, uncompletedContent);
      return this.sendCompleted(type, content);
    }
    /**
     * Send an event over the channel with the content having gone through `completeContent` already.
     * @param {string} type the event type
     * @param {object} content
     * @returns {Promise} the promise of the request
     */

  }, {
    key: "sendCompleted",
    value: function () {
      var _sendCompleted = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(type, content) {
        var sendType, response;
        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                sendType = type;

                if (type === _VerificationRequest.REQUEST_TYPE) {
                  sendType = MESSAGE_TYPE;
                }

                _context2.next = 4;
                return this._client.sendEvent(this._roomId, sendType, content);

              case 4:
                response = _context2.sent;

                if (type === _VerificationRequest.REQUEST_TYPE) {
                  this._requestEventId = response.event_id;
                }

              case 6:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function sendCompleted(_x4, _x5) {
        return _sendCompleted.apply(this, arguments);
      }

      return sendCompleted;
    }()
  }, {
    key: "receiveStartFromOtherDevices",
    get: function get() {
      return true;
    }
  }, {
    key: "roomId",
    get: function get() {
      return this._roomId;
    }
    /** The transaction id generated/used by this verification channel */

  }, {
    key: "transactionId",
    get: function get() {
      return this._requestEventId;
    }
  }], [{
    key: "getOtherPartyUserId",
    value: function getOtherPartyUserId(event, client) {
      var type = InRoomChannel.getEventType(event);

      if (type !== _VerificationRequest.REQUEST_TYPE) {
        return;
      }

      var ownUserId = client.getUserId();
      var sender = event.getSender();
      var content = event.getContent();
      var receiver = content.to;

      if (sender === ownUserId) {
        return receiver;
      } else if (receiver === ownUserId) {
        return sender;
      }
    }
  }, {
    key: "canCreateRequest",
    value: function canCreateRequest(type) {
      return type === _VerificationRequest.REQUEST_TYPE;
    }
    /**
     * Extract the transaction id used by a given key verification event, if any
     * @param {MatrixEvent} event the event
     * @returns {string} the transaction id
     */

  }, {
    key: "getTransactionId",
    value: function getTransactionId(event) {
      if (InRoomChannel.getEventType(event) === _VerificationRequest.REQUEST_TYPE) {
        return event.getId();
      } else {
        var relation = event.getRelation();

        if (relation && relation.rel_type === M_REFERENCE) {
          return relation.event_id;
        }
      }
    }
    /**
     * Checks whether this event is a well-formed key verification event.
     * This only does checks that don't rely on the current state of a potentially already channel
     * so we can prevent channels being created by invalid events.
     * `handleEvent` can do more checks and choose to ignore invalid events.
     * @param {MatrixEvent} event the event to validate
     * @param {MatrixClient} client the client to get the current user and device id from
     * @returns {bool} whether the event is valid and should be passed to handleEvent
     */

  }, {
    key: "validateEvent",
    value: function validateEvent(event, client) {
      var txnId = InRoomChannel.getTransactionId(event);

      if (typeof txnId !== "string" || txnId.length === 0) {
        return false;
      }

      var type = InRoomChannel.getEventType(event);
      var content = event.getContent(); // from here on we're fairly sure that this is supposed to be
      // part of a verification request, so be noisy when rejecting something

      if (type === _VerificationRequest.REQUEST_TYPE) {
        if (!content || typeof content.to !== "string" || !content.to.length) {
          _logger.logger.log("InRoomChannel: validateEvent: " + "no valid to " + (content && content.to));

          return false;
        } // ignore requests that are not direct to or sent by the syncing user


        if (!InRoomChannel.getOtherPartyUserId(event, client)) {
          _logger.logger.log("InRoomChannel: validateEvent: " + "not directed to or sent by me: ".concat(event.getSender()) + ", ".concat(content && content.to));

          return false;
        }
      }

      return _VerificationRequest.VerificationRequest.validateEvent(type, event, client);
    }
    /**
     * As m.key.verification.request events are as m.room.message events with the InRoomChannel
     * to have a fallback message in non-supporting clients, we map the real event type
     * to the symbolic one to keep things in unison with ToDeviceChannel
     * @param {MatrixEvent} event the event to get the type of
     * @returns {string} the "symbolic" event type
     */

  }, {
    key: "getEventType",
    value: function getEventType(event) {
      var type = event.getType();

      if (type === MESSAGE_TYPE) {
        var content = event.getContent();

        if (content) {
          var msgtype = content.msgtype;

          if (msgtype === _VerificationRequest.REQUEST_TYPE) {
            return _VerificationRequest.REQUEST_TYPE;
          }
        }
      }

      if (type && type !== _VerificationRequest.REQUEST_TYPE) {
        return type;
      } else {
        return "";
      }
    }
  }]);
  return InRoomChannel;
}();

exports.InRoomChannel = InRoomChannel;

var InRoomRequests = /*#__PURE__*/function () {
  function InRoomRequests() {
    (0, _classCallCheck2["default"])(this, InRoomRequests);
    this._requestsByRoomId = new Map();
  }

  (0, _createClass2["default"])(InRoomRequests, [{
    key: "getRequest",
    value: function getRequest(event) {
      var roomId = event.getRoomId();
      var txnId = InRoomChannel.getTransactionId(event);
      return this._getRequestByTxnId(roomId, txnId);
    }
  }, {
    key: "getRequestByChannel",
    value: function getRequestByChannel(channel) {
      return this._getRequestByTxnId(channel.roomId, channel.transactionId);
    }
  }, {
    key: "_getRequestByTxnId",
    value: function _getRequestByTxnId(roomId, txnId) {
      var requestsByTxnId = this._requestsByRoomId.get(roomId);

      if (requestsByTxnId) {
        return requestsByTxnId.get(txnId);
      }
    }
  }, {
    key: "setRequest",
    value: function setRequest(event, request) {
      this._setRequest(event.getRoomId(), InRoomChannel.getTransactionId(event), request);
    }
  }, {
    key: "setRequestByChannel",
    value: function setRequestByChannel(channel, request) {
      this._setRequest(channel.roomId, channel.transactionId, request);
    }
  }, {
    key: "_setRequest",
    value: function _setRequest(roomId, txnId, request) {
      var requestsByTxnId = this._requestsByRoomId.get(roomId);

      if (!requestsByTxnId) {
        requestsByTxnId = new Map();

        this._requestsByRoomId.set(roomId, requestsByTxnId);
      }

      requestsByTxnId.set(txnId, request);
    }
  }, {
    key: "removeRequest",
    value: function removeRequest(event) {
      var roomId = event.getRoomId();

      var requestsByTxnId = this._requestsByRoomId.get(roomId);

      if (requestsByTxnId) {
        requestsByTxnId["delete"](InRoomChannel.getTransactionId(event));

        if (requestsByTxnId.size === 0) {
          this._requestsByRoomId["delete"](roomId);
        }
      }
    }
  }, {
    key: "findRequestInProgress",
    value: function findRequestInProgress(roomId) {
      var requestsByTxnId = this._requestsByRoomId.get(roomId);

      if (requestsByTxnId) {
        var _iterator = _createForOfIteratorHelper(requestsByTxnId.values()),
            _step;

        try {
          for (_iterator.s(); !(_step = _iterator.n()).done;) {
            var request = _step.value;

            if (request.pending) {
              return request;
            }
          }
        } catch (err) {
          _iterator.e(err);
        } finally {
          _iterator.f();
        }
      }
    }
  }]);
  return InRoomRequests;
}();

exports.InRoomRequests = InRoomRequests;