"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.ToDeviceRequests = exports.ToDeviceChannel = void 0;

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _typeof2 = _interopRequireDefault(require("@babel/runtime/helpers/typeof"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _randomstring = require("../../../randomstring");

var _logger = require("../../../logger");

var _VerificationRequest = require("./VerificationRequest");

var _Error = require("../Error");

var _event = require("../../../models/event");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

/**
 * A key verification channel that sends verification events over to_device messages.
 * Generates its own transaction ids.
 */
var ToDeviceChannel = /*#__PURE__*/function () {
  // userId and devices of user we're about to verify
  function ToDeviceChannel(client, userId, devices) {
    var transactionId = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : null;
    var deviceId = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : null;
    (0, _classCallCheck2["default"])(this, ToDeviceChannel);
    this._client = client;
    this.userId = userId;
    this._devices = devices;
    this.transactionId = transactionId;
    this._deviceId = deviceId;
  }

  (0, _createClass2["default"])(ToDeviceChannel, [{
    key: "isToDevices",
    value: function isToDevices(devices) {
      var _this = this;

      if (devices.length === this._devices.length) {
        var _iterator = _createForOfIteratorHelper(devices),
            _step;

        try {
          var _loop = function _loop() {
            var device = _step.value;

            var d = _this._devices.find(function (d) {
              return d.deviceId === device.deviceId;
            });

            if (!d) {
              return {
                v: false
              };
            }
          };

          for (_iterator.s(); !(_step = _iterator.n()).done;) {
            var _ret = _loop();

            if ((0, _typeof2["default"])(_ret) === "object") return _ret.v;
          }
        } catch (err) {
          _iterator.e(err);
        } finally {
          _iterator.f();
        }

        return true;
      } else {
        return false;
      }
    }
  }, {
    key: "getTimestamp",

    /**
     * @param {MatrixEvent} event the event to get the timestamp of
     * @return {number} the timestamp when the event was sent
     */
    value: function getTimestamp(event) {
      var content = event.getContent();
      return content && content.timestamp;
    }
    /**
     * Changes the state of the channel, request, and verifier in response to a key verification event.
     * @param {MatrixEvent} event to handle
     * @param {VerificationRequest} request the request to forward handling to
     * @param {bool} isLiveEvent whether this is an even received through sync or not
     * @returns {Promise} a promise that resolves when any requests as an anwser to the passed-in event are sent.
     */

  }, {
    key: "handleEvent",
    value: function () {
      var _handleEvent = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee(event, request, isLiveEvent) {
        var _this2 = this;

        var type, content, deviceId, cancelContent, wasStarted, isStarted, isAcceptingEvent, nonChosenDevices, message;
        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                type = event.getType();
                content = event.getContent();

                if (!(type === _VerificationRequest.REQUEST_TYPE || type === _VerificationRequest.READY_TYPE || type === _VerificationRequest.START_TYPE)) {
                  _context.next = 9;
                  break;
                }

                if (!this.transactionId) {
                  this.transactionId = content.transaction_id;
                }

                deviceId = content.from_device; // adopt deviceId if not set before and valid

                if (!this._deviceId && this._devices.includes(deviceId)) {
                  this._deviceId = deviceId;
                } // if no device id or different from addopted one, cancel with sender


                if (!(!this._deviceId || this._deviceId !== deviceId)) {
                  _context.next = 9;
                  break;
                }

                // also check that message came from the device we sent the request to earlier on
                // and do send a cancel message to that device
                // (but don't cancel the request for the device we should be talking to)
                cancelContent = this.completeContent((0, _Error.errorFromEvent)((0, _Error.newUnexpectedMessageError)()));
                return _context.abrupt("return", this._sendToDevices(_VerificationRequest.CANCEL_TYPE, cancelContent, [deviceId]));

              case 9:
                wasStarted = request.phase === _VerificationRequest.PHASE_STARTED || request.phase === _VerificationRequest.PHASE_READY;
                _context.next = 12;
                return request.handleEvent(event.getType(), event, isLiveEvent, false, false);

              case 12:
                isStarted = request.phase === _VerificationRequest.PHASE_STARTED || request.phase === _VerificationRequest.PHASE_READY;
                isAcceptingEvent = type === _VerificationRequest.START_TYPE || type === _VerificationRequest.READY_TYPE; // the request has picked a ready or start event, tell the other devices about it

                if (!(isAcceptingEvent && !wasStarted && isStarted && this._deviceId)) {
                  _context.next = 20;
                  break;
                }

                nonChosenDevices = this._devices.filter(function (d) {
                  return d !== _this2._deviceId;
                });

                if (!nonChosenDevices.length) {
                  _context.next = 20;
                  break;
                }

                message = this.completeContent({
                  code: "m.accepted",
                  reason: "Verification request accepted by another device"
                });
                _context.next = 20;
                return this._sendToDevices(_VerificationRequest.CANCEL_TYPE, message, nonChosenDevices);

              case 20:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function handleEvent(_x, _x2, _x3) {
        return _handleEvent.apply(this, arguments);
      }

      return handleEvent;
    }()
    /**
     * See {InRoomChannel.completedContentFromEvent} why this is needed.
     * @param {MatrixEvent} event the received event
     * @returns {Object} the content object
     */

  }, {
    key: "completedContentFromEvent",
    value: function completedContentFromEvent(event) {
      return event.getContent();
    }
    /**
     * Add all the fields to content needed for sending it over this channel.
     * This is public so verification methods (SAS uses this) can get the exact
     * content that will be sent independent of the used channel,
     * as they need to calculate the hash of it.
     * @param {string} type the event type
     * @param {object} content the (incomplete) content
     * @returns {object} the complete content, as it will be sent.
     */

  }, {
    key: "completeContent",
    value: function completeContent(type, content) {
      // make a copy
      content = Object.assign({}, content);

      if (this.transactionId) {
        content.transaction_id = this.transactionId;
      }

      if (type === _VerificationRequest.REQUEST_TYPE || type === _VerificationRequest.READY_TYPE || type === _VerificationRequest.START_TYPE) {
        content.from_device = this._client.getDeviceId();
      }

      if (type === _VerificationRequest.REQUEST_TYPE) {
        content.timestamp = Date.now();
      }

      return content;
    }
    /**
     * Send an event over the channel with the content not having gone through `completeContent`.
     * @param {string} type the event type
     * @param {object} uncompletedContent the (incomplete) content
     * @returns {Promise} the promise of the request
     */

  }, {
    key: "send",
    value: function send(type) {
      var uncompletedContent = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

      // create transaction id when sending request
      if ((type === _VerificationRequest.REQUEST_TYPE || type === _VerificationRequest.START_TYPE) && !this.transactionId) {
        this.transactionId = ToDeviceChannel.makeTransactionId();
      }

      var content = this.completeContent(type, uncompletedContent);
      return this.sendCompleted(type, content);
    }
    /**
     * Send an event over the channel with the content having gone through `completeContent` already.
     * @param {string} type the event type
     * @param {object} content
     * @returns {Promise} the promise of the request
     */

  }, {
    key: "sendCompleted",
    value: function () {
      var _sendCompleted = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(type, content) {
        var result, remoteEchoEvent;
        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                if (!(type === _VerificationRequest.REQUEST_TYPE)) {
                  _context2.next = 6;
                  break;
                }

                _context2.next = 3;
                return this._sendToDevices(type, content, this._devices);

              case 3:
                result = _context2.sent;
                _context2.next = 9;
                break;

              case 6:
                _context2.next = 8;
                return this._sendToDevices(type, content, [this._deviceId]);

              case 8:
                result = _context2.sent;

              case 9:
                // the VerificationRequest state machine requires remote echos of the event
                // the client sends itself, so we fake this for to_device messages
                remoteEchoEvent = new _event.MatrixEvent({
                  sender: this._client.getUserId(),
                  content: content,
                  type: type
                });
                _context2.next = 12;
                return this._request.handleEvent(type, remoteEchoEvent,
                /*isLiveEvent=*/
                true,
                /*isRemoteEcho=*/
                true,
                /*isSentByUs=*/
                true);

              case 12:
                return _context2.abrupt("return", result);

              case 13:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function sendCompleted(_x4, _x5) {
        return _sendCompleted.apply(this, arguments);
      }

      return sendCompleted;
    }()
  }, {
    key: "_sendToDevices",
    value: function _sendToDevices(type, content, devices) {
      if (devices.length) {
        var msgMap = {};

        var _iterator2 = _createForOfIteratorHelper(devices),
            _step2;

        try {
          for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
            var deviceId = _step2.value;
            msgMap[deviceId] = content;
          }
        } catch (err) {
          _iterator2.e(err);
        } finally {
          _iterator2.f();
        }

        return this._client.sendToDevice(type, (0, _defineProperty2["default"])({}, this.userId, msgMap));
      } else {
        return Promise.resolve();
      }
    }
    /**
     * Allow Crypto module to create and know the transaction id before the .start event gets sent.
     * @returns {string} the transaction id
     */

  }, {
    key: "deviceId",
    get: function get() {
      return this._deviceId;
    }
  }], [{
    key: "getEventType",
    value: function getEventType(event) {
      return event.getType();
    }
    /**
     * Extract the transaction id used by a given key verification event, if any
     * @param {MatrixEvent} event the event
     * @returns {string} the transaction id
     */

  }, {
    key: "getTransactionId",
    value: function getTransactionId(event) {
      var content = event.getContent();
      return content && content.transaction_id;
    }
    /**
     * Checks whether the given event type should be allowed to initiate a new VerificationRequest over this channel
     * @param {string} type the event type to check
     * @returns {bool} boolean flag
     */

  }, {
    key: "canCreateRequest",
    value: function canCreateRequest(type) {
      return type === _VerificationRequest.REQUEST_TYPE || type === _VerificationRequest.START_TYPE;
    }
    /**
     * Checks whether this event is a well-formed key verification event.
     * This only does checks that don't rely on the current state of a potentially already channel
     * so we can prevent channels being created by invalid events.
     * `handleEvent` can do more checks and choose to ignore invalid events.
     * @param {MatrixEvent} event the event to validate
     * @param {MatrixClient} client the client to get the current user and device id from
     * @returns {bool} whether the event is valid and should be passed to handleEvent
     */

  }, {
    key: "validateEvent",
    value: function validateEvent(event, client) {
      if (event.isCancelled()) {
        _logger.logger.warn("Ignoring flagged verification request from " + event.getSender());

        return false;
      }

      var content = event.getContent();

      if (!content) {
        _logger.logger.warn("ToDeviceChannel.validateEvent: invalid: no content");

        return false;
      }

      if (!content.transaction_id) {
        _logger.logger.warn("ToDeviceChannel.validateEvent: invalid: no transaction_id");

        return false;
      }

      var type = event.getType();

      if (type === _VerificationRequest.REQUEST_TYPE) {
        if (!Number.isFinite(content.timestamp)) {
          _logger.logger.warn("ToDeviceChannel.validateEvent: invalid: no timestamp");

          return false;
        }

        if (event.getSender() === client.getUserId() && content.from_device == client.getDeviceId()) {
          // ignore requests from ourselves, because it doesn't make sense for a
          // device to verify itself
          _logger.logger.warn("ToDeviceChannel.validateEvent: invalid: from own device");

          return false;
        }
      }

      return _VerificationRequest.VerificationRequest.validateEvent(type, event, client);
    }
  }, {
    key: "makeTransactionId",
    value: function makeTransactionId() {
      return (0, _randomstring.randomString)(32);
    }
  }]);
  return ToDeviceChannel;
}();

exports.ToDeviceChannel = ToDeviceChannel;

var ToDeviceRequests = /*#__PURE__*/function () {
  function ToDeviceRequests() {
    (0, _classCallCheck2["default"])(this, ToDeviceRequests);
    this._requestsByUserId = new Map();
  }

  (0, _createClass2["default"])(ToDeviceRequests, [{
    key: "getRequest",
    value: function getRequest(event) {
      return this.getRequestBySenderAndTxnId(event.getSender(), ToDeviceChannel.getTransactionId(event));
    }
  }, {
    key: "getRequestByChannel",
    value: function getRequestByChannel(channel) {
      return this.getRequestBySenderAndTxnId(channel.userId, channel.transactionId);
    }
  }, {
    key: "getRequestBySenderAndTxnId",
    value: function getRequestBySenderAndTxnId(sender, txnId) {
      var requestsByTxnId = this._requestsByUserId.get(sender);

      if (requestsByTxnId) {
        return requestsByTxnId.get(txnId);
      }
    }
  }, {
    key: "setRequest",
    value: function setRequest(event, request) {
      this.setRequestBySenderAndTxnId(event.getSender(), ToDeviceChannel.getTransactionId(event), request);
    }
  }, {
    key: "setRequestByChannel",
    value: function setRequestByChannel(channel, request) {
      this.setRequestBySenderAndTxnId(channel.userId, channel.transactionId, request);
    }
  }, {
    key: "setRequestBySenderAndTxnId",
    value: function setRequestBySenderAndTxnId(sender, txnId, request) {
      var requestsByTxnId = this._requestsByUserId.get(sender);

      if (!requestsByTxnId) {
        requestsByTxnId = new Map();

        this._requestsByUserId.set(sender, requestsByTxnId);
      }

      requestsByTxnId.set(txnId, request);
    }
  }, {
    key: "removeRequest",
    value: function removeRequest(event) {
      var userId = event.getSender();

      var requestsByTxnId = this._requestsByUserId.get(userId);

      if (requestsByTxnId) {
        requestsByTxnId["delete"](ToDeviceChannel.getTransactionId(event));

        if (requestsByTxnId.size === 0) {
          this._requestsByUserId["delete"](userId);
        }
      }
    }
  }, {
    key: "findRequestInProgress",
    value: function findRequestInProgress(userId, devices) {
      var requestsByTxnId = this._requestsByUserId.get(userId);

      if (requestsByTxnId) {
        var _iterator3 = _createForOfIteratorHelper(requestsByTxnId.values()),
            _step3;

        try {
          for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
            var request = _step3.value;

            if (request.pending && request.channel.isToDevices(devices)) {
              return request;
            }
          }
        } catch (err) {
          _iterator3.e(err);
        } finally {
          _iterator3.f();
        }
      }
    }
  }, {
    key: "getRequestsInProgress",
    value: function getRequestsInProgress(userId) {
      var requestsByTxnId = this._requestsByUserId.get(userId);

      if (requestsByTxnId) {
        return Array.from(requestsByTxnId.values()).filter(function (r) {
          return r.pending;
        });
      }

      return [];
    }
  }]);
  return ToDeviceRequests;
}();

exports.ToDeviceRequests = ToDeviceRequests;