"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.VerificationRequest = exports.PHASE_DONE = exports.PHASE_CANCELLED = exports.PHASE_STARTED = exports.PHASE_READY = exports.PHASE_REQUESTED = exports.PHASE_UNSENT = exports.READY_TYPE = exports.DONE_TYPE = exports.CANCEL_TYPE = exports.START_TYPE = exports.REQUEST_TYPE = exports.EVENT_PREFIX = void 0;

var _slicedToArray2 = _interopRequireDefault(require("@babel/runtime/helpers/slicedToArray"));

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _toConsumableArray2 = _interopRequireDefault(require("@babel/runtime/helpers/toConsumableArray"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _assertThisInitialized2 = _interopRequireDefault(require("@babel/runtime/helpers/assertThisInitialized"));

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _logger = require("../../../logger");

var _events = require("events");

var _Error = require("../Error");

var _QRCode = require("../QRCode");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

// How long after the event's timestamp that the request times out
var TIMEOUT_FROM_EVENT_TS = 10 * 60 * 1000; // 10 minutes
// How long after we receive the event that the request times out

var TIMEOUT_FROM_EVENT_RECEIPT = 2 * 60 * 1000; // 2 minutes
// to avoid almost expired verification notifications
// from showing a notification and almost immediately
// disappearing, also ignore verification requests that
// are this amount of time away from expiring.

var VERIFICATION_REQUEST_MARGIN = 3 * 1000; // 3 seconds

var EVENT_PREFIX = "m.key.verification.";
exports.EVENT_PREFIX = EVENT_PREFIX;
var REQUEST_TYPE = EVENT_PREFIX + "request";
exports.REQUEST_TYPE = REQUEST_TYPE;
var START_TYPE = EVENT_PREFIX + "start";
exports.START_TYPE = START_TYPE;
var CANCEL_TYPE = EVENT_PREFIX + "cancel";
exports.CANCEL_TYPE = CANCEL_TYPE;
var DONE_TYPE = EVENT_PREFIX + "done";
exports.DONE_TYPE = DONE_TYPE;
var READY_TYPE = EVENT_PREFIX + "ready";
exports.READY_TYPE = READY_TYPE;
var PHASE_UNSENT = 1;
exports.PHASE_UNSENT = PHASE_UNSENT;
var PHASE_REQUESTED = 2;
exports.PHASE_REQUESTED = PHASE_REQUESTED;
var PHASE_READY = 3;
exports.PHASE_READY = PHASE_READY;
var PHASE_STARTED = 4;
exports.PHASE_STARTED = PHASE_STARTED;
var PHASE_CANCELLED = 5;
exports.PHASE_CANCELLED = PHASE_CANCELLED;
var PHASE_DONE = 6;
/**
 * State machine for verification requests.
 * Things that differ based on what channel is used to
 * send and receive verification events are put in `InRoomChannel` or `ToDeviceChannel`.
 * @event "change" whenever the state of the request object has changed.
 */

exports.PHASE_DONE = PHASE_DONE;

var VerificationRequest = /*#__PURE__*/function (_EventEmitter) {
  (0, _inherits2["default"])(VerificationRequest, _EventEmitter);

  var _super = _createSuper(VerificationRequest);

  function VerificationRequest(channel, verificationMethods, client) {
    var _this;

    (0, _classCallCheck2["default"])(this, VerificationRequest);
    _this = _super.call(this);
    (0, _defineProperty2["default"])((0, _assertThisInitialized2["default"])(_this), "_cancelOnTimeout", function () {
      try {
        if (_this.initiatedByMe) {
          _this.cancel({
            reason: "Other party didn't accept in time",
            code: "m.timeout"
          });
        } else {
          _this.cancel({
            reason: "User didn't accept in time",
            code: "m.timeout"
          });
        }
      } catch (err) {
        _logger.logger.error("Error while cancelling verification request", err);
      }
    });
    _this.channel = channel;
    _this.channel._request = (0, _assertThisInitialized2["default"])(_this);
    _this._verificationMethods = verificationMethods;
    _this._client = client;
    _this._commonMethods = [];

    _this._setPhase(PHASE_UNSENT, false);

    _this._eventsByUs = new Map();
    _this._eventsByThem = new Map();
    _this._observeOnly = false;
    _this._timeoutTimer = null;
    _this._accepting = false;
    _this._declining = false;
    _this._verifierHasFinished = false;
    _this._cancelled = false;
    _this._chosenMethod = null; // we keep a copy of the QR Code data (including other user master key) around
    // for QR reciprocate verification, to protect against
    // cross-signing identity reset between the .ready and .start event
    // and signing the wrong key after .start

    _this._qrCodeData = null; // The timestamp when we received the request event from the other side

    _this._requestReceivedAt = null;
    return _this;
  }
  /**
   * Stateless validation logic not specific to the channel.
   * Invoked by the same static method in either channel.
   * @param {string} type the "symbolic" event type, as returned by the `getEventType` function on the channel.
   * @param {MatrixEvent} event the event to validate. Don't call getType() on it but use the `type` parameter instead.
   * @param {MatrixClient} client the client to get the current user and device id from
   * @returns {bool} whether the event is valid and should be passed to handleEvent
   */


  (0, _createClass2["default"])(VerificationRequest, [{
    key: "calculateEventTimeout",
    value: function calculateEventTimeout(event) {
      var effectiveExpiresAt = this.channel.getTimestamp(event) + TIMEOUT_FROM_EVENT_TS;

      if (this._requestReceivedAt && !this.initiatedByMe && this.phase <= PHASE_REQUESTED) {
        var expiresAtByReceipt = this._requestReceivedAt + TIMEOUT_FROM_EVENT_RECEIPT;
        effectiveExpiresAt = Math.min(effectiveExpiresAt, expiresAtByReceipt);
      }

      return Math.max(0, effectiveExpiresAt - Date.now());
    }
    /** The current remaining amount of ms before the request should be automatically cancelled */

  }, {
    key: "otherPartySupportsMethod",

    /** Checks whether the other party supports a given verification method.
     *  This is useful when setting up the QR code UI, as it is somewhat asymmetrical:
     *  if the other party supports SCAN_QR, we should show a QR code in the UI, and vice versa.
     *  For methods that need to be supported by both ends, use the `methods` property.
     *  @param {string} method the method to check
     *  @param {boolean} force to check even if the phase is not ready or started yet, internal usage
     *  @return {bool} whether or not the other party said the supported the method */
    value: function otherPartySupportsMethod(method) {
      var force = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;

      if (!force && !this.ready && !this.started) {
        return false;
      }

      var theirMethodEvent = this._eventsByThem.get(REQUEST_TYPE) || this._eventsByThem.get(READY_TYPE);

      if (!theirMethodEvent) {
        // if we started straight away with .start event,
        // we are assuming that the other side will support the
        // chosen method, so return true for that.
        if (this.started && this.initiatedByMe) {
          var myStartEvent = this._eventsByUs.get(START_TYPE);

          var _content = myStartEvent && myStartEvent.getContent();

          var myStartMethod = _content && _content.method;
          return method == myStartMethod;
        }

        return false;
      }

      var content = theirMethodEvent.getContent();

      if (!content) {
        return false;
      }

      var methods = content.methods;

      if (!Array.isArray(methods)) {
        return false;
      }

      return methods.includes(method);
    }
    /** Whether this request was initiated by the syncing user.
     * For InRoomChannel, this is who sent the .request event.
     * For ToDeviceChannel, this is who sent the .start event
     */

  }, {
    key: "beginKeyVerification",

    /* Start the key verification, creating a verifier and sending a .start event.
     * If no previous events have been sent, pass in `targetDevice` to set who to direct this request to.
     * @param {string} method the name of the verification method to use.
     * @param {string?} targetDevice.userId the id of the user to direct this request to
     * @param {string?} targetDevice.deviceId the id of the device to direct this request to
     * @returns {VerifierBase} the verifier of the given method
     */
    value: function beginKeyVerification(method) {
      var targetDevice = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

      // need to allow also when unsent in case of to_device
      if (!this.observeOnly && !this._verifier) {
        var validStartPhase = this.phase === PHASE_REQUESTED || this.phase === PHASE_READY || this.phase === PHASE_UNSENT && this.channel.constructor.canCreateRequest(START_TYPE);

        if (validStartPhase) {
          // when called on a request that was initiated with .request event
          // check the method is supported by both sides
          if (this._commonMethods.length && !this._commonMethods.includes(method)) {
            throw (0, _Error.newUnknownMethodError)();
          }

          this._verifier = this._createVerifier(method, null, targetDevice);

          if (!this._verifier) {
            throw (0, _Error.newUnknownMethodError)();
          }

          this._chosenMethod = method;
        }
      }

      return this._verifier;
    }
    /**
     * sends the initial .request event.
     * @returns {Promise} resolves when the event has been sent.
     */

  }, {
    key: "sendRequest",
    value: function () {
      var _sendRequest = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee() {
        var methods;
        return _regenerator["default"].wrap(function _callee$(_context) {
          while (1) {
            switch (_context.prev = _context.next) {
              case 0:
                if (!(!this.observeOnly && this._phase === PHASE_UNSENT)) {
                  _context.next = 4;
                  break;
                }

                methods = (0, _toConsumableArray2["default"])(this._verificationMethods.keys());
                _context.next = 4;
                return this.channel.send(REQUEST_TYPE, {
                  methods: methods
                });

              case 4:
              case "end":
                return _context.stop();
            }
          }
        }, _callee, this);
      }));

      function sendRequest() {
        return _sendRequest.apply(this, arguments);
      }

      return sendRequest;
    }()
    /**
     * Cancels the request, sending a cancellation to the other party
     * @param {string?} error.reason the error reason to send the cancellation with
     * @param {string?} error.code the error code to send the cancellation with
     * @returns {Promise} resolves when the event has been sent.
     */

  }, {
    key: "cancel",
    value: function () {
      var _cancel = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2() {
        var _ref,
            _ref$reason,
            reason,
            _ref$code,
            code,
            _args2 = arguments;

        return _regenerator["default"].wrap(function _callee2$(_context2) {
          while (1) {
            switch (_context2.prev = _context2.next) {
              case 0:
                _ref = _args2.length > 0 && _args2[0] !== undefined ? _args2[0] : {}, _ref$reason = _ref.reason, reason = _ref$reason === void 0 ? "User declined" : _ref$reason, _ref$code = _ref.code, code = _ref$code === void 0 ? "m.user" : _ref$code;

                if (!(!this.observeOnly && this._phase !== PHASE_CANCELLED)) {
                  _context2.next = 11;
                  break;
                }

                this._declining = true;
                this.emit("change");

                if (!this._verifier) {
                  _context2.next = 8;
                  break;
                }

                return _context2.abrupt("return", this._verifier.cancel((0, _Error.errorFactory)(code, reason)()));

              case 8:
                this._cancellingUserId = this._client.getUserId();
                _context2.next = 11;
                return this.channel.send(CANCEL_TYPE, {
                  code: code,
                  reason: reason
                });

              case 11:
              case "end":
                return _context2.stop();
            }
          }
        }, _callee2, this);
      }));

      function cancel() {
        return _cancel.apply(this, arguments);
      }

      return cancel;
    }()
    /**
     * Accepts the request, sending a .ready event to the other party
     * @returns {Promise} resolves when the event has been sent.
     */

  }, {
    key: "accept",
    value: function () {
      var _accept = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee3() {
        var methods;
        return _regenerator["default"].wrap(function _callee3$(_context3) {
          while (1) {
            switch (_context3.prev = _context3.next) {
              case 0:
                if (!(!this.observeOnly && this.phase === PHASE_REQUESTED && !this.initiatedByMe)) {
                  _context3.next = 6;
                  break;
                }

                methods = (0, _toConsumableArray2["default"])(this._verificationMethods.keys());
                this._accepting = true;
                this.emit("change");
                _context3.next = 6;
                return this.channel.send(READY_TYPE, {
                  methods: methods
                });

              case 6:
              case "end":
                return _context3.stop();
            }
          }
        }, _callee3, this);
      }));

      function accept() {
        return _accept.apply(this, arguments);
      }

      return accept;
    }()
    /**
     * Can be used to listen for state changes until the callback returns true.
     * @param {Function} fn callback to evaluate whether the request is in the desired state.
     *                      Takes the request as an argument.
     * @returns {Promise} that resolves once the callback returns true
     * @throws {Error} when the request is cancelled
     */

  }, {
    key: "waitFor",
    value: function waitFor(fn) {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        var check = function check() {
          var handled = false;

          if (fn(_this2)) {
            resolve(_this2);
            handled = true;
          } else if (_this2.cancelled) {
            reject(new Error("cancelled"));
            handled = true;
          }

          if (handled) {
            _this2.off("change", check);
          }

          return handled;
        };

        if (!check()) {
          _this2.on("change", check);
        }
      });
    }
  }, {
    key: "_setPhase",
    value: function _setPhase(phase) {
      var notify = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
      this._phase = phase;

      if (notify) {
        this.emit("change");
      }
    }
  }, {
    key: "_getEventByEither",
    value: function _getEventByEither(type) {
      return this._eventsByThem.get(type) || this._eventsByUs.get(type);
    }
  }, {
    key: "_getEventBy",
    value: function _getEventBy(type, byThem) {
      if (byThem) {
        return this._eventsByThem.get(type);
      } else {
        return this._eventsByUs.get(type);
      }
    }
  }, {
    key: "_calculatePhaseTransitions",
    value: function _calculatePhaseTransitions() {
      var transitions = [{
        phase: PHASE_UNSENT
      }];

      var phase = function phase() {
        return transitions[transitions.length - 1].phase;
      }; // always pass by .request first to be sure channel.userId has been set


      var hasRequestByThem = this._eventsByThem.has(REQUEST_TYPE);

      var requestEvent = this._getEventBy(REQUEST_TYPE, hasRequestByThem);

      if (requestEvent) {
        transitions.push({
          phase: PHASE_REQUESTED,
          event: requestEvent
        });
      }

      var readyEvent = requestEvent && this._getEventBy(READY_TYPE, !hasRequestByThem);

      if (readyEvent && phase() === PHASE_REQUESTED) {
        transitions.push({
          phase: PHASE_READY,
          event: readyEvent
        });
      }

      var startEvent;

      if (readyEvent || !requestEvent) {
        var theirStartEvent = this._eventsByThem.get(START_TYPE);

        var ourStartEvent = this._eventsByUs.get(START_TYPE); // any party can send .start after a .ready or unsent


        if (theirStartEvent && ourStartEvent) {
          startEvent = theirStartEvent.getSender() < ourStartEvent.getSender() ? theirStartEvent : ourStartEvent;
        } else {
          startEvent = theirStartEvent ? theirStartEvent : ourStartEvent;
        }
      } else {
        startEvent = this._getEventBy(START_TYPE, !hasRequestByThem);
      }

      if (startEvent) {
        var fromRequestPhase = phase() === PHASE_REQUESTED && requestEvent.getSender() !== startEvent.getSender();
        var fromUnsentPhase = phase() === PHASE_UNSENT && this.channel.constructor.canCreateRequest(START_TYPE);

        if (fromRequestPhase || phase() === PHASE_READY || fromUnsentPhase) {
          transitions.push({
            phase: PHASE_STARTED,
            event: startEvent
          });
        }
      }

      var ourDoneEvent = this._eventsByUs.get(DONE_TYPE);

      if (this._verifierHasFinished || ourDoneEvent && phase() === PHASE_STARTED) {
        transitions.push({
          phase: PHASE_DONE
        });
      }

      var cancelEvent = this._getEventByEither(CANCEL_TYPE);

      if ((this._cancelled || cancelEvent) && phase() !== PHASE_DONE) {
        transitions.push({
          phase: PHASE_CANCELLED,
          event: cancelEvent
        });
        return transitions;
      }

      return transitions;
    }
  }, {
    key: "_transitionToPhase",
    value: function _transitionToPhase(transition) {
      var _this3 = this;

      var phase = transition.phase,
          event = transition.event; // get common methods

      if (phase === PHASE_REQUESTED || phase === PHASE_READY) {
        if (!this._wasSentByOwnDevice(event)) {
          var content = event.getContent();
          this._commonMethods = content.methods.filter(function (m) {
            return _this3._verificationMethods.has(m);
          });
        }
      } // detect if we're not a party in the request, and we should just observe


      if (!this.observeOnly) {
        // if requested or accepted by one of my other devices
        if (phase === PHASE_REQUESTED || phase === PHASE_STARTED || phase === PHASE_READY) {
          if (this.channel.receiveStartFromOtherDevices && this._wasSentByOwnUser(event) && !this._wasSentByOwnDevice(event)) {
            this._observeOnly = true;
          }
        }
      } // create verifier


      if (phase === PHASE_STARTED) {
        var _event$getContent = event.getContent(),
            method = _event$getContent.method;

        if (!this._verifier && !this.observeOnly) {
          this._verifier = this._createVerifier(method, event);

          if (!this._verifier) {
            this.cancel({
              code: "m.unknown_method",
              reason: "Unknown method: ".concat(method)
            });
          } else {
            this._chosenMethod = method;
          }
        }
      }
    }
  }, {
    key: "_applyPhaseTransitions",
    value: function _applyPhaseTransitions() {
      var _this4 = this;

      var transitions = this._calculatePhaseTransitions();

      var existingIdx = transitions.findIndex(function (t) {
        return t.phase === _this4.phase;
      }); // trim off phases we already went through, if any

      var newTransitions = transitions.slice(existingIdx + 1); // transition to all new phases

      var _iterator = _createForOfIteratorHelper(newTransitions),
          _step;

      try {
        for (_iterator.s(); !(_step = _iterator.n()).done;) {
          var transition = _step.value;

          this._transitionToPhase(transition);
        }
      } catch (err) {
        _iterator.e(err);
      } finally {
        _iterator.f();
      }

      return newTransitions;
    }
  }, {
    key: "_isWinningStartRace",
    value: function _isWinningStartRace(newEvent) {
      if (newEvent.getType() !== START_TYPE) {
        return false;
      }

      var oldEvent = this._verifier.startEvent;
      var oldRaceIdentifier;

      if (this.isSelfVerification) {
        // if the verifier does not have a startEvent,
        // it is because it's still sending and we are on the initator side
        // we know we are sending a .start event because we already
        // have a verifier (checked in calling method)
        if (oldEvent) {
          var oldContent = oldEvent.getContent();
          oldRaceIdentifier = oldContent && oldContent.from_device;
        } else {
          oldRaceIdentifier = this._client.getDeviceId();
        }
      } else {
        if (oldEvent) {
          oldRaceIdentifier = oldEvent.getSender();
        } else {
          oldRaceIdentifier = this._client.getUserId();
        }
      }

      var newRaceIdentifier;

      if (this.isSelfVerification) {
        var newContent = newEvent.getContent();
        newRaceIdentifier = newContent && newContent.from_device;
      } else {
        newRaceIdentifier = newEvent.getSender();
      }

      return newRaceIdentifier < oldRaceIdentifier;
    }
  }, {
    key: "hasEventId",
    value: function hasEventId(eventId) {
      var _iterator2 = _createForOfIteratorHelper(this._eventsByUs.values()),
          _step2;

      try {
        for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
          var event = _step2.value;

          if (event.getId() === eventId) {
            return true;
          }
        }
      } catch (err) {
        _iterator2.e(err);
      } finally {
        _iterator2.f();
      }

      var _iterator3 = _createForOfIteratorHelper(this._eventsByThem.values()),
          _step3;

      try {
        for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
          var _event = _step3.value;

          if (_event.getId() === eventId) {
            return true;
          }
        }
      } catch (err) {
        _iterator3.e(err);
      } finally {
        _iterator3.f();
      }

      return false;
    }
    /**
     * Changes the state of the request and verifier in response to a key verification event.
     * @param {string} type the "symbolic" event type, as returned by the `getEventType` function on the channel.
     * @param {MatrixEvent} event the event to handle. Don't call getType() on it but use the `type` parameter instead.
     * @param {bool} isLiveEvent whether this is an even received through sync or not
     * @param {bool} isRemoteEcho whether this is the remote echo of an event sent by the same device
     * @param {bool} isSentByUs whether this event is sent by a party that can accept and/or observe the request like one of our peers.
     *   For InRoomChannel this means any device for the syncing user. For ToDeviceChannel, just the syncing device.
     * @returns {Promise} a promise that resolves when any requests as an anwser to the passed-in event are sent.
     */

  }, {
    key: "handleEvent",
    value: function () {
      var _handleEvent = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee4(type, event, isLiveEvent, isRemoteEcho, isSentByUs) {
        var wasObserveOnly, isDuplicateEvent, oldPhase, newTransitions, newEventWinsRace, shouldGenerateQrCode, lastTransition, phase;
        return _regenerator["default"].wrap(function _callee4$(_context4) {
          while (1) {
            switch (_context4.prev = _context4.next) {
              case 0:
                if (!(this.done || this.cancelled)) {
                  _context4.next = 2;
                  break;
                }

                return _context4.abrupt("return");

              case 2:
                wasObserveOnly = this._observeOnly;

                this._adjustObserveOnly(event, isLiveEvent);

                if (!(!this.observeOnly && !isRemoteEcho)) {
                  _context4.next = 9;
                  break;
                }

                _context4.next = 7;
                return this._cancelOnError(type, event);

              case 7:
                if (!_context4.sent) {
                  _context4.next = 9;
                  break;
                }

                return _context4.abrupt("return");

              case 9:
                // This assumes verification won't need to send an event with
                // the same type for the same party twice.
                // This is true for QR and SAS verification, and was
                // added here to prevent verification getting cancelled
                // when the server duplicates an event (https://github.com/matrix-org/synapse/issues/3365)
                isDuplicateEvent = isSentByUs ? this._eventsByUs.has(type) : this._eventsByThem.has(type);

                if (!isDuplicateEvent) {
                  _context4.next = 12;
                  break;
                }

                return _context4.abrupt("return");

              case 12:
                oldPhase = this.phase;

                this._addEvent(type, event, isSentByUs); // this will create if needed the verifier so needs to happen before calling it


                newTransitions = this._applyPhaseTransitions();
                _context4.prev = 15;

                // only pass events from the other side to the verifier,
                // no remote echos of our own events
                if (this._verifier && !this.observeOnly) {
                  newEventWinsRace = this._isWinningStartRace(event);

                  if (this._verifier.canSwitchStartEvent(event) && newEventWinsRace) {
                    this._verifier.switchStartEvent(event);
                  } else if (!isRemoteEcho) {
                    if (type === CANCEL_TYPE || this._verifier.events && this._verifier.events.includes(type)) {
                      this._verifier.handleEvent(event);
                    }
                  }
                }

                if (!newTransitions.length) {
                  _context4.next = 30;
                  break;
                }

                if (!(isLiveEvent && newTransitions.some(function (t) {
                  return t.phase === PHASE_READY;
                }))) {
                  _context4.next = 24;
                  break;
                }

                shouldGenerateQrCode = this.otherPartySupportsMethod(_QRCode.SCAN_QR_CODE_METHOD, true);

                if (!shouldGenerateQrCode) {
                  _context4.next = 24;
                  break;
                }

                _context4.next = 23;
                return _QRCode.QRCodeData.create(this, this._client);

              case 23:
                this._qrCodeData = _context4.sent;

              case 24:
                lastTransition = newTransitions[newTransitions.length - 1];
                phase = lastTransition.phase;

                this._setupTimeout(phase); // set phase as last thing as this emits the "change" event


                this._setPhase(phase);

                _context4.next = 31;
                break;

              case 30:
                if (this._observeOnly !== wasObserveOnly) {
                  this.emit("change");
                }

              case 31:
                _context4.prev = 31;

                // log events we processed so we can see from rageshakes what events were added to a request
                _logger.logger.log("Verification request ".concat(this.channel.transactionId, ": ") + "".concat(type, " event with id:").concat(event.getId(), ", ") + "content:".concat(JSON.stringify(event.getContent()), " ") + "deviceId:".concat(this.channel.deviceId, ", ") + "sender:".concat(event.getSender(), ", isSentByUs:").concat(isSentByUs, ", ") + "isLiveEvent:".concat(isLiveEvent, ", isRemoteEcho:").concat(isRemoteEcho, ", ") + "phase:".concat(oldPhase, "=>").concat(this.phase, ", ") + "observeOnly:".concat(wasObserveOnly, "=>").concat(this._observeOnly));

                return _context4.finish(31);

              case 34:
              case "end":
                return _context4.stop();
            }
          }
        }, _callee4, this, [[15,, 31, 34]]);
      }));

      function handleEvent(_x, _x2, _x3, _x4, _x5) {
        return _handleEvent.apply(this, arguments);
      }

      return handleEvent;
    }()
  }, {
    key: "_setupTimeout",
    value: function _setupTimeout(phase) {
      var shouldTimeout = !this._timeoutTimer && !this.observeOnly && phase === PHASE_REQUESTED;

      if (shouldTimeout) {
        this._timeoutTimer = setTimeout(this._cancelOnTimeout, this.timeout);
      }

      if (this._timeoutTimer) {
        var shouldClear = phase === PHASE_STARTED || phase === PHASE_READY || phase === PHASE_DONE || phase === PHASE_CANCELLED;

        if (shouldClear) {
          clearTimeout(this._timeoutTimer);
          this._timeoutTimer = null;
        }
      }
    }
  }, {
    key: "_cancelOnError",
    value: function () {
      var _cancelOnError2 = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee5(type, event) {
        var method, isUnexpectedRequest, isUnexpectedReady, reason;
        return _regenerator["default"].wrap(function _callee5$(_context5) {
          while (1) {
            switch (_context5.prev = _context5.next) {
              case 0:
                if (!(type === START_TYPE)) {
                  _context5.next = 6;
                  break;
                }

                method = event.getContent().method;

                if (this._verificationMethods.has(method)) {
                  _context5.next = 6;
                  break;
                }

                _context5.next = 5;
                return this.cancel((0, _Error.errorFromEvent)((0, _Error.newUnknownMethodError)()));

              case 5:
                return _context5.abrupt("return", true);

              case 6:
                isUnexpectedRequest = type === REQUEST_TYPE && this.phase !== PHASE_UNSENT;
                isUnexpectedReady = type === READY_TYPE && this.phase !== PHASE_REQUESTED; // only if phase has passed from PHASE_UNSENT should we cancel, because events
                // are allowed to come in in any order (at least with InRoomChannel). So we only know
                // we're dealing with a valid request we should participate in once we've moved to PHASE_REQUESTED
                // before that, we could be looking at somebody elses verification request and we just
                // happen to be in the room

                if (!(this.phase !== PHASE_UNSENT && (isUnexpectedRequest || isUnexpectedReady))) {
                  _context5.next = 14;
                  break;
                }

                _logger.logger.warn("Cancelling, unexpected ".concat(type, " verification ") + "event from ".concat(event.getSender()));

                reason = "Unexpected ".concat(type, " event in phase ").concat(this.phase);
                _context5.next = 13;
                return this.cancel((0, _Error.errorFromEvent)((0, _Error.newUnexpectedMessageError)({
                  reason: reason
                })));

              case 13:
                return _context5.abrupt("return", true);

              case 14:
                return _context5.abrupt("return", false);

              case 15:
              case "end":
                return _context5.stop();
            }
          }
        }, _callee5, this);
      }));

      function _cancelOnError(_x6, _x7) {
        return _cancelOnError2.apply(this, arguments);
      }

      return _cancelOnError;
    }()
  }, {
    key: "_adjustObserveOnly",
    value: function _adjustObserveOnly(event, isLiveEvent) {
      // don't send out events for historical requests
      if (!isLiveEvent) {
        this._observeOnly = true;
      }

      if (this.calculateEventTimeout(event) < VERIFICATION_REQUEST_MARGIN) {
        this._observeOnly = true;
      }
    }
  }, {
    key: "_addEvent",
    value: function _addEvent(type, event, isSentByUs) {
      if (isSentByUs) {
        this._eventsByUs.set(type, event);
      } else {
        this._eventsByThem.set(type, event);
      } // once we know the userId of the other party (from the .request event)
      // see if any event by anyone else crept into this._eventsByThem


      if (type === REQUEST_TYPE) {
        var _iterator4 = _createForOfIteratorHelper(this._eventsByThem.entries()),
            _step4;

        try {
          for (_iterator4.s(); !(_step4 = _iterator4.n()).done;) {
            var _step4$value = (0, _slicedToArray2["default"])(_step4.value, 2),
                _type = _step4$value[0],
                _event2 = _step4$value[1];

            if (_event2.getSender() !== this.otherUserId) {
              this._eventsByThem["delete"](_type);
            }
          } // also remember when we received the request event

        } catch (err) {
          _iterator4.e(err);
        } finally {
          _iterator4.f();
        }

        this._requestReceivedAt = Date.now();
      }
    }
  }, {
    key: "_createVerifier",
    value: function _createVerifier(method) {
      var startEvent = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;
      var targetDevice = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : null;

      if (!targetDevice) {
        targetDevice = this.targetDevice;
      }

      var _targetDevice = targetDevice,
          userId = _targetDevice.userId,
          deviceId = _targetDevice.deviceId;

      var VerifierCtor = this._verificationMethods.get(method);

      if (!VerifierCtor) {
        _logger.logger.warn("could not find verifier constructor for method", method);

        return;
      }

      return new VerifierCtor(this.channel, this._client, userId, deviceId, startEvent, this);
    }
  }, {
    key: "_wasSentByOwnUser",
    value: function _wasSentByOwnUser(event) {
      return event.getSender() === this._client.getUserId();
    } // only for .request, .ready or .start

  }, {
    key: "_wasSentByOwnDevice",
    value: function _wasSentByOwnDevice(event) {
      if (!this._wasSentByOwnUser(event)) {
        return false;
      }

      var content = event.getContent();

      if (!content || content.from_device !== this._client.getDeviceId()) {
        return false;
      }

      return true;
    }
  }, {
    key: "onVerifierCancelled",
    value: function onVerifierCancelled() {
      this._cancelled = true; // move to cancelled phase

      var newTransitions = this._applyPhaseTransitions();

      if (newTransitions.length) {
        this._setPhase(newTransitions[newTransitions.length - 1].phase);
      }
    }
  }, {
    key: "onVerifierFinished",
    value: function onVerifierFinished() {
      this.channel.send("m.key.verification.done", {});
      this._verifierHasFinished = true; // move to .done phase

      var newTransitions = this._applyPhaseTransitions();

      if (newTransitions.length) {
        this._setPhase(newTransitions[newTransitions.length - 1].phase);
      }
    }
  }, {
    key: "getEventFromOtherParty",
    value: function getEventFromOtherParty(type) {
      return this._eventsByThem.get(type);
    }
  }, {
    key: "invalid",
    get: function get() {
      return this.phase === PHASE_UNSENT;
    }
    /** returns whether the phase is PHASE_REQUESTED */

  }, {
    key: "requested",
    get: function get() {
      return this.phase === PHASE_REQUESTED;
    }
    /** returns whether the phase is PHASE_CANCELLED */

  }, {
    key: "cancelled",
    get: function get() {
      return this.phase === PHASE_CANCELLED;
    }
    /** returns whether the phase is PHASE_READY */

  }, {
    key: "ready",
    get: function get() {
      return this.phase === PHASE_READY;
    }
    /** returns whether the phase is PHASE_STARTED */

  }, {
    key: "started",
    get: function get() {
      return this.phase === PHASE_STARTED;
    }
    /** returns whether the phase is PHASE_DONE */

  }, {
    key: "done",
    get: function get() {
      return this.phase === PHASE_DONE;
    }
    /** once the phase is PHASE_STARTED (and !initiatedByMe) or PHASE_READY: common methods supported by both sides */

  }, {
    key: "methods",
    get: function get() {
      return this._commonMethods;
    }
    /** the method picked in the .start event */

  }, {
    key: "chosenMethod",
    get: function get() {
      return this._chosenMethod;
    }
  }, {
    key: "timeout",
    get: function get() {
      var requestEvent = this._getEventByEither(REQUEST_TYPE);

      if (requestEvent) {
        return this.calculateEventTimeout(requestEvent);
      }

      return 0;
    }
    /**
     * The key verification request event.
     * @returns {MatrixEvent} The request event, or falsey if not found.
     */

  }, {
    key: "requestEvent",
    get: function get() {
      return this._getEventByEither(REQUEST_TYPE);
    }
    /** current phase of the request. Some properties might only be defined in a current phase. */

  }, {
    key: "phase",
    get: function get() {
      return this._phase;
    }
    /** The verifier to do the actual verification, once the method has been established. Only defined when the `phase` is PHASE_STARTED. */

  }, {
    key: "verifier",
    get: function get() {
      return this._verifier;
    }
  }, {
    key: "canAccept",
    get: function get() {
      return this.phase < PHASE_READY && !this._accepting && !this._declining;
    }
  }, {
    key: "accepting",
    get: function get() {
      return this._accepting;
    }
  }, {
    key: "declining",
    get: function get() {
      return this._declining;
    }
    /** whether this request has sent it's initial event and needs more events to complete */

  }, {
    key: "pending",
    get: function get() {
      return !this.observeOnly && this._phase !== PHASE_DONE && this._phase !== PHASE_CANCELLED;
    }
    /** Only set after a .ready if the other party can scan a QR code */

  }, {
    key: "qrCodeData",
    get: function get() {
      return this._qrCodeData;
    }
  }, {
    key: "initiatedByMe",
    get: function get() {
      // event created by us but no remote echo has been received yet
      var noEventsYet = this._eventsByUs.size + this._eventsByThem.size === 0;

      if (this._phase === PHASE_UNSENT && noEventsYet) {
        return true;
      }

      var hasMyRequest = this._eventsByUs.has(REQUEST_TYPE);

      var hasTheirRequest = this._eventsByThem.has(REQUEST_TYPE);

      if (hasMyRequest && !hasTheirRequest) {
        return true;
      }

      if (!hasMyRequest && hasTheirRequest) {
        return false;
      }

      var hasMyStart = this._eventsByUs.has(START_TYPE);

      var hasTheirStart = this._eventsByThem.has(START_TYPE);

      if (hasMyStart && !hasTheirStart) {
        return true;
      }

      return false;
    }
    /** The id of the user that initiated the request */

  }, {
    key: "requestingUserId",
    get: function get() {
      if (this.initiatedByMe) {
        return this._client.getUserId();
      } else {
        return this.otherUserId;
      }
    }
    /** The id of the user that (will) receive(d) the request */

  }, {
    key: "receivingUserId",
    get: function get() {
      if (this.initiatedByMe) {
        return this.otherUserId;
      } else {
        return this._client.getUserId();
      }
    }
    /** The user id of the other party in this request */

  }, {
    key: "otherUserId",
    get: function get() {
      return this.channel.userId;
    }
  }, {
    key: "isSelfVerification",
    get: function get() {
      return this._client.getUserId() === this.otherUserId;
    }
    /**
     * The id of the user that cancelled the request,
     * only defined when phase is PHASE_CANCELLED
     */

  }, {
    key: "cancellingUserId",
    get: function get() {
      var myCancel = this._eventsByUs.get(CANCEL_TYPE);

      var theirCancel = this._eventsByThem.get(CANCEL_TYPE);

      if (myCancel && (!theirCancel || myCancel.getId() < theirCancel.getId())) {
        return myCancel.getSender();
      }

      if (theirCancel) {
        return theirCancel.getSender();
      }

      return undefined;
    }
    /**
     * The cancellation code e.g m.user which is responsible for cancelling this verification
     */

  }, {
    key: "cancellationCode",
    get: function get() {
      var ev = this._getEventByEither(CANCEL_TYPE);

      return ev ? ev.getContent().code : null;
    }
  }, {
    key: "observeOnly",
    get: function get() {
      return this._observeOnly;
    }
    /**
     * Gets which device the verification should be started with
     * given the events sent so far in the verification. This is the
     * same algorithm used to determine which device to send the
     * verification to when no specific device is specified.
     * @returns {{userId: *, deviceId: *}} The device information
     */

  }, {
    key: "targetDevice",
    get: function get() {
      var theirFirstEvent = this._eventsByThem.get(REQUEST_TYPE) || this._eventsByThem.get(READY_TYPE) || this._eventsByThem.get(START_TYPE);

      var theirFirstContent = theirFirstEvent.getContent();
      var fromDevice = theirFirstContent.from_device;
      return {
        userId: this.otherUserId,
        deviceId: fromDevice
      };
    }
  }], [{
    key: "validateEvent",
    value: function validateEvent(type, event, client) {
      var content = event.getContent();

      if (!type || !type.startsWith(EVENT_PREFIX)) {
        return false;
      } // from here on we're fairly sure that this is supposed to be
      // part of a verification request, so be noisy when rejecting something


      if (!content) {
        _logger.logger.log("VerificationRequest: validateEvent: no content");

        return false;
      }

      if (type === REQUEST_TYPE || type === READY_TYPE) {
        if (!Array.isArray(content.methods)) {
          _logger.logger.log("VerificationRequest: validateEvent: " + "fail because methods");

          return false;
        }
      }

      if (type === REQUEST_TYPE || type === READY_TYPE || type === START_TYPE) {
        if (typeof content.from_device !== "string" || content.from_device.length === 0) {
          _logger.logger.log("VerificationRequest: validateEvent: " + "fail because from_device");

          return false;
        }
      }

      return true;
    }
  }]);
  return VerificationRequest;
}(_events.EventEmitter);

exports.VerificationRequest = VerificationRequest;