"use strict";

var _interopRequireWildcard = require("@babel/runtime/helpers/interopRequireWildcard");

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.MatrixHttpApi = MatrixHttpApi;
exports.retryNetworkOperation = retryNetworkOperation;
exports.AbortError = exports.ConnectionError = exports.MatrixError = exports.PREFIX_MEDIA_R0 = exports.PREFIX_IDENTITY_V2 = exports.PREFIX_IDENTITY_V1 = exports.PREFIX_UNSTABLE = exports.PREFIX_R0 = void 0;

var _regenerator = _interopRequireDefault(require("@babel/runtime/regenerator"));

var _asyncToGenerator2 = _interopRequireDefault(require("@babel/runtime/helpers/asyncToGenerator"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _inherits2 = _interopRequireDefault(require("@babel/runtime/helpers/inherits"));

var _possibleConstructorReturn2 = _interopRequireDefault(require("@babel/runtime/helpers/possibleConstructorReturn"));

var _getPrototypeOf2 = _interopRequireDefault(require("@babel/runtime/helpers/getPrototypeOf"));

var _wrapNativeSuper2 = _interopRequireDefault(require("@babel/runtime/helpers/wrapNativeSuper"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _typeof2 = _interopRequireDefault(require("@babel/runtime/helpers/typeof"));

var _contentType = require("content-type");

var utils = _interopRequireWildcard(require("./utils"));

var _logger = require("./logger");

var callbacks = _interopRequireWildcard(require("./realtime-callbacks"));

function _createSuper(Derived) { var hasNativeReflectConstruct = _isNativeReflectConstruct(); return function _createSuperInternal() { var Super = (0, _getPrototypeOf2["default"])(Derived), result; if (hasNativeReflectConstruct) { var NewTarget = (0, _getPrototypeOf2["default"])(this).constructor; result = Reflect.construct(Super, arguments, NewTarget); } else { result = Super.apply(this, arguments); } return (0, _possibleConstructorReturn2["default"])(this, result); }; }

function _isNativeReflectConstruct() { if (typeof Reflect === "undefined" || !Reflect.construct) return false; if (Reflect.construct.sham) return false; if (typeof Proxy === "function") return true; try { Date.prototype.toString.call(Reflect.construct(Date, [], function () {})); return true; } catch (e) { return false; } }

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { (0, _defineProperty2["default"])(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

/*
TODO:
- CS: complete register function (doing stages)
- Identity server: linkEmail, authEmail, bindEmail, lookup3pid
*/

/**
 * A constant representing the URI path for release 0 of the Client-Server HTTP API.
 */
var PREFIX_R0 = "/_matrix/client/r0";
/**
 * A constant representing the URI path for as-yet unspecified Client-Server HTTP APIs.
 */

exports.PREFIX_R0 = PREFIX_R0;
var PREFIX_UNSTABLE = "/_matrix/client/unstable";
/**
 * URI path for v1 of the the identity API
 * @deprecated Use v2.
 */

exports.PREFIX_UNSTABLE = PREFIX_UNSTABLE;
var PREFIX_IDENTITY_V1 = "/_matrix/identity/api/v1";
/**
 * URI path for the v2 identity API
 */

exports.PREFIX_IDENTITY_V1 = PREFIX_IDENTITY_V1;
var PREFIX_IDENTITY_V2 = "/_matrix/identity/v2";
/**
 * URI path for the media repo API
 */

exports.PREFIX_IDENTITY_V2 = PREFIX_IDENTITY_V2;
var PREFIX_MEDIA_R0 = "/_matrix/media/r0";
/**
 * Construct a MatrixHttpApi.
 * @constructor
 * @param {EventEmitter} event_emitter The event emitter to use for emitting events
 * @param {Object} opts The options to use for this HTTP API.
 * @param {string} opts.baseUrl Required. The base client-server URL e.g.
 * 'http://localhost:8008'.
 * @param {Function} opts.request Required. The function to call for HTTP
 * requests. This function must look like function(opts, callback){ ... }.
 * @param {string} opts.prefix Required. The matrix client prefix to use, e.g.
 * '/_matrix/client/r0'. See PREFIX_R0 and PREFIX_UNSTABLE for constants.
 *
 * @param {boolean} opts.onlyData True to return only the 'data' component of the
 * response (e.g. the parsed HTTP body). If false, requests will return an
 * object with the properties <tt>code</tt>, <tt>headers</tt> and <tt>data</tt>.
 *
 * @param {string} opts.accessToken The access_token to send with requests. Can be
 * null to not send an access token.
 * @param {Object=} opts.extraParams Optional. Extra query parameters to send on
 * requests.
 * @param {Number=} opts.localTimeoutMs The default maximum amount of time to wait
 * before timing out the request. If not specified, there is no timeout.
 * @param {boolean} [opts.useAuthorizationHeader = false] Set to true to use
 * Authorization header instead of query param to send the access token to the server.
 */

exports.PREFIX_MEDIA_R0 = PREFIX_MEDIA_R0;

function MatrixHttpApi(event_emitter, opts) {
  utils.checkObjectHasKeys(opts, ["baseUrl", "request", "prefix"]);
  opts.onlyData = opts.onlyData || false;
  this.event_emitter = event_emitter;
  this.opts = opts;
  this.useAuthorizationHeader = Boolean(opts.useAuthorizationHeader);
  this.uploads = [];
}

MatrixHttpApi.prototype = {
  /**
   * Sets the baase URL for the identity server
   * @param {string} url The new base url
   */
  setIdBaseUrl: function setIdBaseUrl(url) {
    this.opts.idBaseUrl = url;
  },

  /**
   * Get the content repository url with query parameters.
   * @return {Object} An object with a 'base', 'path' and 'params' for base URL,
   *          path and query parameters respectively.
   */
  getContentUri: function getContentUri() {
    var params = {
      access_token: this.opts.accessToken
    };
    return {
      base: this.opts.baseUrl,
      path: "/_matrix/media/r0/upload",
      params: params
    };
  },

  /**
   * Upload content to the Home Server
   *
   * @param {object} file The object to upload. On a browser, something that
   *   can be sent to XMLHttpRequest.send (typically a File).  Under node.js,
   *   a Buffer, String or ReadStream.
   *
   * @param {object} opts  options object
   *
   * @param {string=} opts.name   Name to give the file on the server. Defaults
   *   to <tt>file.name</tt>.
   *
   * @param {boolean=} opts.includeFilename if false will not send the filename,
   *   e.g for encrypted file uploads where filename leaks are undesirable.
   *   Defaults to true.
   *
   * @param {string=} opts.type   Content-type for the upload. Defaults to
   *   <tt>file.type</tt>, or <tt>applicaton/octet-stream</tt>.
   *
   * @param {boolean=} opts.rawResponse Return the raw body, rather than
   *   parsing the JSON. Defaults to false (except on node.js, where it
   *   defaults to true for backwards compatibility).
   *
   * @param {boolean=} opts.onlyContentUri Just return the content URI,
   *   rather than the whole body. Defaults to false (except on browsers,
   *   where it defaults to true for backwards compatibility). Ignored if
   *   opts.rawResponse is true.
   *
   * @param {Function=} opts.callback Deprecated. Optional. The callback to
   *    invoke on success/failure. See the promise return values for more
   *    information.
   *
   * @param {Function=} opts.progressHandler Optional. Called when a chunk of
   *    data has been uploaded, with an object containing the fields `loaded`
   *    (number of bytes transferred) and `total` (total size, if known).
   *
   * @return {Promise} Resolves to response object, as
   *    determined by this.opts.onlyData, opts.rawResponse, and
   *    opts.onlyContentUri.  Rejects with an error (usually a MatrixError).
   */
  uploadContent: function uploadContent(file, opts) {
    if (utils.isFunction(opts)) {
      // opts used to be callback
      opts = {
        callback: opts
      };
    } else if (opts === undefined) {
      opts = {};
    } // default opts.includeFilename to true (ignoring falsey values)


    var includeFilename = opts.includeFilename !== false; // if the file doesn't have a mime type, use a default since
    // the HS errors if we don't supply one.

    var contentType = opts.type || file.type || 'application/octet-stream';
    var fileName = opts.name || file.name; // We used to recommend setting file.stream to the thing to upload on
    // Node.js. As of 2019-06-11, this is still in widespread use in various
    // clients, so we should preserve this for simple objects used in
    // Node.js. File API objects (via either the File or Blob interfaces) in
    // the browser now define a `stream` method, which leads to trouble
    // here, so we also check the type of `stream`.

    var body = file;

    if (body.stream && typeof body.stream !== "function") {
      _logger.logger.warn("Using `file.stream` as the content to upload. Future " + "versions of the js-sdk will change this to expect `file` to " + "be the content directly.");

      body = body.stream;
    } // backwards-compatibility hacks where we used to do different things
    // between browser and node.


    var rawResponse = opts.rawResponse;

    if (rawResponse === undefined) {
      if (global.XMLHttpRequest) {
        rawResponse = false;
      } else {
        _logger.logger.warn("Returning the raw JSON from uploadContent(). Future " + "versions of the js-sdk will change this default, to " + "return the parsed object. Set opts.rawResponse=false " + "to change this behaviour now.");

        rawResponse = true;
      }
    }

    var onlyContentUri = opts.onlyContentUri;

    if (!rawResponse && onlyContentUri === undefined) {
      if (global.XMLHttpRequest) {
        _logger.logger.warn("Returning only the content-uri from uploadContent(). " + "Future versions of the js-sdk will change this " + "default, to return the whole response object. Set " + "opts.onlyContentUri=false to change this behaviour now.");

        onlyContentUri = true;
      } else {
        onlyContentUri = false;
      }
    } // browser-request doesn't support File objects because it deep-copies
    // the options using JSON.parse(JSON.stringify(options)). Instead of
    // loading the whole file into memory as a string and letting
    // browser-request base64 encode and then decode it again, we just
    // use XMLHttpRequest directly.
    // (browser-request doesn't support progress either, which is also kind
    // of important here)


    var upload = {
      loaded: 0,
      total: 0
    };
    var promise; // XMLHttpRequest doesn't parse JSON for us. request normally does, but
    // we're setting opts.json=false so that it doesn't JSON-encode the
    // request, which also means it doesn't JSON-decode the response. Either
    // way, we have to JSON-parse the response ourselves.

    var bodyParser = null;

    if (!rawResponse) {
      bodyParser = function bodyParser(rawBody) {
        var body = JSON.parse(rawBody);

        if (onlyContentUri) {
          body = body.content_uri;

          if (body === undefined) {
            throw Error('Bad response');
          }
        }

        return body;
      };
    }

    if (global.XMLHttpRequest) {
      var defer = utils.defer();
      var xhr = new global.XMLHttpRequest();
      upload.xhr = xhr;
      var cb = requestCallback(defer, opts.callback, this.opts.onlyData);

      var timeout_fn = function timeout_fn() {
        xhr.abort();
        cb(new Error('Timeout'));
      }; // set an initial timeout of 30s; we'll advance it each time we get
      // a progress notification


      xhr.timeout_timer = callbacks.setTimeout(timeout_fn, 30000);

      xhr.onreadystatechange = function () {
        switch (xhr.readyState) {
          case global.XMLHttpRequest.DONE:
            callbacks.clearTimeout(xhr.timeout_timer);
            var resp;

            try {
              if (xhr.status === 0) {
                throw new AbortError();
              }

              if (!xhr.responseText) {
                throw new Error('No response body.');
              }

              resp = xhr.responseText;

              if (bodyParser) {
                resp = bodyParser(resp);
              }
            } catch (err) {
              err.http_status = xhr.status;
              cb(err);
              return;
            }

            cb(undefined, xhr, resp);
            break;
        }
      };

      xhr.upload.addEventListener("progress", function (ev) {
        callbacks.clearTimeout(xhr.timeout_timer);
        upload.loaded = ev.loaded;
        upload.total = ev.total;
        xhr.timeout_timer = callbacks.setTimeout(timeout_fn, 30000);

        if (opts.progressHandler) {
          opts.progressHandler({
            loaded: ev.loaded,
            total: ev.total
          });
        }
      });
      var url = this.opts.baseUrl + "/_matrix/media/r0/upload";
      var queryArgs = [];

      if (includeFilename && fileName) {
        queryArgs.push("filename=" + encodeURIComponent(fileName));
      }

      if (!this.useAuthorizationHeader) {
        queryArgs.push("access_token=" + encodeURIComponent(this.opts.accessToken));
      }

      if (queryArgs.length > 0) {
        url += "?" + queryArgs.join("&");
      }

      xhr.open("POST", url);

      if (this.useAuthorizationHeader) {
        xhr.setRequestHeader("Authorization", "Bearer " + this.opts.accessToken);
      }

      xhr.setRequestHeader("Content-Type", contentType);
      xhr.send(body);
      promise = defer.promise; // dirty hack (as per _request) to allow the upload to be cancelled.

      promise.abort = xhr.abort.bind(xhr);
    } else {
      var queryParams = {};

      if (includeFilename && fileName) {
        queryParams.filename = fileName;
      }

      promise = this.authedRequest(opts.callback, "POST", "/upload", queryParams, body, {
        prefix: "/_matrix/media/r0",
        headers: {
          "Content-Type": contentType
        },
        json: false,
        bodyParser: bodyParser
      });
    }

    var self = this; // remove the upload from the list on completion

    var promise0 = promise["finally"](function () {
      for (var i = 0; i < self.uploads.length; ++i) {
        if (self.uploads[i] === upload) {
          self.uploads.splice(i, 1);
          return;
        }
      }
    }); // copy our dirty abort() method to the new promise

    promise0.abort = promise.abort;
    upload.promise = promise0;
    this.uploads.push(upload);
    return promise0;
  },
  cancelUpload: function cancelUpload(promise) {
    if (promise.abort) {
      promise.abort();
      return true;
    }

    return false;
  },
  getCurrentUploads: function getCurrentUploads() {
    return this.uploads;
  },
  idServerRequest: function idServerRequest(callback, method, path, params, prefix, accessToken) {
    if (!this.opts.idBaseUrl) {
      throw new Error("No Identity Server base URL set");
    }

    var fullUri = this.opts.idBaseUrl + prefix + path;

    if (callback !== undefined && !utils.isFunction(callback)) {
      throw Error("Expected callback to be a function but got " + (0, _typeof2["default"])(callback));
    }

    var opts = {
      uri: fullUri,
      method: method,
      withCredentials: false,
      json: true,
      // we want a JSON response if we can
      _matrix_opts: this.opts,
      headers: {}
    };

    if (method === 'GET') {
      opts.qs = params;
    } else if ((0, _typeof2["default"])(params) === "object") {
      opts.json = params;
    }

    if (accessToken) {
      opts.headers['Authorization'] = "Bearer ".concat(accessToken);
    }

    var defer = utils.defer();
    this.opts.request(opts, requestCallback(defer, callback, this.opts.onlyData));
    return defer.promise;
  },

  /**
   * Perform an authorised request to the homeserver.
   * @param {Function} callback Optional. The callback to invoke on
   * success/failure. See the promise return values for more information.
   * @param {string} method The HTTP method e.g. "GET".
   * @param {string} path The HTTP path <b>after</b> the supplied prefix e.g.
   * "/createRoom".
   *
   * @param {Object=} queryParams A dict of query params (these will NOT be
   * urlencoded). If unspecified, there will be no query params.
   *
   * @param {Object} [data] The HTTP JSON body.
   *
   * @param {Object|Number=} opts additional options. If a number is specified,
   * this is treated as `opts.localTimeoutMs`.
   *
   * @param {Number=} opts.localTimeoutMs The maximum amount of time to wait before
   * timing out the request. If not specified, there is no timeout.
   *
   * @param {sting=} opts.prefix The full prefix to use e.g.
   * "/_matrix/client/v2_alpha". If not specified, uses this.opts.prefix.
   *
   * @param {Object=} opts.headers map of additional request headers
   *
   * @return {Promise} Resolves to <code>{data: {Object},
   * headers: {Object}, code: {Number}}</code>.
   * If <code>onlyData</code> is set, this will resolve to the <code>data</code>
   * object only.
   * @return {module:http-api.MatrixError} Rejects with an error if a problem
   * occurred. This includes network problems and Matrix-specific error JSON.
   */
  authedRequest: function authedRequest(callback, method, path, queryParams, data, opts) {
    if (!queryParams) {
      queryParams = {};
    }

    if (this.useAuthorizationHeader) {
      if (isFinite(opts)) {
        // opts used to be localTimeoutMs
        opts = {
          localTimeoutMs: opts
        };
      }

      if (!opts) {
        opts = {};
      }

      if (!opts.headers) {
        opts.headers = {};
      }

      if (!opts.headers.Authorization) {
        opts.headers.Authorization = "Bearer " + this.opts.accessToken;
      }

      if (queryParams.access_token) {
        delete queryParams.access_token;
      }
    } else {
      if (!queryParams.access_token) {
        queryParams.access_token = this.opts.accessToken;
      }
    }

    var requestPromise = this.request(callback, method, path, queryParams, data, opts);
    var self = this;
    requestPromise["catch"](function (err) {
      if (err.errcode == 'M_UNKNOWN_TOKEN') {
        self.event_emitter.emit("Session.logged_out", err);
      } else if (err.errcode == 'M_CONSENT_NOT_GIVEN') {
        self.event_emitter.emit("no_consent", err.message, err.data.consent_uri);
      }
    }); // return the original promise, otherwise tests break due to it having to
    // go around the event loop one more time to process the result of the request

    return requestPromise;
  },

  /**
   * Perform a request to the homeserver without any credentials.
   * @param {Function} callback Optional. The callback to invoke on
   * success/failure. See the promise return values for more information.
   * @param {string} method The HTTP method e.g. "GET".
   * @param {string} path The HTTP path <b>after</b> the supplied prefix e.g.
   * "/createRoom".
   *
   * @param {Object=} queryParams A dict of query params (these will NOT be
   * urlencoded). If unspecified, there will be no query params.
   *
   * @param {Object} [data] The HTTP JSON body.
   *
   * @param {Object=} opts additional options
   *
   * @param {Number=} opts.localTimeoutMs The maximum amount of time to wait before
   * timing out the request. If not specified, there is no timeout.
   *
   * @param {sting=} opts.prefix The full prefix to use e.g.
   * "/_matrix/client/v2_alpha". If not specified, uses this.opts.prefix.
   *
   * @param {Object=} opts.headers map of additional request headers
   *
   * @return {Promise} Resolves to <code>{data: {Object},
   * headers: {Object}, code: {Number}}</code>.
   * If <code>onlyData</code> is set, this will resolve to the <code>data</code>
   * object only.
   * @return {module:http-api.MatrixError} Rejects with an error if a problem
   * occurred. This includes network problems and Matrix-specific error JSON.
   */
  request: function request(callback, method, path, queryParams, data, opts) {
    opts = opts || {};
    var prefix = opts.prefix !== undefined ? opts.prefix : this.opts.prefix;
    var fullUri = this.opts.baseUrl + prefix + path;
    return this.requestOtherUrl(callback, method, fullUri, queryParams, data, opts);
  },

  /**
   * Perform a request to an arbitrary URL.
   * @param {Function} callback Optional. The callback to invoke on
   * success/failure. See the promise return values for more information.
   * @param {string} method The HTTP method e.g. "GET".
   * @param {string} uri The HTTP URI
   *
   * @param {Object=} queryParams A dict of query params (these will NOT be
   * urlencoded). If unspecified, there will be no query params.
   *
   * @param {Object} [data] The HTTP JSON body.
   *
   * @param {Object=} opts additional options
   *
   * @param {Number=} opts.localTimeoutMs The maximum amount of time to wait before
   * timing out the request. If not specified, there is no timeout.
   *
   * @param {sting=} opts.prefix The full prefix to use e.g.
   * "/_matrix/client/v2_alpha". If not specified, uses this.opts.prefix.
   *
   * @param {Object=} opts.headers map of additional request headers
   *
   * @return {Promise} Resolves to <code>{data: {Object},
   * headers: {Object}, code: {Number}}</code>.
   * If <code>onlyData</code> is set, this will resolve to the <code>data</code>
   * object only.
   * @return {module:http-api.MatrixError} Rejects with an error if a problem
   * occurred. This includes network problems and Matrix-specific error JSON.
   */
  requestOtherUrl: function requestOtherUrl(callback, method, uri, queryParams, data, opts) {
    if (opts === undefined || opts === null) {
      opts = {};
    } else if (isFinite(opts)) {
      // opts used to be localTimeoutMs
      opts = {
        localTimeoutMs: opts
      };
    }

    return this._request(callback, method, uri, queryParams, data, opts);
  },

  /**
   * Form and return a homeserver request URL based on the given path
   * params and prefix.
   * @param {string} path The HTTP path <b>after</b> the supplied prefix e.g.
   * "/createRoom".
   * @param {Object} queryParams A dict of query params (these will NOT be
   * urlencoded).
   * @param {string} prefix The full prefix to use e.g.
   * "/_matrix/client/v2_alpha".
   * @return {string} URL
   */
  getUrl: function getUrl(path, queryParams, prefix) {
    var queryString = "";

    if (queryParams) {
      queryString = "?" + utils.encodeParams(queryParams);
    }

    return this.opts.baseUrl + prefix + path + queryString;
  },

  /**
   * @private
   *
   * @param {function} callback
   * @param {string} method
   * @param {string} uri
   * @param {object} queryParams
   * @param {object|string} data
   * @param {object=} opts
   *
   * @param {boolean} [opts.json =true] Json-encode data before sending, and
   *   decode response on receipt. (We will still json-decode error
   *   responses, even if this is false.)
   *
   * @param {object=} opts.headers  extra request headers
   *
   * @param {number=} opts.localTimeoutMs client-side timeout for the
   *    request. Default timeout if falsy.
   *
   * @param {function=} opts.bodyParser function to parse the body of the
   *    response before passing it to the promise and callback.
   *
   * @return {Promise} a promise which resolves to either the
   * response object (if this.opts.onlyData is truthy), or the parsed
   * body. Rejects
   */
  _request: function _request(callback, method, uri, queryParams, data, opts) {
    if (callback !== undefined && !utils.isFunction(callback)) {
      throw Error("Expected callback to be a function but got " + (0, _typeof2["default"])(callback));
    }

    opts = opts || {};
    var self = this;

    if (this.opts.extraParams) {
      queryParams = _objectSpread(_objectSpread({}, queryParams), this.opts.extraParams);
    }

    var headers = utils.extend({}, opts.headers || {});
    var json = opts.json === undefined ? true : opts.json;
    var bodyParser = opts.bodyParser; // we handle the json encoding/decoding here, because request and
    // browser-request make a mess of it. Specifically, they attempt to
    // json-decode plain-text error responses, which in turn means that the
    // actual error gets swallowed by a SyntaxError.

    if (json) {
      if (data) {
        data = JSON.stringify(data);
        headers['content-type'] = 'application/json';
      }

      if (!headers['accept']) {
        headers['accept'] = 'application/json';
      }

      if (bodyParser === undefined) {
        bodyParser = function bodyParser(rawBody) {
          return JSON.parse(rawBody);
        };
      }
    }

    var defer = utils.defer();
    var timeoutId;
    var timedOut = false;
    var req;
    var localTimeoutMs = opts.localTimeoutMs || this.opts.localTimeoutMs;

    var resetTimeout = function resetTimeout() {
      if (localTimeoutMs) {
        if (timeoutId) {
          callbacks.clearTimeout(timeoutId);
        }

        timeoutId = callbacks.setTimeout(function () {
          timedOut = true;

          if (req && req.abort) {
            req.abort();
          }

          defer.reject(new MatrixError({
            error: "Locally timed out waiting for a response",
            errcode: "ORG.MATRIX.JSSDK_TIMEOUT",
            timeout: localTimeoutMs
          }));
        }, localTimeoutMs);
      }
    };

    resetTimeout();
    var reqPromise = defer.promise;

    try {
      req = this.opts.request({
        uri: uri,
        method: method,
        withCredentials: false,
        qs: queryParams,
        qsStringifyOptions: opts.qsStringifyOptions,
        useQuerystring: true,
        body: data,
        json: false,
        timeout: localTimeoutMs,
        headers: headers || {},
        _matrix_opts: this.opts
      }, function (err, response, body) {
        if (localTimeoutMs) {
          callbacks.clearTimeout(timeoutId);

          if (timedOut) {
            return; // already rejected promise
          }
        }

        var handlerFn = requestCallback(defer, callback, self.opts.onlyData, bodyParser);
        handlerFn(err, response, body);
      });

      if (req) {
        // This will only work in a browser, where opts.request is the
        // `browser-request` import. Currently `request` does not support progress
        // updates - see https://github.com/request/request/pull/2346.
        // `browser-request` returns an XHRHttpRequest which exposes `onprogress`
        if ('onprogress' in req) {
          req.onprogress = function (e) {
            // Prevent the timeout from rejecting the deferred promise if progress is
            // seen with the request
            resetTimeout();
          };
        } // FIXME: This is EVIL, but I can't think of a better way to expose
        // abort() operations on underlying HTTP requests :(


        if (req.abort) reqPromise.abort = req.abort.bind(req);
      }
    } catch (ex) {
      defer.reject(ex);

      if (callback) {
        callback(ex);
      }
    }

    return reqPromise;
  }
};
/*
 * Returns a callback that can be invoked by an HTTP request on completion,
 * that will either resolve or reject the given defer as well as invoke the
 * given userDefinedCallback (if any).
 *
 * HTTP errors are transformed into javascript errors and the deferred is rejected.
 *
 * If bodyParser is given, it is used to transform the body of the successful
 * responses before passing to the defer/callback.
 *
 * If onlyData is true, the defer/callback is invoked with the body of the
 * response, otherwise the result object (with `code` and `data` fields)
 *
 */

var requestCallback = function requestCallback(defer, userDefinedCallback, onlyData, bodyParser) {
  userDefinedCallback = userDefinedCallback || function () {};

  return function (err, response, body) {
    if (err) {
      // the unit tests use matrix-mock-request, which throw the string "aborted" when aborting a request.
      // See https://github.com/matrix-org/matrix-mock-request/blob/3276d0263a561b5b8326b47bae720578a2c7473a/src/index.js#L48
      var aborted = err.name === "AbortError" || err === "aborted";

      if (!aborted && !(err instanceof MatrixError)) {
        // browser-request just throws normal Error objects,
        // not `TypeError`s like fetch does. So just assume any
        // error is due to the connection.
        err = new ConnectionError("request failed", err);
      }
    }

    if (!err) {
      try {
        if (response.statusCode >= 400) {
          err = parseErrorResponse(response, body);
        } else if (bodyParser) {
          body = bodyParser(body);
        }
      } catch (e) {
        err = new Error("Error parsing server response: ".concat(e));
      }
    }

    if (err) {
      defer.reject(err);
      userDefinedCallback(err);
    } else {
      var res = {
        code: response.statusCode,
        // XXX: why do we bother with this? it doesn't work for
        // XMLHttpRequest, so clearly we don't use it.
        headers: response.headers,
        data: body
      };
      defer.resolve(onlyData ? body : res);
      userDefinedCallback(null, onlyData ? body : res);
    }
  };
};
/**
 * Attempt to turn an HTTP error response into a Javascript Error.
 *
 * If it is a JSON response, we will parse it into a MatrixError. Otherwise
 * we return a generic Error.
 *
 * @param {XMLHttpRequest|http.IncomingMessage} response response object
 * @param {String} body raw body of the response
 * @returns {Error}
 */


function parseErrorResponse(response, body) {
  var httpStatus = response.statusCode;
  var contentType = getResponseContentType(response);
  var err;

  if (contentType) {
    if (contentType.type === 'application/json') {
      var jsonBody = (0, _typeof2["default"])(body) === 'object' ? body : JSON.parse(body);
      err = new MatrixError(jsonBody);
    } else if (contentType.type === 'text/plain') {
      err = new Error("Server returned ".concat(httpStatus, " error: ").concat(body));
    }
  }

  if (!err) {
    err = new Error("Server returned ".concat(httpStatus, " error"));
  }

  err.httpStatus = httpStatus;
  return err;
}
/**
 * extract the Content-Type header from the response object, and
 * parse it to a `{type, parameters}` object.
 *
 * returns null if no content-type header could be found.
 *
 * @param {XMLHttpRequest|http.IncomingMessage} response response object
 * @returns {{type: String, parameters: Object}?} parsed content-type header, or null if not found
 */


function getResponseContentType(response) {
  var contentType;

  if (response.getResponseHeader) {
    // XMLHttpRequest provides getResponseHeader
    contentType = response.getResponseHeader("Content-Type");
  } else if (response.headers) {
    // request provides http.IncomingMessage which has a message.headers map
    contentType = response.headers['content-type'] || null;
  }

  if (!contentType) {
    return null;
  }

  try {
    return (0, _contentType.parse)(contentType);
  } catch (e) {
    throw new Error("Error parsing Content-Type '".concat(contentType, "': ").concat(e));
  }
}
/**
 * Construct a Matrix error. This is a JavaScript Error with additional
 * information specific to the standard Matrix error response.
 * @constructor
 * @param {Object} errorJson The Matrix error JSON returned from the homeserver.
 * @prop {string} errcode The Matrix 'errcode' value, e.g. "M_FORBIDDEN".
 * @prop {string} name Same as MatrixError.errcode but with a default unknown string.
 * @prop {string} message The Matrix 'error' value, e.g. "Missing token."
 * @prop {Object} data The raw Matrix error JSON used to construct this object.
 * @prop {integer} httpStatus The numeric HTTP status code given
 */


var MatrixError = /*#__PURE__*/function (_Error) {
  (0, _inherits2["default"])(MatrixError, _Error);

  var _super = _createSuper(MatrixError);

  function MatrixError(errorJson) {
    var _this;

    (0, _classCallCheck2["default"])(this, MatrixError);
    errorJson = errorJson || {};
    _this = _super.call(this, "MatrixError: ".concat(errorJson.errcode));
    _this.errcode = errorJson.errcode;
    _this.name = errorJson.errcode || "Unknown error code";
    _this.message = errorJson.error || "Unknown message";
    _this.data = errorJson;
    return _this;
  }

  return MatrixError;
}( /*#__PURE__*/(0, _wrapNativeSuper2["default"])(Error));
/**
 * Construct a ConnectionError. This is a JavaScript Error indicating
 * that a request failed because of some error with the connection, either
 * CORS was not correctly configured on the server, the server didn't response,
 * the request timed out, or the internet connection on the client side went down.
 * @constructor
 */


exports.MatrixError = MatrixError;

var ConnectionError = /*#__PURE__*/function (_Error2) {
  (0, _inherits2["default"])(ConnectionError, _Error2);

  var _super2 = _createSuper(ConnectionError);

  function ConnectionError(message) {
    var _this2;

    var cause = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : undefined;
    (0, _classCallCheck2["default"])(this, ConnectionError);
    _this2 = _super2.call(this, message + (cause ? ": ".concat(cause.message) : ""));
    _this2._cause = cause;
    return _this2;
  }

  (0, _createClass2["default"])(ConnectionError, [{
    key: "name",
    get: function get() {
      return "ConnectionError";
    }
  }, {
    key: "cause",
    get: function get() {
      return this._cause;
    }
  }]);
  return ConnectionError;
}( /*#__PURE__*/(0, _wrapNativeSuper2["default"])(Error));

exports.ConnectionError = ConnectionError;

var AbortError = /*#__PURE__*/function (_Error3) {
  (0, _inherits2["default"])(AbortError, _Error3);

  var _super3 = _createSuper(AbortError);

  function AbortError() {
    (0, _classCallCheck2["default"])(this, AbortError);
    return _super3.call(this, "Operation aborted");
  }

  (0, _createClass2["default"])(AbortError, [{
    key: "name",
    get: function get() {
      return "AbortError";
    }
  }]);
  return AbortError;
}( /*#__PURE__*/(0, _wrapNativeSuper2["default"])(Error));
/**
 * Retries a network operation run in a callback.
 * @param  {number}   maxAttempts maximum attempts to try
 * @param  {Function} callback    callback that returns a promise of the network operation. If rejected with ConnectionError, it will be retried by calling the callback again.
 * @return {any} the result of the network operation
 * @throws {ConnectionError} If after maxAttempts the callback still throws ConnectionError
 */


exports.AbortError = AbortError;

function retryNetworkOperation(_x, _x2) {
  return _retryNetworkOperation.apply(this, arguments);
}

function _retryNetworkOperation() {
  _retryNetworkOperation = (0, _asyncToGenerator2["default"])( /*#__PURE__*/_regenerator["default"].mark(function _callee2(maxAttempts, callback) {
    var attempts, lastConnectionError;
    return _regenerator["default"].wrap(function _callee2$(_context2) {
      while (1) {
        switch (_context2.prev = _context2.next) {
          case 0:
            attempts = 0;
            lastConnectionError = null;

          case 2:
            if (!(attempts < maxAttempts)) {
              _context2.next = 21;
              break;
            }

            _context2.prev = 3;

            if (!(attempts > 0)) {
              _context2.next = 6;
              break;
            }

            return _context2.delegateYield( /*#__PURE__*/_regenerator["default"].mark(function _callee() {
              var timeout;
              return _regenerator["default"].wrap(function _callee$(_context) {
                while (1) {
                  switch (_context.prev = _context.next) {
                    case 0:
                      timeout = 1000 * Math.pow(2, attempts);

                      _logger.logger.log("network operation failed ".concat(attempts, " times,") + " retrying in ".concat(timeout, "ms..."));

                      _context.next = 4;
                      return new Promise(function (r) {
                        return setTimeout(r, timeout);
                      });

                    case 4:
                    case "end":
                      return _context.stop();
                  }
                }
              }, _callee);
            })(), "t0", 6);

          case 6:
            _context2.next = 8;
            return callback();

          case 8:
            return _context2.abrupt("return", _context2.sent);

          case 11:
            _context2.prev = 11;
            _context2.t1 = _context2["catch"](3);

            if (!(_context2.t1 instanceof ConnectionError)) {
              _context2.next = 18;
              break;
            }

            attempts += 1;
            lastConnectionError = _context2.t1;
            _context2.next = 19;
            break;

          case 18:
            throw _context2.t1;

          case 19:
            _context2.next = 2;
            break;

          case 21:
            throw lastConnectionError;

          case 22:
          case "end":
            return _context2.stop();
        }
      }
    }, _callee2, null, [[3, 11]]);
  }));
  return _retryNetworkOperation.apply(this, arguments);
}