"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.CallEventHandler = void 0;

var _classCallCheck2 = _interopRequireDefault(require("@babel/runtime/helpers/classCallCheck"));

var _createClass2 = _interopRequireDefault(require("@babel/runtime/helpers/createClass"));

var _defineProperty2 = _interopRequireDefault(require("@babel/runtime/helpers/defineProperty"));

var _logger = require("../logger");

var _call = require("./call");

var _event = require("../@types/event");

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

// Don't ring unless we'd be ringing for at least 3 seconds: the user needs some
// time to press the 'accept' button
var RING_GRACE_PERIOD = 3000;

var CallEventHandler = /*#__PURE__*/function () {
  function CallEventHandler(client) {
    var _this = this;

    (0, _classCallCheck2["default"])(this, CallEventHandler);
    (0, _defineProperty2["default"])(this, "client", void 0);
    (0, _defineProperty2["default"])(this, "calls", void 0);
    (0, _defineProperty2["default"])(this, "callEventBuffer", void 0);
    (0, _defineProperty2["default"])(this, "candidateEventsByCall", void 0);
    (0, _defineProperty2["default"])(this, "evaluateEventBuffer", function () {
      if (_this.client.getSyncState() === "SYNCING") {
        // don't process any events until they are all decrypted
        if (_this.callEventBuffer.some(function (e) {
          return e.isBeingDecrypted();
        })) return;
        var ignoreCallIds = new Set(); // inspect the buffer and mark all calls which have been answered
        // or hung up before passing them to the call event handler.

        var _iterator = _createForOfIteratorHelper(_this.callEventBuffer),
            _step;

        try {
          for (_iterator.s(); !(_step = _iterator.n()).done;) {
            var ev = _step.value;

            if (ev.getType() === _event.EventType.CallAnswer || ev.getType() === _event.EventType.CallHangup) {
              ignoreCallIds.add(ev.getContent().call_id);
            }
          } // now loop through the buffer chronologically and inject them

        } catch (err) {
          _iterator.e(err);
        } finally {
          _iterator.f();
        }

        var _iterator2 = _createForOfIteratorHelper(_this.callEventBuffer),
            _step2;

        try {
          for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
            var e = _step2.value;

            if (e.getType() === _event.EventType.CallInvite && ignoreCallIds.has(e.getContent().call_id)) {
              // This call has previously been answered or hung up: ignore it
              continue;
            }

            try {
              _this.handleCallEvent(e);
            } catch (e) {
              _logger.logger.error("Caught exception handling call event", e);
            }
          }
        } catch (err) {
          _iterator2.e(err);
        } finally {
          _iterator2.f();
        }

        _this.callEventBuffer = [];
      }
    });
    (0, _defineProperty2["default"])(this, "onEvent", function (event) {
      // any call events or ones that might be once they're decrypted
      if (event.getType().indexOf("m.call.") === 0 || event.isBeingDecrypted()) {
        // queue up for processing once all events from this sync have been
        // processed (see above).
        _this.callEventBuffer.push(event);
      }

      if (event.isBeingDecrypted() || event.isDecryptionFailure()) {
        // add an event listener for once the event is decrypted.
        event.once("Event.decrypted", function () {
          if (event.getType().indexOf("m.call.") === -1) return;

          if (_this.callEventBuffer.includes(event)) {
            // we were waiting for that event to decrypt, so recheck the buffer
            _this.evaluateEventBuffer();
          } else {
            // This one wasn't buffered so just run the event handler for it
            // straight away
            try {
              _this.handleCallEvent(event);
            } catch (e) {
              _logger.logger.error("Caught exception handling call event", e);
            }
          }
        });
      }
    });
    this.client = client;
    this.calls = new Map(); // The sync code always emits one event at a time, so it will patiently
    // wait for us to finish processing a call invite before delivering the
    // next event, even if that next event is a hangup. We therefore accumulate
    // all our call events and then process them on the 'sync' event, ie.
    // each time a sync has completed. This way, we can avoid emitting incoming
    // call events if we get both the invite and answer/hangup in the same sync.
    // This happens quite often, eg. replaying sync from storage, catchup sync
    // after loading and after we've been offline for a bit.

    this.callEventBuffer = [];
    this.candidateEventsByCall = new Map();
    this.client.on("sync", this.evaluateEventBuffer);
    this.client.on("event", this.onEvent);
  }

  (0, _createClass2["default"])(CallEventHandler, [{
    key: "stop",
    value: function stop() {
      this.client.removeListener("sync", this.evaluateEventBuffer);
      this.client.removeListener("event", this.onEvent);
    }
  }, {
    key: "handleCallEvent",
    value: function handleCallEvent(event) {
      var content = event.getContent();
      var call = content.call_id ? this.calls.get(content.call_id) : undefined; //console.info("RECV %s content=%s", event.getType(), JSON.stringify(content));

      if (event.getType() === _event.EventType.CallInvite) {
        if (event.getSender() === this.client.credentials.userId) {
          return; // ignore invites you send
        }

        if (event.getLocalAge() > content.lifetime - RING_GRACE_PERIOD) {
          return; // expired call
        }

        if (call && call.state === _call.CallState.Ended) {
          return; // stale/old invite event
        }

        if (call) {
          _logger.logger.log("WARN: Already have a MatrixCall with id ".concat(content.call_id, " but got an ") + "invite. Clobbering.");
        }

        call = (0, _call.createNewMatrixCall)(this.client, event.getRoomId(), {
          forceTURN: this.client._forceTURN
        });

        if (!call) {
          _logger.logger.log("Incoming call ID " + content.call_id + " but this client " + "doesn't support WebRTC"); // don't hang up the call: there could be other clients
          // connected that do support WebRTC and declining the
          // the call on their behalf would be really annoying.


          return;
        }

        call.callId = content.call_id;
        call.initWithInvite(event);
        this.calls.set(call.callId, call); // if we stashed candidate events for that call ID, play them back now

        if (this.candidateEventsByCall.get(call.callId)) {
          var _iterator3 = _createForOfIteratorHelper(this.candidateEventsByCall.get(call.callId)),
              _step3;

          try {
            for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
              var ev = _step3.value;
              call.onRemoteIceCandidatesReceived(ev);
            }
          } catch (err) {
            _iterator3.e(err);
          } finally {
            _iterator3.f();
          }
        } // Were we trying to call that user (room)?


        var existingCall;

        var _iterator4 = _createForOfIteratorHelper(this.calls.values()),
            _step4;

        try {
          for (_iterator4.s(); !(_step4 = _iterator4.n()).done;) {
            var thisCall = _step4.value;
            var isCalling = [_call.CallState.WaitLocalMedia, _call.CallState.CreateOffer, _call.CallState.InviteSent].includes(thisCall.state);

            if (call.roomId === thisCall.roomId && thisCall.direction === _call.CallDirection.Outbound && isCalling) {
              existingCall = thisCall;
              break;
            }
          }
        } catch (err) {
          _iterator4.e(err);
        } finally {
          _iterator4.f();
        }

        if (existingCall) {
          // If we've only got to wait_local_media or create_offer and
          // we've got an invite, pick the incoming call because we know
          // we haven't sent our invite yet otherwise, pick whichever
          // call has the lowest call ID (by string comparison)
          if (existingCall.state === _call.CallState.WaitLocalMedia || existingCall.state === _call.CallState.CreateOffer || existingCall.callId > call.callId) {
            _logger.logger.log("Glare detected: answering incoming call " + call.callId + " and canceling outgoing call " + existingCall.callId);

            existingCall.replacedBy(call);
            call.answer();
          } else {
            _logger.logger.log("Glare detected: rejecting incoming call " + call.callId + " and keeping outgoing call " + existingCall.callId);

            call.hangup(_call.CallErrorCode.Replaced, true);
          }
        } else {
          this.client.emit("Call.incoming", call);
        }
      } else if (event.getType() === _event.EventType.CallAnswer) {
        if (!call) {
          return;
        }

        if (event.getSender() === this.client.credentials.userId) {
          if (call.state === _call.CallState.Ringing) {
            call.onAnsweredElsewhere(content);
          }
        } else {
          call.onAnswerReceived(event);
        }
      } else if (event.getType() === _event.EventType.CallCandidates) {
        if (event.getSender() === this.client.credentials.userId) {
          return;
        }

        if (!call) {
          // store the candidates; we may get a call eventually.
          if (!this.candidateEventsByCall.has(content.call_id)) {
            this.candidateEventsByCall.set(content.call_id, []);
          }

          this.candidateEventsByCall.get(content.call_id).push(event);
        } else {
          call.onRemoteIceCandidatesReceived(event);
        }
      } else if ([_event.EventType.CallHangup, _event.EventType.CallReject].includes(event.getType())) {
        // Note that we also observe our own hangups here so we can see
        // if we've already rejected a call that would otherwise be valid
        if (!call) {
          // if not live, store the fact that the call has ended because
          // we're probably getting events backwards so
          // the hangup will come before the invite
          call = (0, _call.createNewMatrixCall)(this.client, event.getRoomId());

          if (call) {
            call.callId = content.call_id;
            call.initWithHangup(event);
            this.calls.set(content.call_id, call);
          }
        } else {
          if (call.state !== _call.CallState.Ended) {
            if (event.getType() === _event.EventType.CallHangup) {
              call.onHangupReceived(content);
            } else {
              call.onRejectReceived(content);
            }

            this.calls["delete"](content.call_id);
          }
        }
      } else if (event.getType() === _event.EventType.CallSelectAnswer) {
        if (!call) return;

        if (event.getContent().party_id === call.ourPartyId) {
          // Ignore remote echo
          return;
        }

        call.onSelectAnswerReceived(event);
      } else if (event.getType() === _event.EventType.CallNegotiate) {
        if (!call) return;

        if (event.getContent().party_id === call.ourPartyId) {
          // Ignore remote echo
          return;
        }

        call.onNegotiateReceived(event);
      }
    }
  }]);
  return CallEventHandler;
}();

exports.CallEventHandler = CallEventHandler;